/******************************************************************************

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#ifndef _FMDB_H_
#define _FMDB_H_
#include "SCUtil.h"
#include "GMI.h" 
#include "pmZoltanCallbacks.h"
#include "pmMigration.h"
#include "pmUtility.h"
#include "pmModel.h"
#include "mFMDB.h"
#include "mPart.h"
#include "mEntity.h"
#include "mException.h"
#include "mVertex.h"
#include "ParUtil.h"
#include "FMDBInternals.h"
#include "pmMigrationCallbacks.h"
#include "FMDB_cint.h" 
#include "modeler.h"
#include "FMDBfwd.h"
#include "mEntityContainer.h"
#include "iUtil.h"

#ifdef __cplusplus
extern "C" {

  typedef class mMesh *  pMeshMdl;
  typedef class mPart  * pPart;  

  typedef EntitySet<mEntity>* pEntSet;
  typedef mEntityGroup* pPartSet;
  typedef pGEntity pGeomEnt;

  typedef class mEntity * pMeshEnt;
  typedef class mEntity * pMeshRgn;
  typedef class mEntity * pMeshFace;
  typedef class mEntity * pMeshEdge;
  typedef class mEntity * pMeshVtx;
  typedef class mEntity * pNode;

  typedef class mAttachableDataContainer* pPtnObj;

  typedef std::pair<int, pMeshEnt> pairRemote;
  typedef class migrationCB MigrCB;
  struct entSetIter
   {
     short int isList;
     void* iter;
   };

  // typedef for iterator; 
  typedef std::list<pEntSet>::iterator  pMeshSetIter;
  typedef std::vector<pEntSet>::iterator  pPartSetIter;
  typedef GenIterator<mPartEntityContainer::iter, mEntity>* pPartEntIter;
  typedef entSetIter* pSetEntIter;

#else   /* not __cplusplus */

#endif /* not __cplusplus */

#ifndef FMDB_PARALLEL
typedef void* MPI_Comm;
#endif

#define MAX_STRING_LENGTH 2048
#define FMDB_NONE -1
  enum FMDB_EntType { FMDB_VERTEX=0,     /*1*/ FMDB_EDGE,    /*2*/ FMDB_FACE,
                              /*3*/ FMDB_REGION, /*4*/ FMDB_ALLTYPE};

  enum FMDB_EntTopo { FMDB_POINT=0,     /*1*/ FMDB_LINE,    /*2*/ FMDB_POLYGON, 
                              /*3*/ FMDB_TRI,   /*4*/ FMDB_QUAD,    /*5*/ FMDB_POLYHEDRON,
                              /*6*/ FMDB_TET,   /*7*/ FMDB_HEX,     /*8*/ FMDB_PRISM, 
                              /*9*/ FMDB_PYRAMID, /*10*/FMDB_SEPTA, /*11*/FMDB_ALLTOPO};

  enum FMDB_EntStatus { FMDB_INTER=0, /*1*/ FMDB_BDRY,  /*2*/ FMDB_GHOST};

  enum FMDB_SetType { FMDB_SSET=0,    /*1*/ FMDB_LSET,  /*2*/ FMDB_PSET};
 
  enum FMDB_LBMethod {FMDB_RCB=0, /*1*/ FMDB_RIB, /*2*/ FMDB_GRAPH, /*3*/ FMDB_HYPERGRAPH, /*4*/ FMDB_PARMETIS};

  enum FMDB_LBApproach {FMDB_PARTITION=0, /*1*/ FMDB_REPARTITION, /*2*/ FMDB_REFINE, /*3*/ PartKway, /*4*/ PartGeom, /*5*/ PartGeomKWay, /*6*/ AdaptiveRepart, /*7*/ RefineKway};

//************************************
//************************************
//      1- SYSTEM-LEVEL FUNCTIONS
//************************************
//************************************

int SCUTIL_Init(MPI_Comm comm);
int SCUTIL_Finalize(void);

int SCUTIL_SplitComm (int numProcGrp);
int SCUTIL_SetComm (MPI_Comm comm);
int SCUTIL_GetComm (MPI_Comm *comm);
int SCUTIL_CommSize();
int SCUTIL_CommRank();
int SCUTIL_CommGrp();
int SCUTIL_GrpSize();
int SCUTIL_GrpRank();

int SCUTIL_Sync(void);

int SCUTIL_GetCurMem(double *mem);
int SCUTIL_DspCurMem(const char* msg);

int SCUTIL_ResetRsrc();
int SCUTIL_GetRsrcDiff(double *time, double *mem);  // resource increase since reset
int SCUTIL_DspRsrcDiff(const char* msg);
int SCUTIL_DspRsrc(const char* msg, double time, double mem);
int SCUTIL_DspSysInfo();

//************************************
//************************************
//      2- TAG FUNCTIONS
//************************************
//************************************

int FMDB_Mesh_CreateTag (pMeshMdl mesh, const char* tagName, int tagType, int tagSize, pTag& tag);
int FMDB_Mesh_DelTag (pMeshMdl mesh, pTag tag, int forceDel);
int FMDB_Mesh_FindTag (pMeshMdl mesh, const char* tagName, pTag& tag);
int FMDB_Mesh_HasTag (pMeshMdl mesh, const pTag tag, int *has);
int FMDB_Mesh_GetTag (pMeshMdl mesh, std::vector<pTag>& tags);

int FMDB_Tag_InUse (pMeshMdl mesh, const pTag tag, int* use); 
int FMDB_Tag_GetType (const pTag tag, int* type);
int FMDB_Tag_GetName (const pTag tag, char* tag_name, int name_len);
int FMDB_Tag_GetSize (const pTag tag, int* size);
int FMDB_Tag_GetByte (const pTag tag, int* byte);

//************************************
//************************************
//      3- MESH-PART FUNCTIONS
//************************************
//************************************

//************************************
//  3.1 Mesh/Part Management 
//************************************

int FMDB_Mesh_Create(pGeomMdl geom, pMeshMdl& mesh);
int FMDB_Mesh_Del(pMeshMdl mesh);

int FMDB_Mesh_LoadSerialFromFile (pMeshMdl mesh, const char* fileName);
int FMDB_Mesh_LoadFromFile (pMeshMdl mesh, const char* fileName, int distributed);
int FMDB_Mesh_WriteToFile (pMeshMdl mesh, const char* fileName, int distributed);

// temporarily unavailable since we do not allow various number parts per process
//int FMDB_Mesh_CreatePart (pMeshMdl mesh, pPart& part);
int FMDB_Mesh_DelPart (pMeshMdl mesh, pPart part);

int FMDB_Mesh_IsPtnSynced(pMeshMdl mesh, int* sync);
int FMDB_Mesh_SyncPtn (pMeshMdl mesh);

//************************************
//  3.2 Mesh Information 
//************************************
int FMDB_Mesh_SetLastError (pMeshMdl mesh, int error_type);
int FMDB_Mesh_GetLastError (pMeshMdl mesh, int *error_type);
int FMDB_Mesh_IsEmpty (pMeshMdl mesh, int* empty);
int FMDB_Mesh_GetGeomMdl (pMeshMdl mesh, pGeomMdl& geom);
int FMDB_Mesh_GetDim (pMeshMdl mesh, int* dim);

int FMDB_Mesh_SetNumPart (pMeshMdl mesh, int numPart);
int FMDB_Mesh_GetNumPart (pMeshMdl mesh, int* numPart);
int FMDB_Mesh_GetPart (pMeshMdl mesh, int ith, pPart& part); 

int FMDB_Mesh_GetAllPart (pMeshMdl mesh, int procId, std::vector<pPart>& vecPart);
int FMDB_Mesh_GetAllPartID (pMeshMdl  mesh, int procId, std::vector<int>& vecPartID); 

int FMDB_Mesh_GetNumSet (pMeshMdl  mesh,  int* numEntSet);
int FMDB_Mesh_GetSet (pMeshMdl  mesh, std::vector<pEntSet>& vecEntset);

int FMDB_Mesh_GetNborProcID (pMeshMdl mesh, std::vector<int>& procRank);

//************************************
//  3.3 Part Information 
//************************************

int FMDB_Part_GlobID (pPart part);
int FMDB_Part_LocalID (pPart part);
int FMDB_PartID_CommRank (int partID);
int FMDB_PartID_LocalID (int partID);

int FMDB_Part_GetDim (pPart part, int* dim);
int FMDB_Part_GetNumSet (pPart part, int* numEntSet);
int FMDB_Part_GetSet (pPart part, std::vector<pEntSet>& vecEntset);

int FMDB_Part_FindEnt (pPart part, pMeshEnt ent, int* found);
int FMDB_Part_GetNumEnt (pPart part, int type, int topo, int* numEnt);
int FMDB_Part_GetNumPartBdry (pPart part, int destPart, int type, int topo, int* numEnt);
int FMDB_Part_GetNumGhost (pPart part, int type, int* numEnt);

int FMDB_Part_GetNumNborPart (pPart part, int type, int* numNbor);
int FMDB_Part_GetNborPartID (pPart part, int type, std::vector<int>& partID);

//************************************
//  3.4 Part Traversal
//************************************

int FMDB_PartEntIter_Init (pPart part, int type, int topo, pPartEntIter& iter);
int FMDB_PartEntIter_InitRevClas (pPart part, pGeomEnt geomEnt, int type, pPartEntIter& iter);
int FMDB_PartEntIter_InitPartBdry (pPart part, int destPart, int type, int topo, pPartEntIter& iter);

int FMDB_PartEntIter_GetNext(pPartEntIter iter, pMeshEnt& meshEnt);
int FMDB_PartEntIter_Del(pPartEntIter iter);
int FMDB_PartEntIter_IsEnd(pPartEntIter iter, int *isEnd);
int FMDB_PartEntIter_Reset(pPartEntIter iter);

//************************************
//  3.5 ID Management
//************************************
int FMDB_Mesh_SetSetID (pMeshMdl  mesh);
int FMDB_Mesh_DelSetID (pMeshMdl  mesh);

int FMDB_Part_SetEntID (pPart part, int type);
int FMDB_Part_DelEntID (pPart part, int type);
int FMDB_Part_SetSetID (pPart part);
int FMDB_Part_DelSetID (pPart part);

// ************************************
//  3.6 Part Tag 
//************************************

int FMDB_Part_DelTag (pPart part, pTag tag);
int FMDB_Part_HasTag (pPart part, pTag tag, int* hasTag);
int FMDB_Part_GetTag (pPart part, std::vector<pTag>& tags);

int FMDB_Part_SetByteTag (pPart part, pTag tag, const void* data, int data_size);
int FMDB_Part_GetByteTag (pPart part, pTag tag, void** data, int* data_size);
int FMDB_Part_SetIntTag (pPart part, pTag tag, const int data);
int FMDB_Part_GetIntTag (pPart part, pTag tag, int* data);
int FMDB_Part_SetDblTag (pPart part, pTag tag, const double data);
int FMDB_Part_GetDblTag (pPart part, pTag tag, double* data);
int FMDB_Part_SetEntTag (pPart part, pTag tag, const pMeshEnt data);
int FMDB_Part_GetEntTag (pPart part, pTag tag, pMeshEnt* data);
int FMDB_Part_SetSetTag (pPart part, pTag tag, const pEntSet data);
int FMDB_Part_GetSetTag (pPart part, pTag tag, pEntSet* data);

int FMDB_Part_SetIntArrTag (pPart part, pTag tag, const int* data, int data_size);
int FMDB_Part_GetIntArrTag (pPart part, pTag tag, int** data, int* data_size);
int FMDB_Part_SetDblArrTag (pPart part, pTag tag, const double* data, int data_size);
int FMDB_Part_GetDblArrTag (pPart part, pTag tag, double** data, int* data_size);
int FMDB_Part_SetEntArrTag (pPart part, pTag tag, const pMeshEnt* data, int data_size);
int FMDB_Part_GetEntArrTag (pPart part, pTag tag, pMeshEnt** data, int* data_size);
int FMDB_Part_SetSetArrTag (pPart part, pTag tag, const pEntSet* data, int data_size);
int FMDB_Part_GetSetArrTag (pPart part, pTag tag, pEntSet** data, int* data_size);

//************************************
//  3.7 Distributed Mesh Functions
//************************************
// global/local load balancing
int FMDB_Mesh_SetPtnParam (pMeshMdl mesh, int ptnMethod, int ptnApproach,
                       double imbalance_tol, int debug_level);
int FMDB_Mesh_LocalPtn(pMeshMdl mesh, pmMigrationCallbacks& migrCB);
int FMDB_Mesh_SerialGlobPtn(pMeshMdl mesh, int ptnMethod, int ptnApproach,
                      double imbalance_tol, int debug_level);
int FMDB_Mesh_GlobPtn(pMeshMdl mesh, pmMigrationCallbacks& migrCB, int ptnMethod, int ptnApproach,
                       double imbalance_tol, int debug_level);

// migration
int FMDB_Mesh_Migr (pMeshMdl mesh, pmMigrationCallbacks& migrCB,
                    map<pMeshEnt, std::pair<int, int> >& entToMigr,
                    map<pPartSet, std::pair<int, int> >& psetToMigr);

int FMDB_Mesh_MigrAdv (pMeshMdl mesh, 
		       pmMigrationCallbacks& migrCB,
                       map<pMeshEnt, std::pair<int, int> >& entToMigr,
                       map<pPartSet, std::pair<int, int> >& psetToMigr,
		       std::vector<pMeshEnt>* newEnt, 
		       std::vector<pMeshEnt>* rmvEnt,
		       std::vector<pPartSet>* newPSet, 
		       std::vector<pPartSet>* rmvPSet);

int FMDB_Mesh_PullPartBdryToOwner (pMeshMdl mesh, 
		       pmMigrationCallbacks& migrCB,
                       std::list<pMeshEnt>& partBdryEnt,
		       std::vector<pMeshEnt>* newEnt, 
		       std::vector<pMeshEnt>* rmvEnt,
		       std::vector<pPartSet>* newPSet, 
		       std::vector<pPartSet>* rmvPSet);

int FMDB_Mesh_PullPartBdryForSnapping (pMeshMdl mesh, 
		       pmMigrationCallbacks& migrCB,
                       std::list<pMeshEnt>& partBdryVtx,
                       std::list<std::pair<pEntity,pEntity> > & vtxEntPair,
		       std::vector<pMeshEnt>* newEnt, 
		       std::vector<pMeshEnt>* rmvEnt,
		       std::vector<pPartSet>* newPSet, 
		       std::vector<pPartSet>* rmvPSet);

int FMDB_Ent_Migr (pMeshMdl mesh, pmMigrationCallbacks& migrCB,
                   pMeshEnt ent, int sourcePartid, int destPartid,
                   std::vector<pMeshEnt> *newEnt, std::vector<pMeshEnt> *oldEnt,
                   std::vector<MPI_Request*> &requests);

int FMDB_Mesh_Merge (pMeshMdl mesh, pmMigrationCallbacks& migrCB);

// ghosting
int FMDB_Mesh_CreateGhost (pMeshMdl mesh, pmMigrationCallbacks& migrCB, 
                           int ghostType, int brgType, 
                           int numLayer, int includeCopy);

int FMDB_Mesh_DelGhost (pMeshMdl mesh);
int FMDB_Mesh_GetGhostInfo (pMeshMdl mesh, std::vector<int>& ghostinfo);

//************************************
//  3.7 Misc
//************************************
int FMDB_Mesh_Verify (pMeshMdl mesh, int* isValid);
int FMDB_Mesh_DspNumEnt (pMeshMdl mesh);
int FMDB_Mesh_DspStat (pMeshMdl mesh);

// ************************************
// ************************************
//      4- ENTITY FUNCTIONS      
// ************************************
// ************************************

//************************************
// 4.1 Entity Management
//************************************
/** create a mesh vertex, edge, face and region */
int FMDB_Vtx_Create (pPart part, pGeomEnt geomEnt, double* xyz,
                     double* param, pMeshEnt& meshVtx);
int FMDB_Edge_Create (pPart part, pGeomEnt geomEnt, pMeshEnt meshVtx1, pMeshEnt meshVtx2, pMeshEnt& meshEdge);
int FMDB_Face_Create (pPart part, pGeomEnt geomEnt, int topo, pMeshEnt* downEnt, int* dirs, pMeshFace& meshFace);

int FMDB_Rgn_Create (pPart part, pGeomEnt geomEnt, int topo, int numDownEnt,
                     pMeshEnt* downEnt, pMeshEnt& meshRgn);

int FMDB_Ent_Find (int topo, pMeshEnt* downEnts, int numDownEnts, pMeshEnt& meshEnt);

int FMDB_Ent_Del (pPart part, pMeshEnt meshEnt);

//************************************
// 4.2 Serial Entity Information
//************************************

int FMDB_Ent_ID (pMeshEnt meshEnt);
std::string FMDB_Ent_StrID(pMeshEnt meshEnt);

int FMDB_Ent_GetType(pMeshEnt meshEnt, int* type);
int FMDB_Ent_GetTopo(pMeshEnt meshEnt, int* topo);

int FMDB_Ent_SetGeomClas (pMeshEnt meshEnt, pGeomEnt geomEnt);
int FMDB_Ent_GetGeomClas (pMeshEnt meshEnt, pGeomEnt& geomEnt);
int FMDB_Ent_GetGeomClasType (pMeshEnt meshEnt, int* geomType);
int FMDB_GeomEnt_GetNumRevClas (pGeomEnt geomEnt, pPart part, int type, int* numEnt);

int FMDB_Ent_GetNumAdj (pMeshEnt meshEnt, int tgtType, int* numAdj);
int FMDB_Ent_GetAdj(pMeshEnt meshEnt, int tgtType, int dir, std::vector<pMeshEnt>& vecAdj);
int FMDB_Ent_Get2ndAdj (pMeshEnt meshEnt, int brgType, int tgtType, std::vector<pMeshEnt>& vecAdj);

//************************************
// 4.3 Vertex Functions
//************************************
int FMDB_Vtx_GetLoadPartOrderPair(pPart part, pMeshEnt meshVtx, int* load_partid, int* load_order);
 
int FMDB_Vtx_SetCoord (pMeshEnt meshVtx, double* xyz);

int FMDB_Vtx_GetCoord (pMeshEnt meshVtx, double** xyz);

int FMDB_Vtx_SetParamCoord (pMeshVtx meshVtx, double* xyz);

int FMDB_Vtx_GetParamCoord1D(pMeshVtx p, double* param);

int FMDB_Vtx_GetParamCoord2D(pMeshVtx p, double* param1, double* param2);

//************************************
// 4.4 Edge Functions 
//************************************

int FMDB_Edge_SetNode (pMeshEnt meshEdge, pNode node);

int FMDB_Edge_GetNode (pMeshEnt meshEdge, int* numNode);

int FMDB_Edge_GetNumNode (pMeshEnt meshEdge, int n, pNode& node);


//************************************
// 4.5 Face Functions
//************************************


//************************************
// 4.6 Region Functions
//************************************


//************************************
// 4.7 Entity Tag
//************************************

int FMDB_Ent_DelTag (pMeshEnt meshEnt, pTag tag);
int FMDB_Ent_HasTag (pMeshEnt meshEnt, pTag tag, int* hasTag);
int FMDB_Ent_GetTag (pMeshEnt meshEnt, std::vector<pTag>& tags);

int FMDB_Ent_SetByteTag (pMeshMdl mesh, pMeshEnt, pTag, const void*, int);
int FMDB_Ent_GetByteTag (pMeshMdl mesh, pMeshEnt, pTag, void**, int*);
int FMDB_Ent_SetIntTag (pMeshMdl mesh, pMeshEnt, pTag, const int);
int FMDB_Ent_GetIntTag (pMeshMdl mesh, pMeshEnt, pTag, int*);
int FMDB_Ent_SetDblTag (pMeshMdl mesh, pMeshEnt, pTag, const double);
int FMDB_Ent_GetDblTag (pMeshMdl mesh, pMeshEnt, pTag, double*);
int FMDB_Ent_SetEntTag (pMeshMdl mesh, pMeshEnt, pTag, const pMeshEnt);
int FMDB_Ent_GetEntTag (pMeshMdl mesh, pMeshEnt, pTag, pMeshEnt*);
int FMDB_Ent_SetSetTag (pMeshMdl mesh, pMeshEnt, pTag, const pEntSet);
int FMDB_Ent_GetSetTag (pMeshMdl mesh, pMeshEnt, pTag, pEntSet*);

int FMDB_Ent_SetIntArrTag (pMeshMdl mesh, pMeshEnt, pTag, const int* data, int data_size);
int FMDB_Ent_GetIntArrTag (pMeshMdl mesh, pMeshEnt, pTag, int** data, int* data_size);
int FMDB_Ent_SetDblArrTag (pMeshMdl mesh, pMeshEnt, pTag, const double* data, int data_size);
int FMDB_Ent_GetDblArrTag (pMeshMdl mesh, pMeshEnt, pTag, double** data, int* data_size);
int FMDB_Ent_SetEntArrTag (pMeshMdl mesh, pMeshEnt, pTag, const pMeshEnt* data, int data_size);
int FMDB_Ent_GetEntArrTag (pMeshMdl mesh, pMeshEnt, pTag, pMeshEnt** data, int* data_size);
int FMDB_Ent_SetSetArrTag (pMeshMdl mesh, pMeshEnt, pTag, const pEntSet* data, int data_size);
int FMDB_Ent_GetSetArrTag (pMeshMdl mesh, pMeshEnt, pTag, pEntSet** data, int* data_size);

//*******************************************
// 4.8 Parallel-specific Entity Information 
//*******************************************
// entity type information - internal, part bdry, ghost, ghosted

int FMDB_Ent_IsOnPartBdry (pMeshEnt meshEnt, int* isPrtBdry);
int FMDB_Ent_IsGhost(pMeshEnt meshEnt, pPart part, int* isGhost);
int FMDB_Ent_IsGhosted (pMeshEnt meshEnt, pPart part, int* ghosted);

// owner part information
int FMDB_Ent_GetOwnPartID(pMeshEnt meshEnt, pPart part, int* partID);
int FMDB_Ent_GetOwnEnt(pMeshEnt meshEnt, pPart part, pMeshEnt& owner);

// remote copy information

int FMDB_Ent_GetNumRmt (pMeshEnt meshEnt, int* numRmt);
int FMDB_Ent_GetAllRmt (pMeshEnt meshEnt, std::vector<std::pair<int, pMeshEnt> >& vecRmt);
int FMDB_Ent_GetRmt(pMeshEnt& meshEnt, int destPart, pMeshEnt& rmtEnt);
int FMDB_Ent_SetRmt(pMeshEnt meshEnt, int partID, pMeshEnt rmtEnt);
int FMDB_Ent_RmvRmt (pMeshEnt meshEnt, int partID);
int FMDB_Ent_ClrRmt (pMeshEnt meshEnt);

int FMDB_Ent_SetPtnTopo (pMeshEnt meshEnt, pPart part);
int FMDB_Ent_GetResPart(pMeshEnt meshEnt, std::vector<int> resPartId);

// ghosting information
int FMDB_Ent_GetNumGhost (pMeshEnt meshEnt, pPart part, int* numGhost);
int FMDB_Ent_GetAllGhost (pMeshEnt meshEnt, pPart part, std::vector<std::pair<int, pMeshEnt> >& vecGhost);
int FMDB_Ent_GetGhost(pMeshEnt& meshEnt, int partID, pMeshEnt& ghostEnt);

// weight control per entity topology
int FMDB_Topo_SetWeight (pMeshMdl mesh, int topo, double weight);
int FMDB_Topo_DelWeight (pMeshMdl mesh, int topo);

// weight control for individual entity
int FMDB_Ent_SetWeight (pMeshEnt ent, double weight);
int FMDB_Ent_DelWeight (pMeshEnt ent);
int FMDB_Ent_GetWeight (pMeshEnt ent, double* weight);

//************************************
//  4.9 Misc
//************************************

int FMDB_Ent_DspInfo (pPart part, pMeshEnt ent);
int FMDB_Ent_DspAllAdj (pMeshEnt ent);
int FMDB_Topo_GetType(int topo);
int FMDB_Topo_GetNumDownAdj (int topo, int downType);
int FMDB_Topo_GetTopoFromFaces(int numFace, pMeshEnt*);


// ************************************
// ************************************
//      5- ENTITY SET FUNCTIONS      
// ************************************
// ************************************

//************************************
//  5.1 Entity Set Management
//************************************

int FMDB_Set_Create (pMeshMdl mesh, pPart part, int set_type, pEntSet& entSet);
int FMDB_Set_Del (pMeshMdl mesh, pPart part, pEntSet entSet);
int FMDB_Mesh_GetNumSet (pMeshMdl mesh, int* numEntSet);
int FMDB_Mesh_GetSet (pMeshMdl mesh, std::vector<pEntSet>& vecSet);
int FMDB_Part_GetNumSet (pPart part, int* numSet);  
int FMDB_Part_GetSet (pPart part, std::vector<pEntSet>& vecSet); 

int FMDB_Set_AddEnts (pEntSet entSet, vector<pMeshEnt> meshEnt);
int FMDB_Set_AddEnt (pEntSet entSet, pMeshEnt meshEnt);
int FMDB_Set_RmvEnt (pEntSet entSet, pMeshEnt meshEnt);
int FMDB_Set_Clr (pEntSet entSet);

//************************************
//  5.2 Entity Set Information
//************************************

int FMDB_Set_ID (pEntSet entSet);
int FMDB_Set_GetNumEnt (pEntSet entSet, int *num);
int FMDB_Set_FindEnt (pEntSet entSet, pMeshEnt meshEnt, int* found);
int FMDB_Set_IsEmpty (pEntSet entSet, int* isEmpty);
int FMDB_Set_GetType (pEntSet entSet, int* type);

int FMDB_Ent_GetPSet (pMeshEnt meshEnt, pEntSet& entSet);   // belonging P-set

//************************************
//  5.3 Traversal Over Entity Set
//************************************
int FMDB_MeshSetIter_Init (pMeshMdl mesh, pMeshSetIter& iter);  // NPO entity set traversal
int FMDB_MeshSetIter_GetNext (pMeshMdl mesh, pMeshSetIter& iter, pEntSet& entSet);
int FMDB_MeshSetIter_IsEnd (pMeshMdl mesh, pMeshSetIter iter, int* isEnd);
int FMDB_MeshSetIter_Reset (pMeshMdl mesh, pMeshSetIter& iter);
int FMDB_MeshSetIter_Del(pMeshSetIter iter);

int FMDB_PartSetIter_Init (pPart part, pPartSetIter& iter);  // PO entity set traversal
int FMDB_PartSetIter_GetNext (pPart part, pPartSetIter& iter, pEntSet& entSet);
int FMDB_PartSetIter_IsEnd (pPart part, pPartSetIter iter, int* isEnd);
int FMDB_PartSetIter_Reset (pPart part, pPartSetIter& iter);
int FMDB_PartSetIter_Del(pPartSetIter iter);


//************************************
//  5.3 Entity Set Traversal
//************************************

int FMDB_SetEntIter_Init (pEntSet entSet, int type, int topo, pSetEntIter& iter); 
int FMDB_SetEntIter_GetNext (pSetEntIter& iter, pMeshEnt& meshEnt);
int FMDB_SetEntIter_IsEnd (pSetEntIter iter, int* isEnd);
int FMDB_SetEntIter_Reset (pSetEntIter iter);
int FMDB_SetEntIter_Del (pSetEntIter iter);

//************************************
//  5.4 Entity Set Tag
//************************************

int FMDB_Set_DelTag (pEntSet entSet, pTag tag);
int FMDB_Set_HasTag (pEntSet entSet, pTag tag, int* hasTag);
int FMDB_Set_GetTag (pEntSet entSet, std::vector<pTag>& tags);

int FMDB_Set_SetByteTag (pEntSet entSet, pTag tag, const void* data, int data_size);
int FMDB_Set_GetByteTag (pEntSet entSet, pTag tag, void** data, int* data_size);
int FMDB_Set_SetIntTag (pEntSet entSet, pTag tag, const int data);
int FMDB_Set_GetIntTag (pEntSet entSet, pTag tag, int* data);
int FMDB_Set_SetDblTag (pEntSet entSet, pTag tag, const double data);
int FMDB_Set_GetDblTag (pEntSet entSet, pTag tag, double* data);
int FMDB_Set_SetEntTag (pEntSet entSet, pTag tag, const pMeshEnt data);
int FMDB_Set_GetEntTag (pEntSet entSet, pTag tag, pMeshEnt* data);
int FMDB_Set_SetSetTag (pEntSet entSet, pTag tag, const pEntSet data);
int FMDB_Set_GetSetTag (pEntSet entSet, pTag tag, pEntSet* data);

int FMDB_Set_SetIntArrTag (pEntSet entSet, pTag tag, const int* data, int data_size);
int FMDB_Set_GetIntArrTag (pEntSet entSet, pTag tag, int** data, int* data_size);
int FMDB_Set_SetDblArrTag (pEntSet entSet, pTag tag, const double* data, int data_size);
int FMDB_Set_GetDblArrTag (pEntSet entSet, pTag tag, double** data, int* data_size);
int FMDB_Set_SetEntArrTag (pEntSet entSet, pTag tag, const pMeshEnt* data, int data_size);
int FMDB_Set_GetEntArrTag (pEntSet entSet, pTag tag, pMeshEnt** data, int* data_size);
int FMDB_Set_SetSetArrTag (pEntSet entSet, pTag tag, const pEntSet* data, int data_size);
int FMDB_Set_GetSetArrTag (pEntSet entSet, pTag tag, pEntSet** data, int* data_size);

//************************************
//  5.5 Entity Set @ Distributed Mesh
//************************************
int FMDB_Set_GetWeight (pEntSet pset, double* weight);

//************************************
//  5.6 Misc
//************************************

int FMDB_Set_DspInfo (pEntSet entSet);


// ************************************
// ************************************
//      6- PARTITION OBJECT FUNCTIONS
// ************************************
// ************************************

int FMDB_Ent_GetAdjPtnObj (pMeshEnt ent, std::vector<pPtnObj>& adjPtnObj);
int FMDB_PtnObj_GetType (pPtnObj ptn_obj, int* type);

#ifdef __cplusplus
}
#endif

#endif /* ifndef _FMDB_H_ */

