/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "FMDB.h"
#include "mPart.h"
#include "mEntity.h"
#include "mTet.h"
#include "mHex.h"
#include "mPyramid.h"
#include "mPrism.h"

int FMDB_Ent_Del (pPart part, pMeshEnt ent)
{
  // remove from upward adj of one-level downward entities
  int type = ent->getLevel();
  if (type>0)
    for(int i=0;i<ent->size(type-1);i++)
        ent->get(type-1,i)->del(ent);

  // remove from part
  part->DEL(ent);
  return SCUtil_SUCCESS;
}

int FMDB_Vtx_Create (pPart part, pGeomEnt geomEnt, double* xyz,
                     double* param, pMeshVtx& meshVtx)
{
  meshVtx = part->createVertex (xyz[0],xyz[1],xyz[2],geomEnt);
  if(param)
    meshVtx->attachVector (FMDB_Util::Instance()->getParametric(), 
                     SCOREC::Util::mVector (*param,*(param+1),*(param+2)) );
  return SCUtil_SUCCESS;
}

int FMDB_Edge_Create (pPart part, pGeomEnt geomEnt, pMeshVtx meshVtx1, pMeshVtx meshVtx2, 
                      pMeshEdge& meshEdge)
{
  pMeshEnt vertices[2];
  vertices[0]=meshVtx1;
  vertices[1]=meshVtx2;

  meshEdge = (pMeshEnt)NULL;
  FMDB_Ent_Find(FMDB_LINE, vertices, 2, meshEdge);
  if (meshEdge)
   return SCUtil_ENTITY_ALREADY_EXISTS;
  meshEdge = (pMeshEnt)part->createEdge ((mVertex*)meshVtx1, (mVertex*)meshVtx2, geomEnt);
  // update vertex->edge adjacency
  meshVtx1->add (meshEdge);
  meshVtx2->add (meshEdge);
  return SCUtil_SUCCESS;
}

int FMDB_Face_Create (pPart part, pGeomEnt geomEnt, int topo, pMeshEnt* downEnts, 
                      int* dirs, pMeshFace& meshFace)
{
  meshFace = (pMeshEnt)NULL;
  int numDownEnts = FMDB_Topo_GetNumDownAdj(topo, downEnts[0]->getLevel());
  int numEdges = FMDB_Topo_GetNumDownAdj(topo, FMDB_EDGE);

  switch (downEnts[0]->getLevel())
  {
    case FMDB_VERTEX: { // create face with vertices
                      pMeshEnt* edges = new pMeshEnt[numEdges];
                      for (int i=0; i<numEdges; i++)
                        if (i!=topo-1)
                          FMDB_Edge_Create (part, geomEnt, downEnts[i], downEnts[i+1], edges[i]);
                        else
                          FMDB_Edge_Create (part, geomEnt, downEnts[0], downEnts[i], edges[i]);

                      FMDB_Ent_Find(topo, edges, numEdges, meshFace);
                      if(meshFace)
                      {
                        delete [] edges;
                        return SCUtil_ENTITY_ALREADY_EXISTS;
                      }
                      if (topo==FMDB_TRI)
                        meshFace = (pMeshEnt)part->createFaceWithEdges((mEdge*)edges[0],
                                                   (mEdge*)edges[1],(mEdge*)edges[2], geomEnt, dirs);
                      else if (topo == FMDB_QUAD)
                        meshFace =  (pMeshEnt)part->createFaceWithEdges((mEdge*)edges[0],
                                    (mEdge*)edges[1], (mEdge*)edges[2],(mEdge*)edges[3],geomEnt,dirs);

                      // update edge->face adjacency
                      for(int i=0;i<numEdges;i++)
                        edges[i]->add(meshFace);
                      delete [] edges;
                    }
                    break;
    case FMDB_EDGE: {  // create face with edges
                      FMDB_Ent_Find(topo, downEnts, numEdges, meshFace);
                      if(meshFace)
                        return SCUtil_ENTITY_ALREADY_EXISTS;
                      if (topo == FMDB_TRI)
                        meshFace = (pMeshEnt)part->createFaceWithEdges((mEdge*) downEnts[0],
                                             (mEdge*)downEnts[1], (mEdge*)downEnts[2],geomEnt,dirs);
                      else if (topo==FMDB_QUAD)
                        meshFace = (pMeshEnt)part->createFaceWithEdges((mEdge*) downEnts[0],
                                       (mEdge*)downEnts[1], (mEdge*)downEnts[2], (mEdge*)downEnts[3],
                                       geomEnt,dirs);
                      else
                        return SCUtil_INVALID_ENTITY_COUNT;
                    // update edge->face adjacency
                    for (int i=0; i<numDownEnts;i++)
                    downEnts[i]->add(meshFace);
                  }
                  break;
     default: return SCUtil_INVALID_ENTITY_COUNT;
  }
  return SCUtil_SUCCESS;
}

int FMDB_Edge_Find (pMeshEnt vtx1, pMeshEnt vtx2, pMeshEnt& meshEnt)
{
  meshEnt = E_exist(vtx1, vtx2);
  if (meshEnt) 
    return SCUtil_SUCCESS;
  else
    return SCUtil_ENTITY_NOT_FOUND;
}

int FMDB_Face_Find (pMeshEnt* ents, int num_ents, pMeshEnt& meshEnt)
{
  if (num_ents==3)
    meshEnt = F_exist(ents[0]->getLevel(), ents[0], ents[1], ents[2], NULL);
  else
    meshEnt = F_exist(ents[0]->getLevel(), ents[0], ents[1], ents[2], ents[3]);
  if (meshEnt) 
    return SCUtil_SUCCESS;
  else
    return SCUtil_ENTITY_NOT_FOUND;
}

int FMDB_Rgn_Find (pMeshEnt* ents, int num_ents, pMeshEnt& meshEnt)
{
  switch (num_ents)
  { 
    case 4: meshEnt = R_exist(ents[0]->getLevel(), ents[0], ents[1], ents[2], ents[3], NULL, NULL);
            break;
    case 5: meshEnt = R_exist(ents[0]->getLevel(), ents[0], ents[1], ents[2], ents[3], ents[4], NULL);
            break;
    case 6: meshEnt = R_exist(ents[0]->getLevel(), ents[0], ents[1], ents[2], ents[3], ents[4], ents[5]);
            break;
    default: return SCUtil_NOT_SUPPORTED;
  }
  if (meshEnt) 
    return SCUtil_SUCCESS;
  else
    return SCUtil_ENTITY_NOT_FOUND;
}

void findEnt(int entType, mEntity** downEnts, int downType, int numDownEnts, mEntity*& meshEnt);
int FMDB_Ent_Find(int topo, pMeshEnt* downEnts, int numDownEnts, pMeshEnt& meshEnt)
{
  if(topo > FMDB_PYRAMID || topo < FMDB_LINE)
   return SCUtil_INVALID_ARGUMENT;

  meshEnt = (pMeshEnt)0;
  int entType = FMDB_Topo_GetType(topo);

  if (entType==1)
    return FMDB_Edge_Find (downEnts[0], downEnts[1], meshEnt);
  if (entType==2)
    return FMDB_Face_Find (downEnts, numDownEnts, meshEnt);
  if (entType==3 &&  numDownEnts<7)
    return FMDB_Rgn_Find (downEnts, numDownEnts, meshEnt);

  int downType = downEnts[0]->getLevel();
  int numDownAdj = FMDB_Topo_GetNumDownAdj(topo, entType-1);

  if(downType < (entType-1))
  {
   pMeshEnt tmpDownEnt[numDownAdj];
   
   if(entType == FMDB_REGION)
   {
    pMeshEnt faces[6];
    switch(topo)
    {
      case FMDB_TET:
                   {
                    for (int i=0; i<numDownAdj; i++)
                     {
                      faces[i]=(pMeshEnt)0;
                      for (int j=0; j<FMDB_TRI; j++)
                          {
                            if (downType==FMDB_VERTEX)
                              tmpDownEnt[j]=downEnts[mTet::Tfv[i][j]];
                            else // downType == FMDB_EDGE
                              tmpDownEnt[j]=downEnts[mTet::Tfe[i][j]];
                          }
			FMDB_Ent_Find(FMDB_TRI, tmpDownEnt, 3, faces[i]);	
                        if(!faces[i])
                         return SCUtil_ENTITY_NOT_FOUND;
                     }
                   }
      break;
      case FMDB_HEX:
                  {
                   for (int i=0; i<numDownAdj; i++)
                    {
                      faces[i]=(pMeshEnt)0;
                      for (int j=0; j<FMDB_QUAD; j++)
                          {
                            if (downType==FMDB_VERTEX)
                              tmpDownEnt[j]=downEnts[mHex::Hfv[i][j]];
                            else // downType == FMDB_EDGE
                              tmpDownEnt[j]=downEnts[mHex::Hfe[i][j]];
                          }
                        FMDB_Ent_Find(FMDB_QUAD, tmpDownEnt, 4, faces[i]);
                        if(!faces[i])
                         return SCUtil_ENTITY_NOT_FOUND;
                    }
                  }
      break;
      case FMDB_PRISM:
                   {
                    for (int i=0; i<numDownAdj; i++)
                     {
		      int faceTopo, numDownFace;

                      if(i==0 || i==4)
                       {
                        faceTopo = FMDB_TRI;
                        numDownFace=3;
                       }
                      else
                       {
                        faceTopo = FMDB_QUAD; 
                        numDownFace=4;
                       }

                      for (int j=0; j<faceTopo; j++)
                          {
                            if (downType==FMDB_VERTEX)
                              tmpDownEnt[j]=downEnts[mPrism::pmfv[i][j]];
                            else // downType == FMDB_EDGE
                              tmpDownEnt[j]=downEnts[mPrism::pmfe[i][j]];
                          }

                      FMDB_Ent_Find(faceTopo, tmpDownEnt, numDownFace, faces[i]);
                      if(!faces[i])
                         return SCUtil_ENTITY_NOT_FOUND;
                     }
                   }
      break;
     case FMDB_PYRAMID:
                     {
     		      for (int i=0; i<numDownAdj; i++)
                      {
                        int faceTopo, numDownFace;
                         if(!i)
                          {
                           faceTopo = FMDB_QUAD;
                           numDownFace = 4;
                          }
                         else
                          {
                           faceTopo = FMDB_TRI;
                           numDownFace = 3;
                          }
                        for (int j=0; j<faceTopo; j++)
                          {
                            if (downType==FMDB_VERTEX)
                              tmpDownEnt[j]=downEnts[mPyramid::prfv[i][j]];
                            else // downType == FMDB_EDGE
                              tmpDownEnt[j]=downEnts[mPyramid::prfe[i][j]];
                          }
                        FMDB_Ent_Find(faceTopo, tmpDownEnt, numDownFace, faces[i]);
                        if(!faces[i])
                          return SCUtil_ENTITY_NOT_FOUND;
                      }                 
                     }
      break;
     }
     findEnt(FMDB_REGION, faces, FMDB_FACE, numDownAdj, meshEnt);
    }
   else if(entType == FMDB_FACE)
    {
    pMeshEnt edges[4];
    for(int i = 0; i < numDownAdj; i++)
     {
     edges[i]=(pMeshEnt)0;
     if(i!=topo-1) 
     {
      tmpDownEnt[0]=downEnts[i];
      tmpDownEnt[1]=downEnts[i+1];
      findEnt(FMDB_EDGE, tmpDownEnt, FMDB_VERTEX, 2, edges[i]);
     }
     else
      {
       tmpDownEnt[0]=downEnts[0];
       tmpDownEnt[1]=downEnts[i];
       findEnt(FMDB_EDGE, tmpDownEnt, FMDB_VERTEX, 2, edges[i]);
      }
     if(!edges[i])
       return SCUtil_ENTITY_NOT_FOUND;
     }
    findEnt(FMDB_FACE, edges, FMDB_EDGE, numDownAdj, meshEnt);
    //delete [] edges;
    }
  //delete [] tmpDownEnt;
  }// if(downType < (entType-1))  
  else
    findEnt(entType, downEnts, downType, numDownEnts, meshEnt);
  if(!meshEnt)
    return SCUtil_ENTITY_NOT_FOUND;
  else
    return SCUtil_SUCCESS;
}

void findEnt(int entType, mEntity** downEnts, int downType, int numDownEnts, mEntity*& meshEnt)
{
  vector<pMeshEnt> vecAdjEnt;
  FMDB_Ent_GetAdj(downEnts[0], entType, 0, vecAdjEnt);

  for(vector<pMeshEnt>::iterator itVec = vecAdjEnt.begin(); itVec != vecAdjEnt.end(); itVec++)
  {
    pMeshEnt upEnt = *itVec;
    vector<pMeshEnt> vecDownEnts;
    FMDB_Ent_GetAdj(upEnt, downType, 0, vecDownEnts);
    int count = 0;
    for (vector<pMeshEnt>::iterator itDown = vecDownEnts.begin(); itDown != vecDownEnts.end(); itDown++)
    {
      pMeshEnt pDownEnt = *itDown;
      bool bEntFound = false;
      for(int j = 0; j < numDownEnts; j++)
        if(downEnts[j] == pDownEnt)
          bEntFound = true;
      if(!bEntFound)
        break;
      ++count;
      if (count == numDownEnts)
        meshEnt = upEnt;
    } // for itDown
    vecDownEnts.clear();
  }  // for itVec
  vecAdjEnt.clear();
}

