/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "FMDB.h"
#include "mPart.h"
#include "mMesh.h"

//************************************
//  5.1 Entity Set Management
//************************************

int FMDB_Set_Create (pMeshMdl mesh, pPart part, int set_type, pEntSet& entSet)
{
  int mesh_dim;
  FMDB_Mesh_GetDim(mesh, &mesh_dim);

  switch (set_type)
  { 
    case FMDB_SSET:
    case FMDB_LSET: entSet = mEntitySet::createEntitySet(mesh, set_type);
                    mesh->addEntSet(entSet);
                    break;
    case FMDB_PSET: entSet = part->createPSet(mesh_dim);
                    break;
    default: return SCUtil_INVALID_SET_TYPE;
  }
  return SCUtil_SUCCESS;
}

int FMDB_Set_Del (pMeshMdl mesh, pPart part, pEntSet entSet)
{
  if (entSet->getContainerType()== mAttachableDataContainer::PSET)
    part->deletePSet((mEntityGroup*)entSet, 1); 
  else
    mesh->removeEntSet((pEntSet)entSet);
   return SCUtil_SUCCESS; 
}

/* Given an entity set handle and a mesh entity handle, add the entity into the entity set.*/
int FMDB_Set_AddEnt (pEntSet entSet, pMeshEnt meshEnt)
{
  if (entSet->getContainerType()== mAttachableDataContainer::PSET)
    ((mEntityGroup*)entSet)->addEntity(meshEnt);
  else
    entSet->addEntity(meshEnt);
  return SCUtil_SUCCESS;
}

int FMDB_Set_AddEnts (pEntSet entSet, vector<pMeshEnt> meshEnts)
{
  vector<pMeshEnt>::iterator iter = meshEnts.begin(); 
  if (entSet->getContainerType()== mAttachableDataContainer::PSET)
    for(; iter!= meshEnts.end(); ++iter)
      ((mEntityGroup*)entSet)->addEntity(*iter); 
  else
    for(; iter!= meshEnts.end(); ++iter)
      entSet->addEntity(*iter);
   return SCUtil_SUCCESS; 
}

/* Given an entity set and a mesh entity handle, remove the entity from the entity set */
int FMDB_Set_RmvEnt (pEntSet /* inout */ entSet, pMeshEnt /* in */ meshEnt)
{
  if (entSet->getContainerType()== mAttachableDataContainer::PSET)
    ((mEntityGroup*)entSet)->removeEntity(meshEnt);
  else
    entSet->removeEntity(meshEnt);
  return SCUtil_SUCCESS;
}

int FMDB_Set_Clr (pEntSet entSet)
{
  if (entSet->getContainerType()== mAttachableDataContainer::PSET)
    ((mEntityGroup*)entSet)->clear(); 
  else if(entSet->isList())
    ((mEntitySetUnordered*)entSet)->clear(); 
  else
    ((mEntitySetOrdered*)entSet)->clear();
  return SCUtil_SUCCESS; 
}

//************************************
//  5.2 Entity Set Information
//************************************

int FMDB_Set_ID (pEntSet eset)
{
  int id;
  eset->getAttachedInt(FMDB_Util::Instance()->getId(), &id);
  if (id>=0) return id;
  else return -1;
}

int FMDB_Set_IsEmpty (pEntSet eset, int* isEmpty)
{
   *isEmpty = eset->isEmpty(); 
   return SCUtil_SUCCESS; 
}

int FMDB_Set_GetType (pEntSet eset, int* type)
{
  if (eset->getContainerType()== mAttachableDataContainer::PSET)
    *type = FMDB_PSET;
  else if (eset->isList())
    *type = FMDB_LSET;
  else 
    *type = FMDB_SSET;
  return SCUtil_SUCCESS; 
}

int FMDB_Set_GetNumEnt (pEntSet eset, int *num)
{
  *num = eset->numEntity(); 
  return SCUtil_SUCCESS;
}

int FMDB_Set_FindEnt (pEntSet eset, pMeshEnt ent, int* found)
{
  *found = (int)(eset->existEntity(ent)); 
  return SCUtil_SUCCESS;
}

int FMDB_Set_GetWeight (pEntSet entSet, double* wgt)
{
  // return error if ent_set is not p-set
  if (entSet->getContainerType()!= mAttachableDataContainer::PSET)
    return SCUtil_INVALID_SET_TYPE;

  double weight, sum=0.0;
  for (mEntityGroup::iter ent_it=((mEntityGroup*)entSet)->begin(); ent_it!=((mEntityGroup*)entSet)->end(); ++ent_it)
  {  
    if ((*ent_it)->getAttachedDouble(FMDB_Util::Instance()->getWeight(), &weight))
      sum+=weight;
    else
      sum+=1.0;
  }
  *wgt = sum;
  return SCUtil_SUCCESS;
}

/*
int FMDB_Part_PO_BuildPrismStacks(pPart part)
{
  if (entSet->getContainerType()!= mAttachableDataContainer::PSET)
    return SCUtil_INVALID_SET_TYPE;

  int mesh_dim;
  FMDB_Mesh_GetDim(part->getMesh(), &mesh_dim);
// construct entity groups 
  pFace oface, face; 
  pRegion oreg, reg; 
  mEntityGroup* entGrp;

  pPartEntIter iter; 
  int isEnd, iterEnd = FMDB_PartEntIter_Init (part, 2, FMDB_ALLTOPO, iter);
  int loop_counter=0;
  int inter_counter=0; 
  while (!iterEnd)
  {
    iterEnd = FMDB_PartEntIter_GetNext(iter, face); 
    if(face->getType()==mEntity::TRI && GEN_type(F_whatIn(face))==2) {                  // on model face
      inter_counter++; 
    
      reg = F_region(face, 0);
      if(!reg)
	 reg = F_region(face, 1);
      if(!reg) {
	 cout<<__func__<<" failed: no region is connected to a surface face. "<<endl; 
	 return SCUtil_FAILURE; 
      }

      oface = face; 
      oreg = reg; 

      if(reg->getType() == mEntity::PRISM) {
	 entGrp = (mEntityGroup*) (part->createPSet(mesh_dim)); 
	 ++loop_counter; 
      }
      else
      {
	 continue; 
      }

      while (reg->getType()== mEntity::PRISM) {
   
	entGrp->addEntity(reg);                   // add an entity to the entity group
      
	pPList flist = R_faces(reg, 0);
	void* it=0; 
	while((face = (pFace)PList_next(flist, &it))) {
	 
	  if(face!=oface && face->getType()==mEntity::TRI) {
            
	    int i=0; 
	    for(; i< F_numRegions(face); i++) {
	      reg = F_region(face, i); 
	      if(reg!=oreg)
		break; 
	    }
	    
	    if(!reg || i==F_numRegions(face)) {
	      cout<<__func__<<" failed: the number of regions connected to a interior face is not correct. "<<endl;
	      return SCUtil_FAILURE;
	    }
	    oface = face; 
	    oreg = reg; 
	    break;   
	  }
	}
	PList_delete(flist);
      }
      
    }  
  }
  
  FMDB_PartEntIter_IsEnd(iter, &isEnd);
  FMDB_PartEntIter_Del (iter);
  return SCUtil_SUCCESS; 
}
*/


