/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "FMDB.h"
 
int FMDB_Set_HasTag (pEntSet set, pTag tag, int* hasTag)
{
  *hasTag=0;
  if (Taggable_HasTag(static_cast<pTaggable>(set), tag))
    *hasTag=1;
  return SCUtil_SUCCESS;
}

int FMDB_Set_DelTag (pEntSet set, pTag tag)
{
  Taggable_DelTag(static_cast<pTaggable>(set), tag);
  return SCUtil_SUCCESS;
}

int FMDB_Set_GetTag (pEntSet set, std::vector<pTag>& tags)
{
  Taggable_GetTag (static_cast<pTaggable>(set), tags);
  return SCUtil_SUCCESS;
}

int FMDB_Set_SetByteTag (pEntSet set, pTag tag, const void* data, int data_size)
{
#ifdef DEBUG
  int exist=0;
  FMDB_Mesh_HasTag (set->getMesh(), tag, &exist);
  if (!exist)
    return SCUtil_INVALID_TAG_HANDLE;
#endif

  if(!data || !data_size)
    return SCUtil_NULL_ARRAY;
  
  int tag_size=Tag_GetSize(tag), tag_type = Tag_GetType(tag);
  switch (tag_type)
  {
    case SCUtil_BYTE: Taggable_SetByteArrData<char>(static_cast<pTaggable>(set), tag, data, data_size);
                      break;
    case SCUtil_INT: if (tag_size==1)  // single int
                       Taggable_SetByteData<int>(static_cast<pTaggable>(set), tag, data, data_size);
                     else  // int array
                       Taggable_SetByteArrData<int>(static_cast<pTaggable>(set), tag, data, data_size);
                     break;
    case SCUtil_DBL: if (tag_size==1)  // single double
                        Taggable_SetByteData<double>(static_cast<pTaggable>(set), tag, data, data_size);
                     else  // double array
                        Taggable_SetByteArrData<double>(static_cast<pTaggable>(set), tag, data, data_size);
                     break;
    case SCUtil_ENT: if (tag_size==1)  // single ent
                        Taggable_SetByteData<pMeshEnt>(static_cast<pTaggable>(set), tag, data, data_size);
                     else  // ent array
                        Taggable_SetByteArrData<pMeshEnt>(static_cast<pTaggable>(set), tag, data, data_size);
                     break;
    case SCUtil_SET: if (tag_size==1)  // single set
                        Taggable_SetByteData<pEntSet>(static_cast<pTaggable>(set), tag, data, data_size);
                     else  // set array
                        Taggable_SetByteArrData<pEntSet>(static_cast<pTaggable>(set), tag, data, data_size);
                     break;
    default: break;
  } // switch
  return SCUtil_SUCCESS;
}

int FMDB_Set_GetByteTag (pEntSet set, pTag tag, void** data, int* data_size)
{
#ifdef DEBUG
  int exist=0;
  FMDB_Mesh_HasTag (set->getMesh(), tag, &exist);
  if (!exist)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  int tag_size = Tag_GetSize(tag);
  int tag_type = Tag_GetType(tag);

  switch (tag_type)
  {
    case SCUtil_BYTE: Taggable_GetByteArrData<char>(static_cast<pTaggable>(set), tag, data, data_size);
                      break;
    case SCUtil_INT: if (tag_size==1)  // int
                       Taggable_GetByteData<int>(static_cast<pTaggable>(set), tag, data, data_size);
                     else // int array
                       Taggable_GetByteArrData<int>(static_cast<pTaggable>(set), tag, data, data_size);
                     break;
    case SCUtil_DBL: if (tag_size==1)  // double
                       Taggable_GetByteData<double>(static_cast<pTaggable>(set), tag, data, data_size);
                     else // double array
                       Taggable_GetByteArrData<double>(static_cast<pTaggable>(set), tag, data, data_size);
                     break;
    case SCUtil_ENT: if (tag_size==1)  // ent
                       Taggable_GetByteData<pMeshEnt>(static_cast<pTaggable>(set), tag, data, data_size);
                     else // ent array
                       Taggable_GetByteArrData<pMeshEnt>(static_cast<pTaggable>(set), tag, data, data_size);
                     break;
    case SCUtil_SET: if (tag_size==1)  // set
                       Taggable_GetByteData<pEntSet>(static_cast<pTaggable>(set), tag, data, data_size);
                     else // set array
                       Taggable_GetByteArrData<pEntSet>(static_cast<pTaggable>(set), tag, data, data_size);
                     break;
    default: break;
  }  // switch
  return SCUtil_SUCCESS;
}

int FMDB_Set_SetIntTag (pEntSet set, pTag tag, const int data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_INT) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<int>(static_cast<pTaggable>(set), tag, &data);
}

int FMDB_Set_GetIntTag (pEntSet set, pTag tag, int* data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_INT) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_GetData<int>(static_cast<pTaggable>(set), tag, data);
}

int FMDB_Set_SetDblTag (pEntSet set, pTag tag, const double data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_DBL) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<double>(static_cast<pTaggable>(set), tag, &data);
}

int FMDB_Set_GetDblTag (pEntSet set, pTag tag, double* data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_DBL) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_GetData<double>(static_cast<pTaggable>(set), tag, data);
}

int FMDB_Set_SetEntTag (pEntSet set, pTag tag, const pMeshEnt data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_ENT) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<pMeshEnt>(static_cast<pTaggable>(set), tag, &data);
}

int FMDB_Set_GetEntTag (pEntSet set, pTag tag, pMeshEnt *data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_ENT) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_GetData<pMeshEnt>(static_cast<pTaggable>(set), tag, data);
}

int FMDB_Set_SetSetTag (pEntSet set, pTag tag, const pEntSet data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_SET) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<pEntSet>(static_cast<pTaggable>(set), tag, &data);
}

int FMDB_Set_GetSetTag (pEntSet set, pTag tag, pEntSet *data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_SET) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_GetData<pEntSet>(static_cast<pTaggable>(set), tag, data);
}

// array tag
int FMDB_Set_SetIntArrTag (pEntSet set, pTag tag, const int* data, int data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_INT) || Tag_GetSize(tag)!=data_size)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<int>(static_cast<pTaggable>(set), tag, data);
}

int FMDB_Set_GetIntArrTag (pEntSet set, pTag tag, int** data, int* data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_INT))
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  *data_size=Tag_GetSize(tag);
  return Taggable_GetData<int>(static_cast<pTaggable>(set), tag, *data);
}

int FMDB_Set_SetDblArrTag (pEntSet set, pTag tag, const double* data, int data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_DBL) || Tag_GetSize(tag)!=data_size)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<double>(static_cast<pTaggable>(set), tag, data);
}

int FMDB_Set_GetDblArrTag (pEntSet set, pTag tag, double** data, int* data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_DBL))
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  *data_size=Tag_GetSize(tag);
  return Taggable_GetData<double>(static_cast<pTaggable>(set), tag, *data);
}

int FMDB_Set_SetEntArrTag (pEntSet set, pTag tag, const pMeshEnt* data, int data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_ENT) || Tag_GetSize(tag)!=data_size)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<pMeshEnt>(static_cast<pTaggable>(set), tag, data);
}

int FMDB_Set_GetEntArrTag (pEntSet set, pTag tag, pMeshEnt** data, int* data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_ENT))
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  *data_size=Tag_GetSize(tag);
  return Taggable_GetData<pMeshEnt>(static_cast<pTaggable>(set), tag, *data);
}

int FMDB_Set_SetSetArrTag (pEntSet set, pTag tag, const pEntSet* data, int data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_SET) || Tag_GetSize(tag)!=data_size)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<pEntSet>(static_cast<pTaggable>(set), tag, data);
}

int FMDB_Set_GetSetArrTag (pEntSet set, pTag tag, pEntSet** data, int* data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(set->getMesh()), tag, SCUtil_SET))
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  *data_size=Tag_GetSize(tag);
  return Taggable_GetData<pEntSet>(static_cast<pTaggable>(set), tag, *data);
}
