/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "FMDB.h"
#include "FMDB_Iterator.h"
#include "mMesh.h"

//************************************
//  3.4 Part Traversal
//************************************
/* Given a part and an entity type (0 to 3), get a pointer to the ﬁrst mesh entity of the type.
   If no entity exists or type is invalid, null is gotton for Iter. */
int FMDB_PartEntIter_Init (
        pPart part, int type, int topo, pPartEntIter& iter)
{
  int mesh_dim;
  FMDB_Mesh_GetDim(part->getMesh(), &mesh_dim);

  if(type == FMDB_ALLTYPE)
  {
    if (topo==FMDB_ALLTOPO)
    {
      if (part->size(0)+part->size(1)+part->size(2)+part->size(3)==0)
        return SCUtil_FAILURE;
       iter = new mPartIterator(part->beginall(FMDB_VERTEX), part->endall(mesh_dim), FMDB_ALLTYPE, FMDB_ALLTOPO,
                             (void*)part, &processingMeshFilter);
    }
    else // specific topo
    {
      if (part->size(FMDB_Topo_GetType(topo))==0)
        return SCUtil_FAILURE;
       iter = new mPartIterator(part->beginall(FMDB_Topo_GetType(topo)), part->endall(FMDB_Topo_GetType(topo)),
              FMDB_Topo_GetType(topo), topo, (void*)part, &processingMeshFilter);
    }
  } 
  else  // (type == FMDB_ALLTYPE)
  {
    if (part->size(type)==0)
      return SCUtil_FAILURE;
    iter = new mPartIterator(part->beginall(type), part->endall(type), type, topo,(void*)part, &processingMeshFilter);

  }
  if(iter->end())
    return SCUtil_FAILURE;
  return SCUtil_SUCCESS;
}

int FMDB_PartEntIter_InitPartBdry(pPart part, int destPart, int type, int topo, pPartEntIter& iter)
{
  int mesh_dim;
  FMDB_Mesh_GetDim(part->getMesh(), &mesh_dim);

  if (type > FMDB_ALLTYPE || type < FMDB_VERTEX)
    return SCUtil_INVALID_ENTITY_TYPE;

  part_pid_struct* part_pid = new part_pid_struct; 
  part_pid->part = part; 
  part_pid->pid = destPart; 

   if(type == FMDB_ALLTYPE)
   {
     if (topo==FMDB_ALLTOPO)
     {
       if (part->size(0)+part->size(1)+part->size(2)+part->size(3)==0)
         return SCUtil_FAILURE;
       iter = new mPartIterator(part->beginall(FMDB_VERTEX), part->endall(mesh_dim), FMDB_ALLTYPE, FMDB_ALLTOPO, (void*)part_pid, &processingPartBdryEntFilter);
     }
     else // specific topo
     {
       if (part->size(FMDB_Topo_GetType(topo))==0)
         return SCUtil_FAILURE;

       iter = new mPartIterator(part->beginall(FMDB_Topo_GetType(topo)), part->endall(FMDB_Topo_GetType(topo)), 
              FMDB_Topo_GetType(topo), topo, (void*)part_pid, &processingPartBdryEntFilter);
     }
   }
   else 
   {
     if (part->size(type)==0 || type==mesh_dim)
       return SCUtil_FAILURE;
     iter = new mPartIterator(part->beginall(type), part->endall(type), type, topo, 
                             (void*)part_pid, &processingPartBdryEntFilter);
   }
  if(iter->end())
    return SCUtil_FAILURE;
  return SCUtil_SUCCESS; 
}

/* Given a part and a geometric entity, get a pointer to the first equal order mesh entity
classified on the geometric model entity. If no mesh entity exists, null is gotton for Iter. */
int FMDB_PartEntIter_InitRevClas (
	pPart /* in */ part,
	pGeomEnt /* in */ geomEnt,
        int /* in */  type,
	pPartEntIter& /* out */ iter)
{
  int mesh_dim;
  FMDB_Mesh_GetDim(part->getMesh(), &mesh_dim);

  part_gent_struct* part_gent = new part_gent_struct;
  part_gent->part = part;
  part_gent->gent = geomEnt;

  if (type == FMDB_ALLTYPE)
  {
    if (part->size(0)+part->size(1)+part->size(2)+part->size(3)==0)
      return SCUtil_FAILURE;
    iter = new mPartIterator(part->beginall(FMDB_VERTEX), part->endall(mesh_dim), 
                               FMDB_ALLTYPE, FMDB_ALLTOPO, (void*)part_gent, &processingGeomFilter);
  }
  else
  {
    if (part->size(type)==0)
      return SCUtil_FAILURE;
    iter = new mPartIterator(part->beginall(type), part->endall(type), type, 
                            FMDB_ALLTOPO,(void*)part_gent, &processingGeomFilter);  
  }
  if(iter->end())
    return SCUtil_FAILURE;
  return SCUtil_SUCCESS;
}

/* get next element of the mesh iterator */
int FMDB_PartEntIter_GetNext(pPartEntIter iter, pMeshEnt& ent)
{
   if(iter->end())
     return SCUtil_FAILURE;
   else
     ent = **iter;
   iter->next();
   return SCUtil_SUCCESS;

}

/* Check if the iterator has reached its end */
int FMDB_PartEntIter_IsEnd(pPartEntIter iter, int * isEnd)
{
  if(iter->end())
     *isEnd = 1;
  else
    *isEnd = 0;
  return SCUtil_SUCCESS;
}                                                                                                                 
int FMDB_PartEntIter_Del (pPartEntIter iter)
{
  if (iter) delete iter;
  iter=NULL;
  return SCUtil_SUCCESS;
}

int FMDB_PartEntIter_Reset (pPartEntIter iter)
{
  iter->reset();
  if(iter->end())
   return SCUtil_FAILURE;
  return SCUtil_SUCCESS;
}


//************************************
//  5.3 Entity Set Traversal Over Mesh/Part 
//************************************

int FMDB_MeshSetIter_Init (pMeshMdl mesh, pMeshSetIter& iter)  // NPO entity set traversal
{
   iter = mesh->beginEntSet(); 
   return SCUtil_SUCCESS; 
}

int FMDB_MeshSetIter_GetNext (pMeshMdl mesh, pMeshSetIter& iter, pEntSet& entSet)
{
   if(iter!=mesh->endEntSet())
   {
      entSet = *iter; 
      ++iter;
   }
   else
      return SCUtil_FAILURE; 
   return SCUtil_SUCCESS; 
}

int FMDB_MeshSetIter_IsEnd (pMeshMdl mesh, pMeshSetIter iter, int* isEnd)
{
   if(iter==mesh->endEntSet())
      *isEnd = 1; 
   else
      *isEnd = 0; 
   return SCUtil_SUCCESS; 
}

int FMDB_MeshSetIter_Reset (pMeshMdl mesh, pMeshSetIter& iter)
{
   iter = mesh->beginEntSet();
   return SCUtil_SUCCESS;
}

int FMDB_MeshSetIter_Del (pMeshSetIter iter)
{ 
  return SCUtil_SUCCESS;
}

int FMDB_PartSetIter_Init (pPart part, pPartSetIter& iter)  // PO entity set traversal
{
   iter = part->beginPSet(); 
   return SCUtil_SUCCESS; 
}

int FMDB_PartSetIter_GetNext (pPart part, pPartSetIter& iter, pEntSet& entSet)
{
   if(iter!=part->endPSet())
   {
      entSet = (pEntSet)(*iter); 
      ++iter; 
   }
   else
      return SCUtil_FAILURE;
   return SCUtil_SUCCESS; 
}

int FMDB_PartSetIter_IsEnd (pPart part, pPartSetIter iter, int* isEnd)
{
   if(iter==part->endPSet())
      *isEnd = 1;
   else
      *isEnd = 0;
   return SCUtil_SUCCESS;
}

int FMDB_PartSetIter_Reset (pPart part, pPartSetIter& iter)
{
   iter = part->beginPSet();
   return SCUtil_SUCCESS;
}

int FMDB_PartSetIter_Del (pPartSetIter iter)
{ 
  return SCUtil_SUCCESS;
}


//*****************************************
//  5.4 Entity Traversal Over Entity Set
//*****************************************

/* Given an entity set and an entity type (0 to 4), get a pointer to the first mesh entity of type.
If no entity exists or type is invalid, null is gotton for Iter.*/

int FMDB_SetEntIter_Init (pEntSet /* in */ entSet, int /* in */ type, int topo, pSetEntIter& /* out */ iter)
{
  iter = new entSetIter;
  if(entSet->isList())
  {
    iter->isList=1;
    iter->iter = new mEntSetUIterator(((mEntitySetUnordered*)entSet)->begin(), ((mEntitySetUnordered*)entSet)->end(),type, topo, (void*)entSet, &processingEntitySetUFilter);
    if(((pEntSetUIterator)iter->iter)->end())
      return SCUtil_FAILURE;
  }
  else
  {
    iter->isList=0;
    iter->iter = new mEntSetOIterator(((mEntitySetOrdered*)entSet)->begin(), ((mEntitySetOrdered*)entSet)->end() ,type, topo, (void*)entSet, &processingEntitySetOFilter );
    if(((pEntSetOIterator)iter->iter)->end())
      return SCUtil_FAILURE;
  }
  return SCUtil_SUCCESS; 
}

/* Gets next entity in the entity set. If there are no more entities, the function returns error (1)*/
int FMDB_SetEntIter_GetNext (pSetEntIter& iter, pMeshEnt& ent)
{
  if(iter->isList)	
  {
    if(((pEntSetUIterator)iter->iter)->end())
      return SCUtil_FAILURE;
    ent = **((pEntSetUIterator)iter->iter);
    ((pEntSetUIterator)iter->iter)->next();
  }
  else
  {
    if(((pEntSetOIterator)iter->iter)->end())
      return SCUtil_FAILURE;
    ent = **((pEntSetOIterator)iter->iter); 
    ((pEntSetOIterator)iter->iter)->next();
  }
  return SCUtil_SUCCESS;
}

int FMDB_SetEntIter_IsEnd(pSetEntIter iter, int* isEnd)
{
  if(iter->isList)
  {
    ((pEntSetUIterator)iter->iter)->valid_check();
    if(((pEntSetUIterator)iter->iter)->end())
      *isEnd = 1;
    else
      *isEnd = 0;
  }
  else
  {  
    if (((pEntSetOIterator)iter->iter)->end())
      *isEnd = 1;
    else
      *isEnd = 0;
  }
  return SCUtil_SUCCESS;
}

/* Resets the iterator*/
int FMDB_SetEntIter_Reset (pSetEntIter iter)
{
  if (iter->isList)
  {
    ((pEntSetUIterator)iter->iter)->reset();
    if (((pEntSetUIterator)iter->iter)->end())
      return SCUtil_FAILURE;
  }
  else
  {
    ((pEntSetOIterator)iter->iter)->reset();
    if (((pEntSetOIterator)iter->iter)->end())
      return SCUtil_FAILURE;
  }
  return SCUtil_SUCCESS;
}

int FMDB_SetEntIter_Del(pSetEntIter iter)
{
  if (iter) delete iter;
  iter=NULL;
  return SCUtil_SUCCESS;
}
