/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "FMDB.h"
 
int FMDB_Part_HasTag (pPart part, pTag tag, int* hasTag)
{
  if (Taggable_HasTag(static_cast<pTaggable>(part), tag))
    *hasTag=1;
  else
    *hasTag=0;
  return SCUtil_SUCCESS;
}

int FMDB_Part_DelTag (pPart part, pTag tag)
{
  Taggable_DelTag(static_cast<pTaggable>(part), tag);
  return SCUtil_SUCCESS;
}

int FMDB_Part_GetTag (pPart part, std::vector<pTag>& tags)
{
  Taggable_GetTag (static_cast<pTaggable>(part), tags);
  return SCUtil_SUCCESS;
}

int FMDB_Part_SetByteTag (pPart part, pTag tag, const void* data, int data_size)
{
#ifdef DEBUG
  int exist=0;
  FMDB_Mesh_HasTag (part->getMesh(), tag, &exist);
  if (!exist)
    return SCUtil_INVALID_TAG_HANDLE;
#endif

  if(!data || !data_size)
    return SCUtil_NULL_ARRAY;

  int tag_size=Tag_GetSize(tag), tag_type = Tag_GetType(tag);
  switch (tag_type)
  {
    case SCUtil_BYTE: Taggable_SetByteArrData<char>(static_cast<pTaggable>(part), tag, data, data_size);
                      break;
    case SCUtil_INT: if (tag_size==1)  // single int
                       Taggable_SetByteData<int>(static_cast<pTaggable>(part), tag, data, data_size);
                     else  // int array
                       Taggable_SetByteArrData<int>(static_cast<pTaggable>(part), tag, data, data_size);
                     break;
    case SCUtil_DBL: if (tag_size==1)  // single double
                        Taggable_SetByteData<double>(static_cast<pTaggable>(part), tag, data, data_size);
                     else  // double array
                        Taggable_SetByteArrData<double>(static_cast<pTaggable>(part), tag, data, data_size);
                     break;
    case SCUtil_ENT: if (tag_size==1)  // single ent
                        Taggable_SetByteData<pMeshEnt>(static_cast<pTaggable>(part), tag, data, data_size);
                     else  // ent array
                        Taggable_SetByteArrData<pMeshEnt>(static_cast<pTaggable>(part), tag, data, data_size);
                     break;
    case SCUtil_SET: if (tag_size==1)  // single set
                        Taggable_SetByteData<pEntSet>(static_cast<pTaggable>(part), tag, data, data_size);
                     else  // set array
                        Taggable_SetByteArrData<pEntSet>(static_cast<pTaggable>(part), tag, data, data_size);
                     break;
    default: break;
  } // switch
  return SCUtil_SUCCESS;
}

int FMDB_Part_GetByteTag (pPart part, pTag tag, void** data, int* data_size)
{
#ifdef DEBUG
  int exist=0;
  FMDB_Mesh_HasTag (part->getMesh(), tag, &exist);
  if (!exist)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  int tag_size = Tag_GetSize(tag);
  int tag_type = Tag_GetType(tag);

  switch (tag_type)
  {
    case SCUtil_BYTE: Taggable_GetByteArrData<char>(static_cast<pTaggable>(part), tag, data, data_size);
                      break;
    case SCUtil_INT: if (tag_size==1)  // int
                       Taggable_GetByteData<int>(static_cast<pTaggable>(part), tag, data, data_size);
                     else // int array
                       Taggable_GetByteArrData<int>(static_cast<pTaggable>(part), tag, data, data_size);
                     break;
    case SCUtil_DBL: if (tag_size==1)  // double
                       Taggable_GetByteData<double>(static_cast<pTaggable>(part), tag, data, data_size);
                     else // double array
                       Taggable_GetByteArrData<double>(static_cast<pTaggable>(part), tag, data, data_size);
                     break;
    case SCUtil_ENT: if (tag_size==1)  // ent
                       Taggable_GetByteData<pMeshEnt>(static_cast<pTaggable>(part), tag, data, data_size);
                     else // ent array
                       Taggable_GetByteArrData<pMeshEnt>(static_cast<pTaggable>(part), tag, data, data_size);
                     break;
    case SCUtil_SET: if (tag_size==1)  // set
                       Taggable_GetByteData<pEntSet>(static_cast<pTaggable>(part), tag, data, data_size);
                     else // set array
                       Taggable_GetByteArrData<pEntSet>(static_cast<pTaggable>(part), tag, data, data_size);
                     break;
    default: break;
  }  // switch
  return SCUtil_SUCCESS;
}

int FMDB_Part_SetIntTag (pPart part, pTag tag, const int data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_INT) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<int>(static_cast<pTaggable>(part), tag, &data);
}

int FMDB_Part_GetIntTag (pPart part, pTag tag, int* data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_INT) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_GetData<int>(static_cast<pTaggable>(part), tag, data);
}

int FMDB_Part_SetDblTag (pPart part, pTag tag, const double data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_DBL) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<double>(static_cast<pTaggable>(part), tag, &data);
}

int FMDB_Part_GetDblTag (pPart part, pTag tag, double* data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_DBL) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_GetData<double>(static_cast<pTaggable>(part), tag, data);
}

int FMDB_Part_SetEntTag (pPart part, pTag tag, const pMeshEnt data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_ENT) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<pMeshEnt>(static_cast<pTaggable>(part), tag, &data);
}

int FMDB_Part_GetEntTag (pPart part, pTag tag, pMeshEnt *data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_ENT) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_GetData<pMeshEnt>(static_cast<pTaggable>(part), tag, data);
}

int FMDB_Part_SetSetTag (pPart part, pTag tag, const pEntSet data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_SET) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<pEntSet>(static_cast<pTaggable>(part), tag, &data);
}

int FMDB_Part_GetSetTag (pPart part, pTag tag, pEntSet *data)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_SET) || Tag_GetSize(tag)!=1)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_GetData<pEntSet>(static_cast<pTaggable>(part), tag, data);
}

// array tag
int FMDB_Part_SetIntArrTag (pPart part, pTag tag, const int* data, int data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_INT) || Tag_GetSize(tag)!=data_size)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<int>(static_cast<pTaggable>(part), tag, data);
}

int FMDB_Part_GetIntArrTag (pPart part, pTag tag, int** data, int* data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_INT))
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  *data_size=Tag_GetSize(tag);
  return Taggable_GetData<int>(static_cast<pTaggable>(part), tag, *data);
}

int FMDB_Part_SetDblArrTag (pPart part, pTag tag, const double* data, int data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_DBL) || Tag_GetSize(tag)!=data_size)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<double>(static_cast<pTaggable>(part), tag, data);
}

int FMDB_Part_GetDblArrTag (pPart part, pTag tag, double** data, int* data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_DBL))
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  *data_size=Tag_GetSize(tag);
  return Taggable_GetData<double>(static_cast<pTaggable>(part), tag, *data);
}

int FMDB_Part_SetEntArrTag (pPart part, pTag tag, const pMeshEnt* data, int data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_ENT) || Tag_GetSize(tag)!=data_size)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<pMeshEnt>(static_cast<pTaggable>(part), tag, data);
}

int FMDB_Part_GetEntArrTag (pPart part, pTag tag, pMeshEnt** data, int* data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_ENT))
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  *data_size=Tag_GetSize(tag);
  return Taggable_GetData<pMeshEnt>(static_cast<pTaggable>(part), tag, *data);
}

int FMDB_Part_SetSetArrTag (pPart part, pTag tag, const pEntSet* data, int data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_SET) || Tag_GetSize(tag)!=data_size)
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  return Taggable_SetData<pEntSet>(static_cast<pTaggable>(part), tag, data);
}

int FMDB_Part_GetSetArrTag (pPart part, pTag tag, pEntSet** data, int* data_size)
{
#ifdef DEBUG
  if (TagHolder_CheckTag(reinterpret_cast<pTagHolder>(part->getMesh()), tag, SCUtil_SET))
    return SCUtil_INVALID_TAG_HANDLE;
#endif
  *data_size=Tag_GetSize(tag);
  return Taggable_GetData<pEntSet>(static_cast<pTaggable>(part), tag, *data);
}
