/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/ 
#include "FMDB.h"
#include "mMesh.h"
#include "mPart.h"
#include <iomanip>

char string_owned_entity[][30] = {"owned vertex", "owned edge", "owned face", "region"};
char string_entity[][30] = {"owned+non-owned vertex", "owned+non-owned edge", "owned+non-owned face"};
char string_LBMethod[][30] = {"RCB", "RIB", "GRAPH", "HYPERGRAPH", "ParMETIS"};
char string_LBApproach[][30] = {"Partition", "Repartition", "Refine", "PartKway", "PartGeom", "PartGeomKWay", "AdaptiveRepart", "RefineKway"};

int FMDB_Mesh_DspNumEnt (pMeshMdl mesh)
{
  SCUTIL_Sync();
  if (!SCUTIL_CommRank()) cout<<"\n[MESH SIZE]\n";
  {
   for (mMesh::partIter pit=mesh->partBegin(); pit!=mesh->partEnd(); ++pit)
     (*pit)->printNumEntities();
  }
  SCUTIL_Sync();
  if (!SCUTIL_CommRank()) cout<<"\n";
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_DspStat (pMeshMdl mesh)
{
#ifdef FMDB_PARALLEL
  // compute imbalance ratio, # pset, #po-ents
  int mesh_dim = ParUtil::Instance()->get_maxMeshDim();
  int numpart = ParUtil::Instance()->getTgtNumParts();
  int arr_dim=9, psetsize;
  // 0-#p-set, 1-#ents in p-set, 2-#owned vertex, 3-#owned edge, 4-#owned face, 5-#region
  // 6-#vertex, 7-#edge, 8-#face
  int *in_arr = new int[arr_dim*numpart]; 
  double *db_arr = new double[2*numpart];   // node_weight, edge weight per part

  if (!SCUTIL_CommRank())
  {
   cout<<"\n[MESH STATISTICS]\n";
   cout<<"#process: "<<SCUTIL_CommSize()<<", #part per process: "<<numpart <<", #total part: "<<SCUTIL_CommSize()*numpart<<endl;
   cout<<"LB_Method: "<<string_LBMethod[mesh->LB_Method]; 
   cout<<", LB_Approach: "<<string_LBApproach[mesh->LB_Approach];
   cout<<setprecision(5)<<", imbalance_tol: "<<(mesh->imbalance_tol-1)*100<<"%\n";
  }

  double weight, node_weight, edge_weight;
  mEntity* ent;
  for (mMesh::partIter pit=mesh->partBegin(); pit!=mesh->partEnd(); ++pit)
  {
    node_weight=0.0, edge_weight = 0.0;
    psetsize=0;

    in_arr[((*pit)->getId()%numpart)*arr_dim] = (*pit)->numPSet();

    for(mPart::psetIter pset_it=(*pit)->beginPSet(); pset_it!=(*pit)->endPSet(); ++pset_it) 
       psetsize+=(*pset_it)->numEntity();
    in_arr[((*pit)->getId()%numpart)*arr_dim+1] = psetsize;

    for (int dim=0; dim<=2; ++dim)
    {
      int numEnt=0;
      for (mPart::iterall ent_it=(*pit)->beginall(dim);ent_it!=(*pit)->endall(dim);++ent_it)
      {
        ent = *ent_it;
        if (ent->getOwnPartID(*pit)==(*pit)->getId())
          ++numEnt;
        if (mesh_dim==2 && dim==2) 
        {
          FMDB_Ent_GetWeight(ent, &weight);
          node_weight+=weight; 
        }
      }
      in_arr[((*pit)->getId()%numpart)*arr_dim+2+dim]=numEnt;      
      in_arr[((*pit)->getId()%numpart)*arr_dim+6+dim]=(*pit)->size(dim);      
    }
    in_arr[((*pit)->getId()%numpart)*arr_dim+5]=(*pit)->size(3);      

    if (mesh_dim==3)
    {
      for (mPart::iterall ent_it=(*pit)->beginall(3);ent_it!=(*pit)->endall(3);++ent_it)
      {
        FMDB_Ent_GetWeight(*ent_it, &weight);
        node_weight+=weight; 
      }
    } // if (mesh_dim==3)
    db_arr[((*pit)->getId()%numpart)*2] = node_weight;
    db_arr[((*pit)->getId()%numpart)*2+1] = edge_weight;
  } // for (mMesh::partIter pit=mesh->partBegin()

  int *out_arr = new int[arr_dim*numpart*SCUTIL_CommSize()];
  double *weight_arr = new double[2*numpart*SCUTIL_CommSize()];

  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::All_Reduce);
  int msg_size = sizeof(int)*arr_dim*numpart+sizeof(double)*2*numpart;

  CM->set_fixed_msg_size(msg_size);
  void* msg_send = CM->alloc_msg(msg_size);
  CM->set_tag(0);

  int* s_int =  (int*)((char*)msg_send);
  for (int i=0; i<arr_dim*numpart; ++i)
    s_int[i] = in_arr[i];

  double* s_double = (double*)((char*)msg_send+sizeof(int)*arr_dim*numpart);
  for (int i=0; i<2*numpart; ++i)
    s_double[i] = db_arr[i];

  CM->send(0, (void*)msg_send);
  CM->finalize_send();
  CM->free_msg(msg_send);

  void* msg_recv;
  int pid_from;
  // receive phase
  while(CM->receive(msg_recv, &pid_from))
  {
    s_int = (int*)((char*)msg_recv);
    for (int i=0; i<arr_dim*numpart; ++i)
      out_arr[pid_from*arr_dim*numpart+i]=s_int[i];    

    s_double = (double*)((char*)msg_recv+sizeof(int)*arr_dim*numpart);
    for (int i=0; i<2*numpart; ++i)
      weight_arr[pid_from*2*numpart+i]=s_double[i];    

    CM->free_msg(msg_recv);
  } //  while (CM->receive(msg_recv, &pid_from))
 
  // compute avg & max #pset and #poent  
  if (!SCUTIL_CommRank())
  {
    // variables related to #ents and #p-set
    int sum_num_pset=0, sum_size_pset=0, part_max_num_pset=0, part_max_num_poent=0, part_min_num_pset=in_arr[0], part_min_num_poent=in_arr[mesh_dim+2];
    int proc_max_num_pset=0, proc_max_num_poent=0;
    int local_sum_num_pset[SCUTIL_CommSize()], local_sum_size_pset[SCUTIL_CommSize()], local_max_num_pset[SCUTIL_CommSize()], 
        local_max_num_poent[SCUTIL_CommSize()], local_min_num_pset[SCUTIL_CommSize()], local_min_num_poent[SCUTIL_CommSize()],
        sum_num_owned_ent[4], local_sum_num_owned_ent[4][SCUTIL_CommSize()], sum_num_ent[4], local_sum_num_ent[4][SCUTIL_CommSize()];
    double local_avg_num_poent[SCUTIL_CommSize()], local_avg_num_pset[SCUTIL_CommSize()], local_avg_size_pset[SCUTIL_CommSize()];

    // variables related to weights
    double proc_sum_weight=0.0, proc_max_weight=0.0, part_max_weight=0.0, part_min_weight=db_arr[0];
    double local_sum_weight[SCUTIL_CommSize()], local_max_weight[SCUTIL_CommSize()], 
           local_min_weight[SCUTIL_CommSize()], local_avg_weight[SCUTIL_CommSize()];

    for (int i=0; i<4; ++i)
    {
      sum_num_owned_ent[i]=0;
      sum_num_ent[i]=0;
      for (int j=0; j<SCUTIL_CommSize(); ++j)
      {
        local_sum_num_owned_ent[i][j]=0;
        local_sum_num_ent[i][j]=0;
      }
    } //for (int i=0; i<4; ++i)

    for (int i=0; i<SCUTIL_CommSize(); ++i) // for each proc
    {     
      local_sum_weight[i] = 0.0;
      local_max_weight[i] = 0.0;
      local_min_weight[i] = weight_arr[i*numpart*2];
      local_avg_weight[i] = 0.0;

      local_sum_num_pset[i]=0;
      local_sum_size_pset[i]=0;
      local_max_num_pset[i]=0;
      local_max_num_poent[i]=0;
      local_min_num_pset[i]=out_arr[i*numpart*arr_dim];
      local_min_num_poent[i]=out_arr[i*numpart*arr_dim+mesh_dim+2];
      local_avg_num_poent[i]=0.0;
      local_avg_num_pset[i]=0.0;

      for (int j=0; j<numpart; ++j) // for each part on proc
      {
        // local info - weight
        local_sum_weight[i] += weight_arr[(i*numpart+j)*2];
        if (local_max_weight[i]<weight_arr[(i*numpart+j)*2]) local_max_weight[i]=weight_arr[(i*numpart+j)*2];
        if (local_min_weight[i]>weight_arr[(i*numpart+j)*2]) local_min_weight[i]=weight_arr[(i*numpart+j)*2];
        if (part_max_weight<weight_arr[(i*numpart+j)*2]) part_max_weight=weight_arr[(i*numpart+j)*2];
        if (part_min_weight>weight_arr[(i*numpart+j)*2]) part_min_weight=weight_arr[(i*numpart+j)*2];

        // local info - weight
        // global info - #p-set & #ents
        sum_num_pset+=out_arr[(i*numpart+j)*arr_dim];
        sum_size_pset+=out_arr[(i*numpart+j)*arr_dim+1];
        for (int k=0; k<4; ++k)
          sum_num_owned_ent[k]+=out_arr[(i*numpart+j)*arr_dim+k+2];
        for (int k=0; k<3; ++k)
          sum_num_ent[k]+=out_arr[(i*numpart+j)*arr_dim+k+6];

        if (part_max_num_pset<out_arr[(i*numpart+j)*arr_dim]) part_max_num_pset = out_arr[(i*numpart+j)*arr_dim];
        if (part_min_num_pset>out_arr[(i*numpart+j)*arr_dim]) part_min_num_pset = out_arr[(i*numpart+j)*arr_dim];
        if (part_max_num_poent<out_arr[(i*numpart+j)*arr_dim+mesh_dim+2]) part_max_num_poent = out_arr[(i*numpart+j)*arr_dim+mesh_dim+2];
        if (part_min_num_poent>out_arr[(i*numpart+j)*arr_dim+mesh_dim+2]) part_min_num_poent = out_arr[(i*numpart+j)*arr_dim+mesh_dim+2];
        // local info - #p-set & #ents
        local_sum_num_pset[i]+=out_arr[(i*numpart+j)*arr_dim];
        local_sum_size_pset[i]+=out_arr[(i*numpart+j)*arr_dim+1];
        for (int k=0; k<4; ++k)
          local_sum_num_owned_ent[k][i]+=out_arr[(i*numpart+j)*arr_dim+k+2];
        for (int k=0; k<3; ++k)
          local_sum_num_ent[k][i]+=out_arr[(i*numpart+j)*arr_dim+k+6];
        if (local_max_num_pset[i]<out_arr[(i*numpart+j)*arr_dim]) local_max_num_pset[i] = out_arr[(i*numpart+j)*arr_dim];
        if (local_min_num_pset[i]>out_arr[(i*numpart+j)*arr_dim]) local_min_num_pset[i] = out_arr[(i*numpart+j)*arr_dim];
        if (local_max_num_poent[i]<out_arr[(i*numpart+j)*arr_dim+mesh_dim+2]) local_max_num_poent[i] = out_arr[(i*numpart+j)*arr_dim+mesh_dim+2];
        if (local_min_num_poent[i]>out_arr[(i*numpart+j)*arr_dim+mesh_dim+2]) local_min_num_poent[i] = out_arr[(i*numpart+j)*arr_dim+mesh_dim+2];
      }   // for (int j=0; j<numpart; ++j)

      local_avg_num_poent[i]=((double)(local_sum_num_owned_ent[mesh_dim][i]))/((double)(numpart));
      local_avg_num_pset[i] = ((double)(local_sum_num_pset[i]))/((double)(numpart));
      if (local_sum_num_pset[i]) 
        local_avg_size_pset[i] = ((double)(local_sum_size_pset[i]))/((double)(local_sum_num_pset[i]));

      // global-weight
      local_avg_weight[i] =  local_sum_weight[i]/((double)(numpart));
      proc_sum_weight += local_sum_weight[i];
      if (proc_max_weight<local_sum_weight[i]) 
        proc_max_weight = local_sum_weight[i];

    } // for (int i=0; i<SCUTIL_CommSize(); ++i)
    int proc_min_num_pset=local_sum_num_pset[0], proc_min_num_poent=local_sum_num_owned_ent[mesh_dim][0];
    double proc_min_weight = local_sum_weight[0];
    for (int i=0; i<SCUTIL_CommSize(); ++i)
    {
      if (proc_max_num_pset<local_sum_num_pset[i]) proc_max_num_pset = local_sum_num_pset[i];
      if (proc_min_num_pset>local_sum_num_pset[i]) proc_min_num_pset = local_sum_num_pset[i];
      if (proc_max_num_poent<local_sum_num_owned_ent[mesh_dim][i]) proc_max_num_poent = local_sum_num_owned_ent[mesh_dim][i];
      if (proc_min_num_poent>local_sum_num_owned_ent[mesh_dim][i]) proc_min_num_poent = local_sum_num_owned_ent[mesh_dim][i];
      // global - min weight
      if (proc_min_weight>local_sum_weight[i]) proc_min_weight=local_sum_weight[i];
    }

    // print process count statistics     
    cout<<"\n-- Per-Process Count Statistics --\n";
    cout<<"Global: \n";
    double avg_num_pset = ((double)(sum_num_pset))/((double)SCUTIL_CommSize());
    double avg_size_pset = 0.0;
    if (sum_size_pset)
      avg_size_pset = ((double)(sum_size_pset))/((double)sum_num_pset);
    cout<<"\t#p-set: avg "<<avg_num_pset<<", max "<<proc_max_num_pset<<", min "<< proc_min_num_pset<<endl;
    cout<<"\tavg p-set size (#ent): "<<avg_size_pset<<"\n\n";

    double avg_num_poent = ((double)(sum_num_owned_ent[mesh_dim]))/((double)SCUTIL_CommSize());
    double proc_avg_weight = proc_sum_weight/((double)SCUTIL_CommSize());
    cout<<setprecision(4)<<"\t#po-ent & weight: avg "<<avg_num_poent<<"-"<<proc_avg_weight
        <<", max "<<proc_max_num_poent<<"-"<<proc_max_weight
        <<", min "<<proc_min_num_poent<<"-"<<proc_min_weight<<endl;

    cout<<setprecision(4)<<"\tImbalance ratio [max/avg weight]: "<<((proc_max_weight/proc_avg_weight)-1)*100<<" %\n";
    if (proc_min_weight) 
      cout<<setprecision(4)<<"\tImbalance ratio [avg/min weight]: "<<((proc_avg_weight/proc_min_weight)-1)*100<<" %\n\n";
    else
      cout<<"\tImbalance ratio [avg/min weight]: N/A - min weight is zero\n\n";

    cout<<setprecision(4)<<"\tImbalance ratio [max/avg #po-ent]: "<<((((double)(proc_max_num_poent))/avg_num_poent)-1)*100<<" %\n";
    if (proc_min_num_poent) 
      cout<<setprecision(4)<<"\tImbalance ratio [avg/min #po-ent]: "<<((avg_num_poent/((double)(proc_min_num_poent)))-1)*100<<" %\n\n";
    else
      cout<<"\tImbalance ratio [avg/min #po-ent]: N/A - min #po-ent is zero\n\n";
    if (sum_num_pset)
    {
      cout<<"#p-set & size: "<<sum_num_pset<<"-"<<sum_size_pset<<"\n\t";
      for (int i=0; i<SCUTIL_CommSize(); ++i)
      { 
        cout<<"(p"<<i<<": "<<local_sum_num_pset[i]<<"-"<< local_sum_size_pset[i]<<") ";
        if (i!=0&&i%5==4)
          cout<<"\n\t";
      }
      cout<<"\n\n";
    } // if (sum_num_pset)

    // print # owned+non-owned entities
    for (int dim=0; dim<mesh_dim; ++dim)
    {
      cout<<"#"<<string_entity[dim]<<": "<<sum_num_ent[dim]<<"\n\t"; 
      for (int i=0; i<SCUTIL_CommSize(); ++i)
      { 
        cout<<"(p"<<i<<": "<<local_sum_num_ent[dim][i]<<") ";
        if (i!=0&&i%8==7)
          cout<<"\n\t";
      }
      cout<<endl;
    } // dim
    cout<<endl;
    // print # owned entities
    for (int dim=0; dim<mesh_dim+1; ++dim)
    {
      if (mesh_dim==2 && dim==2)
         cout<<"#face & weight: "<<sum_num_owned_ent[dim]<<"-"<<proc_sum_weight<<"\n\t"; 
      else if (mesh_dim==3 && dim==3)
         cout<<"#"<<string_owned_entity[dim]<<" & weight: "<<sum_num_owned_ent[dim]
             <<"-"<<proc_sum_weight<<"\n\t";
      else
         cout<<"#"<<string_owned_entity[dim]<<": "<<sum_num_owned_ent[dim]<<"\n\t"; 
      for (int i=0; i<SCUTIL_CommSize(); ++i)
      { 
        if (dim==mesh_dim)
          cout<<"(p"<<i<<": "<<local_sum_num_owned_ent[dim][i]<<"-"<<local_sum_weight[i]<<") ";
        else
          cout<<"(p"<<i<<": "<<local_sum_num_owned_ent[dim][i]<<") ";
        if (i!=0&&i%8==7)
          cout<<"\n\t";
      }
      cout<<endl;
    } // dim

    // print part count statistics
    if (numpart>1)
    {       
      cout<<"\n-- Per-Part Count Statistics --\n";
      cout<<"Global: \n";
      double avg_num_pset = ((double)(sum_num_pset))/((double)(SCUTIL_CommSize()*numpart));
      double avg_size_pset = 0.0;
      if (sum_size_pset) 
        avg_size_pset = ((double)(sum_size_pset))/((double)sum_num_pset);
      cout<<"\t#p-set: avg "<<avg_num_pset<<", max "<<part_max_num_pset<<", min "<< part_min_num_pset<<endl;
      cout<<"\tavg p-set size (#ent): "<<avg_size_pset<<"\n\n";

      double avg_num_poent = ((double)(sum_num_owned_ent[mesh_dim]))/((double)(SCUTIL_CommSize()*numpart));
      double part_avg_weight = proc_sum_weight/((double)(SCUTIL_CommSize()*numpart));
      cout<<setprecision(4)<<"\t#po-ent & weight: avg "<<avg_num_poent<<"-"<<part_avg_weight
          <<", max "<<part_max_num_poent<<"-"<<part_max_weight
          <<", min "<<part_min_num_poent<<"-"<<part_min_weight<<endl;

      cout<<setprecision(4)<<"\tImbalance ratio [max/avg weight]: "<<((part_max_weight/part_avg_weight)-1)*100<<" %\n";
      if (part_min_weight)
        cout<<setprecision(4)<<"\tImbalance ratio [avg/min weight]: "<<((part_avg_weight/part_min_weight)-1)*100<<" %\n\n";
      else
        cout<<"\tImbalance ratio [avg/min weight]: N/A - min weight is zero\n\n";

      cout<<setprecision(4)<<"\tImbalance ratio [max/avg #po-ent]: "<<((((double)(part_max_num_poent))/avg_num_poent)-1)*100<<" %\n";
      if (part_min_num_poent)
        cout<<setprecision(4)<<"\tImbalance ratio [avg/min #po-ent]: "<<((avg_num_poent/((double)(part_min_num_poent)))-1)*100<<" %\n\n";
      else
        cout<<"\tImbalance ratio [avg/min #po-ent]: N/A - min #po-ent is zero\n\n";

      if (sum_num_pset) 
      {
        cout<<"Local (per process) avg/max/min #p-set and avg size:\n\t";
        for (int i=0; i<SCUTIL_CommSize();++i)
        {
          cout<<"(p"<<i<<": avg "<<local_avg_num_pset[i]<<", max "<<local_max_num_pset[i]
              <<", min "<<local_min_num_pset[i]<<", size "<<local_avg_size_pset[i]<<") ";
          if (i!=0&&i%2==1)
            cout<<"\n\t";
        }
      cout<<"\n\n";
      } // if (sum_num_pset) 

      cout<<"Local (per process) #po-ent & weight:\n\t";
      for (int i=0; i<SCUTIL_CommSize();++i)
      {  
        cout<<setprecision(4)<<"(p"<<i<<": avg "<<local_avg_num_poent[i]<<"-"<<local_avg_weight[i]
            <<", max "<<local_max_num_poent[i]<<"-"<<local_max_weight[i]
            <<", min "<<local_min_num_poent[i]<<"-"<<local_min_weight[i]<<") ";
        if (i!=0&&i%2==1)
          cout<<"\n\t";
      } // for int i=0
      cout<<"\n\n";

      // print local imbalance ratio per weight
      cout<<"Local (per process) imbalance ratio [max/avg weight]:\n\t";
      for (int i=0; i<SCUTIL_CommSize();++i)
      {
        cout<<"(p"<<i<<": "<<((local_max_weight[i]/local_avg_weight[i])-1)*100<<"%) ";
        if (i!=0&&i%8==7)
          cout<<"\n\t";
      }
      cout<<"\n\n";

      cout<<"Local (per process) imbalance ratio [avg/min weight]: \n\t";
      for (int i=0; i<SCUTIL_CommSize();++i)
      {
        if (local_min_weight[i])
          cout<<"(p"<<i<<": "<<((local_avg_weight[i]/local_min_weight[i])-1)*100<<"%) ";
        else
          cout<<"(p"<<i<<": N/A) ";
        if (i!=0&&i%8==7)
          cout<<"\n\t";
      }
      cout<<"\n\n";

      // print local imbalance ratio per # po-ent
      cout<<"Local (per process) imbalance ratio [max/avg #po-ent]:\n\t";
      for (int i=0; i<SCUTIL_CommSize();++i)
      {
        cout<<"(p"<<i<<": "<<((((double)(local_max_num_poent[i]))/local_avg_num_poent[i])-1)*100<<"%) ";
        if (i!=0&&i%8==7)
          cout<<"\n\t";
      }
      cout<<"\n\n";

      cout<<"Local (per process) imbalance ratio [avg/min #po-ent]: \n\t";
      for (int i=0; i<SCUTIL_CommSize();++i)
      {
        if (local_min_num_poent[i])
          cout<<"(p"<<i<<": "<<((local_avg_num_poent[i]/((double)(local_min_num_poent[i])))-1)*100<<"%) ";
        else
          cout<<"(p"<<i<<": N/A) ";
        if (i!=0&&i%8==7)
          cout<<"\n\t";
      }
      cout<<"\n\n";

      if (sum_num_pset)
      {
        cout<<"#p-set & size: "<<sum_num_pset<<"-"<<sum_size_pset<<"\n\t";
        for (int i=0; i<SCUTIL_CommSize(); ++i)
        { 
          cout<<"(p"<<i<<": ";
          for (int j=0; j<numpart;++j)
          {
            if (out_arr[(i*numpart+j)*arr_dim])
              cout<<out_arr[(i*numpart+j)*arr_dim]<<"-"<<out_arr[(i*numpart+j)*arr_dim+1];
            else 
              cout<<"0-0";
            if (j<numpart-1) cout<<" ";
          }
          cout<<") ";
          if (i!=0&&i%2==1)
            cout<<"\n\t";
        } // i
      cout<<"\n\n";
      }//    if (sum_num_pset)

      // print # (owned+non-owned) entities
      for (int dim=0; dim<mesh_dim; ++dim)
      {
        cout<<"#"<<string_entity[dim]<<": "<<sum_num_ent[dim]<<"\n\t"; 
        for (int i=0; i<SCUTIL_CommSize(); ++i)
        { 
          cout<<"(p"<<i<<": ";
          for (int j=0; j<numpart;++j)
          {
            cout<<out_arr[(i*numpart+j)*arr_dim+dim+6];   
            if (j<numpart-1) cout<<" ";
          }
          cout<<") ";
          if (i!=0&&i%2==1)
            cout<<"\n\t";
        } // i
        cout<<"\n";
      } // dim
      cout<<endl;
      // print # owned entities
      for (int dim=0; dim<mesh_dim+1; ++dim)
      {
        if (mesh_dim==2 && dim==2)
          cout<<"#face & weight: "<<sum_num_owned_ent[dim]<<"-"<<proc_sum_weight<<"\n\t";
        else if (mesh_dim==3 && dim==3)
          cout<<"#"<<string_owned_entity[dim]<<" & weight: "<<sum_num_owned_ent[dim]
              <<"-"<<proc_sum_weight<<"\n\t";
        else
           cout<<"#"<<string_owned_entity[dim]<<": "<<sum_num_owned_ent[dim]<<"\n\t"; 
        for (int i=0; i<SCUTIL_CommSize(); ++i)
        { 
          cout<<"(p"<<i<<": ";
          for (int j=0; j<numpart;++j)
          {
            if (dim==mesh_dim) cout<<out_arr[(i*numpart+j)*arr_dim+dim+2]<<"-"<<weight_arr[(i*numpart+j)*2];   
            else
              cout<<out_arr[(i*numpart+j)*arr_dim+dim+2];
            if (j<numpart-1) cout<<" ";
          }
          cout<<") ";
          if (i!=0&&i%2==1)
            cout<<"\n\t";
        } // i
        cout<<"\n";
      } // dim
    } // if (numpart>1)
  }  //if (!SCUTIL_CommRank())
  delete [] in_arr;
  delete [] out_arr;
  delete [] db_arr;
  delete [] weight_arr;
  SCUTIL_Sync();
  if (!SCUTIL_CommRank()) cout<<"\n";
  return SCUtil_SUCCESS;
#else
  cout<<__func__<<" not implemented yet\n";
  return SCUtil_FAILURE;
#endif
}

int FMDB_Mesh_DspAll (pMeshMdl mesh)
{
  ParUtil::Instance()->Barrier(__LINE__,__FILE__);
  if (!SCUTIL_CommRank())
    cout<<"\n***  FMDB_Mesh_DspInfo( #parts/proc: "<<mesh->getNumPart()<<")  ***\n";
  for (mMesh::partIter pit=mesh->partBegin(); pit!=mesh->partEnd(); ++pit)
    (*pit)->printAll();
  ParUtil::Instance()->Barrier(__LINE__,__FILE__);
  return SCUtil_SUCCESS;
}

// entity set
int FMDB_Set_DspInfo (pEntSet ent)
{
  cout<<__func__<<" not implemented yet\n";
  return SCUtil_FAILURE;
}

// entity
void printVertex (pPart part, pMeshEnt vtx);
void printNonVertex (pPart part, pMeshEnt ent);

int FMDB_Ent_DspInfo (pPart part, pMeshEnt ent)
{
  if (!ent->getLevel()) // vertex
    printVertex (part, ent);
  else  // non-vertex
    printNonVertex (part, ent);

  return SCUtil_SUCCESS;
}

int FMDB_Ent_DspAllAdj (pMeshEnt ent)
{
  cout<<__func__<<" not implemented yet\n";
  return SCUtil_FAILURE;
}

// ENTITY PRINTING HELPER FUNCTIONS //
void printVertex (pPart part, pMeshEnt vtx)
{
  double xyz[3] = {0.0, 0.0, 0.0};
  ((mVertex*)vtx)->getCoord(xyz[0],xyz[1],xyz[2]);

#ifndef FMDB_PARALLEL      
      cout<<"en_info: "<<vtx->getUid()
          <<"["<<xyz[0]<<","<<xyz[1]<<","<<xyz[2]<<"], "
	  <<"(iD "<<vtx->getId()
 	  <<", mem "<<vtx
	  <<", gLevel "<<GEN_type(vtx->getClassification())
  	  <<", gTag "<<GEN_tag(vtx->getClassification())
	  <<")\n";
#else
  if (vtx->getPClassification())
  { 
    cout<<"PartBdry -- [part "<<part->getId()<<"|proc "<<ParUtil::Instance()->rank()
          <<"] en_info: "<<vtx->getUid()
          <<"["<<xyz[0]<<","<<xyz[1]<<","<<xyz[2]<<"], "
          <<"(iD "<<vtx->getId() 
   	  <<", mem "<<vtx
	  <<", gLevel "<<GEN_type(vtx->getClassification())
  	  <<", gTag "<<GEN_tag(vtx->getClassification())
	  <<", peId "<<vtx->getPClassification()->getId()
	  <<", numRCs="<<vtx->getNumRemoteCopies()
	  <<")\n ";
    }
    else
    {
      cout<<"Int -- [part "<<part->getId()<<"|proc "<<ParUtil::Instance()->rank()
          <<"] en_info: "<<vtx->getUid()
          <<"["<<xyz[0]<<","<<xyz[1]<<","<<xyz[2]<<"], "
          <<"(iD "<<vtx->getId()
	  <<", mem "<<vtx
	  <<", gLevel "<<GEN_type(vtx->getClassification())
  	  <<", gTag "<<GEN_tag(vtx->getClassification())
          <<")\n";
    }
#endif  // PARALLEL
}

void printNonVertex(pPart part, pMeshEnt ent) 
{
#ifndef FMDB_PARALLEL
      cout<<"en_info: "<<ent->getUid()
	  <<" (mem "<<ent
	  <<", gLevel "<<GEN_type(ent->getClassification())
  	  <<", gTag "<<GEN_tag(ent->getClassification())
          <<")\n";
#else
    if (ent->getPClassification())
    { 
      cout<<"PartBdry -- [part "<<part->getId()<<"|proc "<<ParUtil::Instance()->rank()
          <<"] en_info: "<<ent->getUid()
	  <<" (mem "<<ent
	  <<", gLevel "<<GEN_type(ent->getClassification())
  	  <<", gTag "<<GEN_tag(ent->getClassification())
 	  <<", peId "<<ent->getPClassification()->getId()
          <<", numRCs="<<ent->getNumRemoteCopies()
          <<")\n";
    }
    else
    {
      cout<<"Int -- [part "<<part->getId()<<"|proc "<<ParUtil::Instance()->rank()
          <<"] en_info: "<<ent->getUid()
	  <<" (mem "<<ent
	  <<", gLevel "<<GEN_type(ent->getClassification())
  	  <<", gTag "<<GEN_tag(ent->getClassification())
          <<")\n";
    }
#endif
}
