/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "FMDB.h"
#include "oldFMDB.h"
#include "mTet.h"
#include "mHex.h"
#include "mPyramid.h"
#include "mPrism.h"

#ifdef CHECK_RGN
#ifdef DEBUG
int check_rgn (pPart part, pRegion r, int topo, int nFace, pFace* faces)
{
  std::vector<mEntity*> vertices;
  switch (topo)
  {
    case FMDB_TET:    {
		        vertices.push_back(faces[0]->get(FMDB_VERTEX, 0));
		        if (R_faceDir(r, 0))
		        {
		          vertices.push_back(faces[0]->get(FMDB_VERTEX, 2));
		          vertices.push_back(faces[0]->get(FMDB_VERTEX, 1));
                        }
                        else
		        {
                          vertices.push_back(faces[0]->get(FMDB_VERTEX, 1));
		          vertices.push_back(faces[0]->get(FMDB_VERTEX, 2));
                        }
		        for (int i=0; i<faces[1]->size(FMDB_VERTEX); ++i)
		          if (std::find(vertices.begin(), vertices.end(), faces[1]->get(FMDB_VERTEX, i)) == vertices.end())
		          {
		            vertices.push_back(faces[1]->get(FMDB_VERTEX, i));
		            break;
		          }
                        break;
		      } // tet
    case FMDB_PYRAMID:{
		        if (R_faceDir(r, 0))
		        {
		          vertices.push_back(faces[0]->get(FMDB_VERTEX, 0));
		          vertices.push_back(faces[0]->get(FMDB_VERTEX, 3));
		          vertices.push_back(faces[0]->get(FMDB_VERTEX, 2));
		          vertices.push_back(faces[0]->get(FMDB_VERTEX, 1));
		        }
		        else
		          for (int i=0; i<faces[0]->size(FMDB_VERTEX); ++i)
		            vertices.push_back(faces[0]->get(FMDB_VERTEX, i));
		        for (int i=0; i<faces[1]->size(FMDB_VERTEX); ++i)
		          if (std::find(vertices.begin(), vertices.end(), faces[1]->get(FMDB_VERTEX, i)) == vertices.end())
		          {
		            vertices.push_back(faces[1]->get(FMDB_VERTEX, i));
		            break;
		          }
		        break;
		      } // pyramid
    case FMDB_PRISM:  {
		        vertices.push_back(faces[0]->get(FMDB_VERTEX, 0));
		        if (R_faceDir(r, 0))
		        {
		          vertices.push_back(faces[0]->get(FMDB_VERTEX, 2));
		          vertices.push_back(faces[0]->get(FMDB_VERTEX, 1));
                        }
                        else
		        {
                          vertices.push_back(faces[0]->get(FMDB_VERTEX, 1));
		          vertices.push_back(faces[0]->get(FMDB_VERTEX, 2));
                        }
                        for (int i=0; i<faces[4]->size(FMDB_VERTEX); ++i)
		          vertices.push_back(faces[4]->get(FMDB_VERTEX, i));
			break;
		      } // prism
    case FMDB_HEX:    return SCUtil_SUCCESS; // pass temporarily
                      {
                        for (int i=0; i<faces[0]->size(FMDB_VERTEX); ++i)
		          vertices.push_back(faces[0]->get(FMDB_VERTEX, i));
                        for (int i=0; i<faces[5]->size(FMDB_VERTEX); ++i)
		          vertices.push_back(faces[5]->get(FMDB_VERTEX, i));
			break;
                      } // hex
    default: break;
  } // switch (topo)

  // printing all
/*
  cout<<"[part "<<part->getId()<<"] "<<__func__<<" "<<r->getUid() << "(topo="<<topo<<", face dir=";
  for (int j=0; j<nFace; ++j)
    cout<<R_faceDir(r, j)<<" ";
  cout<<")\n";
  for (int i=0; i<r->size(FMDB_VERTEX); ++i)
  {
    if (r->get(FMDB_VERTEX, i)!=vertices.at(i))
      cout<<"\t"<<i<<"'th vertex order mismatch - face-based: "<<vertices.at(i)->getUid()
          <<", vertex-based: "<<r->get(FMDB_VERTEX, i)->getUid()<<endl;
    else
      cout<<"\t"<<i<<"'th vertex order match\n";
  }
*/
  // printing errors only
  for (int i=0; i<r->size(FMDB_VERTEX); ++i)
  {
    if (r->get(FMDB_VERTEX, i)!=vertices.at(i))
    {
      cout<<"[part "<<part->getId()<<"] "<<__func__<<" "<<r->getUid() << "(topo="<<topo<<", face dir=";
      for (int j=0; j<nFace; ++j)
        cout<<R_faceDir(r, j)<<" ";
      cout<<")\n";
      cout<<"\t"<<i<<"'th vertex order mismatch - face-based: "<<vertices.at(i)->getUid()
          <<", vertex-based: "<<r->get(FMDB_VERTEX, i)->getUid()<<endl;
      return SCUtil_FAILURE;
    }
  }
  return SCUtil_SUCCESS;
}
#endif
#endif

pRegion create_rgn_with_faces(pPart part, int nFace, pFace *faces, int *dirs, pGEntity gent)
{
  pRegion r=NULL;
#ifdef DEBUG
  int topo;
#endif
  switch (nFace)
  {
    case 4: r = (pRegion)(part->createTetWithFaces ((mFace*)faces[0],(mFace*)faces[1],
				(mFace*)faces[2],(mFace*)faces[3],gent));
#ifdef DEBUG
            topo = FMDB_TET;
#endif
            break;
    case 6: r = (pRegion)(part->createHexWithFaces ((mFace*)faces[0],(mFace*)faces[1],
				(mFace*)faces[2],(mFace*)faces[3],(mFace*)faces[4],(mFace*)faces[5], gent));
#ifdef DEBUG
            topo = FMDB_HEX;
#endif
	    break;
    case 5: if (faces[0]->size(1) == 3) 
            {
              r = (pRegion)(part->createPrismWithFaces ((mFace*)faces[0],(mFace*)faces[1],
				(mFace*)faces[2],(mFace*)faces[3],(mFace*)faces[4],gent)); 
#ifdef DEBUG
              topo = FMDB_PRISM;
#endif
            }
            else // (faces[0]->size(1) == 4) 
            {
              r = (pRegion)(part->createPyramidWithFaces ((mFace*)faces[0],(mFace*)faces[1],
				(mFace*)faces[2],(mFace*)faces[3],(mFace*)faces[4],gent));
#ifdef DEBUG
              topo = FMDB_PYRAMID;
#endif
            }
            break;
    default: std::cerr<<__func__<<" ERROR: wrong # faces - "<<nFace<<endl;
             break;
  } // switch (nFace)

#ifdef CHECK_RGN
#ifdef DEBUG
  check_rgn(part, r, topo, nFace, faces);
#endif
#endif

  if (r)
    for (int i=0; i<nFace; ++i)
      faces[i]->add(r);
  return r;
}

pRegion create_rgn_with_vertices(pPart part, int nVertex, pVertex* vertices, int *dirs, pGEntity gent)
{
  mEntity* tmpEdges[4];
  switch (nVertex)
  {
    case 4: 
    {
      mVertex* v1 = (mVertex*)vertices[0];
      mVertex* v2 = (mVertex*)vertices[1];
      mVertex* v3 = (mVertex*)vertices[2];
      mVertex* v4 = (mVertex*)vertices[3];
      mEntity* edges[6];
      mEntity* faces[4];

      edges[0] = (mEntity*)E_exist(v1, v2);
       if (!edges[0])
            edges[0]=M_createE(part,v1,v2, gent);

      edges[1] = (mEntity*)E_exist(v2, v3);
      if (!edges[1])
            edges[1]=M_createE(part,v2,v3,gent);

      edges[2] = (mEntity*)E_exist(v3, v1);
      if (!edges[2])
         edges[2]=M_createE(part,v3,v1,gent);

      edges[3]=(mEntity*)E_exist(v1, v4);
      if (!edges[3])
       {
        edges[3]=M_createE(part, v1,v4,gent);
       }

      edges[4]=(mEntity*)E_exist(v2,v4);
      if (!edges[4])
        {
         edges[4]=M_createE(part,v2,v4,gent);
        }

      edges[5]=(mEntity*)E_exist(v3,v4);
      if (!edges[5])
        {
        edges[5]=M_createE(part, v3, v4,gent);
        }

      faces[0]=(mEntity*)F_exist(1, edges[0], edges[1], edges[2], 0);
      if(!faces[0])
        {
        faces[0]= M_createF(part, 3, edges, 0, gent);
        }
      faces[1]=(mEntity*)F_exist(1, edges[0], edges[4], edges[3], 0);
      if(!faces[1])
          {
            tmpEdges[0]=edges[0];
            tmpEdges[1]=edges[4];
            tmpEdges[2]=edges[3];
            faces[1]= M_createF(part, 3, tmpEdges, 0, gent);
          }

      faces[2]=(mEntity*)F_exist(1, edges[1], edges[5], edges[4], 0);
          if(!faces[2])
          {
            tmpEdges[0]=edges[1];
            tmpEdges[1]=edges[5];
            tmpEdges[2]=edges[4];
            faces[2]= M_createF(part, 3, tmpEdges, 0, gent);
          }
      faces[3]=(mEntity*)F_exist(1, edges[2], edges[5], edges[3], 0);
          if(!faces[3])
          {
            tmpEdges[0]=edges[2];
            tmpEdges[1]=edges[5];
            tmpEdges[2]=edges[3];
            faces[3]= M_createF(part, 3, tmpEdges, 0, gent);
          }

      return create_rgn_with_faces(part, 4, faces, 0, gent);
    } // case 4
    case 5: //pyramid
    {
      mVertex* v1 = (mVertex*)vertices[0];
      mVertex* v2 = (mVertex*)vertices[1];
      mVertex* v3 = (mVertex*)vertices[2];
      mVertex* v4 = (mVertex*)vertices[3];
      mVertex* v5 = (mVertex*)vertices[4];
      mEntity* faces[5];
      mEntity* edges[8];
      edges[0]=(mEntity*)E_exist(v1, v2);
          if (!edges[0])
           {
            edges[0]=M_createE(part,v1, v2, gent);
           }

          edges[1]=(mEntity*)E_exist(v2, v3);
          if (!edges[1])
           {
            edges[1]=M_createE(part,v2,v3, gent);
           }

          edges[2]=(mEntity*)E_exist(v3, v4);
          if (!edges[2])
            {
            edges[2]=M_createE(part, v3, v4, gent);
            }

          edges[3]=(mEntity*)E_exist(v4, v1);
          if (!edges[3])
           {
            edges[3]=M_createE(part, v4, v1, gent);
           }

          edges[4]=(mEntity*)E_exist(v1, v5);
          if (!edges[4])
           {
            edges[4]=M_createE(part, v1, v5, gent);
           }

          edges[5]=(mEntity*)E_exist(v2, v5);
          if (!edges[5])
           {
            edges[5]=M_createE(part, v2, v5, gent);
           }

          edges[6]=(mEntity*)E_exist(v3, v5);
          if (!edges[6])
           {
            edges[6]=M_createE(part, v3, v5, gent);
           }

          edges[7]=(mEntity*)E_exist(v4, v5);
          if (!edges[7])
           {
            edges[7]=M_createE(part, v4,v5,gent);
           }

         faces[0]=(mEntity*)F_exist(1, edges[0], edges[1], edges[2], edges[3]);
          if(!faces[0])
            {
            faces[0]= M_createF(part, 4, edges, 0, gent);
            }
         faces[1]=(mEntity*)F_exist(1, edges[0], edges[5], edges[4], 0);
          if(!faces[1])
          {
            tmpEdges[0]=edges[0];
            tmpEdges[1]=edges[5];
            tmpEdges[2]=edges[4];
            faces[1]= M_createF(part, 3, tmpEdges, 0, gent);
          }
         faces[2]=(mEntity*)F_exist(1, edges[1], edges[6], edges[5], 0);
          if(!faces[2])
          {
            tmpEdges[0]=edges[1];
            tmpEdges[1]=edges[6];
            tmpEdges[2]=edges[5];
            faces[2]= M_createF(part, 3, tmpEdges, 0, gent);
          }
        faces[3]=(mEntity*)F_exist(1, edges[2], edges[7], edges[6], 0);
          if(!faces[3])
          {
            tmpEdges[0]=edges[2];
            tmpEdges[1]=edges[7];
            tmpEdges[2]=edges[6];
            faces[3]= M_createF(part, 3, tmpEdges, 0, gent);
          }
        faces[4]=(mEntity*)F_exist(1, edges[3], edges[4], edges[7], 0);
          if(!faces[4])
          {
            tmpEdges[0]=edges[3];
            tmpEdges[1]=edges[4];
            tmpEdges[2]=edges[7];
            faces[4]= M_createF(part, 3, tmpEdges, 0, gent);
          }
     return create_rgn_with_faces(part, 5, faces, 0, gent);
    } // case 5
    case 6:   // prism
    {
       mEntity* faces[5];
       mEntity* edges[9];
       mVertex* v1 = (mVertex*)vertices[0];
       mVertex* v2 = (mVertex*)vertices[1];
       mVertex* v3 = (mVertex*)vertices[2];
       mVertex* v4 = (mVertex*)vertices[3];
       mVertex* v5 = (mVertex*)vertices[4];
       mVertex* v6 = (mVertex*)vertices[5];

      edges[0]=(mEntity*)E_exist(v1, v2);
      if (!edges[0])
        {
         edges[0]=M_createE(part, v1, v2, gent);
        }
      edges[1]=(mEntity*)E_exist(v2, v3);
      if (!edges[1])
       {
        edges[1]=M_createE(part, v2, v3, gent);
       }
       edges[2]=(mEntity*)E_exist(v3, v1);
      if (!edges[2])
       {
        edges[2]=M_createE(part, v3, v1, gent);
       }

       edges[3]=(mEntity*)E_exist(v1, v4);
      if (!edges[3])
       {
        edges[3]=M_createE(part, v1, v4, gent);
       }

        edges[4]=(mEntity*)E_exist(v2, v5);
       if (!edges[4])
        {
        edges[4]=M_createE(part, v2, v5, gent);
        }

         edges[5]=(mEntity*)E_exist(v3, v6);
       if (!edges[5])
        {
        edges[5]=M_createE(part, v3, v6, gent);
        }
         edges[6]=(mEntity*)E_exist(v4, v5);
      if (!edges[6])
       {
        edges[6]=M_createE(part, v4, v5, gent);
       }

       edges[7]=(mEntity*)E_exist(v5, v6);
      if (!edges[7])
       {
        edges[7]=M_createE(part, v5, v6, gent);
       }

      edges[8]=(mEntity*)E_exist(v6, v4);
      if (!edges[8])
       {
        edges[8]=M_createE(part, v6, v4, gent);
       }
 
      faces[0]=(mEntity*)F_exist(1, edges[0], edges[1], edges[2], 0);
      if(!faces[0])
           {
            faces[0]= M_createF(part, 3, edges, 0, gent);
           }

      faces[1]=(mEntity*)F_exist(1, edges[0], edges[4], edges[6], edges[3]);
      if(!faces[1])
          {
            tmpEdges[0]=edges[0];
            tmpEdges[1]=edges[4];
            tmpEdges[2]=edges[6];
            tmpEdges[3]=edges[3];
            faces[1]= M_createF(part, 4, tmpEdges, 0, gent);
          }

      faces[2]=(mEntity*)F_exist(1, edges[1], edges[5], edges[7], edges[4]);
      if(!faces[2])
          {
            tmpEdges[0]=edges[1];
            tmpEdges[1]=edges[5];
            tmpEdges[2]=edges[7];
            tmpEdges[3]=edges[4];
            faces[2]= M_createF(part, 4, tmpEdges, 0, gent);
          }

      faces[3]=(mEntity*)F_exist(1, edges[2], edges[3], edges[8], edges[5]);
      if(!faces[3])
          {
            tmpEdges[0]=edges[2];
            tmpEdges[1]=edges[3];
            tmpEdges[2]=edges[8];
            tmpEdges[3]=edges[5];
            faces[3]= M_createF(part, 4, tmpEdges, 0, gent);
          }
      faces[4]=(mEntity*)F_exist(1, edges[6], edges[7], edges[8], 0);
      if(!faces[4])
          {
            tmpEdges[0]=edges[6];
            tmpEdges[1]=edges[7];
            tmpEdges[2]=edges[8];
            faces[4]= M_createF(part, 3, tmpEdges, 0, gent);
      }
      return create_rgn_with_faces(part, 5, faces, 0, gent);
    } // case 6
    case 8: // hex
    {
      mEntity* faces[6];
      mEntity* edges[12];

      mVertex* v1 = (mVertex*)vertices[0];
      mVertex* v2 = (mVertex*)vertices[1];
      mVertex* v3 = (mVertex*)vertices[2];
      mVertex* v4 = (mVertex*)vertices[3];
      mVertex* v5 = (mVertex*)vertices[4];
      mVertex* v6 = (mVertex*)vertices[5];
      mVertex* v7 = (mVertex*)vertices[6];
      mVertex* v8 = (mVertex*)vertices[7];

       edges[0]=(mEntity*)E_exist(v1,v2);
          if (!edges[0])
           {
            edges[0]=M_createE(part,v1,v2,gent);
           }
        edges[1]=(mEntity*)E_exist(v2,v3);
          if (!edges[1])
           {
            edges[1]=M_createE(part,v2,v3,gent);
           }
       edges[2]=(mEntity*)E_exist(v3,v4);
          if (!edges[2])
           {
            edges[2]=M_createE(part,v3,v4,gent);
           }
       edges[3]=(mEntity*)E_exist(v1,v4);
          if (!edges[3])
           {
            edges[3]=M_createE(part,v1,v4,gent);
           }
       edges[4]=(mEntity*)E_exist(v1,v5);
          if (!edges[4])
           {
            edges[4]=M_createE(part,v1,v5,gent);
           }
       edges[5]=(mEntity*)E_exist(v2,v6);
          if (!edges[5])
           {
            edges[5]=M_createE(part,v2,v6,gent);
           }
       edges[6]=(mEntity*)E_exist(v3,v7);
          if (!edges[6])
           {
            edges[6]=M_createE(part,v3,v7,gent);
           }
        edges[7]=(mEntity*)E_exist(v4,v8);
          if (!edges[7])
           {
            edges[7]=M_createE(part,v4,v8,gent);
           }
        edges[8]=(mEntity*)E_exist(v5,v6);
          if (!edges[8])
           {
            edges[8]=M_createE(part,v5,v6,gent);
           }
        edges[9]=(mEntity*)E_exist(v6,v7);
          if (!edges[9])
           {
            edges[9]=M_createE(part,v6,v7,gent);
           }
        edges[10]=(mEntity*)E_exist(v7,v8);
          if (!edges[10])
           {
            edges[10]=M_createE(part,v7,v8,gent);
           }
        edges[11]=(mEntity*)E_exist(v8,v5);
          if (!edges[11])
           {
            edges[11]=M_createE(part,v8,v5,gent);
           }

        faces[0]=(mEntity*)F_exist(1, edges[3], edges[2], edges[1], edges[0]);
        faces[1]=(mEntity*)F_exist(1, edges[0], edges[5], edges[8], edges[4]);
        faces[2]=(mEntity*)F_exist(1, edges[1], edges[6], edges[9], edges[5]);
        faces[3]=(mEntity*)F_exist(1, edges[2], edges[7], edges[10], edges[6]);
        faces[4]=(mEntity*)F_exist(1, edges[3], edges[4], edges[11], edges[7]);
        faces[5]=(mEntity*)F_exist(1, edges[8], edges[9], edges[10], edges[11]);
        
        if(faces[0] && faces[1] && faces[2] && faces[3] && faces[4] && faces[5])
           return (pRegion)0;

        if(!faces[0])
          {
            tmpEdges[0]=edges[3];
            tmpEdges[1]=edges[2];
            tmpEdges[2]=edges[1];
            tmpEdges[3]=edges[0];
            faces[0]= M_createF(part, 4, edges, 0, gent);
          } 

         if(!faces[1])
           {
            tmpEdges[0]=edges[0];
            tmpEdges[1]=edges[5];
            tmpEdges[2]=edges[8];
            tmpEdges[3]=edges[4];
            faces[1]= M_createF(part, 4, tmpEdges, 0, gent);
          }

         if(!faces[2])
          {
            tmpEdges[0]=edges[1];
            tmpEdges[1]=edges[6];
            tmpEdges[2]=edges[9];
            tmpEdges[3]=edges[5];
            faces[2]= M_createF(part, 4, tmpEdges, 0, gent);
          }
         if(!faces[3])
           {
            tmpEdges[0]=edges[2];
            tmpEdges[1]=edges[7];
            tmpEdges[2]=edges[10];
            tmpEdges[3]=edges[6];
            faces[3]= M_createF(part, 4, tmpEdges, 0, gent);
          }
          if(!faces[4])
           {
            tmpEdges[0]=edges[3];
            tmpEdges[1]=edges[4];
            tmpEdges[2]=edges[11];
            tmpEdges[3]=edges[7];
            faces[4]= M_createF(part, 4, tmpEdges, 0, gent);
          }
         if(!faces[5])
           {
            tmpEdges[0]=edges[8];
            tmpEdges[1]=edges[9];
            tmpEdges[2]=edges[10];
            tmpEdges[3]=edges[11];
            faces[5]= M_createF(part, 4, tmpEdges, 0, gent);
          }
         return create_rgn_with_faces(part, 6, faces, 0, gent);
    } // hex
    default: std::cerr<<__func__<<" ERROR: wrong # vertices - "<<nVertex<<endl;
             break;
  }// switch (nVertices)
}

pRegion create_rgn_with_edges (pPart part, int nEdges, pMeshEnt* edges, int *dirs, pGeomEnt gent)
{
  pRegion rgn;
  mEntity* tmpDownEnt[4];
  int numFaces;
  switch (nEdges)
  {
    case 6: numFaces=4; break;
    case 8: numFaces=5; break;
    case 9: numFaces=5; break;
    case 12: numFaces=6; break;
    default: return NULL;
  }
  pMeshEnt* faces = new pMeshEnt[numFaces];

  switch (nEdges) 
  { 
    case 6: { // FMDB_TET
	      for (int i=0; i<numFaces; ++i)
	      {
	        for (int j=0; j<FMDB_TRI; j++)
	          tmpDownEnt[j]=edges[mTet::Tfe[i][j]];
	        FMDB_Face_Create(part, gent, FMDB_TRI, tmpDownEnt, 0, faces[i]);
	      }
              break;
            }
    case 8: { // FMDB_PYRAMID
	      for (int i=0; i<numFaces; i++)
	      {
	        if (!i) // i==0
	          FMDB_Face_Create(part, gent,FMDB_QUAD, edges, 0, faces[i]);
	        else
 	        {
	          for (int j=0; j<FMDB_TRI; j++)
	            tmpDownEnt[j]=edges[mPyramid::prfe[i][j]];
	          FMDB_Face_Create(part, gent,FMDB_TRI, tmpDownEnt, 0, faces[i]);
	        }
	      }
              break;
	    }
    case 9: { //topo==FMDB_PRISM
	      for (int i=0; i<numFaces; i++)
	      {
	        if (i==0 || i==4)
  	        {
	          for (int j=0; j<FMDB_TRI; j++)
 	            tmpDownEnt[j]=edges[mPrism::pmfe[i][j]];
	          FMDB_Face_Create(part, gent, FMDB_TRI, tmpDownEnt, 0, faces[i]);
	        }
      	 	else
	        {
	          for (int j=0; j<FMDB_QUAD; j++)
	            tmpDownEnt[j]=edges[mPrism::pmfe[i][j]];
	          FMDB_Face_Create(part, gent, FMDB_QUAD, tmpDownEnt, 0, faces[i]);
	        } 
	      }
              break;
            }
    case 12:{ // FMDB_HEX
	      for (int i=0; i<numFaces; i++)
	      {
	        for (int j=0; j<FMDB_QUAD; j++)
	          tmpDownEnt[j]=edges[mHex::Hfe[i][j]];
	        FMDB_Face_Create(part, gent, FMDB_QUAD, tmpDownEnt, 0, faces[i]);
	      }
            break;
	    }
    default: break;
  }
  rgn = create_rgn_with_faces(part, numFaces, faces, 0, gent);
  delete [] faces;  
  return rgn;
}

int FMDB_Rgn_Create (pPart part, pGeomEnt geomEnt, int topo, int numDownEnts,
                     pMeshEnt* downEnts, pMeshEnt& meshRgn)
{
  int downType = downEnts[0]->getLevel();
  switch (downType)
  {
    case 2: meshRgn = (pMeshEnt)create_rgn_with_faces(part, numDownEnts, downEnts, 0, geomEnt);
            break;
    case 1: meshRgn = (pMeshEnt)create_rgn_with_edges(part, numDownEnts, downEnts, 0, geomEnt);
            break;
    case 0: meshRgn = (pMeshEnt)create_rgn_with_vertices(part, numDownEnts, downEnts, 0, geomEnt);
            break;
    default: meshRgn=NULL;
             return SCUtil_FAILURE;
  }
  return SCUtil_SUCCESS;
}

// backward compatibility - declared in oldFMDB.h
pRegion M_createR(pPart part, int nFace, pFace *faces, int *dirs, pGEntity gent)
{
  return create_rgn_with_faces(part, nFace, faces, 0, gent);
}

pRegion M_createRV(pPart part, int nVertex, pVertex* vertices, int *dirs, pGEntity gent)
{
  return create_rgn_with_edges(part, nVertex, vertices, 0, gent);
}


