/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "SCUtil.h"
#include "GMI.h"
#include "FMDB.h"
#include "ParUtil.h"

int SCUTIL_Init(MPI_Comm communicator)
{
  ParUtil::Instance()->init(communicator);
  return SCUtil_SUCCESS;
}

int SCUTIL_SplitComm (int numProcGrp)
{
// this function won't change communicator where # COMM = WORLD_COMM_size/procGrpSize
// since combined with SCUTIL_SetComm, it results in part boundary errors
//  if (ParUtil::Instance()->splitComm(numProc))
//    return SCUtil_INVALID_ARGUMENT; 
// Instead, it will simply do process grouping based on the input
  if (SCUTIL_CommSize()%numProcGrp!=0 || numProcGrp<1)
    return SCUtil_INVALID_ARGUMENT;
  FMDB_Util::Instance()->setProcGrpSize(SCUTIL_CommSize()/numProcGrp);
  return SCUtil_SUCCESS;
}

int SCUTIL_SetComm (MPI_Comm communicator)
{
  cout<<__func__<<" is temporarily unavailable\n";
  return SCUtil_NOT_SUPPORTED;
// this function temporarily unavailable since it results in part boundary errors 
// if called in the middle of operation. Need fault-tolerant mechanism to provide this feature.
//  ParUtil::Instance()->setComm(communicator);
//  return SCUtil_SUCCESS;
}

int SCUTIL_GetComm (MPI_Comm *communicator)
{
  *communicator = ParUtil::Instance()->getComm();
  return SCUtil_SUCCESS;
}

int SCUTIL_Finalize(void)
{
  ParUtil::Instance()->Finalize(1);
  return SCUtil_SUCCESS;
}

int SCUTIL_CommSize()
{
  return ParUtil::Instance()->size();
}

int SCUTIL_CommRank()
{
  return ParUtil::Instance()->rank();
}

int SCUTIL_CommGrp()
{
#ifdef FMDB_PARALLEL
// this temporarily turned off since SCUTIL_SplitComm doesn't do communicator split.
//  return ParUtil::Instance()->group();
// Instead, the group id is computed manually based on process group size stored in FMDB_Util
  return SCUTIL_CommRank()/FMDB_Util::Instance()->getProcGrpSize();
#else
  return 0;
#endif
}

int SCUTIL_GrpSize()
{
  return FMDB_Util::Instance()->getProcGrpSize();
}

int SCUTIL_GrpRank()
{
  // returns rank within process group
  return SCUTIL_CommRank()%FMDB_Util::Instance()->getProcGrpSize();
}

int SCUTIL_Sync(void)
{
  ParUtil::Instance()->Barrier(__LINE__,__FILE__); 
  return SCUtil_SUCCESS;
}

int SCUTIL_GetCurMem(double *mem)
{ 
  *mem = ParUtil::Instance()->getCurMem();  // MB
  return SCUtil_SUCCESS;
}

int SCUTIL_DspCurMem(const char* msg)
{ 
  ParUtil::Instance()->report (msg, 0, ParUtil::Instance()->getCurMem());
  return SCUtil_SUCCESS;
}


int SCUTIL_ResetRsrc()
{ 
  ParUtil::Instance()->reset();
  return SCUtil_SUCCESS;
}

int SCUTIL_GetRsrcDiff(double *time, double *mem)
{ 
  *time = ParUtil::Instance()->getTimeDiff();  // sec
  *mem = ParUtil::Instance()->getMemDiff();  // MB
  return SCUtil_SUCCESS;
}

int SCUTIL_DspRsrcDiff(const char* msg)
{ 
  ParUtil::Instance()->report (msg, ParUtil::Instance()->getTimeDiff(),
                               ParUtil::Instance()->getMemDiff());
  return SCUtil_SUCCESS;
}

int SCUTIL_DspRsrc(const char* msg, double time, double mem)
{
  ParUtil::Instance()->report (msg, time, mem);
  return SCUtil_SUCCESS;
}

int SCUTIL_DspSysInfo()
{ 
  ParUtil::Instance()->sysinfo();
  return SCUtil_SUCCESS;
}

