/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "FMDB.h"
#include "mMesh.h"

int FMDB_Mesh_CreateTag (pMeshMdl mesh, const char* tag_name, int tag_type, int tag_size, pTag& tag)
{
  pTag tmpTag;
  if (!FMDB_Mesh_FindTag (mesh, tag_name, tmpTag))
  {
    tag = tmpTag;
    return SCUtil_TAG_ALREADY_EXISTS;
  }

  tag = TagHolder_CreateTag(static_cast<pTagHolder>(mesh), tag_name, tag_type, tag_size); 
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_DelTag (pMeshMdl mesh, pTag tag, int forceDel)
{ 
  int has_tag;
  FMDB_Mesh_HasTag (mesh, tag, &has_tag);
  if (!has_tag) return SCUtil_INVALID_TAG_HANDLE;

  pPart part;

  if (forceDel)
  {
    for (mMesh::partIter pit=mesh->partBegin(); pit!=mesh->partEnd(); ++pit)
    {
      part = *pit;
      Taggable_DelTag(static_cast<pTaggable>(part), tag);

      // entity
      for (int i=0; i<=part->getDimension(); ++i)
        for (mPart::iterall ent_it=part->beginall(i); ent_it!=part->endall(i); ++ent_it)
          Taggable_DelTag(static_cast<pTaggable>(*ent_it), tag);
     // part set
     std::vector<pEntSet> entsets;
     FMDB_Part_GetSet (part, entsets);
     for (std::vector<pEntSet>::iterator set_it=entsets.begin(); set_it!=entsets.end(); ++set_it)
       Taggable_DelTag(static_cast<pTaggable>(*set_it), tag);
       //(*set_it)->deleteTagData(tag); 
    }  // for each part

    // entity set
    for (std::list<pEntSet>::iterator eset_it = mesh->beginEntSet();
         eset_it!=mesh->endEntSet(); ++eset_it)
      Taggable_DelTag(static_cast<pTaggable>(*eset_it), tag);
  }  // forceDel

  // if forceDel==0 and tag still in use, FMDB_Mesh_Del crashes
  TagHolder_DelTag(static_cast<pTagHolder>(mesh), tag); 
  delete tag; // delete tag handle
  return  SCUtil_SUCCESS;
}

int FMDB_Mesh_FindTag (pMeshMdl mesh, const char* tag_name, pTag& tag)
{
  return TagHolder_FindTag (static_cast<pTagHolder>(mesh), tag_name, tag);
}

int FMDB_Mesh_HasTag (pMeshMdl mesh, const pTag tag, int *exist)
{ 
  return TagHolder_HasTag (static_cast<pTagHolder>(mesh), tag, exist);
}

int FMDB_Tag_InUse (pMeshMdl mesh, const pTag tag, int *use)
{
  // mesh ent set
  for (std::list<pEntSet>::iterator eset_it = mesh->beginEntSet();
         eset_it!=mesh->endEntSet(); ++eset_it)
    if (Taggable_HasTag(static_cast<pTaggable>(*eset_it), tag))
    {  *use=1; return SCUtil_SUCCESS; }

  pPart part;
  for (mMesh::partIter pit=mesh->partBegin(); pit!=mesh->partEnd(); ++pit)
  {
    part = *pit;
    // part
    if (Taggable_HasTag(static_cast<pTaggable>(part), tag))
    {  *use=1; return SCUtil_SUCCESS; }

    // part ent set
     std::vector<pEntSet> entsets;
     FMDB_Part_GetSet (part, entsets);
     for (std::vector<pEntSet>::iterator set_it=entsets.begin(); set_it!=entsets.end(); ++set_it)
       if (Taggable_HasTag(static_cast<pTaggable>(*set_it), tag))
     {  *use=1; return SCUtil_SUCCESS; } 

    // entity
    for (int i=0; i<=part->getDimension(); ++i)
      for (mPart::iterall ent_it=part->beginall(i); ent_it!=part->endall(i); ++ent_it)
        if (Taggable_HasTag(static_cast<pTaggable>(*ent_it), tag))
        {  *use=1; return SCUtil_SUCCESS; } 
  } // for part
  
  *use = 0;
  return SCUtil_SUCCESS;
}

int FMDB_Tag_GetName (const pTag tag, char* tag_name, int name_len)
{
  Tag_GetName(tag, tag_name, name_len);
  return SCUtil_SUCCESS;
}

int FMDB_Tag_GetType (const pTag tag, int* type)
{
  *type = Tag_GetType(tag);
  return SCUtil_SUCCESS;
}

int FMDB_Tag_GetSize (const pTag tag, int* size)
{
  *size = Tag_GetSize(tag);
  return SCUtil_SUCCESS;
}

int FMDB_Tag_GetByte (pTag tag, int* byte)
{
  switch (Tag_GetType(tag))
  {
    case SCUtil_BYTE: *byte=Tag_GetSize(tag); break;
    case SCUtil_INT:  *byte=Tag_GetSize(tag)*sizeof(int); break;
    case SCUtil_DBL:  *byte=Tag_GetSize(tag)*sizeof(double); break;
    case SCUtil_ENT:  *byte=Tag_GetSize(tag)*sizeof(void*); break;
    case SCUtil_SET:  *byte=Tag_GetSize(tag)*sizeof(void*); break;
    default: break;
  }
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_GetTag (pMeshMdl mesh, std::vector<pTag>& tags)
{
  TagHolder_GetTag (static_cast<pTagHolder>(mesh), tags);
  return  SCUtil_SUCCESS;
}


