/******************************************************************************

  (c) 2004-2012 Scientific Computation Research Center,
      Rensselaer Polytechnic Institute. All rights reserved.

  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.

*******************************************************************************/
#ifdef FMDB_PARALLEL
#define TAG_NAME_SIZE 64

#include <stdio.h>
#include <iostream>

#include <vector>
#include <set>
#include <map>
#include <string>
#include <mpi.h>
#include <cstring>

#include <malloc.h>
#include <mcheck.h>

#include "FMDB.h"
//for BG performance study
//include "rts.h"
#include "mMesh.h"
#include "mPart.h"
#include "mEntity.h"
#include "mVertex.h"

using std::set;
using std::pair;
using std::map;
using std::cout;
using std::endl;


struct packedGhostVertex
{
  int iD;
  double coord[3] __attribute__((aligned (16)));
  short int numRC;
}__attribute__((aligned (16)));

struct packedGhostEnt
{
  short int type;
  short int gTag;
  short int gType;
  short int topo;
  short int numTags;
  mEntity* sender;
  mEntity* ownerEnt;
  int ownerPid;
  int numDE;
}__attribute__((aligned (16)));

struct packedTag
{
   char tagName[TAG_NAME_SIZE];
}__attribute__((aligned (16)));

// sub-function used by createGhost
void createGhostMapping(pPart part, pMeshEnt ent, int pid, short ghostStatus, pMeshEnt ghostCopy=NULL);
void getGhostEnts(pMeshMdl mesh, int procId, int bDim, int gDim, int include_copy, set<pMeshEnt>* entitiesToGhost, int numLayer, pTag tagVisited);
void setGhostDestination(pMeshMdl mesh, int procId, pMeshEnt bridgeEnt, int gDim, pTag tag, int numLayer, set<pMeshEnt>* entitiesToGhost, set<pMeshEnt>& tagEnts);
void getVertices(pPart part, int procId, int gDim, pMeshEnt ent, int destId, set<pMeshEnt>& pVertices);
void eliminateDuplicateEnts(pPart, int procId, set<pMeshEnt>& entitiesToGhost);
void exchangeGhostEntities(pMeshMdl mesh, int partId, int dim, int gDim, pmMigrationCallbacks &cb, set<pMeshEnt>* entitiesToGhost, int num);
void exchangeEntities2Bounce(pPart part, list<rc_struct_2>& entitiesToBounce, int gDim, int dim);
void pack_Ghost_MEntity(pMeshMdl mesh, pMeshEnt ent, int, pmMigrationCallbacks& cb, void*& msg, int &);
void unpack_Ghost_MEntity(pMeshMdl mesh, int ghostDim, void *msg_recv, int msg_size, int pid, pmMigrationCallbacks &cb, list<rc_struct_2>& entitiesToBounce);
void removeEntities(pPart part, set<pMeshEnt>* entitiesToRemove, int gDim);
void removeLyrEnts(pMeshMdl mesh, int procId, int gDim, int bDim, set<pMeshEnt> inAdjs, int lyr, set<pMeshEnt>* entitiesToRemove, set<pMeshEnt>& out, pTag tag, set<pMeshEnt>&, double*);
void deleteDownwardAdj(pPart part, pMeshEnt pEnt, set<pMeshEnt>* entitiesToRemove, int gDim);
void getLayerEnts(pMeshMdl mesh, int procId, int gDim, int bDim, int lyr, int includeCopy, pTag tag, set<pMeshEnt>* entitiesToGhost, set<pMeshEnt>& tagEnts);

void createGhostMapping(pPart part, pMeshEnt ent, int pid, short ghostStatus, pMeshEnt gCopy)
{
  mGhost* ghost=NULL;
  if(ent->getGhostIndex() >= 0)
   {
     ghost=part->getGhost(ent->getGhostIndex());
     if(ghost)
      {
       ghost->addGPs(pid);
       if(gCopy)
        ghost->addGhostCopy(pid, gCopy);
      }
   }
  else
  {
    ghost=new mGhost();
    ghost->addGPs(pid);
    ent->setGhostIndex(part->getGhostVecSize());
    ghost->setOwnerEnt(ent);
    part->addGhost(ghost);
    ghost->setGhostStatus(ghostStatus);
    if(gCopy)
    {
      ghost->addGhostCopy(pid, gCopy);
    }
  }
}
int createGhost(pMeshMdl mesh, pmMigrationCallbacks & migrCB, int gDim, int bDim, int numLayer, int include_copy)
{
  int procId; // MPI rank
  std::vector<pPart> vecPart; 
  char* tagName;
  int tagSize=1, tagType = 0;
  pTag tagVisited;
  double t1, t3;
#ifdef DEBUG
  long lLocalMove = 0, lTotalMove = 0, lInc = 0;
  double localTime, totalTimeMax, totalTimeAvg;
#endif
  //Get current process id 
  procId = SCUTIL_CommRank();
  
  //Get all parts in the current mesh instance. Right now we have one part per mesh instance 
  FMDB_Mesh_GetAllPart(mesh, procId, vecPart);
  int numProc=SCUTIL_CommSize();

  //Visited tag is for numLayer > 1 only
  tagName = (char*)"visited";
  if(numLayer > 1)	
    FMDB_Mesh_CreateTag(mesh, tagName, tagType, tagSize, tagVisited);
  /*
  // performance measurement code on BG/L
  size_t memSizeB;
  rts_getavailablememory(&memSizeB);
  if(!procId)
    cout << "\n Memory allocated before ghosting:" << memSizeB<< endl;
  */ 
  pPart part = (pPart)vecPart.at(0);
  //Store the top level entity to be ghosted and its downward adjacent vertices
  set<pMeshEnt>* entitiesToGhost = new set<pMeshEnt>[1+numLayer];
    
  t1 = MPI_Wtime();
  getGhostEnts(mesh, procId, bDim, gDim, include_copy, entitiesToGhost, numLayer, tagVisited);
  //first exchange vertices
  exchangeGhostEntities(mesh, procId, 0, gDim, migrCB, &entitiesToGhost[0],0);
 
  // Now exchange ghost entities layer by layer
  exchangeGhostEntities(mesh, procId, gDim, gDim, migrCB, entitiesToGhost, numLayer);

  t3 = MPI_Wtime();
  /* //memory measurement on BG/L
  size_t memSize, maxMem, memAvail;
  rts_getavailablememory(&memSize);
  memAvail=memSize - memSizeB;
  MPI_Allreduce(&memAvail,&maxMem,1,MPI_INT,MPI_MAX,MPI_COMM_WORLD);
  if(!procId)
    cout << "\n Memory after ghost entity collection allocated " << maxMem << endl;
  */
#ifdef DEBUG
  localTime = t3 - t1;   
#endif
  pMeshEnt ent=NULL;
  mGhost* ghost=NULL;
  for(int i = 1; i <= numLayer; i++ )
   {
    for(set<pMeshEnt>::iterator setIter = entitiesToGhost[i].begin(); setIter != entitiesToGhost[i].end(); setIter++ )
     {
       ent=*setIter;
       ghost=part->getGhost(ent->getGhostIndex());
#ifdef DEBUG
       lLocalMove += ghost->getNumGPs();
#endif     
    }
#ifdef DEBUG
    lInc += entitiesToGhost[i].size();
#endif
   }
#ifdef DEBUG
  // uncomment it for performance information
  MPI_Allreduce(&localTime,&totalTimeMax,1,MPI_DOUBLE,MPI_MAX, MPI_COMM_WORLD);
  MPI_Allreduce(&localTime,&totalTimeAvg,1,MPI_DOUBLE,MPI_SUM, MPI_COMM_WORLD);
  MPI_Allreduce(&lLocalMove,&lTotalMove,1,MPI_LONG,MPI_SUM, MPI_COMM_WORLD);
   
  totalTimeAvg=totalTimeAvg/numProc;    
#endif
  //Deallocate memory 
  for(int i = 0; i < (1 + numLayer); i++ )
    entitiesToGhost[i].clear();
  delete[] entitiesToGhost;
  mesh->addGhostInfo(gDim, bDim, numLayer); 
  vecPart.clear();    
#ifdef DEBUG_
  if(procId==0)
   cout << "\n[Create Ghost] type:" << gDim << ", bridge:" << bDim << ", include_copy:" << include_copy << ", layer: " << numLayer << ", #ghosts:" << lTotalMove << ", max time:" << totalTimeMax << "(sec), avg time: " << totalTimeAvg <<"(sec)\n";
#endif
  return SCUtil_SUCCESS;
}

// collect entities of dimension 'gDim' adjacent to part boundary entities of dimension 'bDim'. 
// Set the destination part ids of the entities to be ghosted. 
void getGhostEnts(pMeshMdl mesh, int procId, int bDim, int gDim, int include_copy, set<pMeshEnt>* entitiesToGhost, int numLayer, pTag tag)
{
  int iIsEnd = 0;
  std::vector<pPart> vecPart;
  pPart part;
  pPartEntIter it;
  pMeshEnt pBrgEnt;
  int pDim;
  set<pMeshEnt> tagEnts;
 
  FMDB_Mesh_GetAllPart(mesh, procId, vecPart);
  part = vecPart.at(0);
  //Step 1 First get all part boundary entities of bridge dimension 'bdim'
  iIsEnd = FMDB_PartEntIter_InitPartBdry(part, -1, bDim, FMDB_ALLTOPO, it);
  if(iIsEnd != 0)
    return;

  while(!iIsEnd)
   {
    //Get next entity on part boundary having dimension bdim
    iIsEnd = FMDB_PartEntIter_GetNext(it, pBrgEnt);
    if(iIsEnd) 
      break;
    //If include_copy=0 and the owner part id is different than current part id, skip the bridge entity
    if((include_copy == 1) || (pBrgEnt->getOwner() == procId))
      setGhostDestination(mesh, procId, pBrgEnt, gDim, tag, numLayer, entitiesToGhost, tagEnts); 

    if(numLayer > 1)
     {
       FMDB_Ent_SetIntTag(mesh, pBrgEnt, tag, 1);
       tagEnts.insert(pBrgEnt);
     }
   } // end while

  FMDB_Part_GetDim(part, &pDim);
  for(int lyr = 2; lyr <= numLayer; lyr++ )
   {
    getLayerEnts(mesh, procId, gDim, bDim, lyr, include_copy, tag, entitiesToGhost, tagEnts);
    /// Keep attaching tags to process the next layer, if this is the last layer, stop attaching tags.  
    if(lyr < numLayer)
     {
      for(set<pMeshEnt>::iterator setIter = entitiesToGhost[lyr].begin(); setIter != entitiesToGhost[lyr].end(); setIter++ )
        FMDB_Ent_SetIntTag(mesh, (*setIter), tag, 1);
     }
   }// for lyr

   // Force delete is an expensive operation so delete tags manually
  if(numLayer > 1)
  {
   for(int i = 1; i <= numLayer; i++ )
    {
     for(set<pMeshEnt>::iterator setIter = entitiesToGhost[i].begin(); setIter != entitiesToGhost[i].end(); setIter++ )
       FMDB_Ent_DelTag(*setIter, tag);
   }
   for(set<pMeshEnt>::iterator itSet = tagEnts.begin(); itSet != tagEnts.end(); itSet++)
     FMDB_Ent_DelTag(*itSet, tag);

   tagEnts.clear();
   FMDB_Mesh_DelTag(mesh, tag, 0);
  }
 
  // Now eliminate duplicate entities.  
  eliminateDuplicateEnts(part, procId, entitiesToGhost[0]);
  if(gDim < pDim)
   {
    for(int lyr = 1; lyr <= numLayer; lyr++)
      eliminateDuplicateEnts(part, procId, entitiesToGhost[lyr]);
   }
  vecPart.clear();
}

// Entity collection procedure for number of layers > 1
void getLayerEnts(pMeshMdl mesh, int procId, int gDim, int bDim, int lyr, int include_copy, pTag tagVisited, set<pMeshEnt>* entitiesToGhost, set<pMeshEnt>& tagEnts)
{
  int dir = 0, gId;
  pMeshEnt pEnt, pBrgEnt, pUpEnt;
  mGhost* gInfo;
  std::vector<pPart> vecPart;
  pPart part;
  int iHasTag = -1;
 
  FMDB_Mesh_GetAllPart(mesh, procId, vecPart);
  part = vecPart.at(0);
  for(set<pMeshEnt>::iterator setIter = entitiesToGhost[lyr - 1].begin(); setIter != entitiesToGhost[lyr - 1].end(); setIter++ )
   {
    //For each entity of gDim, get downward adjacent entities of type bdim and check if they are visited or not
    vector<pMeshEnt> vecDownAdjEnt;
    pEnt= (pMeshEnt)(*setIter);
    gInfo=part->getGhost(pEnt->getGhostIndex());
    FMDB_Ent_GetAdj(pEnt, bDim, dir, vecDownAdjEnt); 

    for(vector<pMeshEnt>::iterator itAdj = vecDownAdjEnt.begin(); itAdj != vecDownAdjEnt.end(); itAdj++ )
     {
      pBrgEnt = (pMeshEnt)(*itAdj);
      FMDB_Ent_HasTag(pBrgEnt, tagVisited, &iHasTag);
   
      if(iHasTag == 1)
        continue;
      FMDB_Ent_SetIntTag(mesh, pBrgEnt, tagVisited, 1);
      tagEnts.insert(pBrgEnt);
   
      if(include_copy == 0 && (pBrgEnt->getOwner() != procId))
        continue;
      vector<pMeshEnt> vecUpAdjEnt;
      FMDB_Ent_GetAdj(pBrgEnt, gDim, dir, vecUpAdjEnt);
   
      for(vector<pMeshEnt>::iterator itUpAdj = vecUpAdjEnt.begin(); itUpAdj != vecUpAdjEnt.end(); itUpAdj++ )
       {
         pUpEnt = (pMeshEnt)(*itUpAdj);
         FMDB_Ent_HasTag(pUpEnt, tagVisited, &iHasTag);
         if(iHasTag == 1)
           continue;
         for(set<int>::iterator gpiter = gInfo->gpBegin(); gpiter != gInfo->gpEnd(); gpiter++)
          {
           gId = *gpiter; 
           if(pUpEnt->getRemoteCopy(gId))
              continue;
           createGhostMapping(part, pUpEnt, gId, HAS_GHOSTS);
           getVertices(part, procId, gDim, pUpEnt, gId, entitiesToGhost[0]);
         }
      
       //if(part->getGhost(pUpEnt->getGhostIndex()) > 0)
       entitiesToGhost[lyr].insert(pUpEnt);
       } // itUpAdj
     vecUpAdjEnt.clear();
    } // end for itAdj
  vecDownAdjEnt.clear();
 } // end for setIter
}
//If different parts want to send the same entity information to a destination part, 
// this method will communicate with them so that only a single part 
// sends the entity information to the destination part. 

void eliminateDuplicateEnts(pPart part, int partId, set<pMeshEnt>& gEnts)
{
  //ipcomman setting for send/receive messages
  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::Neighbors);
  CM->set_tag(0);
  int msg_size = sizeof(pMeshEnt) + sizeof(short int);
  CM->set_fixed_msg_size(msg_size);
  void *msg_send = CM->alloc_msg(msg_size);
  short int* s_pid;
  int iNumSent = 0, iNumRcvd = 0;
 
  // fmdb variables 
  pMeshEnt pDownEnt, pRmtCopy;
  pMeshEnt* s_ent;
  int destId, pid;

  for(set<pMeshEnt>::iterator itSet = gEnts.begin(); itSet != gEnts.end(); itSet++ )
   {
     pDownEnt = *itSet;
     //if the entity is not a remote copy or its ghost destinations have already been deleted 
     //in previous duplicate elimination rounds then skip it 
     if(!pDownEnt->getPClassification() || pDownEnt->getGhostIndex() == -1)
       continue;
  
     mGhost* gInfo=part->getGhost(pDownEnt->getGhostIndex()); 
     vector<pairRemote> vecRmtCopy;
     FMDB_Ent_GetAllRmt (pDownEnt, vecRmtCopy);

     for(mGhost::GPIter gpiter = gInfo->gpBegin(); gpiter != gInfo->gpEnd(); gpiter++)
      {
       destId = *gpiter;

       //If there exist other remote copies then check with other remote copies if they want to send the entity to the same destination.
       for(vector<pairRemote>::iterator It = vecRmtCopy.begin(); It != vecRmtCopy.end(); It++ )
        {
          pid = It->first;
          pRmtCopy = It->second;
          s_ent = (pMeshEnt*)msg_send;
          s_ent[0] = pRmtCopy;
          s_pid = (short int*)((char*)msg_send + sizeof(pMeshEnt));
          s_pid[0] = destId;
          CM->send(pid, (void*)msg_send);
          ++iNumSent;
        }
      } 
    vecRmtCopy.clear();
  }// end for itSet

  CM->finalize_send();
  CM->free_msg(msg_send);
 
 // receive phase begins
 void *msg_recv;
 int pid_from, sendTo;

 while(CM->receive(msg_recv, &pid_from))
 {
   ++iNumRcvd;
   pMeshEnt* s_ent = (mEntity**)msg_recv;
   pMeshEnt pRmtEnt = s_ent[0];

   s_pid = (short int*)((char*)msg_recv + sizeof(mEntity*));
   sendTo = *s_pid;

   CM->free_msg(msg_recv);	
   if(pid_from < partId)
    {
      mGhost* ghost=part->getGhost(pRmtEnt->getGhostIndex());
      if(ghost)
      {
       ghost->deleteGPs(sendTo);
       //If after deletion, it has no more destination part ids then delete that entity from entitiesToSend
       if(ghost->getNumGPs()==0)
       {
         gEnts.erase(pRmtEnt);	
         pRmtEnt->setGhostIndex(-1);
       }
      }
    }//if pid_from<partId
  }
}
// Decides the destination part ids of the entity being ghosted. 
// If entity copies already exists on all parts on which the bridge entity exists then it does not insert it in the ghosting list.
void setGhostDestination(pMeshMdl mesh, int procId, pMeshEnt bridgeEnt, int gDim, pTag tag, int numLayer, set<pMeshEnt>* entitiesToGhost, set<pMeshEnt>& tagEnts)
{
  std::vector<pPart> vecPart;
  vector<pMeshEnt> vecAdjEnt;
  int dir = 0, prtBId, iHasTag = -1;
  pPart part;
  pMeshEnt pAdjEnt, pRemoteCopy;
  pairRemote pr;
 
  FMDB_Mesh_GetAllPart(mesh, procId, vecPart);
  part = vecPart.at(0);
 
  //For each bridge entity, get adjacent entities of ghost dimension 'dim' */
  FMDB_Ent_GetAdj (bridgeEnt, gDim, dir, vecAdjEnt);
    
  // If any part does not belong to residence parts of pAdjEnt and belongs to residence parts of bridge entity bridgeEnt, 
  // we need a ghost on that part
  for (vector<pMeshEnt>::iterator It = vecAdjEnt.begin(); It != vecAdjEnt.end(); It++ )
   {
     pAdjEnt = (pMeshEnt)(*It);
     // Get remote copies of the bridge entity
     vector<pairRemote> vecBrgRmtCopy;
   
     for(vector<pairRemote>::iterator itBrgRmtCpy = vecBrgRmtCopy.begin(); itBrgRmtCpy != vecBrgRmtCopy.end(); itBrgRmtCpy++)
      {
       pr = (pairRemote)(*itBrgRmtCpy);
       prtBId = pr.first;
				
       pRemoteCopy = pAdjEnt->getRemoteCopy(prtBId);
       if(!pRemoteCopy)
        {
         createGhostMapping(part, pAdjEnt, prtBId, HAS_GHOSTS);
         entitiesToGhost[1].insert(pAdjEnt);
			
         // Get downward adjacencies of the ghost entity
        getVertices(part, procId, gDim, pAdjEnt, prtBId, entitiesToGhost[0]);
        }
     } //end itBrgRmtCpy
      
   vecBrgRmtCopy.clear();
   // Attach visited tag for processing subsequent layers
   if(numLayer > 1)
   {
     FMDB_Ent_HasTag(pAdjEnt, tag, &iHasTag);
     if(iHasTag != 1 && (pAdjEnt->getGhostIndex() != -1))
       FMDB_Ent_SetIntTag(mesh, pAdjEnt, tag, 1);
   }
  } // end vecAdjEnt
 vecAdjEnt.clear();
} // end function

// Collects the downward adjacent entities of the entity to be ghosted 'ghostEnt'
void getVertices(pPart part, int partId, int gDim, pMeshEnt ghostEnt, int destId, set<pMeshEnt>& pVertices)
{
  vector<pMeshEnt> vecDownEnts;
  pMeshEnt pDownEnt;
  FMDB_Ent_GetAdj(ghostEnt,0,0,vecDownEnts);

  //Get downward adjacencies of the ghost entity
  for (vector<pMeshEnt>::iterator veciter = vecDownEnts.begin(); veciter!=vecDownEnts.end();++veciter)
   {
    // We dont want to send duplicate remote copies to the same destination part. 
    // Niether we want to send the entity if it already has a remote copy on the destination part
    pDownEnt = *veciter;
    // If the vertex already exists on remote part in the form of remote copy then skip it
    if(pDownEnt->getRemoteCopy(destId))
      continue;
    createGhostMapping(part, pDownEnt, destId, HAS_GHOSTS);
    pVertices.insert(pDownEnt);
   } // end adj vec
  vecDownEnts.clear();
} // end function 

void exchangeGhostEntities(pMeshMdl mesh, int partId,int dim, int gDim, pmMigrationCallbacks &cb, set<pMeshEnt>* entitiesToSend, int size)
{
  /* ipcomman settings for send/receive message. We need neighborhood communication only */
  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::Neighbors);
  CM->set_tag(0);
  CM->set_fixed_msg_size(0);
  int num_sent = 0, num_recvd = 0;
  list<rc_struct_2> entitiesToBounce;
  unsigned long totalSize = 0;
  pMeshEnt pEnt;
  mGhost* ghost;
  int destId, msg_size,lstart=1;
  if(size == 0)
    lstart = 0;

  vector<pPart> vecPart;
  FMDB_Mesh_GetAllPart(mesh, partId, vecPart);
  pPart part=vecPart.at(0);

  for(int i = lstart; i <= size; i++)
   {
     for(set<pMeshEnt>::iterator gcIter=entitiesToSend[i].begin(); gcIter!=entitiesToSend[i].end(); gcIter++)
      {
        pEnt = *gcIter;
        ghost=part->getGhost(pEnt->getGhostIndex());
        if(ghost->getNumGPs() == 0)
          continue;

        for(mGhost::GPIter gpiter = ghost->gpBegin(); gpiter != ghost->gpEnd(); gpiter++)
 	  {
	   destId = *gpiter;
	   void *msg_send;
	   pack_Ghost_MEntity(mesh, pEnt, destId, cb, msg_send, msg_size);
	   // Send message to destId
	   CM->send(destId, msg_send, msg_size);
	   totalSize += msg_size;
	   num_sent++;
	   CM->free_msg(msg_send);
         }
      } // end for set
   }
  CM->finalize_send();
  // receive phase begins
  void *msg_recv;
  int pid_from;
  while(int rc = CM->receive(msg_recv, &pid_from))
  {
    num_recvd++;
    unpack_Ghost_MEntity(mesh, dim, msg_recv, rc, pid_from, cb, entitiesToBounce);
    CM->free_msg(msg_recv);
  }
  // We only bounce back information about vertices and reg
  exchangeEntities2Bounce(part, entitiesToBounce, gDim, dim);
}

// **********************************************
void exchangeEntities2Bounce(pPart part, list<rc_struct_2>& entitiesToBounce, int gDim, int dim)
// **********************************************
{
  list<rc_struct_2> entitiesToBroadcast;
  int msg_size = 2*sizeof(pMeshEnt);
  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_tag(0);
  CM->set_fixed_msg_size(msg_size);
  mEntity** msg_send = (mEntity**)CM->alloc_msg(msg_size);
  int iNumRecvd = 0;
  list<rc_struct_2>::const_iterator rcvIter;

  if(dim == 0)
    CM->set_comm_validation(IPComMan::All_Reduce);
   else 
    CM->set_comm_validation(IPComMan::Neighbors);

  for (rcvIter=entitiesToBounce.begin(); rcvIter!=entitiesToBounce.end();++rcvIter)
   {
     msg_send[0] = rcvIter->src_ent;
     msg_send[1] = rcvIter->src_ent;
     CM->send(rcvIter->dest_partid, (void*)msg_send);
   }
  CM->finalize_send();
  CM->free_msg(msg_send);

  // receive phase begins
  void *msg_recv;
  int pid_from;
  mEntity** castbuf;
  int dir = 0;

  while(CM->receive(msg_recv, &pid_from))
   {
    iNumRecvd++;
    castbuf = (mEntity**)msg_recv;
   
    //For vertices, we need to check if parts having their remote copies have destId in their destinations.*/
    if(dim == 0)
    {
      vector<pMeshEnt> vecAdjEnt;
      bool bIsThere = false;
      mGhost* ghost=NULL;
      pMeshEnt pVtx = castbuf[1];

      // 1.1 For each bridge entity, get adjacent entities of ghost dimension 'dim' */
      FMDB_Ent_GetAdj(pVtx, gDim, dir, vecAdjEnt);
 
      // Get adjacencies of ghostType using the bridge. If the isGhost flag is set then we need to remove that entity*/
      for (vector<pMeshEnt>::iterator It = vecAdjEnt.begin(); It != vecAdjEnt.end(); It++ )
       {
         pMeshEnt pAdjEnt = (pMeshEnt)(*It);
         ghost=part->getGhost(pAdjEnt->getGhostIndex());
        if(ghost)
        {
          if(ghost->findPidInGPs(pid_from))
            {
    	      bIsThere = true;
	      break;
           }
       }
      }//end for it
    if(bIsThere)
     {
      createGhostMapping(part, pVtx, pid_from, HAS_GHOSTS, castbuf[0]);
     }
    vecAdjEnt.clear();
   }
  else if(dim == gDim)
   {
    createGhostMapping(part, castbuf[1], pid_from, HAS_GHOSTS, castbuf[0]);
   }
  CM->free_msg(msg_recv);
  }
}

void pack_Ghost_MEntity(mMesh* mesh, mEntity* ent, int pid, pmMigrationCallbacks &cb, void *&msg_send, int &msg_size)
{
  char *buf_user;
  int user_size=0;
  int entity_size=0;
  IPComMan *CM = ParUtil::Instance()->ComMan();

  pPart part;
  FMDB_Mesh_GetPart(mesh, 0, part);
  /* User may have some data to send to the other process
    These data are of size user_size */
  buf_user = (char*)cb.getUserData(ent, pid, user_size);
  if(!buf_user)
    user_size = 0;
  int entDim = ent->getLevel();
  int n_bytes = 0;
  pMeshEnt ownerEnt;
  FMDB_Ent_GetOwnEnt(ent, part, ownerEnt);
  std::vector<pTag>::iterator tagIter;
  entity_size = sizeof(packedGhostEnt);

  vector<pTag> double_tags, tags;
  int has_tag=0, tag_size, tag_type;
  pTag tag;

  int numDE=0;
  FMDB_Ent_HasTag(ent, NULL, &has_tag);
  if(entDim==0)
   {
    entity_size += sizeof(packedGhostVertex) + ent->getNumRemoteCopies() * (sizeof(int) + sizeof(pMeshEnt));
   }
  else
   {
    numDE = ent->size(0);
    entity_size +=  numDE*sizeof(mEntity*);
   }
  if(has_tag)
   {
     FMDB_Ent_GetTag(ent, tags);
     for (tagIter = tags.begin(); tagIter != tags.end(); ++tagIter)
     {
      tag=*tagIter;
      FMDB_Tag_GetType (tag, &tag_type);
      FMDB_Tag_GetSize(tag, &tag_size);
      if(tag_type!=SCUtil_DBL)// FMDB_DBL for field data
       continue;
      entity_size+=sizeof(packedTag);// for tag name
      entity_size+=tag_size*sizeof(double);
      double_tags.push_back(tag);
     }
   }
  msg_size = entity_size + user_size;
  msg_send = CM->alloc_msg(msg_size);
  /* Pack the remote copies of vertex. We want to notify back the ghost information to all remote copies of vertices. */
  if (entDim == 0)
   {
    int numRC=ent->getNumRemoteCopies();
   // fixed length data
    packedGhostEnt *castbuf = (packedGhostEnt*)((char*)msg_send + n_bytes);
    castbuf->sender = ent;   
    castbuf->ownerEnt = ownerEnt;
    castbuf->ownerPid = ent->getOwner();
    castbuf->numDE=0;
    castbuf->numTags=double_tags.size();
    pGEntity gent = ent->getClassification();
    if(gent!=0)
     {
      castbuf->gTag = GEN_tag(ent->getClassification());
      castbuf->gType = GEN_type(ent->getClassification());
     }
  else
     {
      castbuf->gTag = 1;
      castbuf->gType = 3;
     }
   n_bytes += sizeof(packedGhostEnt);

  packedGhostVertex *castvtx=(packedGhostVertex*)((char*)msg_send + n_bytes); 
  double xyz[3];
  ((mVertex*)ent)->getCoord(xyz[0], xyz[1], xyz[2]);
  for(int i=0;i<3;i++)
   castvtx->coord[i] = xyz[i];
 
  castvtx->iD = ent->getId();
  castvtx->numRC = numRC;
  n_bytes += sizeof(packedGhostVertex);

  if(numRC > 0)
   {
    int *RCs_pid;
    pMeshEnt* RCs_ent;	
       
    for (mEntity::RCIter rciter=ent->rcBegin();rciter!=ent->rcEnd();++rciter)
    {
      RCs_pid = (int*)((char*)msg_send + n_bytes);
     *RCs_pid = rciter->first;
      n_bytes += sizeof(int);
      RCs_ent = (mEntity**)((char*)msg_send + n_bytes);
      *RCs_ent = rciter->second;
      n_bytes += sizeof(mEntity*);
    }
   }
  }
  else
  {
  // fixed length data
  packedGhostEnt *castbuf = (packedGhostEnt*)((char*)msg_send + n_bytes);
  castbuf->sender = ent;
  castbuf->ownerEnt = ownerEnt;
  castbuf->ownerPid = ent->getOwner();
  castbuf->topo = ent->getTopo();
  castbuf->type = ent->getType();
  castbuf->numDE = numDE;
  castbuf->numTags=double_tags.size();
  pGEntity gent = ent->getClassification();
  if(gent!=0)
   {
     castbuf->gTag = GEN_tag(ent->getClassification());
     castbuf->gType = GEN_type(ent->getClassification());
   }
  else
  {
    castbuf->gTag = 1;
    castbuf->gType = 3;
  }
  n_bytes += sizeof(packedGhostEnt);

  mEntity** DEs = (mEntity**)((char*)msg_send + n_bytes);
  mEntity* vtx;
  for (int i=0; i < numDE; i++)
  {
    vtx = ent->get(0,i)->getRemoteCopy(pid);
    if(vtx)
    {
     DEs[i] = vtx;                    // remote copy
    }
    else
    { 
      pMeshEnt tmpEnt=ent->get(0,i);
      mGhost* ghost=part->getGhost(tmpEnt->getGhostIndex());
      if(ghost)
        DEs[i]=ghost->getGhostCopy(pid);
      else
        cout << "\n FMDB Error: Ghost not found " << endl;
     }
  }
    n_bytes += numDE*sizeof(mEntity*);
  }

 if(double_tags.size()>0)
   {
    char name[TAG_NAME_SIZE-1];
    pTag tag;
    double* dbl_val;
    for (tagIter = double_tags.begin(); tagIter != double_tags.end(); ++tagIter)
     {
        tag=*tagIter;

        packedTag* entTag=(packedTag*)((char*)msg_send + n_bytes);
        FMDB_Tag_GetName(tag, name, TAG_NAME_SIZE-1);
	FMDB_Tag_GetSize(tag, &tag_size);

        strncpy(entTag->tagName, name, TAG_NAME_SIZE-1);
	entTag->tagName[TAG_NAME_SIZE-1]='\0';
        n_bytes+=sizeof(packedTag);

	if(tag_size==1)
	 {
	  dbl_val=(double*)((char*)msg_send + n_bytes);
	  double tag_val;
	  FMDB_Ent_GetDblTag(mesh, ent, tag, &tag_val);
	  *dbl_val=tag_val;
	  n_bytes+=sizeof(double);
	 }
	else
        {
	  double* dbl_val_arr=new double[tag_size];
          FMDB_Ent_GetDblArrTag(mesh, ent, tag, &dbl_val_arr, &tag_size);
	  for(int i=0; i<tag_size; i++)
	   {
	    dbl_val=(double*)((char*)msg_send + n_bytes);     
	    *dbl_val=dbl_val_arr[i];
	    n_bytes+=sizeof(double);
          }
	 delete[] dbl_val_arr;
        }
     }
   }

 if(user_size)
 {
   memcpy((void*)((char*)msg_send + entity_size),buf_user,user_size);
   cb.deleteUserData(buf_user);
 }
}

void unpack_Ghost_MEntity(mMesh* mesh, int ghostDim, void *msg_recv, int msg_size, int pid,
                  pmMigrationCallbacks &cb, list<rc_struct_2>& entitiesToBounce)
{
  int n_bytes = 0;
  int mypid = ParUtil::Instance()->rank();
  mEntity *ent, *sender, *ownerEnt;
  int ownerPid;
  packedGhostEnt* castbuf = (packedGhostEnt*)msg_recv;
  n_bytes+=sizeof(packedGhostEnt);

  vector<pPart> vecPart;
  FMDB_Mesh_GetAllPart(mesh, mypid, vecPart);
  pPart part=vecPart.at(0);
  int numTags=0, tag_size=0;
  numTags = castbuf->numTags;

  if (castbuf->numDE==0)
  {
   // STEP 1: create entity
   packedGhostVertex *cast_vtx = (packedGhostVertex*)((char*)msg_recv + n_bytes);
   n_bytes += sizeof(packedGhostVertex);
   ent = part->createVertex(cast_vtx->iD, cast_vtx->coord[0], cast_vtx->coord[1], cast_vtx->coord[2],
                                  part->getGEntity(castbuf->gTag, castbuf->gType));
   sender = castbuf->sender;
   ownerPid = castbuf->ownerPid;
   ownerEnt = castbuf->ownerEnt;
   short int numRC = cast_vtx->numRC;
   if (numRC>0)
    {
      int RCs_pid;
      pMeshEnt RCs_ent;
      for (short int i=0; i < numRC; i++)
      {
       RCs_pid = *(int*)((char*)msg_recv + n_bytes);
       n_bytes += sizeof(int);
       RCs_ent = *(mEntity**)((char*)msg_recv + n_bytes);
       n_bytes += sizeof(mEntity*);
       entitiesToBounce.push_back(rc_struct_2(ent, RCs_pid, RCs_ent));	  
      }
    }
  }
 else
 {
   pGEntity g = part->getGEntity(castbuf->gTag, castbuf->gType);
   sender = castbuf->sender;
   ownerPid = castbuf->ownerPid;
   ownerEnt = castbuf->ownerEnt;

   // STEP 1: create an entity
   int numDE = castbuf->numDE;
   mEntity** DEs = (mEntity**)((char*)msg_recv + n_bytes);
   int ierr;
   switch (castbuf->topo)
   {
   case FMDB_LINE:
   {
    ierr=FMDB_Edge_Create(part,g,DEs[0],DEs[1],ent);
   }
   break;
   case FMDB_TRI:
   case FMDB_QUAD:
   {
     ierr=FMDB_Face_Create(part,g,castbuf->topo,DEs,0,ent);
   }
   break;
   case FMDB_TET:
   case FMDB_HEX:
   case FMDB_PRISM:
   case FMDB_PYRAMID:
   {
    ierr=FMDB_Rgn_Create(part,g,castbuf->topo,numDE,DEs,ent);
   }
   break;
   default :
   {  
    cout<<"unpack_Ghost_MEntity error: unpacking not done for this type of mesh entity\n";
    return;
   }
   }
#ifdef DEBUG
  if(ierr == SCUtil_ENTITY_ALREADY_EXISTS)
   cout << "["<<SCUTIL_CommRank()<<"] "<<__func__<<" WARNING: Entity "<<FMDB_Ent_StrID(ent)<<" already exist\n";
#endif
  n_bytes += numDE*sizeof(mEntity*);
 }
  if(numTags>0)
   {
    for(int i=0; i<numTags; i++)
    {
      pTag tag;
      packedTag* taggedEnt=(packedTag*)((char*)msg_recv + n_bytes);
      FMDB_Mesh_FindTag(mesh, taggedEnt->tagName, tag);
      n_bytes += sizeof(packedTag);
      if(tag)
      {
        FMDB_Tag_GetSize(tag, &tag_size);
        if(tag_size==1)
	 {
	  double* dbl_val=(double*)((char*)msg_recv + n_bytes);
	  FMDB_Ent_SetDblTag(mesh, ent, tag, *dbl_val);
	  n_bytes+=sizeof(double);
	 }
       else
       {
         double* dbl_val_arr=new double[tag_size];
         for(int i=0; i<tag_size; i++)
          {
	    dbl_val_arr[i]=*(double*)((char*)msg_recv + n_bytes);	
	    n_bytes+=sizeof(double);
          }
         FMDB_Ent_SetDblArrTag(mesh, ent, tag, dbl_val_arr, tag_size);
	 delete[] dbl_val_arr;
       }
      }
      else
        cout << "\n FMDB Warning: Tag does not exist " << endl;
      }
    }
 // set ghost flag of downward adjacent entities
  vector<pMeshEnt> vecAdjEnt;
  if(ent->getLevel() > 1)
   {
    for(int lvl = 1; lvl < ent->getLevel(); lvl++)
    {
     FMDB_Ent_GetAdj (ent, lvl, 0, vecAdjEnt);
     for (vector<pMeshEnt>::iterator It = vecAdjEnt.begin(); It != vecAdjEnt.end(); It++)
      if(!(*It)->getPClassification())
       {
        (*It)->setGhostIndex(part->getGhostVecSize());
       }
     vecAdjEnt.clear();
    }
   }
  //Add sender to ghost copy
  createGhostMapping(part, ent, ownerPid, IS_GHOST, ownerEnt);

  //Add sender to remote copy and store entity to bounce
  if (ent->getLevel() == 0)                                      
    entitiesToBounce.push_back(rc_struct_2(ent, pid, sender));
  else
   {
    entitiesToBounce.push_back(rc_struct_2(ent, ownerPid, ownerEnt));
   }

  IPComMan *CM = ParUtil::Instance()->ComMan();
 // Process user attached data
  cb.receiveUserData(ent,pid,CM->get_tag(),(void*)((char*)msg_recv + n_bytes));
}

/* deletes existing ghost entities from dimension gDim to 0*/
void deleteGhostEntities(pMeshMdl mesh)
{
  double t1, t2;
  /* Get current process id */
   int procId=SCUTIL_CommRank(), mDim;
  /* Get all parts in the current mesh instance. Right now we have one part per mesh instance */
  std::vector<pPart> vecPart;
  FMDB_Mesh_GetAllPart(mesh, procId, vecPart);
  pPart part = vecPart.at(0);
  std::vector<mGhost*> vecGhost=part->getGhostVec();
  mGhost* ghost=NULL;
  mGhost* downGhost=NULL;
  pMeshEnt meshEnt=NULL;
  FMDB_Mesh_GetDim(mesh, &mDim);
  set<pMeshEnt>* entitiesToRemove = new set<pMeshEnt>[mDim + 1];
  int numProc=SCUTIL_CommSize();

  t1 = MPI_Wtime();
  for(std::vector<mGhost*>::iterator iter=vecGhost.begin(); iter!=vecGhost.end(); ++iter)
    {
      ghost=*iter;
      meshEnt=ghost->getOwner();
      meshEnt->setGhostIndex(-1); 
      if(ghost->isGhost())
        entitiesToRemove[meshEnt->getLevel()].insert(meshEnt);
      for(int dim = 1; dim < meshEnt->getLevel(); dim++)
       {
         vector<pMeshEnt> vecDownEnts;
         FMDB_Ent_GetAdj(meshEnt, dim, 0, vecDownEnts);
         for (vector<pMeshEnt>::iterator veciter = vecDownEnts.begin(); veciter!=vecDownEnts.end();++veciter)
          {
            pEntity pDownEnt = *veciter;	
            downGhost=part->getGhost(pDownEnt->getGhostIndex());
	    if(downGhost && downGhost->isGhost())
             entitiesToRemove[dim].insert(pDownEnt);
            pDownEnt->setGhostIndex(-1);
         }
       }
    }
  removeEntities(part, entitiesToRemove, mDim);
  std::vector<mGhost*>::iterator iter=vecGhost.begin();
  while(iter!=vecGhost.end())
  {
    ghost=*iter;
    iter++;
    delete ghost;
  }
  mesh->clearGhostInfo();    // clear ghosting history

#ifdef DEBUG
  t2 = MPI_Wtime();
  double maxTime = 0, avgTime, localTime = t2 - t1;
  MPI_Allreduce(&localTime,&maxTime,1,MPI_DOUBLE,MPI_MAX,MPI_COMM_WORLD);
  MPI_Allreduce(&localTime,&avgTime,1,MPI_DOUBLE,MPI_SUM, MPI_COMM_WORLD);
  avgTime=avgTime/numProc;

  if(!procId)
   cout << "\n[Delete Ghost] max deletion time:" << maxTime << "(sec), avg time:" << avgTime 
        <<"(sec)\n";
#endif
}

void removeEntities(pPart part, set<pMeshEnt>* entitiesToRemove, int meshDim)
{
 for(int dim = meshDim; dim >= 0; dim-- )  
 {
  for(set<pMeshEnt>::iterator it = entitiesToRemove[dim].begin(); it != entitiesToRemove[dim].end(); it++ )
   FMDB_Ent_Del(part, *it);
 } // end for dim
} // end fn
#endif
