/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "iMeshP.h"
#include "FMDB.h"
#include "mpi.h"
#include "iMeshP_req.h"

void iMeshP_waitForRequest(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            iMeshP_RequestHandle request,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  int numReq = request->reqs_size;
  MPI_Status *status = (MPI_Status*) malloc(sizeof(MPI_Status));
  
  for (int i = 0; i < numReq; i++)
    MPI_Wait(request->reqs[i], status);
}

void iMeshP_waitForAnyRequest(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            iMeshP_RequestHandle *requests,
            int requests_size,
            int *index,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  int numReq, flag;
  bool isDone;
  MPI_Status status;
  MPI_Request *req;
  
  while(1) {
    for (int i = 0; i < requests_size; i++) 
    {
      isDone = true;
      numReq = (requests[i])->reqs_size;
      for (int j = 0; j < numReq; j++)
      { 
        req = ((requests[i])->reqs)[j];

        MPI_Test(req, &flag, &status);
        
        if (!flag) 
          isDone = false;
      }

      if (isDone) {
        *index = i;
        return;
      }
    }
  }
}


void iMeshP_waitForAllRequests(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            iMeshP_RequestHandle *requests,
            int requests_size,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
  
  int numReq;
  MPI_Status *status;
 
  for (int i = 0; i < requests_size; i++) {
    numReq = (requests[i])->reqs_size;
    status = (MPI_Status*) malloc(sizeof(MPI_Status));
    for (int j = 0; j < numReq; j++)
      MPI_Wait(requests[i]->reqs[j], status);
  }
  
}   


void iMeshP_waitForRequestEnt(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            iMeshP_RequestHandle request,
            iBase_EntityHandle **out_entities,
            int *out_entities_allocated,
            int *out_entities_size,
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
  
  int numReq = request->reqs_size;
  MPI_Status *status = (MPI_Status*) malloc(sizeof(MPI_Status));
  
  for (int i = 0; i < numReq; i++)
    MPI_Wait(request->reqs[i], status);

  std::set<iBase_EntityHandle*> entities;
  
  for (int i = 0; i < request->ents_size; i++)
    entities.insert((iBase_EntityHandle*)(request->modified_entities[i]));

  if ( *out_entities == 0 || *out_entities_allocated == 0)
  {
    *out_entities = (iBase_EntityHandle*)(calloc(entities.size(), sizeof(iBase_EntityHandle) ));
    if (*out_entities == 0)
    {
      *err = iBase_MEMORY_ALLOCATION_FAILED;
      FMDB_Mesh_SetLastError((pMeshMdl)instance, *err);
      return;
    }
    *out_entities_allocated = entities.size();
  }
  else if (*out_entities_allocated < entities.size())
  {
    *err = iBase_BAD_ARRAY_SIZE;
    FMDB_Mesh_SetLastError((pMeshMdl)instance, *err);
    return;
  }

  int count = 0;
  std::set<iBase_EntityHandle*>::iterator it;
  for (it = entities.begin(); it != entities.end(); it++)
  {
    (*out_entities)[count] = (iBase_EntityHandle)(*it);
    count ++;
  }

  *out_entities_size = entities.size();
}

void iMeshP_testRequest(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            iMeshP_RequestHandle request,
            int *completed,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  int numReq = request->reqs_size;
  MPI_Status *status = (MPI_Status*) malloc(sizeof(MPI_Status));
  int flag = 0;
  
  for (int i = 0; i < numReq; i++)
  {
    MPI_Test(request->reqs[i], &flag, status);
    if (flag)
    {
      *completed  = flag;
      return;
    }
  }
  *completed = 0;
}


void iMeshP_pollForRequests(
            iMesh_Instance instance,
            iMeshP_PartitionHandle partition,
            iMeshP_RequestHandle **requests_completed,
            int *requests_completed_allocated,
            int *requests_completed_size,
            int *err)
{
  *err = iBase_NOT_SUPPORTED; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
}
