/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "iMeshP.h"
#include "FMDB.h"

void iMeshP_addGhostOf(
            iMesh_Instance instance, 
            const iMeshP_PartitionHandle partition,
            const iMeshP_Part target_part_id,
            iBase_EntityHandle entity_to_copy,
            iMeshP_RequestHandle *request,
            int *err)
{
  *err = iBase_NOT_SUPPORTED; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
  return;
}

void iMeshP_rmvGhostOf(
            iMesh_Instance instance, 
            const iMeshP_PartitionHandle partition,
            const iMeshP_Part target_part_id,
            iBase_EntityHandle copy_to_purge,
            int *err)
{
  *err = iBase_NOT_SUPPORTED; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
  return;
}


void iMeshP_createGhostEntsAll(
            iMesh_Instance instance,
            iMeshP_PartitionHandle partition,
            int ghost_type,
            int bridge_type,
            int num_layers,
            int include_copies,
            int *err)
{
  MigrCB cb;
  *err = FMDB_Mesh_CreateGhost ((pMeshMdl)instance, cb, ghost_type, bridge_type, 
                         num_layers, include_copies);
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
  return;
}

void iMeshP_deleteGhostEntsAll(
            iMesh_Instance instance,
            iMeshP_PartitionHandle partition,
            int *err)
{
  *err = FMDB_Mesh_DelGhost ((pMeshMdl)instance);
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  return;
}

void iMeshP_ghostEntInfo(
            const iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            int *ghost_rules_allocated, 
            int *ghost_rules_size, 
            int **ghost_type,
            int **bridge_type,
            int **num_layers,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  std::vector<int> ghost_rules;
  FMDB_Mesh_GetGhostInfo ((pMeshMdl)instance, ghost_rules);
  int numGhostRules = ghost_rules.size()/3;

  if( *ghost_type==0 || *ghost_rules_allocated==0 ) 
  {
    *ghost_type = (int*)(calloc(numGhostRules, sizeof(int)));
    *bridge_type = (int*)(calloc(numGhostRules, sizeof(int)));
    *num_layers = (int*)(calloc(numGhostRules, sizeof(int)));
    if (*ghost_type==0 || *bridge_type==0 ||*num_layers==0) 
    {
      *err = iBase_MEMORY_ALLOCATION_FAILED;
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }
    *ghost_rules_allocated= numGhostRules;
  }
  else if (*ghost_rules_allocated < numGhostRules)
  {
    *err = iBase_BAD_ARRAY_SIZE;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  *ghost_rules_size = numGhostRules;
  for (int counter=0; counter<numGhostRules; ++counter)
  {
    (*ghost_type)[counter]= ghost_rules.at(counter*3);
    (*bridge_type)[counter]= ghost_rules.at(counter*3+1);
    (*num_layers)[counter]= ghost_rules.at(counter*3+2);
  }
  return;
}
