/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "iMeshP.h"
#include "FMDB.h"
#include "pmDataExchanger.h"
#include "iMeshP_req.h"

//////////////////////////////////////////////
class idExchanger_vtx : public pmDataExchanger
{
public:
  virtual int msg_size();
  virtual void fill_buffer(mEntity *e, mEntity*remoteEnt, void *&msg_send);
  virtual void receiveData(void *msg_recv, int pid_from);
};
  
int idExchanger_vtx :: msg_size()
{
  return 3*sizeof(double) + sizeof(mEntity*);
}

void idExchanger_vtx :: fill_buffer(mEntity *e, mEntity*remoteEnt, void *&msg_send)
{
  double* xyz = (double*)msg_send; 
  V_coord(e, xyz);
  mEntity** s_ent = (mEntity**)((char*)msg_send + sizeof(double)*3);
  *s_ent = remoteEnt; 
}

void idExchanger_vtx :: receiveData(void *msg_recv, int pid_from)
{
  double* xyz = (double*)msg_recv;
  mEntity** s_ent = (mEntity**)((char*)msg_recv + sizeof(double)*3);
  ((mVertex*)(*s_ent))->setCoord(xyz[0], xyz[1], xyz[2]);
}

class rcExchanger : public pmDataExchanger
{
  pMeshDataId id;
  int m_size;

public :
  rcExchanger(pMeshDataId i): id(i), m_size(0) {}
  virtual int msg_size();
  virtual void fill_buffer (mEntity *e, int, int, mEntity *remoteEnt, void *&msg_send);
  virtual void receiveData (void *msg_recv, int pid_from);
};

int rcExchanger::msg_size ()
{
  return m_size;
}

void rcExchanger::fill_buffer (mEntity *e, int src, int dest, mEntity *remoteEnt, void *&msg_send)
{
  pPList ents;
  void* tmp_ptr;
  if (!EN_getDataPtr(e, id, &tmp_ptr))
    throw 1;
  ents = (pPList) tmp_ptr;

  int enities_size = PList_size(ents);
  m_size = sizeof(pEntity) * (enities_size + 1);

  msg_send = ParUtil::Instance()->ComMan()->alloc_msg(m_size); 

  pEntity* entities = (pEntity*)msg_send;
  int icount = 0;
  entities[icount++] = remoteEnt;
  for(int i = 0; i < enities_size; i++) {
    entities[icount++] = (pEntity)PList_item(ents, i);
  }
}

void rcExchanger::receiveData (void *msg_recv, int pid_from)
{  
  pEntity* entities = (pEntity*)msg_recv;
  pEntity ent;

  int icount = 0;
  pEntity e = entities[icount++];
  pPList ents;
  void* tmp_ptr;
  if (!EN_getDataPtr(e, id, &tmp_ptr))
     throw 1;
  ents = (pPList) tmp_ptr;

  int enities_size = PList_size(ents);
  for(int i = 0; i < enities_size; i++)
  {
    ent = (pEntity)PList_item(ents, i);
    ent->addRemoteCopy(pid_from, entities[icount++]);
    ent->setPClassification(e->getPClassification());
  }
}

void updatePartBdry(list<pEntity> &ents_to_updt, pMeshDataId id)
{
  rcExchanger cb(id);
  genericDataExchanger(ents_to_updt.begin(), ents_to_updt.end(), cb);
}
//////////////////////////////////////////////


void iMeshP_exchEntArrToPartsAll(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iBase_EntityHandle *entities,
            const int entities_size,
            const iMeshP_Part *target_part_ids,
            int command_code,  
            int update_ghost,
            iMeshP_RequestHandle *request,
            int *err)
{
  MigrCB cb;
  std::map<mEntity*, std::pair<int, int> > EntToMigr;
  std::map<pPartSet, std::pair<int, int> > SetToMigr;
  for(int i=0; i<entities_size; i++) 
  {
    if( ((int*)target_part_ids)[i]==ParUtil::Instance()->rank() )
      continue; 
    EntToMigr.insert(std::map<mEntity*, std::pair<int, int> >::value_type(((mEntity**)entities)[i], std::make_pair(SCUTIL_CommRank(), ((int*)target_part_ids)[i])));
  }
  *err = FMDB_Mesh_Migr((pMeshMdl)instance, cb, EntToMigr, SetToMigr);
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  return;
}

void iMeshP_migrateEntity(
            iMesh_Instance instance, 
            const iMeshP_PartitionHandle partition,
            iMeshP_PartHandle part,
            iBase_EntityHandle local_entity,
            iMeshP_RequestHandle *request,
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  /// get mesh dimension
  int meshDim;
  FMDB_Mesh_GetDim((pMeshMdl)instance, &meshDim);

  // allocate request
  if (request == 0)
  {
    request = (iMeshP_RequestHandle*) malloc(sizeof(iMeshP_RequestHandle));
    if (request == 0)
    {
      *err = iBase_MEMORY_ALLOCATION_FAILED;
      FMDB_Mesh_SetLastError((pMeshMdl)instance, *err);
      return;
    }
  }

  std::vector<pMeshEnt> newEnts[meshDim + 1];
  std::vector<pMeshEnt> rmvEnts[meshDim + 1];
  std::vector<MPI_Request*> newReqs;

  // make appropriate call to FMDB
  MigrCB cb;
  *err = FMDB_Ent_Migr((pMeshMdl) instance, cb, (pMeshEnt) local_entity, SCUTIL_CommRank()/*source_partid*/, 
                SCUTIL_CommRank()/*dest_partid*/, newEnts, rmvEnts, newReqs);
  if (*err!=iBase_SUCCESS)
  {
    FMDB_Mesh_SetLastError((pMeshMdl)instance, *err);
    return;
  }

  std::vector<pMeshEnt> modEnts;
  for (int i = FMDB_VERTEX; i <= meshDim; i++)
  {
    std::copy(newEnts[i].begin(), newEnts[i].end(), std::back_inserter(modEnts));
  }

  // build iMeshP_Request
  (*request)->reqs = (MPI_Request**) malloc(sizeof(MPI_Request*) * newReqs.size());
  (*request)->reqs_size = newReqs.size();
  (*request)->modified_entities = (iBase_EntityHandle*) malloc(sizeof(iBase_EntityHandle) * (modEnts.size()));
  (*request)->ents_size = modEnts.size();

  // add all entities
  for (int i = 0; i < modEnts.size(); i++)
    ((*request)->modified_entities)[i] = (iBase_EntityHandle)(modEnts[i]);

  // add all MPI_Requests
  for (int i = 0; i < newReqs.size(); i++)
    ((*request)->reqs)[i] = newReqs[i];
}

void iMeshP_updateVtxCoords(
            iMesh_Instance instance, 
            const iMeshP_PartitionHandle partition,
            const iBase_EntityHandle local_vertex_handle,
            int *err)
{
  *err = iBase_NOT_SUPPORTED; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
/*
  mEntity* ent = (mEntity*) local_vertex_handle; 
  if( !EN_duplicate(ent)) {
    *err = iBase_INVALID_ENTITY_HANDLE; 
    return; 
  }

  idExchanger_vtx cb; 
  list<mEntity*> entities; 

  entities.push_back(ent);  
  
  genericDataExchanger(entities.begin(), entities.end(), cb);
*/
}

void iMeshP_replaceOnPartBdry(
            iMesh_Instance instance, 
            const iMeshP_PartitionHandle partition,
            const iBase_EntityHandle *old_entities,
            const int old_entities_size,
            const iBase_EntityHandle *new_entities,
            const int new_entities_size,
            const int *offset,
            const int offset_size,
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  // check the sizes of the arrays 
  if( offset_size!= old_entities_size+1 ) {
    *err = iBase_BAD_ARRAY_SIZE; 
    return; 
  }

  for(int i=1; i<offset_size; i++) {
    if( (offset[i]-offset[i-1])<0 ) {
      *err = iBase_BAD_ARRAY_SIZE; 
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return; 
    }
  }
  if( offset[offset_size-1]!=new_entities_size ){
    *err = iBase_BAD_ARRAY_SIZE; 
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return; 
  }

  // check if the entities are on any part boundary
  mEntity* ent; 
//   for(int i=0; i<old_entities_size; i++) {
   
//     ent = (mEntity*)old_entities[i]; 
//     if( ent==0 || !EN_duplicate(ent)) {
//       *err = iBase_INVALID_ENTITY_HANDLE; 
//       return; 
//     }
//   }

  
  pMeshDataId id = MD_newMeshDataId("new_entities_list");


  list<pEntity> entities; 
  list<pEntity>* nlist; 
  for(int i=0; i<old_entities_size; i++) {
    ent = (mEntity*)old_entities[i];  
    entities.push_back(ent);

    nlist = new list<mEntity*>; 
    for(int j=offset[i]; j<offset[i+1] ; j++) {
      nlist->push_back( (mEntity*)new_entities[j] );
    }
    EN_attachDataPtr(ent, id, nlist);
    
  }

  updatePartBdry(entities, id);

  // delete lists and attached data 
  void* tmp_ptr;  
  for(int i=0; i<old_entities_size; i++) {
    ent = (mEntity*)old_entities[i];  

    if (!EN_getDataPtr(ent, id, &tmp_ptr))
     throw 1;
    nlist = (list<pEntity>*) tmp_ptr; 

    delete nlist; 
    EN_deleteData(ent, id); 
  }

  MD_deleteMeshDataId(id);
}
 
void iMeshP_syncMeshAll(
            iMesh_Instance instance, 
            iMeshP_PartitionHandle partition,
            int *err)
{
  *err = FMDB_Mesh_SyncPtn((pMeshMdl)instance); 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
  return;
}
                            














