/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#ifndef _FMDB_C_INTERNALS_H_
#define _FMDB_C_INTERNALS_H_

#include <stdio.h>
#ifdef SIM
#include "MeshSim.h"
#include "MeshSimInternal.h"
#else
#include "oldFMDB.h"
#include "FMDBInternals.h"
#endif

#include <vector>
#include <list>
#include <iostream>

#ifdef __cplusplus
extern "C" {
#endif

  int M_create_ids(pMesh);  
  int M_remove_ids(pMesh);  

  int M_dimension(pMesh);
  enum mType {VERTEX,EDGE,TRI,QUAD,HEX,PRISM,PYRAMID,TET};
  mType   M_GetElementType   (pEntity pE);
  void    M_GetVertices      (pEntity pE, std::vector<pVertex> &);
  int     R_edgeDir(pRegion , int n);
  double  M_SizeMin (pEntity);

  /// returns the maximal vertex id for all partitions
  int     M_MaximumVertexId (pMesh);
  /// returns how many remote entities exists for a mesh entity
  int     M_numRemoteCopies (pMesh, pEntity);
  pPList  M_getRemoteCopies (pMesh, pEntity);
  int     M_getRemotePID (void *);
  int     M_OwnerProc (pMesh,pEntity);

  void V_setWhatInType (pMesh pm, pVertex, int);//  categorize a mesh vertex to a model entity type
  void E_setWhatInType (pMesh pm, pEdge, int);  //  categorize a mesh edge to a model entity type
  void F_setWhatInType (pMesh pm, pFace, int);  //  categorize a mesh face to a model entity type

  /// retrieve some entities from model tag, id and entity dimensions
  void M_getEntitiesClassified (int gEntityDim, 
				int gEntityTag, 
				int mEntityDim, 
				std::list<pEntity> & myList); 

  int E_getWhatInType (pEntity e);
  int E_getWhatInTag (pEntity e);

#ifndef SIM

// 
//    Added by Eunyoung Seol - Oct 05, 2002
//    Modified 		     - Jun 10, 2003
// 

// **************************
// mesh loading
// **************************
  void M_setRepresentationFlag(pMesh, int i);

// **************************
// mesh info
// **************************
  int M_getMaxDim(pMesh);
  int M_NumUniqueVertices(pMesh);
  int M_NumUniqueEdges(pMesh);
  int M_NumUniqueFaces(pMesh);
  int M_NumUniqueRegions(pMesh);
  
// **************************
//  common boundary entity 
// **************************
  pEntity EN_getRemoteCopy(pMesh, pEntity, int);
  void EN_getRemoteCopies(pMesh,pEntity,
                         std::vector<std::pair<pEntity,int> >&);
  void EN_addRemoteCopy(pMesh,pEntity,int,pEntity);
  void EN_deleteRemoteCopy(pMesh,pEntity);
  pmEntity* EN_getCommonBdry(pEntity);
  void EN_setCommonBdry(pEntity, pmEntity*);
  
// **************************
//  entity info  
// **************************
  int EN_ownerProc(pEntity); 
  bool EN_onPB(pMesh,pEntity ent);      // return true if the entity is on PB
  bool EN_onCB(pEntity); 
  const char* EN_getUid(pEntity ent);	// return unique id

// **************************
// Mesh migration
// **************************
  // load balancer that uses default loadbalancer callback that FMDB defined
  int M_LoadBalance3(pMesh);  
  
  // load balancer that uses load balancer callback that is given by input
  int M_LoadBalance2(pMesh,FMDB_LoadBalancerCallbacks &);

  void M_setEntityOwnership(pMesh);
  
// **************************
// general parallel util
// **************************
  void M_sync();	//synchronization
  int M_Pid();		// return pid
  int M_NumPE();	// return number of processes
  int P_numPE();
  void P_sync();
  // Unify the information max and min through all processors
  void M_unify(double* max, double* min); 
  void M_MergeArray(std::vector<int>&); 
  int M_getMaxNum(int num);

// **************************
// miscellaneous - asked by skocak, May 01, 2002
// **************************
  void M_AssignUniqueRange (pMesh pm, 
		           pMeshDataId tag1,
			   pMeshDataId tag2, 
			   std::vector<pEntity> &l1,
			   std::vector<int> &l2, 
			   int initialId);

#endif   /* ifndef SIM */

#ifdef __cplusplus
}
#endif
#endif
