/******************************************************************************

  (c) 2004-2012 Scientific Computation Research Center,
      Rensselaer Polytechnic Institute. All rights reserved.

  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.

*******************************************************************************/
#ifndef _MGHOST_H_
#define _MGHOST_H_

#include "mEntity.h"
#include "ParUtil.h"


using std::cout;

enum{IS_GHOST, HAS_GHOSTS};
class mGhost
 {
 #ifdef FMDB_PARALLEL

  protected:
   typedef std::map<int, mEntity*> ghostCopyMap;
   ghostCopyMap* theGhostCopies;
   std::set<int>* tempGPs;
   mEntity* ownerEnt;
   short ghostStatus;

  public:
   typedef std::map<int, mEntity*>::iterator GCIter;
   typedef std::set<int>::iterator GPIter;
   
  mGhost();

  void setGPs(std::set<int>& tmpGPs)
   {
      if(tempGPs==NULL)
        tempGPs = new std::set<int>;

      for(std::set<int>::iterator itSet = tmpGPs.begin(); itSet != tmpGPs.end(); itSet++)
        tempGPs->insert(*itSet);
   }
  void addGPs(int i)
   {
       if(tempGPs==NULL)
         tempGPs = new std::set<int>;
       tempGPs->insert(i);
    }
   void deleteGPs(int i)
     {
       if(tempGPs==NULL)
         return;
       tempGPs->erase(i);
       if(tempGPs->empty())
         SAFE_DELETE(tempGPs);
     }
    void clearGPs()
     {
       if(tempGPs==NULL)
         return;
       if(!tempGPs->empty()) tempGPs->clear();
     }
    int getNumGPs()
     {
       if(tempGPs==NULL)
         return 0;
       return tempGPs->size();
     }
   bool findPidInGPs(int pid)
     {
       if(tempGPs==NULL)
         return false;
       return (std::binary_search(tempGPs->begin(), tempGPs->end(),pid));
     }
   int getNumGhostCopies()
     {
       if(theGhostCopies==NULL)
         return 0;
       return theGhostCopies->size();
     }
   mEntity* getOwner()
     {
      return ownerEnt;
     }	
   void setOwnerEnt(mEntity* ent)
     {
      ownerEnt=ent;
     }
   void setGhostStatus(short gStatus)
     {
     if(gStatus == IS_GHOST || gStatus ==HAS_GHOSTS)
      ghostStatus=gStatus;
     }
   bool isGhost()
    {
      if(ghostStatus==IS_GHOST)
       return true;
      return false;
    }
   bool hasGhosts()
    {
      if(ghostStatus==HAS_GHOSTS)
       return true;
      return false;
    }

   std::set<int>& getGPs() { return *tempGPs; }
   virtual GPIter gpBegin() { return tempGPs->begin(); }
   virtual GPIter gpEnd() { return tempGPs->end(); }
   void printGPs();
   mEntity* getGhostCopy(int);                                // global part id
   void getGhostCopies(std::map<int, mEntity*>&);
   void addGhostCopy(int, mEntity*);
   void deleteGhostCopy(int);
   void deleteGhostCopy(GCIter rciter);
   void clearGhostCopies();
   virtual GCIter gcBegin() { return theGhostCopies->begin(); }
   virtual GCIter gcEnd() { return theGhostCopies->end(); }
   void printGCs();
 #endif
 };
#endif
