/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#ifndef _MMESH_H_
#define _MMESH_H_
#include <vector>
#include "GenTag.h"
#include "mPart.h"
#include "mEntitySet.h"

using std::vector;

class mMesh: public TagHolder
{
private:
  pGModel theGeomModel; 
  // tags for ITAPS (temporary - must be revisited)
  std::map<char*, unsigned int> name_tag_map;
  std::map<unsigned int, std::pair<int, int> > tag_att_map; // pair: type and size

public:
  std::vector<mPart*>*  parts;
  short int MRM[4][4];  // MRM for itaps 
#ifdef FMDB_PARALLEL
  int LB_Method; 
  int LB_Approach;
  double imbalance_tol; 
  int debug_level;
#endif
  typedef vector<mPart*>::iterator partIter;
  mMesh(pGModel geom); 
  ~mMesh();
  void clearPart();
  void addPart(mPart* part);  
  void delPart(mPart* part);
  unsigned int getNumPart() {return parts->size(); }
  pGModel getGeomModel()  {  return theGeomModel;  }
  partIter partBegin()  {  return parts->begin(); }
  partIter partEnd()  {  return  parts->end();  }
  mPart* getPart(int ith);

  // tag
  bool findTagID (unsigned int tag_id);
  int getTagID (const char *tag_name, unsigned int* tag_id);
  int nameTagID (unsigned int tag_id, char* name);
  int typeTagID (unsigned int tag_id, int* type);
  int sizeTagID (unsigned int tag_id, int* size);
  int createTagID (const char *tag_name, int type, int size, unsigned int* tag_id);
  int deleteTagID (unsigned int tag_id);
  void getAllTagID (std::vector<unsigned int>& ids);
  
  // entity set
  std::list<EntitySet<mEntity> * > allEntSets;   // this can be improved through dynamic memory allocation
  void addEntSet(EntitySet<mEntity> *p) { allEntSets.push_back(p);}    
  void removeEntSet(EntitySet<mEntity> * p); 
  std::list<EntitySet<mEntity>* >::iterator beginEntSet() {return allEntSets.begin(); }
  std::list<EntitySet<mEntity>* >::iterator endEntSet() {return allEntSets.end(); }
  bool hasEntSet(EntitySet<mEntity> *p)  
  {
    std::list<EntitySet<mEntity>* >::iterator loc = std::find(allEntSets.begin(), allEntSets.end(),p); 
    if(loc==allEntSets.end())
      return false; 
    return true; 
  }
  void clearEntSet(); 

#ifdef FMDB_PARALLEL
  // ghosting
  std::vector<int> vecGhostRule;

  ///Functions to query ghost information in the part
  void addGhostInfo(int gType, int bType, int numLyr);
  void clearGhostInfo();
  void getGhostInfo(std::vector<int>&);
#endif
  // last error happened in mesh
  int last_error;
};

// function to create/delete ghost entities per mesh
int createGhost(mMesh* mesh, pmMigrationCallbacks & migrCB, int gDim, int bDim, int numLayer, int include_copy);
void deleteGhostEntities(mMesh* mesh);
#endif // _MMODEL_H_
