/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#ifndef MMIGRATEUTIL_H
#define MMIGRATEUTIL_H

#include "mPart.h"
#include "mEntity.h"
#include "mFMDB.h"
#include "mAttachableDataContainer.h"
#include "FMDB_LoadBalancer.h"
#include "mException.h"
#include <assert.h>
#include <vector>
#include <set>
#include <list>
#include "ParUtil.h"


int getPid_withBalance(mPart* mesh, mEntity* ent, std::vector<int>&);

#ifdef FMDB_PARALLEL
const int TAGVERTEX    = 222;
const int TAGNONVERTEX = 333;
#endif


static void attachIntVector (mEntity *e, bool mark, unsigned int mk,int part)
{
  if(!mark)
    {
      if(e->getData(mk))e->deleteData(mk);
    }
  else
    {
      mAttachableIntVector *avec;
      if(e->getData(mk))
	{
	  avec = (mAttachableIntVector*)e->getData(mk);
	}
      else
	{
	  avec = new mAttachableIntVector;
	  e->attachData(mk,avec);
	}
      for(int i=0;i<avec->v.size();i++)if(avec->v[i] == part)return;
      avec->v.push_back(part);
    }
}

/*
  e is going to be migrated
  markAllSubs tag entities that are going to be migrated
  e is initially an entity that is to be migrated, we look
  after all sub-entities and tag them also.
*/

static void markAllSubs(mEntity *e, bool mark, int part)

{
  int n = e->getLevel();
  
  unsigned int tagMark = FMDB_Util::Instance()->lookupMeshDataId("_dest");

  attachIntVector (e, mark, tagMark, part);

  std::set<mEntity*> theWholeFamily;
  e->getAllEntitiesOfAllDimensions(theWholeFamily);
  for(std::set<mEntity*>::iterator it2 = theWholeFamily.begin(); 
      it2 != theWholeFamily.end();++it2)
    {
      mEntity *sub = *it2;
      attachIntVector (sub, mark, tagMark, part);
    }
}

static void markAllSubs_oneLevel(mEntity *e, bool mark, int part)
{

  int n = e->getLevel();
 
  unsigned int tagMark = FMDB_Util::Instance()->lookupMeshDataId("_dest");
 
  attachIntVector (e, mark, tagMark, part);
 
  std::set<mEntity*> theWholeFamily;
    e->getAllEntitiesOfAllDimensions_oneLevel(theWholeFamily);
  for(std::set<mEntity*>::iterator it2 = theWholeFamily.begin();
      it2 != theWholeFamily.end();++it2)
    {
      mEntity *sub = *it2;
      attachIntVector (sub, mark, tagMark, part);
    }

}

static int getNbDests(mEntity *e)
{
  unsigned int tagMark = FMDB_Util::Instance()->lookupMeshDataId("_dest");
  if(e->getData(tagMark))
    {
      mAttachableIntVector *avec;
      avec = (mAttachableIntVector*)e->getData(tagMark);
      return avec->v.size();
    }
  else
    {
      return 0;
    }
}

static int getDest(mEntity *e, int i)
{
  unsigned int tagMark = FMDB_Util::Instance()->lookupMeshDataId("_dest");
  if(e->getData(tagMark))
    {
      mAttachableIntVector *avec;
      avec = (mAttachableIntVector*)e->getData(tagMark);
      return avec->v[i];
    }
  else
    {
      return -1;
    }
}

static void packVertices (mEntity *e, int *verts, int &k)
{
  for(int i=0;i<e->size(0);i++)
    verts[k++] = e->get(0,i)->getId();
}

#endif
