/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#ifdef FMDB_PARALLEL

#ifndef PM_MIGRATE_UTIL_H
#define PM_MIGRATE_UTIL_H

#include <list>
#include <vector>
#include <map>
#include <set>
#include "mPart.h"
#include "mEntity.h"
#include "pmUtility.h"
#include "pmDataExchanger.h"
#include "pmMigrationCallbacks.h"

#define REGULAR_MIGRATION 0
#define LOAD_BALANCE 1

// **********************************************
//     structure used for communication
// **********************************************
struct packedEnt
{
  mEntity *sender;
  short int numDE;
  short int gTag;
  short int gType;
  short int numRC;
  short int numBP;
#ifdef MATCHING
  int numMatch;
#endif
}__attribute__((aligned (8)));

struct packedVertex
{
  double coord[3];
  int iD;
  short int numParam;
#ifdef MATCHING
  int numMatch;
#endif
}__attribute__((aligned (8))); 

void getDownEntities(mEntity* ent, std::vector<mEntity*>&);
void exchangeEntitiesToBounce(std::list<rc_struct_2>&);
void exchangeEntitiesToBroadcast(std::list<rc_struct_2>&);

// static 1-to-m partitioning functions on 1 proc
int serial_migratePtnObject (mMesh* mesh);
// static n-to-m partitioning functions on N proc (n<m)
int static_migratePtnObject (mMesh* mesh);

int static_getGlobPtnInfo (mMesh* mesh, int ptnMethod, int ptnApproach,
                    double imbalance_tol, int debug_level);

// dynamic n-to-m partitioning functions (n<m)

int preprocPSet (mMesh* mesh); // called by partition graph construction
int postprocPSet (mMesh* mesh); // called by partition graph construction

int pullPartBdryToOwner(mMesh* mesh, pmMigrationCallbacks &cb,
                    std::list<mEntity*>& part_bdry_entities,
                    std::vector<mEntity*>* newEnt,
                    std::vector<mEntity*>* rmvEnt,
                    std::vector<mEntityGroup*>* newSet,
                    std::vector<mEntityGroup*>* rmvSet);

int pullPartBdryToOwner(mMesh* mesh, pmMigrationCallbacks &cb,
 		    std::list<mEntity*>& vtsOnCB, 
		    std::list<std::pair<mEntity*,mEntity*> > &vtfcPairs,
                    std::vector<mEntity*>* newEnt,
                    std::vector<mEntity*>* rmEnt,
                    std::vector<mEntityGroup*>* newSet,
                    std::vector<mEntityGroup*>* rmvSet);

int migratePtnObject (mMesh* mesh, pmMigrationCallbacks& cb, 
		    std::map<mEntity*, std::pair<int, int> >& EntToMigr,
                    std::map<mEntityGroup*, std::pair<int, int> >& SetToMigr,
                    int debug_level, std::vector<mEntity*>* newEnt, std::vector<mEntity*>* rmvEnt,
	    	    std::vector<mEntityGroup*>* newPSet, std::vector<mEntityGroup*>* rmvPSet);

int getGlobPtnInfo (mMesh* mesh, int ptnMethod, int ptnApproach,
                    double imbalance_tol, int debug_level,
                    std::map<mEntity*, std::pair<int, int> >& EntToMigr,
                    std::map<mEntityGroup*, std::pair<int, int> >& SetToMigr);
                       
void packEntity(mEntity *ent, int pid, int src, pmMigrationCallbacks &cb, void *&msg_send, int &msg_size, int meshDim);

void unpackEntity(mPart* part, int meshDim, void *msg_recv, 
                    pmMigrationCallbacks &cb, int debug_level, std::vector<mEntity*>* newEnt, 
                    std::vector<mEntityGroup*>* newPSet, std::vector<rc_struct_3>& entitiesToBounce,
                    std::map<pair<mEntityGroup*, int>, mEntityGroup*>& entGrpMap);

void collectEntitiesToExchg_and_reset(mMesh*, std::map<mEntity*, std::pair<int, int> >& EntToMigr,
             std::map<mEntity*,int>* entitiesToExchg);

void updatePClas_and_collectEntitiesToRemove(std::map<mEntity*,int>* entitiesToExchg, 
			std::map<mEntity*,int>* entitiesToRemove);                    

void removeUnusedEntities(mMesh* mesh, pmMigrationCallbacks &cb, int dim, std::map<mEntity*, int>* entitiesToRemove);

void unifyEntityList(mPart* part, std::list<mEntity*>& entities);

int migrateEntity (mMesh* mesh, 
             mEntity* ent, int source_partid, int dest_partid, 
             pmMigrationCallbacks& cb,
             std::vector<mEntity*>* newEnt, 
             std::vector<mEntity*>* rmvEnt, 
             std::vector<MPI_Request*>& requests);

// old functions
void pmLoadBalance(mPart* mesh, pmMigrationCallbacks &cb);

void pmLoadBalance(std::vector<mPart*>& meshes, pmMigrationCallbacks &cb);  // load balance on multiple parts per proc

int pmMerge(mPart* mesh, pmMigrationCallbacks &cb);

int pmPartMigration(mPart* mesh, int dimToMove,std::list<mEntity*>& entities,
                    pmMigrationCallbacks &cb,
                    std::list<mEntity*>* rmE, std::list<mEntity*>* newE
                    , std::vector<mEntityGroup*>& rmEG, std::vector<mEntityGroup*>& newEG 
                    ); 

int pmPartMigration(std::vector<mPart*>&meshes,                             // migrate a whole part
		    int source_pid, 
		    int target_rank, 
		    pmMigrationCallbacks &cb); 

int pmPartMigration(mPart* mesh, pmMigrationCallbacks &cb,
 		    std::list<mEntity*>& vtsOnCB, 
		    std::list<std::pair<mEntity*,mEntity*> > &vtfcPairs,
                    std::list<mEntity*>* rmE, std::list<mEntity*>* newE
                    , std::vector<mEntityGroup*>& rmEG, std::vector<mEntityGroup*>& newEG 
 	           ); 

int migrateMeshEntities(mPart* mesh, std::map<mEntity*,int>& POtoMove, pmMigrationCallbacks& cb);
int migrateMeshEntities(mPart*mesh, std::map<mEntity*,int>& POtoMove, pmMigrationCallbacks& cb,
                        std::list<mEntity*>* rmE, std::list<mEntity*>* newE
                        , std::vector<mEntityGroup*>& rmEG, std::vector<mEntityGroup*>& newEG
		       );
					  

int migrateMeshEntities(std::vector<mPart*>& meshes,  
                        std::map<mEntity*,int*>& POtoMove,
		        pmMigrationCallbacks& cb);

int migrateMeshEntities(std::vector<mPart*>& meshes, std::map<mEntity*,int*>& POtoMove,
                        pmMigrationCallbacks& cb,
                        std::list<mEntity*>* rmE, std::list<mEntity*>* newE
			, std::vector<mEntityGroup*>& rmEG, std::vector<mEntityGroup*>& newEG
		       );

int migratePtnObjects (mMesh* mesh, std::map<mEntity*,int*>& EntToMigr,
			std::map<mEntityGroup*,int*>& SetToMigr, pmMigrationCallbacks& cb,
                        std::list<mEntity*>* rmE, std::list<mEntity*>* newE,
			std::vector<mEntityGroup*>& rmEG, std::vector<mEntityGroup*>& newEG);

// migration helper functions
void getFamily(mEntity* ent, std::list<mEntity*>& family);
void printFamily(mEntity* ent);

void getEntitiesToHandle(std::map<mEntity*,int>& entitiesToMove,
 			 std::set<mEntity*>* entitiesToHandle);        

void getEntitiesToHandle(std::map<mEntity*, int*>& entitiesToMove,
			 std::map<mEntity*, int>* entitiesToHandle);           // local_part_id  

void getEntityGroupsToHandle(std::map<mEntity*,int*>& entitiesToMove,
                             std::map<mEntityGroup*, int*>& egsToHandle);

void setBPs(int, std::map<mEntity*,int*>& POToMove, std::map<mEntity*, int>* entitiesToHandle); 

void exchangeBPs(std::set<mEntity*>* entitiesToHandle); 
void exchangeBPs(map<mEntity*, int>* entitiesToHandle); 

 void getEntitiesToHandle_and_setBPs(mPart* part, std::map<mEntity*,int*>& POToMove, std::map<mEntity*,int>* entitiesToHandle);
void getEntsToHandle_setBPs_nbrs(int meshDim, std::map<mEntity*, int*>& POToMove, std::map<mEntity*, int>* entitiesToHandle);

void updatePC_and_collectEntitiesToRemove(map<mEntity*, int>* entitiesToHandle,
					  map<mEntity*, int>* entitiesToRemove);

void removeUnusedEntities(mPart*, pmMigrationCallbacks &cb, int, std::set<mEntity*>*);
void removeUnusedEntities(std::vector<mPart*>& meshes, pmMigrationCallbacks &cb, int dim,
                           map<mEntity*, int>* entitiesToRemove);

void exchangeEntities(std::vector<mPart*>& meshes, int, pmMigrationCallbacks &cb,    
                     int, std::map<mEntity*,int>&, std::list<mEntity*>*
		     , std::vector<mEntityGroup*>& newEG
		     ); 

void unifyEntitiesToMove(mPart* mesh, std::list<mEntity*>& entities);
typedef std::map<mEntity*, int> entityPidMap;

void getPOsToMoveFromPartitionVector(mPart *mesh, int from,
                		int *partitionVector, int delta_id, 
				std::map<mEntity*, int>& entitiesToMove);

void getPOsToMove(mPart* mesh, std::list<mEntity*>& entities,int ent_dim, std::map<mEntity*,int>& entitiesToMove);
void getPOsToMove_nbrs(mPart* mesh, std::list<mEntity*>& entities,int ent_dim, std::map<mEntity*,int>& entitiesToMove);
void getPOsToMove_nbrs2(mPart* mesh, std::list<mEntity*>& entities,int ent_dim, std::map<mEntity*,int>& entitiesToMove);

// new one
void getPOsToMoveForSnap(mPart* mesh, std::list<mEntity*>& vtsOnCB, 
		      std::list<std::pair<mEntity*,mEntity*> > &vtfcPairs,
		      std::map<mEntity*,int>& entitiesToMove);
		      
// ghost entity copy related
// - implemented in pmMigrateGhost.cc

int migrateGhostEntities(mPart* mesh,  int include_copies, pmMigrationCallbacks &cb);

int expandGhostEntities(mPart* mesh, pmMigrationCallbacks &cb, 
			std::vector<mEntity*>& vtsToExpand, int dimToExpand);

void removeGhostEntities(mPart* mesh);

#endif
#endif /* PARALLEL */
