/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#ifdef FMDB_PARALLEL

#ifndef _PM_MODEL_H_
#define _PM_MODEL_H_

#include "pmEntity.h"
#include "mMesh.h"
#include "IPComMan.h"
#include <memory>
#include <iostream>
#include <set>
#include <vector>

#define GHOSTPARTID -111             // must be negative

class mPart;
class mEntity;

class pmModel
{
protected:
  static std::auto_ptr<pmModel> instance;
  mPart* mesh;

  int pmEntId;    // The counter for pmEntity ID number

  std::set<pmEntity*> allPEntities;
  
public:
  typedef std::set<pmEntity*>::iterator PEIter;

  pmModel();
  ~pmModel();
  static pmModel* Instance();

// update ownership of mesh entities
  void updateOwnership();
  void updateOwnership(mMesh* mesh);

// partition entities operators
  inline void addPEntity(pmEntity* pe) { allPEntities.insert(pe); if (pe->getId() > pmEntId) pmEntId = pe->getId(); }
  inline void delPEntity(pmEntity* pe) { delete pe; allPEntities.erase(pe);}

  PEIter peBegin() const
  {  return allPEntities.begin();}
  PEIter peEnd  () const
  {  return allPEntities.end();}
  int getNumPEs() { return allPEntities.size(); }
  pmEntity* getPartitionEntity(std::set<int>& bps, int owner=-1);
  pmEntity* getPartitionEntity(int id);

  void getNborPartId (mPart*, std::set<int>&);
  void getNborProcRank(mMesh*, std::set<int>&);
  void print();
};

#endif
#endif /* PARALLEL */
