# SYNOPSIS
#
#   AX_GMI
#
# DESCRIPTION
#
#   This macro looks for the install directory for the Geometric Modeling 
#   Interface (see http://www.scorec.rpi.edu/software.php and 
#   http://redmine.scorec.rpi.edu/projects/gmi) specified as the argument to 
#   the --with-gmi=<DIR> option.  If the directory is determined to be a valid 
#   install then the GMI_INCLUDES, GMI_LIBS and GMI_LIB_DIR variables are set 
#   to hold the requisite header and library linkages.  If optional meshModel 
#   support is missing then the variable GMI_MESHMODEL_MISSING is set to true; 
#   otherwise it is set to false.  If the header files cannot be found in the 
#   specified directory then the install is defined as invalid and GMI_MISSING 
#   is set to true; otherwise it is set to false.  If the libraries are not found 
#   in the specified directory then GMI_LIBS_MISSING is set to true.
#
#   To compile with GMI, you should use $GMI_INCLUDES.
#   To link with GMI, you should use $GMI_LIBS with LD_FLAGS or $GMI_LIB_DIR 
#   with LDADD.
#

AC_DEFUN([AX_GMI],[
AC_PREREQ(2.59)

AC_ARG_WITH([gmi],
   [AC_HELP_STRING([--with-gmi=DIR],[Specify directory of the GMI install @<:@default=@:>@])],
   [GMI_DIR="$withval"
    DISTCHECK_CONFIGURE_FLAGS="$DISTCHECK_CONFIGURE_FLAGS --with-gmi=\"${withval}\""
   ])

case "x$GMI_DIR" in
    x) AC_SUBST(GMI_MISSING,[true]);;
    xno|xyes)
      AC_MSG_ERROR([--with-gmi requires a GMI install directory as an argument])
      ;;
    x*) 
      GMI_LIBS_MISSING=false
      AS_IF([ test -f ${GMI_DIR}/include/GEntity.h],
                    [SCORECMODEL_INCLUDES="-I${GMI_DIR}/include"], 
                    [AC_MSG_ERROR([$GMI_DIR : not a valid path to a GMI install])])
      AS_IF([ test -f ${GMI_DIR}/include/DiscreteModel.h],
                    [MESHMODEL_INCLUDES="-I${GMI_DIR}/include";
                     AC_SUBST(GMI_MESHMODEL_MISSING,[false])], 
                    [AC_MSG_WARN([$GMI_DIR : does not have meshModel headers installed]); 
                     AC_SUBST(GMI_MESHMODEL_MISSING,[true])])
      AS_IF([ test -f ${GMI_DIR}/lib/libSCORECModel.a],
                    [SCORECMODEL_LIBS="-L${GMI_DIR}/lib -lSCORECModel"], 
                    [AC_MSG_WARN([$GMI_DIR : GMI libraries were not found]);
                     GMI_LIBS_MISSING=true])
      AS_IF([ test -f ${GMI_DIR}/lib/libmeshModel.a],
                    [MESHMODEL_LIBS="-L${GMI_DIR}/lib -lmeshModel"], 
                    [AC_MSG_WARN([$GMI_DIR : does not have meshModel library installed])])
      AC_SUBST(GMI_INCLUDES,"${SCORECMODEL_INCLUDES} ${MESHMODEL_INCLUDES}")
      AC_SUBST([GMI_LIBS_MISSING])
      AS_IF([test "x$GMI_LIBS_MISSING" = "xfalse"],
            [AC_SUBST(GMI_LIB_DIR,"${GMI_DIR}/lib")
             AC_SUBST(GMI_LIBS,"${SCORECMODEL_LIBS} ${MESHMODEL_LIBS}")],
            [AC_SUBST(GMI_LIBS,"")])
      AC_SUBST(GMI_MISSING,[false])
      ;;
esac
])dnl AX_GMI
