# SYNOPSIS
#
#   AX_IPCOMMAN
#
# DESCRIPTION
#
#   This macro looks for the install directory for the Inter Processor 
#   Communication Manager (http://redmine.scorec.rpi.edu/projects/ipcomman) 
#   specified as the argument to the --with-ipcomman=<DIR> option.  If the 
#   directory is determined to be a valid install then the IPCOMMAN_INCLUDES,
#   IPCOMMAN_LIBS, and IPCOMMAN_LIB_DIR variables are set to hold the requisite
#   header and library linkages.  If the directory is determined to *NOT* be a 
#   valid install then IPCOMMAN_MISSING is set to true; otherwise it is set to 
#   false.
#
#   To compile with IPComMan, you should use $IPCOMMAN_INCLUDES.
#   To link with IPComMan, you should use $IPCOMMAN_LIBS with LDFLAGS or 
#   IPCOMMAN_LIB_DIR with LDADD.
#
AC_DEFUN([AX_IPCOMMAN],[
AC_PREREQ(2.59)

AC_ARG_WITH([ipcomman],
   [AC_HELP_STRING([--with-ipcomman=DIR], [Specify directory containing IPComMan @<:@default=NONE@:>@])],
   [IPCOMMAN_DIR="$withval"
    DISTCHECK_CONFIGURE_FLAGS="$DISTCHECK_CONFIGURE_FLAGS --with-ipcomman=\"${withval}\""
   ])
  
case "x$IPCOMMAN_DIR" in
    xno|x)
      AC_SUBST([IPCOMMAN_MISSING],[true])
      ;;
    xyes)
      AC_MSG_ERROR([--with-ipcomman requires an argument])
      ;;
    x*)
      AS_IF([ test -f ${IPCOMMAN_DIR}/include/IPComMan.h ],  
            [IPCOMMAN_INCLUDES="-I${IPCOMMAN_DIR}/include"],
            [AC_MSG_ERROR([$IPCOMMAN_DIR : not a valid IPCOMMAN install - header files not found])])
      AS_IF([ test -f ${IPCOMMAN_DIR}/lib/libipcomman.a],
            [IPCOMMAN_LIBS="-L${IPCOMMAN_DIR}/lib -lipcomman"
             IPCOMMAN_LIB_DIR="${IPCOMMAN_DIR}/lib"], 
            [AC_MSG_ERROR([$IPCOMMAN_DIR : not a valid IPCOMMAN install - libipcomman.a not found])])
      AC_SUBST(IPCOMMAN_LIBS)
      AC_SUBST(IPCOMMAN_LIB_DIR)
      AC_SUBST(IPCOMMAN_INCLUDES)
      AC_SUBST([IPCOMMAN_MISSING],[false])
esac
])dnl AX_IPCOMMAN
