# SYNOPSIS
#
#   AX_ITERATORS
#
# DESCRIPTION
#
#   This macro looks for the install directory for iterator utilities
#   specified as the argument to the --with-iterators=<DIR> option.  If the 
#   directory is determined to be a valid install then the ITERATORS_INCLUDES 
#   are set to hold the requisite headers.  If the header files cannot be found
#   in the specified directory then the install is defined as invalid and 
#   ITERATORS_MISSING is set to true; otherwise it is set to false.
#
#   To compile with iterator utilities, you should use $ITERATORS_INCLUDES.
#
AC_DEFUN([AX_ITERATORS],[
AC_PREREQ(2.59)

AC_ARG_WITH([iterators],
   [AC_HELP_STRING([--with-iterators=DIR], [Specify directory containing the iterator utilities @<:@default=NONE@:>@])],
   [ITERATORS_DIR="$withval"
    DISTCHECK_CONFIGURE_FLAGS="$DISTCHECK_CONFIGURE_FLAGS --with-iterators=\"${withval}\""
   ])

case "x$ITERATORS_DIR" in
    xno|x)
      AC_SUBST([ITERATORS_MISSING],[true])
      ;;
    xyes)
      AC_MSG_ERROR([--with-iterators requires an argument])
      ;;
    x*)
      ITERATORS_MISSING=true
      AS_IF([ test -f ${ITERATORS_DIR}/src/GenIterator.h],
                     [ITERATORS_INCLUDES="-I${ITERATORS_DIR}/src"; ITERATORS_MISSING=false],
                     [])
      AS_IF([ test -f ${ITERATORS_DIR}/include/GenIterator.h],
                     [ITERATORS_INCLUDES="-I${ITERATORS_DIR}/include/"; ITERATORS_MISSING=false],
                     [])
      AS_IF([test "x$ITERATORS_MISSING" = "xtrue"],
            [AC_MSG_ERROR([$ITERATORS_DIR : not a valid iterator utilities install - GenIterator.h not found])],
            [])
      AC_SUBST(ITERATORS_INCLUDES)
esac


])dnl AX_ITERATORS
