# SYNOPSIS
#
#   AX_MPI
#
# DESCRIPTION
#
#   This macro looks for the mpi compiler wrappers in the directory specified 
#   as the argument to the --with-mpi=<DIR> option.  If the directory is 
#   determined to be a valid install then the preference list of mpi compiler 
#   wrapper is set in the F77_LIST, FC_LIST, CXX_LIST, and CC_LIST variables. 
#   If the directory is determined to *NOT* be a valid install then MPI_MISSING
#   is set to true; otherwise it is set to false.
#
#   After the call to AX_MPI the following autoconf macros should be callled 
#   from configure.ac to set the Fortran, Fortran77, C and C++ compilers:
#     AC_PROG_CC( [$CC_LIST] )
#     AC_PROG_CPP
#     AC_PROG_CXX( [$CXX_LIST] )
#     AC_PROG_CXXCPP
#     AC_PROG_FC( [$FC_LIST] )
#     AC_PROG_F77( [$F77_LIST] )
#


AC_DEFUN([AX_MPI],[
AC_PREREQ(2.59)

AC_ARG_WITH( [mpi], AC_HELP_STRING([[--with-mpi(=DIR)]], [Enable parallel support @<:@default=no@:>@]),
             [WITH_MPI="$withval"],[WITH_MPI=no] )
AC_MSG_NOTICE([--with-mpi defined as x$WITH_MPI])
case "x$WITH_MPI" in
  xno)
    CC_LIST="cc gcc cl egcs pgcc"
    CXX_LIST="CC aCC cxx xlC_r xlC c++ g++ pgCC gpp cc++ cl FCC KCC RCC"
    FC_LIST="gfortran ifort pgf90"
    F77_LIST="gfortran ifort pgf77"
    AC_SUBST(MPI_LIBS,[])
    AC_SUBST(MPI_INCLUDES,[])
    AC_SUBST(MPI_MISSING,[true])
    AC_MSG_NOTICE([--with-mpi defined as "no"])
    ;;
  xyes)
    CC_LIST="mpicc mpcc"
    CXX_LIST="mpiCC mpCC mpicxx"
    FC_LIST="mpif90"
    F77_LIST="mpif77"
    AC_SUBST(MPI_LIBS,[])
    AC_SUBST(MPI_INCLUDES, [])
    AC_SUBST(MPI_MISSING,[false])
    ax_mpi_ok=yes
    ;;
  x*)
    AC_SUBST(MPI_LIBS,[${WITH_MPI}/lib])
    if test -z "$MPI_INCLUDES";then 
      AC_SUBST(MPI_INCLUDES,[${WITH_MPI}/include])
      AC_MSG_NOTICE([MPI_INCLUDES is substituted to ${MPIHOME}/include])
    else
      AC_MSG_NOTICE([MPI_INCLUDES is set to ${MPI_INCLUDES}])
    fi
    AS_IF([ ! test -e "${MPI_INCLUDES}/mpi.h" ], [AC_MSG_ERROR([header mpi.h does not exist at ${MPI_INCLUDES}])])
    AC_SUBST(MPI_MISSING,[false])
    if test -z "$CC";then
      AC_MSG_NOTICE([testing for CC executables])
      for prog in mpicc mpcc; do
        if test -x ${WITH_MPI}/bin/$prog; then
          CC="${WITH_MPI}/bin/$prog"
          CC_LIST="$prog"
        fi
      done
    else
      CC_LIST="$CC"
    fi
    if test -z "$CXX";then
      AC_MSG_NOTICE([testing for CXX executables])
      for prog in mpicxx mpiCC mpCC mpicxx; do
        if test -x ${WITH_MPI}/bin/$prog; then
          CXX="${WITH_MPI}/bin/$prog"
          CXX_LIST="$prog"
        fi
      done
    else
      CXX_LIST="$CXX"
    fi
    if test -z "$FC";then
      AC_MSG_NOTICE([testing for FC executables])
      for prog in mpif90; do
        if test -x ${WITH_MPI}/bin/$prog; then
          FC="${WITH_MPI}/bin/$prog"
          FC_LIST="$prog"
        fi
      done
    else
      FC_LIST="$FC"
    fi
    if test -z "$F77";then
      AC_MSG_NOTICE([testing for F77 executables])
      for prog in mpif77; do
        if test -x ${WITH_MPI}/bin/$prog; then
          F77="${WITH_MPI}/bin/$prog"
          F77_LIST="$prog"
        fi
      done
    else
      F77_LIST="$F77"
    fi
    ;;
esac
])dnl AX_MPI
