# SYNOPSIS
#
#   AX_OPTIMIZATION
#
# DESCRIPTION
#
#   This macro allows the user to compile the code in opt mode. The argument 
#   to the --with-opt=<OptLevel>. The OptLevel parameter can be -O2, -O3, -O4, -O5 
#   If the passed parameter is not one of those, no optimization flag will be passed by default
#
#   To compile with the OPTIMIZATION_FLAGS variable add it to the AM_*FLAGS variable 
#   in Makefile.am as follows:
# 
#      AM_CXXFLAGS += $(OPTIMIZATION_FLAGS)
#

AC_DEFUN([AX_OPTIMIZATION],[
AC_PREREQ(2.59)

AC_ARG_WITH([opt],
  [AC_HELP_STRING([--with-opt=OptLevel], [Specify the opt level @<:@default=NONE@:>@])],
  [OPTIMIZATION_LEVEL=$withval
   DISTCHECK_CONFIGURE_FLAGS="$DISTCHECK_CONFIGURE_FLAGS --with-opt=\"${withval}\""
  ])

case "x$OPTIMIZATION_LEVEL" in
    x) AC_SUBST([OPTIMIZATION_MISSING],[true]);;
    x2|x3|x4|x5)
      OPTIMIZATION_FLAGS="-O${OPTIMIZATION_LEVEL} -DOPTIMIZATION"
      AC_SUBST([OPTIMIZATION_FLAGS])
      AC_SUBST([OPTIMIZATION_MISSING],[false])
      ;;
    x*)
      AC_MSG_ERROR([--with-opt requires an optimization level such as 2,3,4 or 5 to be passed as an argument])
      AC_SUBST([OPTIMIZATION_MISSING],[true])
      ;;
esac

])dnl AX_OPTIMIZATION
