# SYNOPSIS
#
#   AX_PARMETIS
#
# DESCRIPTION
#
#   This macro looks for the install directory for Parmetis (see 
#   http://glaros.dtc.umn.edu/gkhome/metis/parmetis/overview ) specified as the 
#   argument to the --with-parmetis=<DIR> option.  If the directory is 
#   determined to be a valid install then the PARMETIS_INCLUDES, 
#   PARMETIS_INCLUDE_DIR, PARMETIS_LIBS, and PARMETIS_LIB_DIR 
#   variables are set to hold the requisite header and 
#   library linkages.  If the directory is determined to *NOT* be a 
#   valid install then PARMETIS_MISSING is set to true; otherwise it is 
#   set to false.
#
#   To compile with PARMETIS, you should use $PARMETIS_INCLUDES.
#   To link with PARMETIS, you can use $PARMETIS_LIBS with *LDFLAGS or
#   $PARMETIS_LIB_DIR/libparmetis.a and $METIS_LIB_DIR/libmetis.a with 
#   *LDADD.
#
AC_DEFUN([AX_PARMETIS],[
AC_PREREQ(2.59)

AC_ARG_WITH([parmetis],
   [AC_HELP_STRING([--with-parmetis=DIR], [Specify directory containing PARMETIS @<:@default=NONE@:>@])],
   [PARMETIS_DIR="$withval"
    DISTCHECK_CONFIGURE_FLAGS="$DISTCHECK_CONFIGURE_FLAGS --with-parmetis=\"${withval}\""
   ])
  
case "x$PARMETIS_DIR" in
    xno|x)
      AC_SUBST([PARMETIS_MISSING],[true])
      ;;
    xyes)
      AC_MSG_ERROR([--with-parmetis requires an argument])
      ;;
    x*)
      HEADER_MISSING=true
      AS_IF([ test -f ${PARMETIS_DIR}/include/parmetis.h],
                     [PARMETIS_INCLUDES="-I${PARMETIS_DIR}/include/";
                      HEADER_MISSING=false, 
                      PARMETIS_INCLUDE_DIR=${PARMETIS_DIR}/include], 
                     [])
      AS_IF([ test -f ${PARMETIS_DIR}/parmetis.h],
                     [PARMETIS_INCLUDES="-I${PARMETIS_DIR}"; 
                      HEADER_MISSING=false; 
                      PARMETIS_INCLUDE_DIR=${PARMETIS_DIR}], 
                     [])
      if [ test "x$HEADER_MISSING" = "xtrue" ]; then
         AC_MSG_ERROR([$PARMETIS_DIR : not a valid parmetis source])
      fi

      PARMETIS_LIB_MISSING=true
      AS_IF([ test -f ${PARMETIS_DIR}/lib/libparmetis.a],
                     [PARMETIS_LIBS="-L${PARMETIS_DIR}/lib/ -lparmetis";
                      PARMETIS_LIB_MISSING=false, 
                      PARMETIS_LIB_DIR=${PARMETIS_DIR}/lib], 
                     [])
      AS_IF([ test -f ${PARMETIS_DIR}/libparmetis.a],
                     [PARMETIS_LIBS="-L${PARMETIS_DIR}/ -lparmetis";
                      PARMETIS_LIB_MISSING=false, 
                      PARMETIS_LIB_DIR=${PARMETIS_DIR}], 
                     [])
      if [ test "x$PARMETIS_LIB_MISSING" = "xtrue" ]; then
        AC_MSG_ERROR([$PARMETIS_DIR : not a valid PARMETIS install - libparmetis.a not found])
      fi

      METIS_LIB_MISSING=true
      AS_IF([ test -f ${PARMETIS_DIR}/lib/libmetis.a],
                     [PARMETIS_LIBS="$PARMETIS_LIBS -L${PARMETIS_DIR}/lib/ -lmetis";
                      METIS_LIB_MISSING=false, 
                      METIS_LIB_DIR=${PARMETIS_DIR}/lib], 
                     [])
      AS_IF([ test -f ${PARMETIS_DIR}/libmetis.a],
                     [PARMETIS_LIBS="$PARMETIS_LIBS -L${PARMETIS_DIR} -lmetis";
                      METIS_LIB_MISSING=false, 
                      METIS_LIB_DIR=${PARMETIS_DIR}], 
                     [])
      if [ test "x$PARMETIS_LIB_MISSING" = "xtrue" ]; then
        AC_MSG_ERROR([$PARMETIS_DIR : not a valid PARMETIS install - libmetis.a not found])
      fi

      AC_SUBST(PARMETIS_LIBS)
      AC_SUBST(PARMETIS_INCLUDES)
      AC_SUBST(PARMETIS_INCLUDE_DIR)
      AC_SUBST(PARMETIS_LIB_DIR)
      AC_SUBST(METIS_INCLUDE_DIR)
      AC_SUBST(METIS_LIB_DIR)
      AC_SUBST([PARMETIS_MISSING],[false])
esac
])dnl AX_PARMETIS
