# SYNOPSIS
#
#   AX_SCORECUTIL
#
# DESCRIPTION
#
#   This macro looks for the install directory of SCORECUtil specified as the 
#   argument to the --with-scorecutil=<DIR> option.  If the directory is 
#   determined to be a valid install then the SCORECUTIL_INCLUDES, SCORECUTIL_LIBS 
#   and SCORECUTIL_LIB_DIR variables are set to hold the requisite header and library 
#   linkages.  If SCORECUTIL_LIBS cannot be found then the variable 
#   SCORECUTIL_LIBS_MISSING is set to true.  If the directory is determined to 
#   *NOT* be a valid install then SCORECUTIL_MISSING is set to true; otherwise
#   false.
#
#   To compile with SCORECUtil, you should use $SCORECUTIL_INCLUDES.
#   To link with SCORECUtil, you should use $SCORECUTIL_LIBS with LD_FLAGS or 
#   $SCORECUTIL_LIB_DIR with LDADD.
#

AC_DEFUN([AX_SCORECUTIL],[
AC_PREREQ(2.59)

AC_ARG_WITH([scorecutil],
  [AC_HELP_STRING([--with-scorecutil=DIR], [Specify directory containing the SCORECUtil hinstall @<:@default=NONE@:>@])],
  [SCORECUTIL_DIR=$withval
   DISTCHECK_CONFIGURE_FLAGS="$DISTCHECK_CONFIGURE_FLAGS --with-scorecutil=\"${withval}\""
  ])

case "x$SCORECUTIL_DIR" in
    x) AC_SUBST([SCORECUTIL_MISSING],[true]);;
    xno|xyes)
      AC_MSG_ERROR([--with-scorecutil requires a SCORECUtil install directory as an argument])
      ;;
    x*)
      AS_IF([ test -f ${SCORECUTIL_DIR}/scorecSSList.h],
                    [SCORECUTIL_INCLUDES="-I${SCORECUTIL_DIR}"], 
                    [])
      if test x"$SCORECUTIL_INCLUDES" = x; then
         AS_IF([ test -f ${SCORECUTIL_DIR}/include/scorecSSList.h],
                       [SCORECUTIL_INCLUDES="-I${SCORECUTIL_DIR}/include"], 
                       [])
      fi
      if test x"$SCORECUTIL_INCLUDES" = x; then
         AS_IF([ test -f ${SCORECUTIL_DIR}/SCORECUtil/include/scorecSSList.h],
                       [SCORECUTIL_INCLUDES="-I${SCORECUTIL_DIR}/SCORECUtil/include"], 
                       [])
      fi              
      if test x"$SCORECUTIL_INCLUDES" = x; then
         AS_IF([ test -f ${SCORECUTIL_DIR}/include/SCORECUtil/scorecSSList.h],
                       [SCORECUTIL_INCLUDES="-I${SCORECUTIL_DIR}/include/SCORECUtil"], 
                       [])
      fi              
      if test x"$SCORECUTIL_INCLUDES" = x; then
         AC_MSG_ERROR([$SCORECUTIL_DIR: not a valid path to SCORECUtil headers])
      fi
 
      AS_IF([ test -f ${SCORECUTIL_DIR}/lib/libSCORECUtil.a],
                    [SCORECUTIL_LIBS="-L${SCORECUTIL_DIR}/lib -lSCORECUtil"], 
                    [AC_MSG_WARN([$SCORECUTIL_HEADERS_DIR: not a valid path to SCORECUtil libraries]); 
                     AC_SUBST(SCORECUTIL_LIBS_MISSING,[true])])
      AC_SUBST(SCORECUTIL_INCLUDES)
      AC_SUBST(SCORECUTIL_LIBS)
      AC_SUBST(SCORECUTIL_LIB_DIR, "${SCORECUTIL_DIR}/lib")
      AC_SUBST([SCORECUTIL_MISSING],[false])
      ;;
esac
]) dnl AX_SCORECUTIL
