# SYNOPSIS
#
#   AX_ZOLTAN
#
# DESCRIPTION
#
#   This macro looks for the install directory for Zoltan (see 
#   http://www.cs.sandia.gov/Zoltan/ ) specified as the argument to the 
#   --with-zoltan=<DIR> option.  If the directory is determined to be a valid 
#   install then the ZOLTAN_INCLUDES, ZOLTAN_INCLUDE_DIR, ZOLTAN_LIBS and 
#   ZOLTAN_LIB_DIR variables are set to hold the requisite header and library 
#   linkages.  If the directory is determined to *NOT* be a valid install then 
#   ZOLTAN_MISSING is set to true; otherwise it is set to false.
#
#   To compile with ZOLTAN, you should use $ZOLTAN_INCLUDES.
#   To link with ZOLTAN, you should use $ZOLTAN_LIBS with *LDFLAGS or 
#   $ZOLTAN_LIB_DIR/libzoltan.a
#
AC_DEFUN([AX_ZOLTAN],[
AC_PREREQ(2.59)

AC_ARG_WITH([zoltan],
   [AC_HELP_STRING([--with-zoltan=DIR], [Specify directory containing ZOLTAN @<:@default=NONE@:>@])],
   [ZOLTAN_DIR="$withval"
    DISTCHECK_CONFIGURE_FLAGS="$DISTCHECK_CONFIGURE_FLAGS --with-zoltan=\"${withval}\""
   ])
  
case "x$ZOLTAN_DIR" in
    xno|x)
      AC_SUBST([ZOLTAN_MISSING],[true])
      ;;
    xyes)
      AC_MSG_ERROR([--with-zoltan requires an argument])
      ;;
    x*)
      AS_IF([ test -f ${ZOLTAN_DIR}/include/zoltan.h],
                     [ZOLTAN_INCLUDES="-I${ZOLTAN_DIR}/include/"], 
                     [AC_MSG_ERROR([$ZOLTAN_DIR : not a valid zoltan source])])
      AS_IF([ test -f ${ZOLTAN_DIR}/lib/libzoltan.a],
                     [ZOLTAN_LIBS="-L${ZOLTAN_DIR}/lib/ -lzoltan"], 
                     [AC_MSG_ERROR([$ZOLTAN_DIR : not a valid ZOLTAN install - libzoltan.a not found])])
      AC_SUBST(ZOLTAN_LIBS)
      AC_SUBST(ZOLTAN_INCLUDES)
      AC_SUBST([ZOLTAN_LIB_DIR],[${ZOLTAN_DIR}/lib])
      AC_SUBST([ZOLTAN_INCLUDE_DIR],[${ZOLTAN_DIR}/lib}])
      AC_SUBST([ZOLTAN_MISSING],[false])
esac
])dnl AX_ZOLTAN
