/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "mMeshIO.h" 
#include "FMDB_Internals.h"
using namespace std; 

//***************************************************
int exportPVTK(const char* fname, int numPtn)
//***************************************************
{
  char realName[256];
  sprintf(realName,"%s.pvtu",fname); 

   FILE* outFile = fopen(realName, "w");  
   if (!outFile)
     return SCUtil_FILE_WRITE_ERROR;

  fprintf(outFile, "<?xml version=\"1.0\"?>\n"); 
  fprintf(outFile, "<VTKFile type=\"PUnstructuredGrid\" version=\"0.1\" byte_order=\"LittleEndian\">\n");
  fprintf(outFile, "<PUnstructuredGrid GhostLevel=\"0\">\n");

  fprintf(outFile, "<PCellData Scalars=\"partId\" >\n");
  fprintf(outFile, "<PDataArray type=\"Int32\" Name=\"partId\" format=\"ascii\"/>\n");    // global part id
  fprintf(outFile, "</PCellData>\n");

  fprintf(outFile, "<PPoints>\n<PDataArray type=\"Float32\" NumberOfComponents=\"3\"/>\n </PPoints>\n"); 

  fprintf(outFile, "<PCells>\n");
  fprintf(outFile, "<PDataArray type=\"Int32\"  Name=\"connectivity\" format=\"ascii\"/>\n");
  fprintf(outFile, "<PDataArray type=\"Int32\" Name=\"offsets\" format=\"ascii\"/>\n");
  fprintf(outFile, "<PDataArray type=\"UInt8\" Name=\"types\" format=\"ascii\"/>\n");
  fprintf(outFile, "</PCells>\n");

  for (int pid=0; pid<numPtn; pid++)
    fprintf(outFile, "<Piece Source=\"%s%d.vtu\"/>\n", fname,pid); 

  fprintf(outFile, "</PUnstructuredGrid>\n</VTKFile>\n");
  fclose(outFile);    
 
  return SCUtil_SUCCESS;
}

//********************************************************
int exportVTK(mPart *part, const char* fname, int partId) 
//********************************************************
{
    FILE* outFile = fopen(fname, "w");
    if (!outFile)
        return SCUtil_FILE_WRITE_ERROR;

    int mDim;
    FMDB_Part_GetDim(part, &mDim);

    mMesh* mesh = part->getMesh();

    int numRgn, numVtc;
    FMDB_Part_GetNumEnt(part, FMDB_VERTEX, FMDB_POINT, &numVtc);
    if (mDim == 3)
        FMDB_Part_GetNumEnt(part, FMDB_REGION, FMDB_ALLTOPO, &numRgn);
    else
        FMDB_Part_GetNumEnt(part, FMDB_FACE, FMDB_ALLTOPO, &numRgn);

    pTag new_tag;
    FMDB_Mesh_CreateTag(mesh, "integer_tag", SCUtil_INT, 1, new_tag);

    fprintf(outFile, "<?xml version=\"1.0\"?>\n");
    fprintf(outFile, "<VTKFile type=\"UnstructuredGrid\" version=\"0.1\" byte_order=\"LittleEndian\">\n");
    fprintf(outFile, "<UnstructuredGrid>\n");

    fprintf(outFile, "<Piece NumberOfPoints=\"%d\"  NumberOfCells=\"%d\">\n", numVtc, numRgn);
    fprintf(outFile, "<Points>\n ");
    fprintf(outFile, "<DataArray type=\"Float32\" NumberOfComponents=\"3\" format=\"ascii\">\n");

    mEntity* ent;
    int i = 0, id, has_tag;
    vector<pMeshEnt> taggedEnt;
    vector<pTag> tagID;
    vector<pTag>::iterator tagIter;

    for (mPart::iterall it = part->beginall(0); it != part->endall(0); ++it) {
        ent = *it;
        double loc[3];
        FMDB_Ent_HasTag(ent, NULL, &has_tag);
        if (has_tag)
            taggedEnt.push_back(ent);
        FMDB_Ent_SetIntTag(mesh, ent, new_tag, i);
        V_coord((pVertex) ent, loc);
        fprintf(outFile, "  %.15f  %.15f  %.15f  \n", loc[0], loc[1], loc[2]);
        ++i;
    }

    // Header for vtk
    fprintf(outFile, "\n</DataArray>\n");
    fprintf(outFile, "</Points>\n");

    FMDB_Mesh_GetTag(mesh, tagID);

    if (taggedEnt.size() > 0) {
        fprintf(outFile, "<PointData>\n");
        exportTagsToVTK(mesh, taggedEnt, outFile);
        fprintf(outFile, "</PointData>\n");
    }
    taggedEnt.clear();
    // Header for vtk
    fprintf(outFile, "<Cells>\n");
    fprintf(outFile, "<DataArray type=\"Int32\"  Name=\"connectivity\" format=\"ascii\">\n");

    int* numV = new int[numRgn+1];
    int* entType = new int[numRgn+1];
    numV[0] = 0;
    int iNum = 0;
    pMeshEnt tagVtx;
    for (mPart::iterall it = part->beginall(mDim); it != part->endall(mDim); ++it) {
        ent = *it;
        FMDB_Ent_HasTag(ent, NULL, &has_tag);
        if (has_tag)
            taggedEnt.push_back(ent);
        iNum++;
        entType[iNum] = M_GetElementType(ent);
        vector<pMeshEnt> vecAdj;
        FMDB_Ent_GetAdj(ent, FMDB_VERTEX, 0, vecAdj);
        numV[iNum] = vecAdj.size() + numV[iNum - 1];
        for (vector<pMeshEnt>::iterator itVec = vecAdj.begin(); itVec != vecAdj.end(); itVec++) {
            tagVtx = *itVec;
            FMDB_Ent_GetIntTag(mesh, tagVtx, new_tag, &id);
            fprintf(outFile, " %d", id);
        }
    }
    // Header for vtk
    fprintf(outFile, "\n</DataArray>\n");

    fprintf(outFile, "<DataArray type=\"Int32\" Name=\"offsets\" format=\"ascii\">\n");

    for (int i = 1; i < numRgn + 1; i++)
        fprintf(outFile, " %d ", numV[i]);

    delete [] numV;

    // Header for vtk
    fprintf(outFile, "\n</DataArray>\n");
    fprintf(outFile, "<DataArray type=\"UInt8\" Name=\"types\" format=\"ascii\">\n");

    for (int i = 1; i < numRgn + 1; i++)
        switch (entType[i]) {
            case TRI:
                fprintf(outFile, " 5 ");
                break;
            case QUAD:
                fprintf(outFile, " 9 ");
                break;
            case TET:
                fprintf(outFile, " 10 ");
                break;
            case PYRAMID:
                fprintf(outFile, " 14 ");
                break;
            case PRISM:
                fprintf(outFile, " 13 ");
                break;
            case HEX:
                fprintf(outFile, " 12  ");
                break;
            default:
#ifdef DEBUG
                cout << "region topology [" << i << "] NOT supported" << endl;
#endif
                fclose(outFile);
                return SCUtil_INVALID_ENTITY_TOPO;
        }

        delete [] entType;

    // Header for vtk
    fprintf(outFile, "\n</DataArray>\n");
    fprintf(outFile, "</Cells>\n");

    if (partId != -1) {
        fprintf(outFile, "<CellData Scalars=\"partId\">\n");

        fprintf(outFile, "<DataArray type=\"Int32\" Name=\"partId\" format=\"ascii\">\n");
        for (int i = 1; i < numRgn + 1; i++)
            fprintf(outFile, " %d ", partId);
        fprintf(outFile, "\n</DataArray>\n");

        fprintf(outFile, "</CellData>\n");
    }

    if (taggedEnt.size() > 0) {
        fprintf(outFile, "<CellData>\n");
        exportTagsToVTK(mesh, taggedEnt, outFile);
        fprintf(outFile, "</CellData>\n");
    }
    fprintf(outFile, "</Piece>\n</UnstructuredGrid>\n</VTKFile>\n");

    int iHasTag = 0;
    for (mPart::iterall it = part->beginall(0); it != part->endall(0); ++it) {
        ent = *it;
        FMDB_Ent_HasTag(ent, new_tag, &iHasTag);
        if (iHasTag)
            FMDB_Ent_DelTag(ent, new_tag);
    }

    int ierr = FMDB_Mesh_DelTag(mesh, new_tag, 0); if ( ierr ) return ierr;

    fclose(outFile);
    return SCUtil_SUCCESS;
}

//********************************************************
int exportTagsToVTK(pMeshMdl mesh, vector<pMeshEnt> const& taggedEnt, FILE* outFile)
//********************************************************
{
  vector<pTag> tagID;
  vector<pTag>::iterator tagIter; 
  pMeshEnt ent;
  int has_tag;
  FMDB_Mesh_GetTag (mesh, tagID);
  // output tagged data
  pTag tag;
  int name_len=100;
  char* tag_name = new char[name_len];
  int tag_type, tag_size, arr_size;
  for (tagIter = tagID.begin(); tagIter != tagID.end(); ++tagIter)
  {
    tag = *tagIter;
    FMDB_Tag_GetName(tag, tag_name, name_len);
    // this tag was created only for computation purposes so no need to write it to the file.
    if(strcmp(tag_name,"integer_tag")==0)
      continue;
    FMDB_Tag_GetType (tag, &tag_type);
    FMDB_Tag_GetSize (tag, &tag_size);
    bool printOnce=false;
    for (std::vector<pMeshEnt>::const_iterator ent_it=taggedEnt.begin();
               ent_it != taggedEnt.end(); ent_it++)
    {
      ent = *ent_it;
      FMDB_Ent_HasTag(ent, tag, &has_tag);
      if(has_tag)
      {
        if (!printOnce)
        {
          if(tag_type == SCUtil_INT)
            fprintf(outFile, "<DataArray type=\"Int32\" Name=\"%s\" NumberOfComponents=\"%d\" format=\"ascii\">\n", tag_name, tag_size);
          else if(tag_type == SCUtil_DBL)
            fprintf(outFile, "<DataArray type=\"Float32\" Name=\"%s\" NumberOfComponents=\"%d\" format=\"ascii\">\n", tag_name, tag_size);
          else
            cout << "\n FMDB Warning: Tag type not supported for vtk files" << endl;
          printOnce=true;
        } //if (!printOnce)
        if(tag_type == SCUtil_INT)
        {
          if(tag_size==1)
          {
            int tag_data;
            FMDB_Ent_GetIntTag(mesh, ent, tag, &tag_data);
            fprintf(outFile, "%d\n", tag_data);
          }
          else
          {
            int* arr_data = new int[tag_size];
            FMDB_Ent_GetIntArrTag (mesh, ent, tag, &arr_data, &arr_size);
            for (int i=0; i<arr_size; ++i)
              fprintf(outFile, "%d ", arr_data[i]);
            fprintf(outFile, "\n");
            delete [] arr_data;
          }
        } // if(tag_type == SCUtil_INT)
        else if(tag_type == SCUtil_DBL)
        {
          if(tag_size==1)
          {
            double tag_data;
            FMDB_Ent_GetDblTag(mesh, ent, tag, &tag_data);
            fprintf(outFile, "%0.15f\n", (float)tag_data);
          }
          else
          {
            double* arr_data = new double[tag_size];
            FMDB_Ent_GetDblArrTag (mesh, ent, tag, &arr_data, &arr_size);
            for (int i=0; i<arr_size; ++i)
              fprintf(outFile, "%0.15f ", (float)arr_data[i]);
            fprintf(outFile,"\n");
            delete [] arr_data;
          }
        }
      }// end has_tag
    } // for ent_it 
    if (printOnce)
      fprintf(outFile, "\n</DataArray>\n");
  }  // for tagIter

  delete [] tag_name;
  return SCUtil_SUCCESS;
}

// fname can be given with any extension.
// Aug 05, 2007, add the functionality to display multiple parts per process.  
//********************************************************
int exportMeshToVTK (pMeshMdl mesh,const char * fname)
//********************************************************
{
  // Remove the .vtk extension from the file name
  char without_extension[256], realName[256];
  snprintf(without_extension,strlen(fname)-3,"%s",fname);

#ifdef FMDB_PARALLEL
  int mypid = ParUtil::Instance()->rank(); 
  int gsize=SCUTIL_CommSize();
  
  vector<pPart> vecPart;
  FMDB_Mesh_GetAllPart(mesh, mypid, vecPart);
  int localNumPtn = vecPart.size(); 
  if(gsize==1 && localNumPtn==1)
  {
    sprintf(realName,"%s.vtu",without_extension);
    exportVTK(vecPart[0], realName, -1); 
  }
  int globalNumPtn; 

  // get the sum of local number of parts from previous processes
  int preSum = 0; 
  MPI_Allreduce(&localNumPtn, &globalNumPtn, 1,MPI_INT,MPI_SUM,ParUtil::Instance()->getComm());
  MPI_Scan(&localNumPtn, &preSum, 1, MPI_INT, MPI_SUM, ParUtil::Instance()->getComm());  
  preSum = preSum-localNumPtn; 

  if(globalNumPtn>1)
  {
    // piece files 
    for (int pid=0; pid<localNumPtn; pid++)                   
    {
      sprintf(realName,"%s%d.vtu",without_extension,preSum+pid);  // Assume the localNumPtn is the same per process. 
      exportVTK(vecPart[pid], realName,  preSum+pid); 
    }  
     
    // PUnstructuredGrid file
    if (mypid == 0)
      exportPVTK(without_extension,globalNumPtn); 
  }
  else
#endif
  {
    pPart part;
    FMDB_Mesh_GetPart(mesh, 0, part); 
    char realName[256];
    sprintf(realName,"%s.vtu",without_extension);
    exportVTK(part, realName,  -1);                // Assume the first mesh in meshes is not empty.   
  }
#ifdef FMDB_PARALLEL
  ParUtil::Instance()->Barrier(1, "finish vtk"); 
#endif
  return SCUtil_SUCCESS;
}

/// Function for writing out a part of a mesh in vtu format without tags
//********************************************************
int write_3Dpart_vtu(pMeshMdl mesh, const char* fname, vector<pRegion> &regs)
//********************************************************
{
  vector<pTag> rgnTagsToWrite;
  vector<pTag> vtxTagsToWrite;
  return write_3Dpart_vtu(mesh, fname, regs, rgnTagsToWrite, vtxTagsToWrite);
}

/// Function for writing out a part of a mesh in vtu format
int write_3Dpart_vtu(pMeshMdl mesh, const char* fname, vector<pRegion> &regs, vector<pTag>& rgnTagsToWrite, vector<pTag>& vtxTagsToWrite)
{
    char realName[256];
    sprintf(realName, "%s.vtu", fname);

    FILE* outFile = fopen(realName, "w");
    if (!outFile) {
        throw new mException(__LINE__, __FILE__, "unable to create vtu File");
    }

    int numRgn = regs.size();

    set<pVertex> verts;
    for (int rgnIdx = 0; rgnIdx < numRgn; rgnIdx++) {
        vector<pMeshEnt> vecAdjVtx;
        vector<pMeshEnt>::iterator adjVtxIter;
        FMDB_Ent_GetAdj(regs[rgnIdx], FMDB_VERTEX, 0, vecAdjVtx);
        for (adjVtxIter = vecAdjVtx.begin();
             adjVtxIter != vecAdjVtx.end();
             adjVtxIter++) {
            verts.insert((pVertex) * adjVtxIter);
        }
    }

    fprintf(outFile, "<?xml version=\"1.0\"?>\n");
    fprintf(outFile, "<VTKFile type=\"UnstructuredGrid\" version=\"0.1\" byte_order=\"LittleEndian\">\n");
    fprintf(outFile, "<UnstructuredGrid>\n");

    fprintf(outFile, "<Piece NumberOfPoints=\"%zu\"  NumberOfCells=\"%d\">\n", verts.size(), numRgn);

    fprintf(outFile, "<PointData>\n");
    // write attached vtx data
    vector<pTag>::iterator tagIter;
    for ( tagIter = vtxTagsToWrite.begin();
          tagIter != vtxTagsToWrite.end();
          tagIter++ ) {
        const int tagLen = 512;
        char tagName[tagLen];
        FMDB_Tag_GetName(*tagIter, tagName, tagLen);
        // write the header
        fprintf(outFile, "<DataArray type=\"Float32\" Name=\"%s\" NumberOfComponents=\"1\" format=\"ascii\">\n", tagName);
        // loop over mesh vtx
        set<pVertex>::iterator vtxIter;
        for (vtxIter = verts.begin(); vtxIter != verts.end(); ++vtxIter) {
            // get the attached data
            double dblData;
            FMDB_Ent_GetDblTag(mesh, *vtxIter, *tagIter, &dblData);
            fprintf(outFile, " %.15f", dblData);
        }
        fprintf(outFile, "\n</DataArray>\n");
    }
    fprintf(outFile, "</PointData>\n");


    fprintf(outFile, "<CellData>\n");
    // write region attached data
    for ( tagIter = rgnTagsToWrite.begin();
          tagIter != rgnTagsToWrite.end();
          tagIter++ ) {
        const int tagLen = 512;
        char tagName[tagLen];
        FMDB_Tag_GetName(*tagIter, tagName, tagLen);
        // write the header
        fprintf(outFile, "<DataArray type=\"Float32\" Name=\"%s\" NumberOfComponents=\"1\" format=\"ascii\">\n", tagName);
        // loop over mesh regions
        for (int rgnIdx = 0; rgnIdx < numRgn; rgnIdx++) {
            // get the attached data
            double dblData;
            FMDB_Ent_GetDblTag(mesh, regs[rgnIdx], *tagIter, &dblData);
            fprintf(outFile, " %.15f", dblData);
        }
        fprintf(outFile, "\n</DataArray>\n");
    }
    fprintf(outFile, "</CellData>\n");

    fprintf(outFile, "<Points>\n ");
    fprintf(outFile, "<DataArray type=\"Float32\" NumberOfComponents=\"3\" format=\"ascii\">\n");

    // set the vertex IDs and write their coordinates
    pTag vertexIDTag;
    FMDB_Mesh_CreateTag(mesh, "vertex ids", SCUtil_INT, 1, vertexIDTag);

    double* vtxCoords = new double[3];
    int vtxID = 0;
    set<pVertex>::iterator vtxIter;
    for (vtxIter = verts.begin(); vtxIter != verts.end(); ++vtxIter) {
        FMDB_Ent_SetIntTag(mesh, *vtxIter, vertexIDTag, vtxID++);
        FMDB_Vtx_GetCoord(*vtxIter, &vtxCoords);
        fprintf(outFile, "  %.15f  %.15f  %.15f  \n", vtxCoords[0], vtxCoords[1], vtxCoords[2]);
    }
    delete [] vtxCoords;

    // Header for vtk
    fprintf(outFile, "\n</DataArray>\n");
    fprintf(outFile, "</Points>\n");

    // Header for vtk
    fprintf(outFile, "<Cells>\n");
    fprintf(outFile, "<DataArray type=\"Int32\"  Name=\"connectivity\" format=\"ascii\">\n");


    // write the vertex IDs for each region and compute the offsets into the vertex data for each region
    int* rgnVtxOffset = new int[numRgn+1];
    rgnVtxOffset[0] = 0;  // the first region indexes into the vertex data at position zero
    for (int rgnIdx = 0; rgnIdx < numRgn; ++rgnIdx) {
        vector<pMeshEnt> vecAdjVtx;
        vector<pMeshEnt>::iterator adjVtxIter;
        FMDB_Ent_GetAdj(regs[rgnIdx], FMDB_VERTEX, 0, vecAdjVtx);
        // compute the offset
        rgnVtxOffset[rgnIdx + 1] = vecAdjVtx.size() + rgnVtxOffset[rgnIdx];
        // write the vertex ids
        for (adjVtxIter = vecAdjVtx.begin(); adjVtxIter != vecAdjVtx.end(); adjVtxIter++) {
            FMDB_Ent_GetIntTag(mesh, *adjVtxIter, vertexIDTag, &vtxID);
            fprintf(outFile, " %d", vtxID);
        }
    }

    // done with the vertex IDs ... delete the tag data
    for (int rgnIdx = 0; rgnIdx < numRgn; ++rgnIdx) {
        vector<pMeshEnt> vecAdjVtx;
        vector<pMeshEnt>::iterator adjVtxIter;
        FMDB_Ent_GetAdj(regs[rgnIdx], FMDB_VERTEX, 0, vecAdjVtx);
        for (adjVtxIter = vecAdjVtx.begin(); adjVtxIter != vecAdjVtx.end(); adjVtxIter++) {
            FMDB_Ent_DelTag(*adjVtxIter, vertexIDTag);
        }
    }
    FMDB_Mesh_DelTag(mesh, vertexIDTag, 0);

    // write the vertex offsets for each region
    fprintf(outFile, "\n</DataArray>\n");
    fprintf(outFile, "<DataArray type=\"Int32\" Name=\"offsets\" format=\"ascii\">\n");
    for (int i = 1; i < numRgn + 1; ++i) {
        fprintf(outFile, " %d ", rgnVtxOffset[i]);
    }
    fprintf(outFile, "\n</DataArray>\n");
    delete [] rgnVtxOffset;

    // write the region types
    fprintf(outFile, "<DataArray type=\"UInt8\" Name=\"types\" format=\"ascii\">\n");
    for (int rgnIdx = 0; rgnIdx < numRgn; ++rgnIdx) {
        int entType = M_GetElementType(regs[rgnIdx]);
        switch (entType) {
            case TRI:
                fprintf(outFile, " 5 ");
                break;
            case QUAD:
                fprintf(outFile, " 9 ");
                break;
            case TET:
                fprintf(outFile, " 10 ");
                break;
            case PYRAMID:
                fprintf(outFile, " 14 ");
                break;
            case PRISM:
                fprintf(outFile, " 13 ");
                break;
            case HEX:
                fprintf(outFile, " 12  ");
                break;
            default:
                cout << "region topology [" << rgnIdx << "] NOT supported" << endl;
                exit(0);
        }
    }
    fprintf(outFile, "\n</DataArray>\n");
    fprintf(outFile, "</Cells>\n");

    fprintf(outFile, "</Piece>\n</UnstructuredGrid>\n</VTKFile>\n");
    fclose(outFile);

  return 0;
}

