/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#ifdef FMDB_PARALLEL
#include "mMeshIO.h"
#include "FMDB.h"
#include <iomanip>
#include <fstream>

// this function assumes one-level representation
//********************************************************
int exportMeshToSMS(pMeshMdl mesh, const char* fName)
//********************************************************
{
  pPart part;
  mEntity* ent;

  unsigned int tagId= MD_lookupMeshDataId("UniqPartBdryEntID");

// STEP 1: Count # part boundary entities owned by the local process
//         and compute initialId to start per process
  vector<mEntity*> ownedPartBdry_ents;                       // entity and local_pid 
  vector<mEntity*> partBdry_ents; 

 for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
  {
    part = *part_it;
    for (int dim=0; dim <3 ; ++dim)
    {
      for (mPart::iterall ent_it=part->beginall(dim);ent_it!=part->endall(dim);++ent_it)
      {
        ent = *ent_it;
        if (ent->getPClassification())  // on part bdry
        {
	  partBdry_ents.push_back(ent);                  
      	  if (ent->getOwnPartID(part)==part->getId())            
            ownedPartBdry_ents.push_back(ent);
	} 
      }  // end of it 
    }  // end of dim 
  } // for std::vector<pPart>::iterator
  
  // do communication to exchange numOwnedPBEnt to know initial part bdry entity id to assign on local process
  int* in = new int;
  int* out = new int[SCUTIL_CommSize()]; // out[i] has local_numOwnedPartBdryEnt of process i on all processes
    *in = ownedPartBdry_ents.size();

  MPI_Allgather(in, 1, MPI_INT, out, 1, MPI_INT, MPI_COMM_WORLD);

  int initialPartBdryID = 0;
  for (int i=0; i<SCUTIL_CommRank();++i)
    initialPartBdryID += out[i];

  int numTotalPartBdryEnt = 0;
  for (int i=0; i<SCUTIL_CommSize();++i)
    numTotalPartBdryEnt += out[i];

  delete in; 
  delete[] out; 

  // loop over ownedPartBdry_ents, send global unique id  to remote copies
  // when the remote copy receives the msg, attach as well

  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::All_Reduce);
  CM->set_tag(0);

  int msg_size = sizeof(pEntity) + sizeof(int); // mem address of remote copy entity and unique part bdry id
  CM->set_fixed_msg_size(msg_size);
  void* msg_send = (void*)CM->alloc_msg(msg_size); 
  int partbdry_uid;

  for (std::vector<pEntity>::iterator owned_partbdry_it=ownedPartBdry_ents.begin();
              owned_partbdry_it!=ownedPartBdry_ents.end();++owned_partbdry_it)
  {
    EN_attachDataInt(*owned_partbdry_it, tagId, initialPartBdryID);
    for (mEntity::RCIter rciter=(*owned_partbdry_it)->rcBegin(); rciter!=(*owned_partbdry_it)->rcEnd();++rciter)
    {
      pEntity* msg_ent = (pEntity*)msg_send; 
      *msg_ent = rciter->second;// remote copy address
      int* msg_int = (int*)((char*)msg_send+sizeof(pEntity));  
      *msg_int = initialPartBdryID;// unique id
      CM->send(rciter->first/ParUtil::Instance()->getTgtNumParts(), (void*)msg_send);
    }
    ++initialPartBdryID;
  }
  CM->finalize_send();
  CM->free_msg(msg_send);

  // receive phase begins -- on non-master
  void* msg_recv;
  int rank_from;
  while (CM->receive(msg_recv, &rank_from))
  {
    pEntity* msg_ent = (pEntity*)msg_recv;
    int* msg_int = (int*)((char*)msg_recv+sizeof(pEntity)); 
    EN_attachDataInt (*msg_ent, tagId, *msg_int);
    CM->free_msg(msg_recv);
  }  // if(rc)

  char outmesh[256];
  char without_extension[256];
  snprintf(without_extension,strlen(fName)-3,"%s",fName);

  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
  {
     part = *part_it;

     sprintf(outmesh,"%s%d.sms",without_extension, part->getId()); 
     std::ofstream ofs(outmesh); 
     
     ofs<<"sms 4\n";  

     ofs<<SCUTIL_CommSize()*ParUtil::Instance()->getTgtNumParts()<<" "<<part->getId()<<" "
	<<pmModel::Instance()->getNumPEs()<<" "<<numTotalPartBdryEnt<<endl;
     pmModel::PEIter peiter=pmModel::Instance()->peBegin();
     pmEntity* pe;
     for(;peiter!=pmModel::Instance()->peEnd();++peiter)
     {
       pe=(*peiter);
       ofs<<pe->getId()<<" "<<pe->getLevel()<<" "<<pe->getOwner()<<" "
	  <<pe->getNumBPs()<<" ";
       for (pmEntity::BPIter bpiter=pe->bpBegin();bpiter!=pe->bpEnd();++bpiter)
	 ofs<<*bpiter<<" ";
       ofs<<endl;
     }

     std::map<pVertex,int> numberid;	
     VIter vIter;
     EIter eIter;
     FIter fIter;
     RIter rIter;
     double loc[3];
     pVertex vertex;
     pRegion region;
     pEdge edge;
     pFace face;

     int gtype;
     int gtag;
     pGEntity gEntity;

     int NbRegions, NbFaces, NbEdges, NbVertices,NbPoints;
     NbRegions=part->size(3);
     NbFaces=part->size(2);
     NbEdges=part->size(1);
     NbVertices=part->size(0);
     NbPoints=NbVertices;

     ofs<< NbRegions << " " << NbFaces << " "<< NbEdges <<" " <<NbVertices << " " << NbPoints <<endl;

// STEP 5: write mesh vertices

     if (NbVertices>0)
     {
       vIter = M_vertexIter(part);
       int i=1;
	    
       while((vertex=VIter_next(vIter)))
       {  
	 numberid.insert(std::make_pair(vertex,i++));
	 V_coord(vertex,loc);
	 gEntity = vertex->getClassification(); 
	 if (gEntity!=0) 
         {
	   gtype=GEN_type(gEntity); 
	   gtag = GEN_tag(gEntity); 
	 }
	 else
         {
	   gtype=3; 
	   gtag=1; 
	 } 
      // write coordinates and gclassification
	 ofs << numberid[vertex]<<" "
          << gtag << " " << gtype << " "
	  << V_numEdges(vertex) << endl;
         ofs<<setprecision(15)<< loc[0] << " " << loc[1] 
	  << " " << loc[2]<<" "; 
      // write parametric
	 if (vertex->getData(FMDB_Util::Instance()->getParametric()))
	 {
	   SCOREC::Util::mVector vec = vertex->getAttachedVector(FMDB_Util::Instance()->getParametric());
	   switch (gtype)
	   {
	     case 1 :  // model edge
	     {  ofs << vec(0); break; }
	     case 2 : // model face
	     {  ofs << vec(0)<<" "<<vec(1)<<" 0"; break;  }
	     default: break;
	   }
	 }
	 else
	 {
	   switch (gtype)
	   {
          case 1 :  // model edge
	  {  ofs <<"0"; break; }
          case 2 : // model face
	  {  ofs <<"0 0 0"; break;  }
	  default: break;
	}
      }
      ofs << endl;
      // write pclassification
      if (vertex->getPClassification()) 
      {
        ofs<<vertex->getPClassification()->getId()<<" "; 
        EN_getDataInt(vertex, tagId, &partbdry_uid);
        ofs<<partbdry_uid<<" "<<vertex->getOwnPartID(part)<<endl;
      }
      else
	ofs<<"-1 -1 -1"<<endl;
    } // while vertex
    VIter_delete(vIter);
  }  // end of vertex

// STEP 6: write mesh edges

  if (NbEdges>0)
  {	
    eIter = M_edgeIter(part);
    int i=1;
    while((edge = EIter_next(eIter)))
    {
      gEntity = edge->getClassification(); 
      if(gEntity!=0) {
	gtype= GEN_type(gEntity); 
	gtag = GEN_tag(gEntity); 
      }
      else{
	gtype=3;
	gtag=1; 
      }
      ofs << gtag << " " << gtype << " " ;
      EN_setID(edge,i++);
      for (int j=0; j<2 ;j++)
      {
	vertex = E_vertex(edge,j);
        ofs << numberid[vertex] << " ";  
      } 		
      //ofs << E_numFaces(edge) << " " << "0"<<endl;
      int npts = E_numPoints(edge); 
      ofs << E_numFaces(edge) << " " << npts <<" "<<endl;

      if(npts) {
	double param[]={0.0, 0.0, 0.0};
        pPoint pt = E_point(edge, 0);
        ofs<<setprecision(15)<<P_x(pt)<<" "<<P_y(pt)<<" "<<P_z(pt);
        switch (E_whatInType(edge))
        {
	  case 1: param[0] = P_param1(pt);
		  ofs<<" "<<param[0];
                  break;
          case 2: P_param2(pt, &param[0], &param[1], 0);
 	 	  ofs<<" "<<param[0]<<" "<<param[1]<<" 0";
                  break;
          default: break;
        }
        ofs<<endl;
      }


      if (edge->getPClassification())
      {
        ofs<<edge->getPClassification()->getId()<<" "; 
        EN_getDataInt(edge, tagId, &partbdry_uid);
        ofs<<partbdry_uid<<" "<<edge->getOwnPartID(part)<<endl;
      }
      else
	ofs<<"-1 -1 -1"<<endl;
    }

    EIter_delete(eIter);
  }  // end of edges
  			
  if (NbFaces>0)
  {	  
    int i=1;
    fIter = M_faceIter(part);
    while((face = FIter_next(fIter)))
    {
      gEntity = face->getClassification(); 
      if(gEntity!=0) {
	gtype= GEN_type(gEntity);
	gtag = GEN_tag(gEntity); 
      }
      else{
	gtype=3;
	gtag=1; 
      }
      ofs << gtag << " " << gtype << " ";
      EN_setID(face,i++);
      int numEdges=F_numEdges(face);
      ofs << numEdges << " ";
      for (int j=0; j<numEdges ;j++)
      {
	edge = F_edge(face,j);
	int ID=EN_id(edge);
	if (!F_edgeDir(face,j)) ID=-ID;
	ofs << ID << " ";
      } 		
      ofs <<"0"<<endl; 
      if (face->getPClassification())
      {
        ofs<<face->getPClassification()->getId()<<" "; 
        EN_getDataInt(face, tagId, &partbdry_uid);
        ofs<<partbdry_uid<<" "<<face->getOwnPartID(part)<<endl;
      }
      else
	ofs<<"-1 -1 -1"<<endl;
    }  // while face
    FIter_delete(fIter);
  }  // end of faces
  	
  if (NbRegions>0)
  {
    int i=1;
    rIter = M_regionIter(part);
    while((region = RIter_next(rIter)))
    { 
      gEntity =(pGEntity)R_whatIn(region);
      if(gEntity!=0) 
	gtag = GEN_tag(gEntity);
      else
	gtag = 1; 
      int numFaces=R_numFaces(region);
      ofs << gtag << " " << numFaces << " ";		
      EN_setID(region,i++);
      for (int j=0; j<numFaces ;j++)
      {		  
	face = R_face(region,j);
	int ID=EN_id(face);
	if (!R_faceDir(region,j)) ID=-ID;
	ofs << ID << " ";
      } 		
      ofs << "0"<<endl;
    }
    RIter_delete(rIter);
  } // end of regions 
  
    int NbPartSets = part->numPSet(); 
    if (NbPartSets) // save Part Set info
    {
      ofs<<"PartSet\n " << NbPartSets << endl;

      mEntitySet* pset;
      int pset_cnt=0;
      for (mPart::psetIter pset_it=part->beginPSet(); pset_it!=part->endPSet(); ++pset_it)
      {
        ofs << 1 << " " <<(*pset_it)->numEntity()<<endl;
        if ((*pset_it)->numEntity()==0) continue;
        for (mEntityGroup::entIter ent_it=((mEntityGroup*)(*pset_it))->begin(); 
                ent_it!=((mEntityGroup*)(*pset_it))->end(); ++ent_it)
        {
          if ((*ent_it)->getLevel())
            ofs << (*ent_it)->getLevel() << " " << EN_id(*ent_it) << endl;
          else
            ofs<< "0 " << numberid[*ent_it] << endl;
        }
      } // for each ent set
    } // if part->numEntGrp()>0
    ofs.close();
  } // end of ipart 

  // STEP 7: deallocate the attached data for ownedPartBdry_ents 
  for (vector<mEntity*>::iterator pbdry_it = partBdry_ents.begin(); pbdry_it!=partBdry_ents.end(); ++pbdry_it) 
    EN_deleteData(*pbdry_it, tagId); 
  MD_deleteMeshDataId(tagId); 

  return SCUtil_SUCCESS;
}

#endif
