/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#ifdef FMDB_PARALLEL
#include <iomanip>
#include <fstream>
#include "mMeshIO.h"
#include "FMDB.h"
#include "pmMigration.h"

struct msg_struct
{
  mEntity* ent;
  int partbdry_uid;
  int source_partid; 
  msg_struct(mEntity* e, int i, int j): ent(e), partbdry_uid(i), source_partid(j){}
};

typedef map<int, mEntity*> PartBdryEntMapType;

int load_part(mPart* part, const char* fName, int, int, PartBdryEntMapType&);

//********************************************************
int import_parallel_multipart_SMS (mMesh* mesh, const char* fName)
//********************************************************
{
// *********************************
// test mesh file availability
// *********************************
  char inmesh[256]; 
  char without_extension[256];
  snprintf(without_extension,strlen(fName)-3,"%s",fName);
  sprintf(inmesh,"%s0.sms",without_extension); 
  FILE* test_in = fopen (inmesh,"r");
  if(!test_in)
  {
    cout<<"["<<SCUTIL_CommRank()<<"] FMDB ERROR: unable to open file \""<<inmesh<<"\"\n";
    fclose (test_in);
    return SCUtil_FILE_NOT_FOUND;  
  }

  int maxNumPart=ParUtil::Instance()->getTgtNumParts();
  int partid_offset = SCUTIL_CommGrp()*SCUTIL_GrpSize()*maxNumPart;

  // SET partition model
  int mypartid=FMDB_Part_GlobID(mesh->getPart(0))-partid_offset;
  sprintf(inmesh,"%s%d.sms",without_extension, mypartid);   
  FILE* pmodel_in = fopen (inmesh,"r");

  if (!pmodel_in)
  {
    fclose(pmodel_in);
    cout<<"["<<SCUTIL_CommRank()<<"] setup pmodel fails due to file open failure\n";
  }

  char line[256];
  int Version, total_parts, my_part, num_pent, num_partbdry_ent;
  int pent_id, pent_dim, pent_owner, pent_num_respart;

  fscanf(pmodel_in,"%s %d",line,&Version);
  fscanf(pmodel_in,"%d %d %d %d",&total_parts, &my_part, &num_pent, &num_partbdry_ent);

  pmModel* pmodel=pmModel::Instance();
  
  for (int i=0; i<num_pent;++i)
  {
    fscanf(pmodel_in,"%d %d %d %d",&pent_id, &pent_dim, &pent_owner, &pent_num_respart);
    set<int> res_part_set;
    
    int res_part;
    for (int j=0; j<pent_num_respart; ++j)
    {  
      fscanf(pmodel_in,"%d",&res_part);
      res_part_set.insert(res_part+partid_offset);
    }
    pmEntity* newPE = new pmEntity(pent_id, res_part_set, pent_dim);   
    pmodel->addPEntity(newPE); 
    newPE->setOwner(pent_owner+partid_offset);
  } 
  fclose (pmodel_in);

// *********************************
// READ individual part 
// ********************************* 
  PartBdryEntMapType PartBdryEntMap[maxNumPart];
  int local_dim=2, global_dim=2;
  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)  
  {
    mypartid=FMDB_Part_GlobID(*part_it)-partid_offset;
    sprintf(inmesh,"%s%d.sms",without_extension, mypartid);   
    load_part (*part_it, inmesh, mypartid, partid_offset, PartBdryEntMap[FMDB_Part_LocalID(*part_it)]);
    if ((*part_it)->size(3)) local_dim=3;
  }

// *********************************
// setup part boundary
// *********************************

  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::All_Reduce);
  int source_partid, msg_size = sizeof(mEntity*)+sizeof(int)*3; 
  CM->set_fixed_msg_size(msg_size);
  void* msg_send = CM->alloc_msg(msg_size);
  int* send_int; 
  mEntity** send_ent; 
  CM->set_tag(0);

  for (int local_partid=0; local_partid<maxNumPart; ++local_partid)
  {
    for (PartBdryEntMapType::iterator pbdry_it=PartBdryEntMap[local_partid].begin(); 
           pbdry_it!=PartBdryEntMap[local_partid].end(); ++pbdry_it) 
    {
      source_partid=SCUTIL_CommRank()*maxNumPart+local_partid;
      for (pmEntity::BPIter partid_it= pbdry_it->second->getPClassification()->bpBegin(); 
             partid_it != pbdry_it->second->getPClassification()->bpEnd(); ++partid_it)
      {
        if (*partid_it!=source_partid)
        {
          send_ent = (mEntity**)msg_send; 
          *send_ent = pbdry_it->second;
          send_int =  (int*)((char*)msg_send + sizeof(mEntity*)); 
          send_int[0] = pbdry_it->first;
          send_int[1] = source_partid;
          send_int[2] = *partid_it; // dest_partid
          CM->send((*partid_it)/maxNumPart, (void*)msg_send);
        } // if
      } // for
    } 
  }
  CM->finalize_send();
  CM->free_msg(msg_send);

  // receive 
  void* msg_recv; 
  int pid_from;
  while (CM->receive(msg_recv, &pid_from))
  {
    mEntity* remote_ent = *(mEntity**)msg_recv;
    int* recv_int = (int*)((char*)msg_recv + sizeof(mEntity*));
#ifdef DEBUG
    if (recv_int[2]-SCUTIL_CommRank()*maxNumPart>=maxNumPart)
      cout<<"["<<SCUTIL_CommRank()<<"] FMDB ERROR: dest_partid="<<recv_int[2]<<", local dest_partid=="<<recv_int[2]-pid_from*maxNumPart<<", num part per proc="<<maxNumPart<<endl;
#endif
    PartBdryEntMap[recv_int[2]-SCUTIL_CommRank()*maxNumPart][recv_int[0]]->addRemoteCopy(recv_int[1], remote_ent);
    CM->free_msg(msg_recv);
  } //  while(int rc = CM->receive(msg_recv, &pid_from))
  
// *********************************
// SET global variable - mesh dimension and num parts
// *********************************
  MPI_Allreduce(&local_dim, &global_dim, 1, MPI_INT, MPI_MAX, ParUtil::Instance()->getComm());
  ParUtil::Instance()->set_maxMeshDim(global_dim);

  ParUtil::Instance()->setCurNumParts(maxNumPart);

// *********************************
// remove pmodel entity with NO classified mesh entities and update neighbor proc info
// *********************************
  vector<pmEntity*> pmEntityToRemove;
  for(pmModel::PEIter peit=pmModel::Instance()->peBegin(); peit!=pmModel::Instance()->peEnd(); ++peit)
    if ((*peit)->getNumRClasEnt()==0)
      pmEntityToRemove.push_back(*peit);

  for (vector<pmEntity*>::iterator peit=pmEntityToRemove.begin(); peit!=pmEntityToRemove.end();++peit)
    pmModel::Instance()->delPEntity(*peit);

  vector<int> NborProc;
  FMDB_Mesh_GetNborProcID (mesh, NborProc);

  set<int>SetNborProc;
  std::copy(NborProc.begin(), NborProc.end(), inserter(SetNborProc, SetNborProc.end()));
  ParUtil::Instance()->ComMan()->set_Nbrs(SetNborProc);


  return SCUtil_SUCCESS;
}

//********************************************************
int load_part (mPart* part, const char* fName, int mypartid, int partid_offset, 
              PartBdryEntMapType& PartBdryEntMap)
//********************************************************
{
  int maxNumPart=ParUtil::Instance()->getTgtNumParts();
  char line[256];
  int i, j, NbRegions=0, NbFaces=0, NbEdges, NbVertices, NbPoints;
  int GEntityType,GEntityId,EntityNbConnections;
  int Dummy, Version=3, VertexId1,VertexId2, Edge1, Face1;
  int nbPts,NbEdgesOnFace,NbFacesOnRegion;
  double x, y, z, u, v;
  int patch, pclas_id, partbdry_uid, owning_partid;
  mEntity* ent;

  FILE* in = fopen (fName,"r");

  if (!in)
  {
    fclose(in);
    return SCUtil_FILE_NOT_FOUND;
  }

// *********************************
// STEP 1: setup format
// *********************************
  fscanf(in,"%s %d",line,&Version);
  typedef GEntity* (*entityBy_FP)(SGModel*,int,int);
  entityBy_FP fp = GM_entityByTag; 

// *********************************
// STEP 2: check number of partitions
// *********************************
  int pid, total, numPEs, totalOwningPartBdryEnts;                         // local totalOwningPartBdryEnts 
  fscanf(in,"%d %d %d %d",&total, &pid, &numPEs, &totalOwningPartBdryEnts);
  if (total>SCUTIL_GrpSize()*maxNumPart || pid+partid_offset != FMDB_Part_GlobID(part))
  {
#ifdef DEBUG
    cout<<"["<<SCUTIL_CommRank()<<"] total #parts="<<total
        <<", SCUTIL_GrpSize()*maxNumPart="<<SCUTIL_GrpSize()*maxNumPart
        <<", fileid+partid_offset="<<pid+partid_offset
        <<", local partid="<< FMDB_Part_GlobID(part)<<endl;
#endif
    return SCUtil_FILE_NOT_FOUND;
  }

// *********************************
// STEP 3: setup pmodel entity map
// *********************************
  int pent_id, pedim, peowner, penumBPs;
  pmModel* pmodel=pmModel::Instance();
  pmEntity* pmodel_ent;
  map<int, pmEntity*> pmodel_ent_map;
  set<int> bps;  
  for (i=0; i<numPEs;++i)
  {
    fscanf(in,"%d %d %d %d",&pent_id, &pedim, &peowner, &penumBPs);
    bps.clear();
    int bp;
    for (int j=0; j<penumBPs;++j)
    {  
      fscanf(in,"%d",&bp);
      bps.insert(bp+partid_offset);
    }
    pmodel_ent_map.insert(std::map<int, pmEntity*>::value_type(pent_id, pmodel->getPartitionEntity(bps)));
  } 

// *********************************
// STEP 4: read # entities per type
// *********************************
  fscanf(in,"%d %d %d %d %d",&NbRegions,&NbFaces,&NbEdges,&NbVertices,&NbPoints);

// *********************************
// STEP 5: read vertices
// *********************************
  std::vector<mVertex*> vertices;
  mVertex* vv;        
  int vid;
  for(i=0;i<NbVertices;i++)
  {
    fscanf(in,"%d",&vid);       
    fscanf(in,"%d",&GEntityId);       
    if (!GEntityId) continue;
    fscanf(in,"%d %d %lf %lf %lf",&GEntityType,&EntityNbConnections,&x,&y,&z); 
    vv = part->createVertex(vid,x,y,z,part->getGEntity(GEntityId,GEntityType,fp)); 
    vertices.push_back(vv);
	    
    switch(GEntityType)
    {
      case 0: break;
      case 1: fscanf(in,"%le",&u);
              vv->attachVector(FMDB_Util::Instance()->getParametric(),SCOREC::Util::mVector (u,0,0) );
              break;
      case 2: fscanf(in,"%le %le %d",&u,&v,&patch);		
  	      vv->attachVector(FMDB_Util::Instance()->getParametric(),SCOREC::Util::mVector (u,v,patch) );
   	      break;
      case 3: break;
      default: return SCUtil_INVALID_ENTITY_TYPE;
    } // switch(GEntityType)

    if (Version==4) 
      fscanf(in,"%d %d %d", &pclas_id, &partbdry_uid, &owning_partid);
    else 
      fscanf(in,"%d %d", &pclas_id, &partbdry_uid);

    if (pclas_id!=-1) // part bdry entity
    {
      vv->setPClassification(pmodel_ent_map[pclas_id]);
      PartBdryEntMap.insert(PartBdryEntMapType::value_type(partbdry_uid, (mEntity*)vv));
    }
  } // for(i=0;i<NbVertices;i++)
   
// *********************************
// STEP 6: read edges
// *********************************
  std::vector<mEntity*> edges;
  mVertex *v1, *v2;
  mEntity* e;
    
  for (i=0;i<NbEdges;i++)
  {
    fscanf(in,"%d",&GEntityId);
    if (!GEntityId) continue;
      
    fscanf(in,"%d %d %d %d %d",&GEntityType, &VertexId1,&VertexId2,&EntityNbConnections,&nbPts); 
    v1 = vertices[VertexId1-1];
    v2 = vertices[VertexId2-1];
    e = M_createE(part, v1, v2, part->getGEntity(GEntityId,GEntityType,fp));
    edges.push_back(e);
 
    for(int j=0;j<nbPts;j++)
    {
      fscanf(in,"%lf %lf %lf ", &x, &y, &z);
      pPoint pt = P_new();
      P_setPos(pt, x, y, z);
      switch(GEntityType)
      {
        case 1: fscanf(in,"%le",&u);
                FMDB_P_setParametricPos(pt, u, 0, 0);
                break;
        case 2: fscanf(in,"%le %le %d",&u,&v,&patch); 
                FMDB_P_setParametricPos(pt, u, v, 0);
                break;
        default: return SCUtil_INVALID_ENTITY_TYPE;
      } // switch(GEntityType)
      E_setPoint(e, pt);
    } // for(int j=0;j<nbPts;j++)

    if (Version==4) 
      fscanf(in,"%d %d %d", &pclas_id, &partbdry_uid, &owning_partid);
    else 
      fscanf(in,"%d %d", &pclas_id, &partbdry_uid);

    if (pclas_id!=-1) // part bdry entity
    {
      e->setPClassification(pmodel_ent_map[pclas_id]);
      PartBdryEntMap.insert(PartBdryEntMapType::value_type(partbdry_uid, (mEntity*)e));
    }
  }  // for(i=0;i<NbEdges;i++)

// *********************************
// STEP 7: read faces
// *********************************
  std::vector<mEntity*> faces;
  mEntity* f;
  mEntity* eg[4];
  
  for (i=0;i<NbFaces;i++)
  {
    fscanf(in,"%d",&GEntityId);
    if (!GEntityId) continue;
    fscanf(in,"%d %d",&GEntityType, &NbEdgesOnFace);
    for (j=0; j<NbEdgesOnFace; j++)
    {
      fscanf(in, "%d", &Edge1);
      eg[j]=edges[abs(Edge1)-1]; 
    }

    fscanf(in, "%d", &nbPts);
    f = M_createF(part, NbEdgesOnFace, eg, 0, part->getGEntity(GEntityId,GEntityType,fp));       
    faces.push_back(f);

    for(int j=0;j<nbPts;j++)
    {
      switch(GEntityType)
      {
	case 0: break;
	case 1: fscanf(in,"%le",&u);
	        break;
	case 2: fscanf(in,"%le %le %d",&u,&v,&patch);
	        break;
        case 3: break;
        default: return SCUtil_INVALID_ENTITY_TYPE;
      } // switch(GEntityType)
    } // for(int j=0;j<nbPts;j++)

    if (Version==4) 
      fscanf(in,"%d %d %d", &pclas_id, &partbdry_uid, &owning_partid);
    else 
      fscanf(in,"%d %d", &pclas_id, &partbdry_uid);

    if (pclas_id!=-1) // part bdry entity
    {
      f->setPClassification(pmodel_ent_map[pclas_id]);
      PartBdryEntMap.insert(PartBdryEntMapType::value_type(partbdry_uid, (mEntity*)f));
    }
  } // for(i=0;i<NbFaces;i++)
  
// *********************************
// STEP 8: read regions
// *********************************
  std::vector<mEntity*> regions;
  mEntity* fc[6];
  mEntity* r;
  for (i=0;i<NbRegions;i++)
  {
    fscanf(in,"%d",&GEntityId);
    if (!GEntityId) continue;
    fscanf(in,"%d",&NbFacesOnRegion);
     
    for(int j=0; j<NbFacesOnRegion; j++)
    {
      fscanf(in,"%d", &Face1);
      fc[j]=faces[abs(Face1)-1];
    }
    fscanf(in,"%d", &Dummy);
      
    r=M_createR(part,NbFacesOnRegion,fc,0,part->getGEntity(GEntityId,3,fp));
    regions.push_back(r);
  } // for(i=0;i<NbRegions;i++)
 
// ************************************
//    STEP 9: Read Entity Sets in Part Level
// ************************************
  int NbEntSets=0, MATCHING=0;
  pEntSet eset;
  if (fscanf(in, "%s", line)!= EOF)
    MATCHING=1;
  if (MATCHING==1 && !strcmp(line,"PartSet"))
  {
    fscanf(in, "%d", &NbEntSets);
    int isList, numEnts, entType, entId;
 
    int eset_cnt=0;
    for (; eset_cnt<NbEntSets; ++eset_cnt)
    {
      fscanf(in, "%d %d", &isList, &numEnts);
      if (!numEnts) continue;
      eset = part->createPSet(part->getDimension() );
      eset->attachInt(FMDB_Util::Instance()->getId(), eset_cnt); // attach ID starting from NbMeshSets
      for (int ent_cnt=0; ent_cnt<numEnts; ++ent_cnt)
      {
        fscanf(in, "%d %d", &entType, &entId);
        switch(entType)
        {
          case 0: ((mEntityGroup*)eset)->addEntity(vertices[entId-1]);
                  break;
          case 1: ((mEntityGroup*)eset)->addEntity(edges[entId-1]);
                  break;
          case 2: ((mEntityGroup*)eset)->addEntity(faces[entId-1]);
                  break;
          case 3: ((mEntityGroup*)eset)->addEntity(regions[entId-1]);
                  break;
          default: break;
        }  // switch (entType)
      } // for ent_cnt
    } // for (; eset_cnt<NbEntSets; ++eset_cnt)

    if (fscanf(in, "%s", line) == EOF)
      MATCHING=0;
  }  // if (MATCHING==1 && strcmp(line,"PartSet")) 
  fclose (in);

  return SCUtil_SUCCESS; 
}
#endif
