/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/ 
#ifdef FMDB_PARALLEL
#include <map>
#include "FMDB.h"
void  printEntToMigr(std::map<mEntity*, int*>& EntToMigr)       
{
  std::map<mEntity*, int*>::const_iterator mapit;  
  int mypid=ParUtil::Instance()->rank();
  for (mapit=EntToMigr.begin(); mapit!=EntToMigr.end();++mapit)
    cout<<"("<<mypid<<") move "<<mapit->first->getUid()<<" to Part "<<(mapit->second)[1]<<" (proc "<<(mapit->second)[2]<<")\n";
}

void pmLoadBalance(mPart* part, pmMigrationCallbacks &cb)
{
   vector<mPart*> meshes; 
   meshes.push_back(part);
   pmLoadBalance(meshes, cb);
}


// *****************************************
void pmLoadBalance(std::vector<mPart*>& parts, pmMigrationCallbacks &cb)
// *****************************************
{
  if (M_globalMinDim(parts[0])==0)
    ParUtil::Instance()->Msg(ParUtil::INFO,"\n*** LOAD BALANCING MAY NOT WORK DUE TO EMPTY MESH ***\n\n");     

  mPart* part = parts[0]; 
  // assume all part meshes have the same dimension
  int from = M_globalMaxDim(part); 
  ParUtil::Instance()->set_maxMeshDim(from);

  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::Neighbors);

/**< STEP 1: get partition objects from Zoltan */
  std::map<mEntity*,int*> EntToMigr;               // EntToMigr: entity_pointer, local_pid + tgt_glb_pid + tgt_proc_rank
  std::map<mEntityGroup*, int*> SetToMigr; 
  cb.partition(parts, EntToMigr, SetToMigr);  

/**< STEP 2: migrate entities */ 
  CM->set_comm_validation(IPComMan::All_Reduce);

//  int numPart    = ParUtil::Instance()->getCurNumParts();    // current_local_num_parts
  int numPart = (int)parts.size(); 
  int usrNumPart = ParUtil::Instance()->getUsrNumParts();    // usr_local_num_parts is set during partition

  // when local_num_part is increased 
  pGModel model = part->getSolidModel();
  if(numPart<usrNumPart)  {
    for(int i= numPart ; i<usrNumPart; ++i) {

      mPart* tmpPart = MS_newMesh(model);
      parts.push_back(tmpPart);
    }
  }
	      
  std::list<mEntity*> rmE[4], newE[4];
  std::vector<mEntityGroup*> rmEG; 
  std::vector<mEntityGroup*> newEG; 
  migratePtnObjects(part->getMesh(), EntToMigr, SetToMigr, cb, rmE, newE, rmEG, newEG); 
}

// *****************************************
int pmMerge(mPart* part, pmMigrationCallbacks &cb)
// *****************************************
{
  if (ParUtil::Instance()->size()==1)  return SCUtil_SUCCESS;

  int mDim = M_globalMaxDim(part); 
#ifdef _DEBUG
  double t1, t2;
#endif  
// STEP 1: get partition objects to move
  std::map<mEntity*,int> EntToMigr;
#ifdef _DEBUG
  t1=ParUtil::Instance()->wTime();
#endif

// determine the processor to merge
  int numPE = ParUtil::Instance()->size();
  int myPid = ParUtil::Instance()->rank();
  int* numR = new int[numPE];
  
  // send phase begins
  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::All_Reduce);

  CM->set_tag(0);
  CM->set_fixed_msg_size(sizeof(mEntity*));
  int* msg_send = (int*)CM->alloc_msg(sizeof(mEntity*));

  int num_sent = 0, num_recvd = 0;
 
  for (int pid=0; pid<numPE;++pid)
  {
    if (pid==P_pid()) continue;
    *msg_send = part->size(mDim);
    CM->send(pid, (void*)msg_send);
    num_sent++;
  }
  CM->finalize_send();
  CM->free_msg(msg_send);  

  // receive phase begins
  void *msg_recv;
  int pid_from;
  while (CM->receive(msg_recv, &pid_from))
  {
    num_recvd++;
    numR[pid_from] = *(int*)msg_recv;
    CM->free_msg(msg_recv);
  }

//  ParUtil::Instance()->Barrier(__LINE__,__FILE__);
  numR[myPid] = part->size(mDim);
  int pidToMerge = myPid;
  int numRMax = part->size(mDim);
  
  for (int k=0; k<numPE; ++k)
    if (numR[k] > numRMax)
    {
      numRMax = numR[k];
      pidToMerge = k;
    }
    
  delete [] numR;

  pidToMerge = P_getMinInt(pidToMerge);
  ParUtil::Instance()->Msg(ParUtil::INFO,"\tPID TO MERGE = %d\n",pidToMerge);

  if (ParUtil::Instance()->rank() != pidToMerge)
    for (mPart::iterall it=part->beginall(mDim); it!=part->endall(mDim); ++it)
    {
  // move all entities to a master processor
      EntToMigr.insert(entityPidMap::value_type(*it,pidToMerge)); 
    }

#ifdef _DEBUG  
  t2=ParUtil::Instance()->wTime();
  ParUtil::Instance()->Msg(ParUtil::INFO,"\t* %f (sec) get entities to move\n",t2-t1);
#endif
// STEP 2: migrate entities

  /// Restore neighbors
  std::vector<int> NborProc;
  FMDB_Mesh_GetNborProcID (part->getMesh(), NborProc);

  std::set<int> SetNborProc;
  std::copy(NborProc.begin(), NborProc.end(), inserter(SetNborProc, SetNborProc.end()));
  ParUtil::Instance()->ComMan()->set_Nbrs(SetNborProc);
  CM->set_comm_validation(IPComMan::Neighbors);

  list<mEntity*> rmE[4];
  list<mEntity*> newE[4];
  std::vector<mEntityGroup*> rmEG;
  std::vector<mEntityGroup*> newEG; 
  return migrateMeshEntities(part, EntToMigr, cb, rmE, newE, rmEG, newEG);
}

int pmPartMigration(std::vector<mPart*>& parts, 
		    int source_pid, 
		    int target_rank, 
		    pmMigrationCallbacks &cb)
{ 
  int numPart, sender, mypid, source_lid; 
  mPart* part; 
  int dim; 

// ********************************************
// STEP 1: initialize localPids if necessary 
// ********************************************
  int size = (ParUtil::Instance()->localPids).size(); 
  if(size==0) 
    for(int ivec=0; ivec<(int)parts.size(); ivec++)
      (ParUtil::Instance()->localPids).push_back(ivec);  

// ********************************************
// STEP 2: get source part and 
//         send request to target process
// ********************************************
  numPart = ParUtil::Instance()->getCurNumParts();    // max local num parts per process 
  sender = source_pid / numPart;
  mypid = ParUtil::Instance()->rank(); 
  source_lid = source_pid % numPart;   
 
  IPComMan *CM = ParUtil::Instance()->ComMan();
  set<int> nbrs; 
  if(mypid==sender)
    nbrs.insert(target_rank);
  if(mypid==target_rank)
    nbrs.insert(sender); 
  CM->set_Nbrs(nbrs);
  
  CM->set_comm_validation(IPComMan::Neighbors);
  CM->set_tag(0); 
  int msg_size=sizeof(int);
  int num_sent=0, num_recvd=0; 
 
  if(mypid==sender) {
    part=NULL; 
    for(int ivec=0; ivec<(int)parts.size(); ivec++)
      if((ParUtil::Instance()->localPids)[ivec]==source_lid) {
	part = parts[ivec];
	break; 
      }
    if(part==NULL)                       // should throw an exception 
      throw new mException(__LINE__,__FILE__,"Invalid argument of source_global_part_id in part migration."); 
      
    FMDB_Part_GetDim(part, &dim);

    int* msg_send = (int*)CM->alloc_msg(msg_size);
    *msg_send = dim; 
    CM->send(target_rank, (void*)msg_send, msg_size);
    num_sent++; 
    CM->free_msg(msg_send);
  }

  CM->finalize_send(); 
  
// ********************************************  
// STEP 3: create new global pid on target process  
// ********************************************   
   void* msg_recv;
   int pid_from=-1;

   // receive phase
   int count=0; 
   while(CM->receive(msg_recv, &pid_from))
   {
      num_recvd++;
   
    // make sure the request is valid, size will not exceed numPart 
    if((ParUtil::Instance()->localPids).size()>0)
      count = (ParUtil::Instance()->localPids).back()+1; 
    if (count>=numPart) 
      throw new mException(__LINE__,__FILE__,"Invalid argument of target_process in part migration.");
    
    // create a new part 
    part=parts[0];                                    
    pGModel model = part->getSolidModel();
    mPart* tmpPart = MS_newMesh(model);
    parts.push_back(tmpPart);
    
    // create a new global pid for the new part 
    (ParUtil::Instance()->localPids).push_back(count); 
    int usrNumPart = ParUtil::Instance()->getUsrNumParts(); 
    ParUtil::Instance()->setUsrNumParts(usrNumPart+1); 
  }
  
   CM->set_tag(0);
   num_sent=0; 
   num_recvd=0; 
 
  if (pid_from !=-1) {
   int* msg_send = (int*)CM->alloc_msg(msg_size);
   *msg_send = count; 
   CM->send_to_nbrs((void*)msg_send, msg_size);
   CM->free_msg(msg_send);
   num_sent++; 
  }
  CM->finalize_send(); 

// ********************************************    
// STEP 4: create EntToMigr 
// ********************************************  
  std::map<mEntity*,int*> EntToMigr;   
  mEntity* ent; 
  
  // receive phase
  while(CM->receive(msg_recv, &pid_from))
  {
    num_recvd++; 
    int* count = (int*)msg_recv; 

    for (mPart::iterall it=part->beginall(dim); it!=part->endall(dim);++it){ 
      ent=*it; 
      int* pid=new int[3]; 
      pid[0]= source_lid; 
      pid[1]= target_rank*numPart+ *count;  
      pid[2]= target_rank; 
      EntToMigr[ent]=pid; 
    }
   
  int usrNumPart = ParUtil::Instance()->getUsrNumParts(); 
  ParUtil::Instance()->setUsrNumParts(usrNumPart-1); 
    
    // marks the part to move  
   vector<int>::iterator vecit; 
   vecit = std::find((ParUtil::Instance()->localPids).begin(), (ParUtil::Instance()->localPids).end(), source_lid);
    //  (ParUtil::Instance()->localPids).erase(vecit);     
   int lid = *vecit; 
   *vecit = lid-(ParUtil::Instance()->localPids).size(); 
  }

// ********************************************  
// STEP 5: call mesh entity migration 
// ********************************************  
  return migrateMeshEntities(parts, EntToMigr, cb);  
}

// *****************************************
int pmPartMigration(mPart* part,
                    int dimToMove,list<mEntity*>& entities,
                    pmMigrationCallbacks &cb,
                    std::list<mEntity*>* rmE,
                    std::list<mEntity*>* newE
                    , std::vector<mEntityGroup*>& rmEG
                    , std::vector<mEntityGroup*>& newEG)
// *****************************************
{
  if (ParUtil::Instance()->size()==1)  return SCUtil_SUCCESS;


//trace_start();
//summary_start();

  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::Neighbors);

// STEP 1: unify origin entities to move
  unifyEntitiesToMove(part, entities);

// STEP 2: get partition objects to move
  std::map<mEntity*,int> EntToMigr;
  getPOsToMove(part, entities, dimToMove, EntToMigr);
//  getPOsToMove_nbrs(part, entities, dimToMove, EntToMigr);
//  getPOsToMove_nbrs2(part, entities, dimToMove, EntToMigr); 

// STEP 3: migrate entities
  return migrateMeshEntities(part, EntToMigr, cb, rmE, newE, rmEG, newEG);
}

int pmPartMigration(mPart* part, pmMigrationCallbacks &cb,
 		    std::list<mEntity*>& vtsOnCB, 
		    std::list<std::pair<mEntity*,mEntity*> > &vtfcPairs,
                    std::list<mEntity*>* rmE,
                    std::list<mEntity*>* newE
                    , std::vector<mEntityGroup*>& rmEG
                    , std::vector<mEntityGroup*>& newEG)
{
  if (ParUtil::Instance()->size()==1)  return SCUtil_SUCCESS;

  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::Neighbors);

// STEP 1: unify CB vertices to move
  unifyEntitiesToMove(part, vtsOnCB);

// STEP 2: get partition objects to move
  std::map<mEntity*,int> EntToMigr;
// old method written by Li
//  getEntitiesToMove(mesh,vtsOnCB, vtfcPairs, EntToMigr); 

// new method written by Seol
// this much improve part quality and doesn't cause crash
  getPOsToMoveForSnap(part,vtsOnCB, vtfcPairs, EntToMigr); 

// STEP 3: migrate entities
  return migrateMeshEntities(part, EntToMigr, cb, rmE, newE, rmEG, newEG);
}


// *****************************************
int migrateMeshEntities(mPart* part,
                        std::map<mEntity*,int>& EntToMigr,
                        pmMigrationCallbacks& cb)
// *****************************************
{
  std::list<mEntity*> rmE[4]; 
  std::list<mEntity*> newE[4]; 
  std::vector<mEntityGroup*> rmEG;
  std::vector<mEntityGroup*> newEG; 
  return migrateMeshEntities(part, EntToMigr, cb, rmE, newE, rmEG, newEG);
}


// *****************************************
int migrateMeshEntities(mPart* part,
                        std::map<mEntity*,int>& EntToMigr,
                        pmMigrationCallbacks& cb,
                        std::list<mEntity*>* rmE,
                        std::list<mEntity*>* newE
                        , std::vector<mEntityGroup*>& rmEG
                        , std::vector<mEntityGroup*>& newEG)				  
// *****************************************
{
   vector<mPart*> parts;
   parts.push_back(part);

   // construct new EntToMigr structure 
   std::map<mEntity*, int*> pos;
   std::map<mEntity*, int>::iterator iter;
   for(iter=EntToMigr.begin(); iter!=EntToMigr.end(); iter++) {
      int* pids=new int[3];
      pids[0]=0;
      pids[1]=iter->second;
      pids[2]=iter->second;
      pos[iter->first]=pids;
   }
   return migrateMeshEntities(parts, pos, cb, rmE, newE, rmEG, newEG);
}


// *****************************************
int migrateMeshEntities(std::vector<mPart*>& parts, 
                        std::map<mEntity*,int*>& EntToMigr,
		        pmMigrationCallbacks& cb)
// *****************************************			
{
  list<mEntity*> rmE[4];
  list<mEntity*> newE[4];

  std::vector<mEntityGroup*> rmEG; 
  std::vector<mEntityGroup*> newEG; 
  return migrateMeshEntities(parts, EntToMigr, cb,rmE,newE, rmEG, newEG);
}

int migrateMeshEntities(std::vector<mPart*>& parts,
                        std::map<mEntity*,int*>& EntToMigr,
                        pmMigrationCallbacks& cb,
                        std::list<mEntity*>* rmE,
                        std::list<mEntity*>* newE,
			std::vector<mEntityGroup*>& rmEG,
                        std::vector<mEntityGroup*>& newEG)
{
   std::map<mEntityGroup*,int*> SetToMigr; 
   getEntityGroupsToHandle(EntToMigr, SetToMigr);
   migratePtnObjects(parts.at(0)->getMesh(), EntToMigr, SetToMigr, cb, rmE, newE, rmEG, newEG);
   return 0;
}

// *****************************************
int migratePtnObjects (pMeshMdl mesh, 
			std::map<mEntity*,int*>& EntToMigr,
			map<mEntityGroup*,int*>& SetToMigr, 
			pmMigrationCallbacks& cb,
                        std::list<mEntity*>* rmE, 
			std::list<mEntity*>* newE,
 			std::vector<mEntityGroup*>& rmEG, 
			std::vector<mEntityGroup*>& newEG)
// *****************************************
{
#ifdef DEBUG
  int localMove=EntToMigr.size() + SetToMigr.size();
  int totalMove;
  MPI_Allreduce(&localMove,&totalMove,1,MPI_INT,MPI_SUM,ParUtil::Instance()->getComm());
  if (totalMove==0)
  {

    ParUtil::Instance()->Msg(ParUtil::INFO,"\n* NO MIGRATION NEEDED\n");
    return SCUtil_SUCCESS;
  }

//  printEntToMigr (EntToMigr);
#endif 

  std::vector<int> NborProc;
  FMDB_Mesh_GetNborProcID (mesh, NborProc);

  std::set<int> SetNborProc;
  std::copy(NborProc.begin(), NborProc.end(), inserter(SetNborProc, SetNborProc.end()));
  ParUtil::Instance()->ComMan()->set_Nbrs(SetNborProc);

  int meshDim;
  FMDB_Mesh_GetDim(mesh, &meshDim);
  std::map<mEntity*,int> entitiesToRemove[4];
  std::map<mEntity*,int> entitiesToHandle[4];
 
#ifdef DEBUG
  double prev_time=ParUtil::Instance()->getCurTime();
  double prev_mem=ParUtil::Instance()->getCurMem();
  double begin_time=prev_time, begin_mem=prev_mem;
#endif
// ********************************************
// STEP 1: reset BPs: unify entitiesToHandle and clear their BPs and PCs
// ********************************************
  std::vector<pPart> parts;
  FMDB_Mesh_GetAllPart(mesh, SCUTIL_CommRank(), parts);
  getEntitiesToHandle_and_setBPs(parts[0], EntToMigr, entitiesToHandle);

#ifdef DEBUG
  SCUTIL_DspRsrc("set bounding partition ids...", ParUtil::Instance()->getCurTime()-prev_time, 
                         ParUtil::Instance()->getCurMem()-prev_mem);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();
#endif

// ********************************************
// STEP 2: update PClassification and collect entities to remove
// ********************************************

  updatePC_and_collectEntitiesToRemove(entitiesToHandle, entitiesToRemove);

#ifdef DEBUG
  SCUTIL_DspRsrc("update PClassification...", ParUtil::Instance()->getCurTime()-prev_time, 
                         ParUtil::Instance()->getCurMem()-prev_mem);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();
#endif

// ********************************************
// STEP 3: update RC when maxNumParts is changed (costly) 
// ********************************************

  int curNumPart = ParUtil::Instance()->getCurNumParts();    // current_local_num_parts
  int tgtNumPart = ParUtil::Instance()->getTgtNumParts();    // target_local_num_parts 
  mEntity* ent; 
  mPart* part; 
  if(curNumPart!=tgtNumPart) {

    for(int ipart=0; ipart<(int)parts.size(); ipart++) {
      part = parts[ipart];
      for(int idim=0; idim<meshDim; idim++)
	for (mPart::iterall it=part->beginall(idim); it!=part->endall(idim);++it)
        { 
	  // update remote copy
     	  ent=*it; 
	  if(ent->getNumRemoteCopies()) {       
	    std::map<int, mEntity*> tmpRemoteCopies;
	    ent->getRemoteCopies(tmpRemoteCopies);
	    ent->clearRemoteCopies();
	    
	    for(mEntity::RCIter rcIter=tmpRemoteCopies.begin(); rcIter!=tmpRemoteCopies.end(); ++rcIter) {
	      int pid=(*rcIter).first; 	 
	      int newPid = (pid/curNumPart)*tgtNumPart + pid%curNumPart;          // target global pid 
	      ent->addRemoteCopy(newPid, (*rcIter).second);          
	    }

	    // update partition classification
	    if(entitiesToHandle[idim].find(ent)==entitiesToHandle[idim].end() ) { 
	      set<int> BPset; 
	      ent->getPidsExist(BPset);
	      BPset.insert(SCUTIL_CommRank()*tgtNumPart+ipart);
	      pmEntity* pe = pmModel::Instance()->getPartitionEntity(BPset);
	      ent->setPClassification(pe);
	    }
	  }
#ifdef MATCHING
	  if(ent->hasMatchEnt()) {

            const std::multimap<int, mEntity*>* tmpMatchEnts;
	    std::multimap<int, mEntity*> tmpMatchEnts2; 
            tmpMatchEnts = ent->getMatchEnts();

	    int pid, newPid; 
            for(std::multimap<int, mEntity*>::const_iterator meIter=tmpMatchEnts->begin(); meIter!=tmpMatchEnts->end(); ++meIter) 
	    {
              pid=(*meIter).first;
              newPid = (pid/curNumPart)*tgtNumPart + pid%curNumPart;          // target global pid
	      tmpMatchEnts2.insert(mEntity::matchEntMap::value_type(newPid, meIter->second));
            } 
	    
	    ent->clearMatchEnts(); 
	    for(std::multimap<int, mEntity*>::iterator meIter=tmpMatchEnts2.begin(); meIter!=tmpMatchEnts2.end(); ++meIter)
	       ent->addMatchEnt(meIter->first, meIter->second);
	       
	  }
#endif
	}
    }
  }

#ifdef DEBUG
  SCUTIL_DspRsrc("correct remote copy info due to #part change...", ParUtil::Instance()->getCurTime()-prev_time, 
                         ParUtil::Instance()->getCurMem()-prev_mem);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();
#endif

// ********************************************
// STEP 4: Migrate entities
// ********************************************

  for (int i=0; i<=meshDim;++i)
    exchangeEntities(parts, meshDim, cb, i, entitiesToHandle[i], newE, newEG);


#ifdef DEBUG
  SCUTIL_DspRsrc("entity exchange...", ParUtil::Instance()->getCurTime()-prev_time, 
                         ParUtil::Instance()->getCurMem()-prev_mem);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();
#endif

// ********************************************
// STEP 5: collect entities removed to return
// ********************************************

  std::map<mEntity*,int>::iterator mapit; 
  for (int i=0; i<=meshDim;++i)
    for(mapit=entitiesToRemove[i].begin(); mapit!=entitiesToRemove[i].end(); mapit++)
      rmE[i].push_back(mapit->first); 

#ifdef DEBUG
  SCUTIL_DspRsrc("collect removed entities...", ParUtil::Instance()->getCurTime()-prev_time, 
                         ParUtil::Instance()->getCurMem()-prev_mem);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();
#endif

// ********************************************
// STEP 6: delete unused entities and P-sets
// ********************************************

  ParUtil::Instance()->setCurNumParts(tgtNumPart);
  removeUnusedEntities(parts, cb, meshDim, entitiesToRemove); 

  std::map<mEntityGroup*, int*>::iterator egIter;
  mEntityGroup::entIter eiter; 
  int pid;
  mEntityGroup* eg; 
  for(egIter=SetToMigr.begin(); egIter!=SetToMigr.end(); ++egIter) 
  {
    eg=egIter->first;
    pid = (egIter->second)[0];        // local part id
    for(eiter=eg->begin(); eiter!=eg->end(); ++eiter) 
      EntToMigr.erase(*eiter);
    
    parts[pid]->deletePSet(eg, 0);
    rmEG.push_back(eg);  // collect removed entity group
    delete[] (egIter->second); 
  }

  std::map<mEntity*,int*>::iterator iter;
  for(iter = EntToMigr.begin(); iter!=EntToMigr.end(); ++iter) 
  {
    int* pid = (*iter).second;
    delete[] pid;
  }

#ifdef DEBUG
  SCUTIL_DspRsrc("remove entities and sets not in use...", ParUtil::Instance()->getCurTime()-prev_time, 
                         ParUtil::Instance()->getCurMem()-prev_mem);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();
#endif

// ********************************************
// STEP 7: delete empty parts and update localPids 
// ********************************************

  int usrNumPart = ParUtil::Instance()->getUsrNumParts(); 
  int size = (int)parts.size(); 
  if(usrNumPart< size)                   // local_num_parts is reduced     
  {
    if((ParUtil::Instance()->localPids).size()==0 )
    {
      for(int i=usrNumPart; i< size; ++i) 
      {
	mPart* tmpPart = parts[i];
	M_delete(tmpPart);
	parts.pop_back(); 
      }
    }
    else
    {
      vector<int>::iterator vecit; 
      int ipart=0; 
      set<mPart*> partSet; 
      set<int> pidSet; 
      for (vecit=(ParUtil::Instance()->localPids).begin(); vecit!=(ParUtil::Instance()->localPids).end(); ++vecit, ++ipart) 
      {
	if (*vecit<0) 
        {
	  partSet.insert(parts[ipart]); 
	  pidSet.insert(*vecit); 
	}
      } // for vecit
      
      vector<mPart*>::iterator partIt; 
      for (set<mPart*>::iterator setit=partSet.begin(); setit!=partSet.end(); ++setit) 
      {
	partIt = std::find(parts.begin(), parts.end(), *setit);
	if(partIt!=parts.end())
	  parts.erase(partIt);
      } // for setit
      
      for(set<int>::iterator setit=pidSet.begin(); setit!=pidSet.end(); ++setit)
      {
	vecit = std::find((ParUtil::Instance()->localPids).begin(), (ParUtil::Instance()->localPids).end(), *setit); 
	if(vecit!=(ParUtil::Instance()->localPids).end())
	  (ParUtil::Instance()->localPids).erase(vecit); 
      } // for setit
    } // else((ParUtil::Instance()->localPids).size()==0 )
  } // if(usrNumPart< size)

#ifdef DEBUG
  SCUTIL_DspRsrc("remove empty parts...", ParUtil::Instance()->getCurTime()-prev_time, 
                         ParUtil::Instance()->getCurMem()-prev_mem);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();
#endif

// ********************************************
// STEP 8: update partition model
// ********************************************
  /// Set neighboring parts based on the new partition model information
  NborProc.clear();
  FMDB_Mesh_GetNborProcID (mesh, NborProc);

  SetNborProc.clear();
  std::copy(NborProc.begin(), NborProc.end(), inserter(SetNborProc, SetNborProc.end()));
  ParUtil::Instance()->ComMan()->set_Nbrs(SetNborProc);
  pmModel::Instance()->updateOwnership(mesh);

#ifdef DEBUG
  SCUTIL_DspRsrc("update partition model...", ParUtil::Instance()->getCurTime()-prev_time, 
                         ParUtil::Instance()->getCurMem()-prev_mem);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();

  if (!SCUTIL_CommRank())
    cout<<"MIGRATION: #moved=" <<totalMove<<", Time="<<prev_time-begin_time<<" (sec), Mem="<<prev_mem - begin_mem<<" (MB)\n";
#endif
  return SCUtil_SUCCESS;
}

#endif
