/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/ 
#ifdef FMDB_PARALLEL
#include <map>
#include "FMDB.h"
void  printPtnObjToMigr (std::map<mEntity*, std::pair<int, int> >& EntToMigr, 
                          std::map<mEntityGroup*, std::pair<int, int> >& SetToMigr)       
{
  std::map<mEntity*, std::pair<int, int> >::const_iterator entit;  
  for (entit=EntToMigr.begin(); entit!=EntToMigr.end();++entit)
    cout<<"("<<SCUTIL_CommRank()<<") migrate "<<entit->first->getUid()<<" on part "<<entit->second.first<<" to Part "
        <<entit->second.second<<")\n";
/*
  std::map<mEntityGroup*, std::pair<int, int> >::const_iterator setit;  
  for (setit=SetToMigr.begin(); setit!=SetToMigr.end();++setit)
    cout<<"("<<SCUTIL_CommRank()<<") migrate "<<setit->first->getUid()<<" on part "<<setit->second.first<<" to Part "
        <<setit->second.second<<")\n";
*/
}

void collectEntitiesToExchg_and_reset(mMesh*, std::map<mEntity*, std::pair<int, int> >& EntToMigr,
             std::map<mEntity*,int>* entitiesToExchg);

void updatePClas_and_collectEntitiesToRemove(std::map<mEntity*,int>* entitiesToExchg, 
			std::map<mEntity*,int>* entitiesToRemove);

void exchangeEntities(mMesh* mesh, int mesh_dim, pmMigrationCallbacks &cb, int ent_dim,
                     std::map<mEntity*,int>&, int debug_level,
                     std::vector<mEntity*>* newEnt, std::vector<mEntityGroup*>* newPSet); 

void bounceEntities(std::vector<rc_struct_3>& entitiesToBounce);
void broadcastEntities(std::vector<rc_struct_2>& entitiesToBroadcast);

void removeUnusedEntities(mMesh* mesh, pmMigrationCallbacks &cb,int dim, 
			 std::map<mEntity*,int>* entitiesToRemove);

// *****************************************
int migratePtnObject (mMesh* mesh, pmMigrationCallbacks& cb,
		      std::map<mEntity*, std::pair<int, int> >& EntToMigr,
                      std::map<mEntityGroup*, std::pair<int, int> >& SetToMigr,
                      int debug_level, std::vector<mEntity*>* newEnt, std::vector<mEntity*>* rmvEnt, 
 		      std::vector<mEntityGroup*>* newPSet,  std::vector<mEntityGroup*>* rmvPSet)
// *****************************************
{
#ifdef _DEBUG_
  printPtnObjToMigr(EntToMigr, SetToMigr);
#endif

  std::vector<int> NborProc;
  FMDB_Mesh_GetNborProcID (mesh, NborProc);

  std::set<int> SetNborProc;
  std::copy(NborProc.begin(), NborProc.end(), inserter(SetNborProc, SetNborProc.end()));
  ParUtil::Instance()->ComMan()->set_Nbrs(SetNborProc);

  int meshDim;
  FMDB_Mesh_GetDim(mesh, &meshDim);
  std::map<mEntity*,int> entitiesToRemove[4];
  std::map<mEntity*,int> entitiesToExchg[4];
 
#ifdef DEBUG
  double cur_time, prev_time=ParUtil::Instance()->getCurTime();
  double cur_mem, prev_mem=ParUtil::Instance()->getCurMem();
  double begin_time=prev_time, begin_mem=prev_mem;
#endif

// ********************************************
// STEP 1: collect entitiesToExchg and clear their residence parts and pclas
// ********************************************
  for (map<mEntityGroup*,std::pair<int, int> >::iterator pset_it = SetToMigr.begin(); 
       pset_it!=SetToMigr.end(); ++pset_it)
    for (mEntityGroup::entIter ent_it=pset_it->first->begin(); 
          ent_it!=pset_it->first->end(); ++ent_it) 
      EntToMigr.insert(std::map<mEntity*, std::pair<int, int> >::value_type(*ent_it, std::make_pair(pset_it->second.first, pset_it->second.second)));

  collectEntitiesToExchg_and_reset(mesh, EntToMigr, entitiesToExchg); 

#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("computing entities' dest part ids...", cur_time-prev_time, cur_mem-prev_mem);
  prev_time=cur_time;
  prev_mem=cur_mem;
#endif

// ********************************************
// STEP 2: update PClassification and collect entities to remove
// ********************************************

  updatePClas_and_collectEntitiesToRemove(entitiesToExchg, entitiesToRemove);

#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("updating PClassification and collect entities to remove...", cur_time-prev_time, cur_mem-prev_mem);
  prev_time=cur_time;
  prev_mem=cur_mem;
#endif


// ********************************************
// STEP 4: Migrate entities
// ********************************************

  for (int i=0; i<=meshDim;++i) 
    exchangeEntities(mesh, meshDim, cb, i, entitiesToExchg[i], debug_level, newEnt, newPSet);

#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("exchanging entities...", cur_time-prev_time, cur_mem-prev_mem);
  prev_time=cur_time;
  prev_mem=cur_mem;
#endif

// ********************************************
// STEP 5: remove unnecessary entities and P-sets
// ********************************************

  if (debug_level)
  {
    std::map<mEntity*,int>::iterator mapit; 
    for (int i=0; i<=meshDim;++i)
      for(mapit=entitiesToRemove[i].begin(); mapit!=entitiesToRemove[i].end(); mapit++)
        rmvEnt[i].push_back(mapit->first); 
  }

  removeUnusedEntities(mesh, cb, meshDim, entitiesToRemove); 

  std::map<mEntityGroup*, std::pair<int, int> >::iterator egIter;
  mEntityGroup::entIter eiter; 
  int src_partid;
  mEntityGroup* eg; 
  for(egIter=SetToMigr.begin(); egIter!=SetToMigr.end(); ++egIter) 
  {
    eg=egIter->first;
    src_partid = egIter->second.first;        // source part id
    
    mesh->getPart(src_partid%mesh->getNumPart())->deletePSet(eg, 0);
    if (debug_level) rmvPSet[src_partid%mesh->getNumPart()].push_back(eg);  // collect removed entity group
  }

#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("removing unnecesarry entities and p-sets...", cur_time-prev_time, cur_mem-prev_mem);
  prev_time=cur_time;
  prev_mem=cur_mem;
#endif

// ********************************************
// STEP 6: update partition model
// ********************************************
  /// Set neighboring parts based on the new partition model information
  NborProc.clear();  SetNborProc.clear();
  FMDB_Mesh_GetNborProcID (mesh, NborProc);
  std::copy(NborProc.begin(), NborProc.end(), inserter(SetNborProc, SetNborProc.end()));
  ParUtil::Instance()->ComMan()->set_Nbrs(SetNborProc);  
  pmModel::Instance()->updateOwnership(mesh);

#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("updating partition model...", cur_time-prev_time, cur_mem-prev_mem);
  prev_time=cur_time;
  prev_mem=cur_mem;

  int isValid;
  FMDB_Mesh_Verify(mesh, &isValid);
  assert(isValid);

  if (!SCUTIL_CommRank())
    cout<<"\nMIGRATION: Time="<<prev_time-begin_time<<" (sec), Mem="<<prev_mem - begin_mem<<" (MB)\n";
#endif
  return SCUtil_SUCCESS;
}

// *****************************************
void collectEntitiesToExchg_and_reset (mMesh* mesh, std::map<mEntity*, std::pair<int, int> >& EntToMigr,
             std::map<mEntity*,int>* entitiesToExchg)
// *****************************************
{
  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::All_Reduce);

  map<mEntity*, std::pair<int, int> >::iterator mapit;
  mEntity* down_ent, *remote_ent;

  for (mapit= EntToMigr.begin();mapit!= EntToMigr.end();++mapit)
  {
    mapit->first->clearBPs();
    mapit->first->setPClassification((pmEntity*)0);
    mapit->first->addBPs(mapit->second.second);  // dest partid
    (entitiesToExchg[mapit->first->getLevel()])[mapit->first]=mapit->second.first;    // source partid
  }

  int ent_count;
  std::vector<mEntity*> DownEnts;
  DownEnts.resize(27);
  for (mapit=EntToMigr.begin();mapit!= EntToMigr.end();++mapit)
  {
    DownEnts.clear();
    getDownEntities(mapit->first, DownEnts);

    for (std::vector<mEntity*>::iterator lit=DownEnts.begin();lit!=DownEnts.end();++lit)
    {
      down_ent = *lit;

      ent_count = entitiesToExchg[down_ent->getLevel()].count(down_ent);
      if (!ent_count)
      {
        down_ent->clearBPs();
        down_ent->setPClassification((pmEntity*)0);
      }
      entitiesToExchg[down_ent->getLevel()][down_ent]=mapit->second.first;        // src local pid

      mAdjacencyContainer upward;
      down_ent->getHigherOrderUpward(ParUtil::Instance()->get_maxMeshDim(), upward);
      for (int i=0; i<upward.size();++i)
      {
        int FMDB_Ent_GetPSet (pMeshEnt meshEnt, pEntSet& entSet); 

        if (upward[i]->getNumBPs()==0)
        {
          down_ent->addBPs(mapit->second.first); 
          break;  // exit for loop
        }
      }
      down_ent->addBPs(mapit->second.second);
    } // for DownEnts
  } // for EntToMigr

  CM->set_tag(100);
  CM->set_fixed_msg_size(0);

  void *msg_send;
  mEntity** s_ent;
  int* s_id; 

  map<mEntity*, int>::iterator eit;
  int msg_size, numBP;
  mEntity* ent;
  for (int dim = 0; dim <= ParUtil::Instance()->get_maxMeshDim(); ++dim)
  {
    for (eit=entitiesToExchg[dim].begin();eit!=entitiesToExchg[dim].end();++eit)
    {
      ent = eit->first;
      if (ent->getNumRemoteCopies()==0)
        continue;

      numBP = ent->getNumBPs();
      for (mEntity::RCIter rciter=ent->rcBegin();rciter!=ent->rcEnd();++rciter)
      {
        msg_size=sizeof(mEntity*) +sizeof(int) + numBP*sizeof(int);
        msg_send = CM->alloc_msg(msg_size);
        
        s_ent = (mEntity**)msg_send; 
        s_id = (int*)((char*)msg_send+sizeof(mEntity*));
	
        *s_ent = rciter->second; 
        *s_id = rciter->first; 

        int *BPs = (int*)((char*)msg_send + sizeof(mEntity*) + sizeof(int));
        int i = 0;
        for (mEntity::BPIter bpiter=ent->bpBegin();bpiter!=ent->bpEnd();++bpiter)
        {
          BPs[i]=*bpiter;
          i++;
        }
        CM->send((rciter->first)/mesh->getNumPart(), (void*)msg_send, msg_size);
        CM->free_msg(msg_send);
      }
    }
  }
  CM->finalize_send();

  map<mEntity*,int> entitiesToBounce;

  // receive phase
  void *msg_recv;
  int pid_from;
  int* BPs;
  while(int rc = CM->receive(msg_recv, &pid_from))
  {
    remote_ent = *((mEntity**)msg_recv); 
    int* remote_partid = (int*)((char*)msg_recv+sizeof(mEntity*)); 
    BPs = (int*)((char*)msg_recv+sizeof(mEntity*)+sizeof(int)); 
    numBP = (rc-sizeof(mEntity*)-sizeof(int)) / INT_SIZE;

    if (!entitiesToExchg[remote_ent->getLevel()].count(remote_ent))
    {
      remote_ent->clearBPs();
      remote_ent->setPClassification((pmEntity*)0);
    }
    mAdjacencyContainer upward;
    remote_ent->getHigherOrderUpward(ParUtil::Instance()->get_maxMeshDim(),upward);
    for (int i=0; i<upward.size();++i)
    {
      if (upward[i]->getNumBPs()==0)
      {
        remote_ent->addBPs(*remote_partid); ///send to rem copies info about mypid
        entitiesToBounce[remote_ent] = *remote_partid;
          break;  // exit for loop
      }
    } // upward
    entitiesToExchg[remote_ent->getLevel()][remote_ent] = *remote_partid;

    for (int i = 0; i < numBP; i++)
      remote_ent->addBPs(BPs[i]);
    CM->free_msg(msg_recv);
  }

  msg_size = sizeof(mEntity*) + sizeof(int);
  CM->set_tag(0);
  CM->set_fixed_msg_size(msg_size);
  msg_send = CM->alloc_msg(msg_size);

  for (map<mEntity*,int>::iterator bounce_iter = entitiesToBounce.begin(); bounce_iter != entitiesToBounce.end(); ++bounce_iter)
  {
    remote_ent = bounce_iter->first;
    for (mEntity::RCIter rciter=remote_ent->rcBegin();rciter!=remote_ent->rcEnd();++rciter)
    {
      s_ent = (mEntity**)msg_send;
      *s_ent = rciter->second;   
      s_id = (int*)((char*)msg_send + sizeof(mEntity*));
      *s_id = bounce_iter->second; // local part id
      CM->send((rciter->first)/mesh->getNumPart(), (void*)msg_send);
    }
  }
  CM->finalize_send();
  CM->free_msg(msg_send);

  // receive phase
  while(CM->receive(msg_recv, &pid_from))
  {
    ent = *(mEntity**)msg_recv;
    s_id = (int*)((char*)msg_recv + sizeof(mEntity*));
    ent->addBPs(*s_id);
    CM->free_msg(msg_recv);
  }  

#ifdef _DEBUG_
  for (int dim = 0; dim <= ParUtil::Instance()->get_maxMeshDim(); ++dim)
    for (eit=entitiesToExchg[dim].begin();eit!=entitiesToExchg[dim].end();++eit)
      eit->first->printBPs(eit->second);
#endif
}

// *****************************************
void getDownEntities(mEntity* ent, std::vector<mEntity*>& family)
// *****************************************
{
  switch (ent->getLevel())
  {
    case FMDB_REGION:/* METHOD 1 */
           for (int i=0; i<ent->size(FMDB_FACE);++i)
           {
             family.push_back(ent->get(2,i));
             for (int j=0; j<ent->get(2,i)->size(1);++j)
               if (std::find(family.begin(), family.end(), ent->get(2,i)->get(1,j))==family.end())
                 family.push_back(ent->get(2,i)->get(1,j));
           }
           for (int j=0; j<ent->size(FMDB_VERTEX);++j)
             family.push_back(ent->get(FMDB_VERTEX,j));
#ifdef DEBUG
          if (ent->size(0)==4 && family.size()!=14)    //only works for tet.
          {
           cout<<"\n("<<SCUTIL_CommRank()<<") "<<__func__<<" FMDB ERROR: #down ents("<<ent->getUid()<<") is "<<family.size()<<" -- should be 14\n";
            for (vector<mEntity*>::iterator it=family.begin();it!=family.end();++it)
              cout<<FMDB_Ent_StrID(*it)<<"  ";
           cout<<"\n";
          }
#endif
           break;
    case FMDB_FACE:// face
          for (int j=0; j<ent->size(FMDB_EDGE);++j)
            family.push_back(ent->get(FMDB_EDGE,j));
          for (int j=0; j<ent->size(FMDB_VERTEX);++j)
            family.push_back(ent->get(FMDB_VERTEX,j));
          break;
    case FMDB_EDGE: // edge
          family.push_back(ent->get(0,0));
          family.push_back(ent->get(0,1));
           break;
    default: break;
  }
}

// ***********************************************************
void updatePClas_and_collectEntitiesToRemove(
			map<mEntity*,int>* entitiesToExchg, 
			map<mEntity*,int>* entitiesToRemove)
// ***********************************************************
{
  for (int dim=0; dim<=3; ++dim)
  {
    for (std::map<mEntity*, int>::const_iterator eit=entitiesToExchg[dim].begin();
         eit!=entitiesToExchg[dim].end();++eit)
    {
      if (!((eit->first)->findPidInBPs(eit->second)))  // not stay in the current part
	(entitiesToRemove[dim])[eit->first]=eit->second;
      else //if (!ent->getPClassification()) 
        eit->first->setPClassificationWithBPs();
    }    
  }
}

void getRemotePartID(mEntity* ent, set<int>&res_parts)
{
  if (ent->getNumRemoteCopies())
    for (mEntity::RCIter rcIter=ent->rcBegin(); rcIter!=ent->rcEnd();++rcIter)
      res_parts.insert(rcIter->first);
}


// **********************************************
void exchangeEntities(mMesh* mesh, int meshDim, pmMigrationCallbacks &cb, int entDim,
                     map<mEntity*,int>& entitiesToExchg, int debug_level,
		     std::vector<mEntity*>* newEnt, std::vector<mEntityGroup*>* newPSet)	  
// **********************************************
{
  mEntity* ent;
  set<int>::iterator piter;
  int src_partid, dest_partid;
  int mypid=ParUtil::Instance()->rank();
  
  map<mEntity*, int>::const_iterator eit; 
  map<int, mEntity*>::iterator rcIter; 
  
  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::All_Reduce);
  CM->set_tag(0);
  CM->set_fixed_msg_size(0);

  set<int> RCs, res_parts, temp; 
  set<int>::iterator bpIter; 
  for(eit=entitiesToExchg.begin(); eit!=entitiesToExchg.end();++eit)
  {
    ent = eit->first; 
    src_partid = eit->second;             // current global pid

    res_parts.clear(); 
    temp.clear(); 

    getRemotePartID(ent, res_parts);
    res_parts.insert(src_partid);
    set<int>::iterator minit=std::min_element(res_parts.begin(), res_parts.end());
    if(src_partid!=(*minit))
        continue;

    set_subtract(*(ent->tempBPs),res_parts, temp);

    for (piter=temp.begin(); piter!=temp.end();++piter)
    {	      
      dest_partid = *piter; 
      void *msg_send;
      int msg_size; 
      packEntity(ent, dest_partid, src_partid, cb, msg_send, msg_size, meshDim);
      CM->send(dest_partid/mesh->getNumPart(), msg_send, msg_size);
      CM->free_msg(msg_send); 
    }
  }
  CM->finalize_send();

  std::map<pair<mEntityGroup*, int>, mEntityGroup*> entGrpMap; 

   // receive phase begins
  void *msg_recv;
  int pid_from;
  mPart* dest_part; 
  std::vector<rc_struct_3> entitiesToBounce;

  while(CM->receive(msg_recv, &pid_from))
  {
    // get the dest part id
    int* tmp_ptr = (int*)msg_recv;  
    dest_partid = *tmp_ptr; 
    
    dest_part = mesh->getPart(dest_partid%mesh->getNumPart()); 	
    unpackEntity(dest_part, meshDim, msg_recv, cb, debug_level, newEnt, newPSet, entitiesToBounce, entGrpMap);
    CM->free_msg(msg_recv);
  }

  if (meshDim > entDim)
    bounceEntities(entitiesToBounce);
}

// **********************************************
void packEntity(mEntity *ent, int dest_partid, int src_partid, pmMigrationCallbacks &cb, void *&msg_send, int &msg_size, int meshDim)
// **********************************************
{
  char *buf_user;
  int user_size;
  int entity_size;

  char* eg_buf_user; 
  int eg_user_size; 

  int tgtNumPart = ParUtil::Instance()->getTgtNumParts(); 
  IPComMan *CM = ParUtil::Instance()->ComMan();

  // User may have some data to send to the other process. These data are of size user_size_part2.
  buf_user = (char*)cb.getUserData(ent, dest_partid/tgtNumPart /*dest_proc*/, user_size);
  if(!buf_user)
    user_size = 0; 

  short int numBP = ent->getNumBPs();
  if (numBP<2) numBP=0;
  short int numRC = ent->getNumRemoteCopies();
  short int entDim = ent->getLevel();
  short int numDE = entDim+1;

  entity_size = 2*sizeof(int) + sizeof(packedEnt) + numRC*(sizeof(int)+sizeof(mEntity*)) + numBP*sizeof(int) ;

  int n_bytes = 0;

  pGEntity gent = ent->getClassification();
  short int numParam=0, gType, gTag;
  if(gent!=0) {
    gType = GEN_type(gent);
    gTag = GEN_tag(gent);
  }
  else{
    gTag = 1;
    gType = 3;
  }

  if (entDim == 0)
  {
    if (ent->getData(FMDB_Util::Instance()->getParametric()))
    {
      if (gType==1) // vertex classified on model edge
        numParam=1;
      else if (gType==2) // vertex classified on model face
        numParam=2;
    }

    entity_size += sizeof(packedVertex) + numParam*sizeof(double);

    msg_size = entity_size + user_size;
    msg_send = CM->alloc_msg(msg_size);

    // dest global pid and src local pid 
    int* tmp_pid; 
    tmp_pid = (int*)msg_send; 
    tmp_pid[0]=dest_partid; 
    tmp_pid[1]=src_partid; 
    n_bytes += 2*sizeof(int); 

    // fixed length data
    packedEnt *entbuf = (packedEnt*)((char*)msg_send + n_bytes);
    entbuf->numDE = numDE;
    entbuf->sender = ent;
    entbuf->gTag = gTag;
    entbuf->gType = gType;
    entbuf->numRC = numRC;
    entbuf->numBP = numBP;
    n_bytes += sizeof(packedEnt);

    // vertex specific parameters
    packedVertex *castbuf = (packedVertex*)((char*)msg_send + n_bytes);
    castbuf->iD = ent->getId();

    double xyz[3];
    ((mVertex*)ent)->getCoord(xyz[0], xyz[1], xyz[2]);
    for(int i=0;i<3;i++) 
       castbuf->coord[i] = xyz[i]; 
    castbuf->numParam = numParam;
    n_bytes += sizeof(packedVertex);

    if (numParam > 0)
    {
      double *Params = (double*)((char*)msg_send + n_bytes);
      SCOREC::Util::mVector vec = ent->getAttachedVector(FMDB_Util::Instance()->getParametric());
      for (int k=0; k<numParam;++k) 
      {
        Params[k] = vec(k);
      }
      n_bytes += numParam*sizeof(double);
    }
  }
  else
  {
    numDE = ent->size(0); 
    entity_size += sizeof(short int) + numDE*sizeof(mEntity*);   

  if(meshDim==ent->getLevel()) 
  { 
    entity_size = entity_size+sizeof(double)+sizeof(mEntityGroup*); 
    void* tmp_ptr; 
    if(EN_getDataPtr(ent, FMDB_Util::Instance()->getPSet(), &tmp_ptr)) 
    {                   // in an entity-group 
      int pos = ((mEntityGroup*)tmp_ptr)->getEntityOrder(ent);

      if(pos==0) {	                                        // start from 0 
	 entity_size += 2*sizeof(int); 
	 eg_buf_user = (char*)cb.getEntGrpUserData((mEntityGroup*)tmp_ptr, dest_partid/tgtNumPart /*dest_proc*/, eg_user_size);
         if(!eg_buf_user)
	 {
	    eg_user_size = 0;
	 }
	 entity_size += eg_user_size;  
      }
      else
         entity_size += sizeof(int);
     }
   }

    msg_size = entity_size + user_size;
    msg_send = CM->alloc_msg(msg_size);

    // dest global pid and src local pid
    int* tmp_pid;
    tmp_pid = (int*)msg_send;
    tmp_pid[0]=dest_partid;
    tmp_pid[1]=src_partid;
    n_bytes += 2*sizeof(int); 

    // fixed length data
    packedEnt *entbuf = (packedEnt*)((char*)msg_send + n_bytes);
    entbuf->numDE = numDE;
    entbuf->sender = ent;
    entbuf->gTag = gTag;
    entbuf->gType = gType;
    entbuf->numRC = numRC;
    entbuf->numBP = numBP;
    n_bytes += sizeof(packedEnt);

    short int *type = (short int*)((char*)msg_send + n_bytes);
    *type = ent->getType();
    n_bytes += sizeof(short int);
 
    mEntity** DEs = (mEntity**)((char*)msg_send + n_bytes);
    for (int i=0; i<numDE;++i)
    {
      DEs[i] = ent->get(0,i)->getRemoteCopy(dest_partid);
      if (DEs[i]==NULL) 
        cout<<"["<<SCUTIL_CommRank()<<"] part "<<src_partid<<" - FMDB FATAL ERROR in creating "<<ent->getUid()<<" on part "<<dest_partid
            <<" due to "<<ent->get(0,i)->getUid()<<" NOT EXISTING on dest part "<<dest_partid<<endl;
    }
    n_bytes += numDE*sizeof(mEntity*);
  }

  // variable length data - RC and BP
  if (numRC>0)
  {
    int *remote_partid;
    mEntity **RCs_ent;
    for (mEntity::RCIter rciter=ent->rcBegin();rciter!=ent->rcEnd();++rciter)
    {
      remote_partid = (int*)((char*)msg_send + n_bytes);
      *remote_partid = rciter->first;
      n_bytes += sizeof(int);
      RCs_ent = (mEntity**)((char*)msg_send + n_bytes);
      *RCs_ent = rciter->second;
      n_bytes += sizeof(mEntity*);
    }
  }

  if (numBP>1)
  {
    int i=0; 
    int* BPs = (int*)((char*)msg_send + n_bytes);
    for (mEntity::BPIter bpiter=ent->bpBegin();bpiter!=ent->bpEnd();++bpiter)
    {
      BPs[i] = *bpiter;
      i++; 
    }
    n_bytes += numBP*sizeof(int);
  }

  if(ent->getLevel()==meshDim) 
  {
     // copy weight
     double* weight=(double*)((char*)msg_send+n_bytes);
     FMDB_Ent_GetWeight(ent, weight); 
     n_bytes+=sizeof(double); 

     // copy p-set
     mEntityGroup** entGrp = (mEntityGroup**)((char*)msg_send + n_bytes); 
     void* tmp_ptr; 
     if(EN_getDataPtr(ent, FMDB_Util::Instance()->getPSet(), &tmp_ptr))                     // entity is in entity-group 
       *entGrp = (mEntityGroup*)tmp_ptr; 
     else 
       *entGrp = 0; 
       
     n_bytes += sizeof(mEntityGroup*);

     if( *entGrp!=0 ) {
       int* pos = (int*)((char*)msg_send+n_bytes); 
       *pos = (*entGrp)->getEntityOrder(ent);
       n_bytes += sizeof(int); 
     
       if(*pos==0) {                                          // the 1st entity in entity-group
	 int* size = (int*)((char*)msg_send + n_bytes);  
	 *size = eg_user_size; 
	 n_bytes += sizeof(int);

	 if(eg_user_size)  {
	    memcpy((void*)((char*)msg_send + n_bytes), eg_buf_user, eg_user_size);
	    cb.deleteEntGrpUserData(eg_buf_user);	    
	    n_bytes += eg_user_size; 
	 }
       }
     }
  }

  if(user_size)
  {
    memcpy((void*)((char*)msg_send + n_bytes),buf_user,user_size);
    cb.deleteUserData(buf_user);
  }
}

// **********************************************
void unpackEntity(mPart* part, int meshDim, void *msg_recv, 
                  pmMigrationCallbacks &cb, int debug_level, vector<mEntity*>* newEnt, 
                  vector<mEntityGroup*>* newPSet, std::vector<rc_struct_3>& entitiesToBounce,
                  std::map<pair<mEntityGroup*, int>, mEntityGroup*>& entGrpMap)
// **********************************************
{
  int n_bytes = 0;
  short int numRC, numBP;

  mEntity *new_ent, *sender;
  int numPart = ParUtil::Instance()->getTgtNumParts(); 

// PRE-STEP: add sender to remote copy and store entity to bounce
  // dest global pid and src local pid 
  int* tmp_pid = (int*)msg_recv; 
  int dest_partid = tmp_pid[0]; 
  int src_partid = tmp_pid[1]; 
  n_bytes += 2*sizeof(int); 

  packedEnt *entbuf = (packedEnt*)((char*)msg_recv + n_bytes);
  n_bytes += sizeof(packedEnt);

  sender = entbuf->sender;
  numRC = entbuf->numRC;
  numBP = entbuf->numBP;

  if (entbuf->numDE == 1)
  {
    // STEP 1: create entity
    packedVertex *castbuf = (packedVertex*)((char*)msg_recv + n_bytes);
    n_bytes += sizeof(packedVertex);

    new_ent = part->createVertex(castbuf->iD, castbuf->coord[0], castbuf->coord[1], castbuf->coord[2],
                                  part->getGEntity(entbuf->gTag, entbuf->gType));
    new_ent->attachInt (FMDB_Util::Instance()->getSourcePart(), src_partid);
    if (debug_level) newEnt[0].push_back(new_ent);    

    if (castbuf->numParam > 0)
    {
      double* Params = (double*)((char*)msg_recv + n_bytes);
      if (castbuf->numParam == 1)
        new_ent->attachVector (FMDB_Util::Instance()->getParametric(),SCOREC::Util::mVector(Params[0],0,0));
      else // if (castbuf.numParam==2)
        new_ent->attachVector (FMDB_Util::Instance()->getParametric(),SCOREC::Util::mVector(Params[0],Params[1],0));
      n_bytes += castbuf->numParam*sizeof(double);
    }
  }
  else
  {
    short int *type = (short int*)((char*)msg_recv + n_bytes);
    n_bytes += sizeof(short int);

    pGEntity g = part->getGEntity(entbuf->gTag, entbuf->gType);

// STEP 1: create an entity
    short int numDE;
    mEntity** DEs = (mEntity**)((char*)msg_recv + n_bytes);
    switch (*type)
    {
      case mEntity::EDGE:
      {
        numDE=2;
#ifdef _DEBUG_
        if (!(mVertex*)DEs[0]) 
          cout<<"["<<SCUTIL_CommRank()<<"] src part "<<src_partid<<", dest partid "<<dest_partid<<" failed getting vertex v1"<<v1<<endl;
        if (!(mVertex*)DEs[1]) 
          cout<<"["<<SCUTIL_CommRank()<<"] src part "<<src_partid<<", dest partid "<<dest_partid<<" failed getting vertex v2"<<v2<<endl;
#endif
        new_ent = M_createE(part, (mVertex*)DEs[0], (mVertex*)DEs[1], g);
        if (debug_level) newEnt[1].push_back(new_ent);
      } break;

      case mEntity::TRI:
      {
        numDE=3;
        mEntity* edges[3];
        edges[0]=(mEntity*)E_exist((mVertex*)DEs[0],(mVertex*)DEs[1]);
        edges[1]=(mEntity*)E_exist((mVertex*)DEs[1],(mVertex*)DEs[2]);
        edges[2]=(mEntity*)E_exist((mVertex*)DEs[0],(mVertex*)DEs[2]);
#ifdef _DEBUG_
        if (!edges[0]) 
          cout<<"["<<SCUTIL_CommRank()<<"] part "<<dest_partid<<" failed getting edge 0 "<<edges[0]<<endl;
        if (!edges[1]) 
          cout<<"["<<SCUTIL_CommRank()<<"] part "<<dest_partid<<" failed getting edge 1 "<<edges[1]<<endl;
        if (!edges[2]) 
          cout<<"["<<SCUTIL_CommRank()<<"] part "<<dest_partid<<" failed getting edge 2 "<<edges[2]<<endl;
#endif
        new_ent = M_createF(part, 3, edges, 0, g);
        if (debug_level) newEnt[2].push_back(new_ent);
      } break;

      case mEntity::QUAD:
      {
	numDE=4; 
	mEntity* edges[4];
	edges[0]=(mEntity*)E_exist((mVertex*)DEs[0],(mVertex*)DEs[1]);
	edges[1]=(mEntity*)E_exist((mVertex*)DEs[1],(mVertex*)DEs[2]);
	edges[2]=(mEntity*)E_exist((mVertex*)DEs[2],(mVertex*)DEs[3]);
	edges[3]=(mEntity*)E_exist((mVertex*)DEs[3],(mVertex*)DEs[0]);               //does the order of vertex matter?
	new_ent = M_createF(part, 4, edges, 0, g);	
	if (debug_level) newEnt[2].push_back(new_ent);
	
      } break;

      case mEntity::TET:
      {
        numDE=4;
        mEntity* faces[4];    
        faces[0]=(mEntity*)F_exist(0,DEs[0],DEs[1],DEs[2],(mEntity*)0);
        faces[1]=(mEntity*)F_exist(0,DEs[0],DEs[1],DEs[3],(mEntity*)0);
        faces[2]=(mEntity*)F_exist(0,DEs[1],DEs[2],DEs[3],(mEntity*)0);
        faces[3]=(mEntity*)F_exist(0,DEs[0],DEs[2],DEs[3],(mEntity*)0);
        new_ent=(mEntity*)M_createR(part, 4, faces, 0, g);
        if (debug_level) newEnt[3].push_back(new_ent);
      } break;

      case mEntity::HEX: 
      {
	numDE=8; 
		
	mEntity* faces[6];
	faces[0]=(mEntity*)F_exist(0,DEs[0],DEs[3],DEs[2],DEs[1]);
	faces[1]=(mEntity*)F_exist(0,DEs[0],DEs[1],DEs[5],DEs[4]);
	faces[2]=(mEntity*)F_exist(0,DEs[1],DEs[2],DEs[6],DEs[5]);
	faces[3]=(mEntity*)F_exist(0,DEs[2],DEs[3],DEs[7],DEs[6]);
	faces[4]=(mEntity*)F_exist(0,DEs[3],DEs[0],DEs[4],DEs[7]);
	faces[5]=(mEntity*)F_exist(0,DEs[4],DEs[5],DEs[6],DEs[7]);
	new_ent=(mEntity*)M_createR(part, 6, faces, 0, g);
	
	if (debug_level) newEnt[3].push_back(new_ent);
      } break;
   
      case mEntity::PRISM: 
      {
	numDE=6; 
	mEntity* faces[5];
	faces[0]=(mEntity*)F_exist(0,DEs[0],DEs[1],DEs[2],(mEntity*)NULL);
	faces[1]=(mEntity*)F_exist(0,DEs[0],DEs[1],DEs[4],DEs[3]);
	faces[2]=(mEntity*)F_exist(0,DEs[1],DEs[2],DEs[5],DEs[4]);
	faces[3]=(mEntity*)F_exist(0,DEs[2],DEs[0],DEs[3],DEs[5]);
	faces[4]=(mEntity*)F_exist(0,DEs[3],DEs[4],DEs[5],(mEntity*)NULL);
	new_ent=(mEntity*)M_createR(part, 5, faces, 0, g);
	
	if (debug_level) newEnt[3].push_back(new_ent);
      } break; 

      case mEntity::PYRAMID: 
      {
	numDE=5; 

	mEntity* faces[5];
	faces[0]=(mEntity*)F_exist(0,DEs[0],DEs[1],DEs[2],DEs[3]);
	faces[1]=(mEntity*)F_exist(0,DEs[0],DEs[1],DEs[4],(mEntity*)0);
	faces[2]=(mEntity*)F_exist(0,DEs[1],DEs[2],DEs[4],(mEntity*)0);
	faces[3]=(mEntity*)F_exist(0,DEs[2],DEs[3],DEs[4],(mEntity*)0);
	faces[4]=(mEntity*)F_exist(0,DEs[3],DEs[0],DEs[4],(mEntity*)0);
	
	new_ent=(mEntity*)M_createR(part, 5, faces, 0, g);
	if (debug_level) newEnt[3].push_back(new_ent);
      } break;

      default :
        throw new mException(__LINE__,__FILE__,
                             "unpacking not done for this type of mesh entity");
    }
    n_bytes += numDE*sizeof(mEntity*);
  }
#ifdef _DEBUG_
  cout<<__func__<<": src_part "<<src_partid<<", dest_part "<<dest_partid<<" created "<<FMDB_Ent_StrID(new_ent)<<"(mem: "<<new_ent<<")\n";
#endif
// STEP 3: copy Remote Copies of sender
  if (numRC>0)
  {
    int remote_partid;
    mEntity *RCs_ent;
    for (short int i=0; i < numRC; i++)
    {
      remote_partid = *(int*)((char*)msg_recv + n_bytes);
      n_bytes += sizeof(int);
      RCs_ent = *(mEntity**)((char*)msg_recv + n_bytes);
      n_bytes += sizeof(mEntity*);
      new_ent->addRemoteCopy(remote_partid, RCs_ent);
#ifdef _DEBUG_
  cout<<"["<<SCUTIL_CommRank()<<"] part "<<dest_partid<<": "<<FMDB_Ent_StrID(new_ent)<<"->addRemoteCopy ("
      <<remote_partid<<", "<<RCs_ent<<")\n";
#endif

    }
  }

// STEP 4: copy bounding partition iDs and set PClassification
  if (numBP > 1)
  {
    int* BPs = (int*)((char*)msg_recv + n_bytes);
    set<int> BPset;
    for (short int i=0; i < numBP; ++i)
      BPset.insert(BPs[i]);
    pmEntity* pe = pmModel::Instance()->getPartitionEntity(BPset);
    new_ent->setPClassification(pe);
    n_bytes += numBP*sizeof(int);
  }

  if (new_ent->getLevel() < meshDim)
  {
    new_ent->addRemoteCopy(src_partid, sender);  
#ifdef _DEBUG_
  cout<<"["<<SCUTIL_CommRank()<<"] part "<<dest_partid<<": "<<FMDB_Ent_StrID(new_ent)<<"->addRemoteCopy ("
      <<src_partid<<", "<<sender<<")\n";
#endif
    entitiesToBounce.push_back(rc_struct_3(dest_partid, new_ent, src_partid, sender));  
  }
  else // new_ent->getLevel()==meshDim 
  {
    // get weight
    double* weight = (double*)((char*)msg_recv + n_bytes);
    FMDB_Ent_SetWeight(new_ent, *weight);
    n_bytes+=sizeof(double);
    
    // get p-set
    void** tmp = (void**)((char*)msg_recv + n_bytes);
    n_bytes += sizeof(mEntityGroup*); 
    
    if((*tmp)!=0) {                                          // in entity-group 
      mEntityGroup* entGrp = (mEntityGroup*)(*tmp);
      
      map<pair<mEntityGroup*, int>, mEntityGroup*>::iterator mapIt = entGrpMap.find(make_pair(entGrp, src_partid)); 
      if(mapIt!=entGrpMap.end()) {                          // already in the map 
         int* pos = (int*)((char*)msg_recv + n_bytes);
         (mapIt->second)->addEntity(new_ent, *pos);
	 n_bytes += sizeof(int);

         if(*pos==0) {
	    int* eg_user_size = (int*)((char*)msg_recv + n_bytes);
	    n_bytes += sizeof(int);

	    if( (*eg_user_size) !=0) {
	       IPComMan *CM = ParUtil::Instance()->ComMan();
	       cb.receiveEntGrpUserData(mapIt->second, src_partid, CM->get_tag(), (void*)((char*)msg_recv + n_bytes));	    
	    }
	    n_bytes += *eg_user_size;
	 }
      }
      else {                                                 // create one item in the map 
	 mEntityGroup* newEntGrp = (mEntityGroup*) part->createPSet(meshDim); 
  	 if (debug_level) newPSet[part->getId()%numPart].push_back(newEntGrp);
	 entGrpMap.insert(make_pair(make_pair(entGrp, src_partid), newEntGrp)); 

         int* pos = (int*)((char*)msg_recv + n_bytes);
         newEntGrp->addEntity(new_ent, *pos);
         n_bytes += sizeof(int);


         if(*pos==0) {
            int* eg_user_size = (int*)((char*)msg_recv + n_bytes);
            n_bytes += sizeof(int);

            if( (*eg_user_size) !=0) {
               IPComMan *CM = ParUtil::Instance()->ComMan();
               cb.receiveEntGrpUserData(newEntGrp, dest_partid, CM->get_tag(), (void*)((char*)msg_recv + n_bytes));
            }
            n_bytes += *eg_user_size;
         }
      }
    }
  }

// STEP 6: process user attached data
  IPComMan *CM = ParUtil::Instance()->ComMan(); 
  cb.receiveUserData(new_ent, src_partid, CM->get_tag(), (void*)((char*)msg_recv + n_bytes));
}

// **********************************************
void bounceEntities(std::vector<rc_struct_3>& entitiesToBounce)
// **********************************************
{
  int msg_size = sizeof(int)+2*sizeof(mEntity*);

  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::Neighbors);

  CM->set_tag(0);
  CM->set_fixed_msg_size(msg_size);
  void* msg_send = CM->alloc_msg(msg_size);

  std::vector<rc_struct_3>::const_iterator rcvIter;
  for (rcvIter=entitiesToBounce.begin(); rcvIter!=entitiesToBounce.end();++rcvIter)
  { 
    mEntity** s_ent = (mEntity**)msg_send; 
    s_ent[0] = rcvIter->src_ent;  // new_ent
    s_ent[1] = rcvIter->dest_ent; // sender entity
    int* s_id = (int*)((char*)msg_send+sizeof(mEntity*)*2);  
    *s_id = rcvIter->src_partid; // new_ent partid
    CM->send(rcvIter->dest_partid/(ParUtil::Instance()->getTgtNumParts()) , (void*)msg_send);
  }

  std::vector<rc_struct_2> entitiesToBroadcast;
  CM->finalize_send();
  CM->free_msg(msg_send);

  // receive phase begins
  void *msg_recv;
  int pid_from;
  while(CM->receive(msg_recv, &pid_from))
  {
    mEntity** s_ent = (mEntity**)msg_recv;
    int* s_id = (int*)((char*)msg_recv+sizeof(mEntity*)*2); 
    s_ent[1]->addRemoteCopy(*s_id, s_ent[0]);
#ifdef _DEBUG_
  cout<<"["<<SCUTIL_CommRank()<<"] "<<__func__<<FMDB_Ent_StrID(s_ent[1])<<"->addRemoteCopy ("
      <<*s_id<<", "<<s_ent[0]<<")\n";
#endif
    entitiesToBroadcast.push_back(rc_struct_2(s_ent[1] /*sender entity*/, *s_id, s_ent[0] /*new entity*/));
    CM->free_msg(msg_recv);
  }

  broadcastEntities(entitiesToBroadcast);
}


// **********************************************
void broadcastEntities(std::vector<rc_struct_2>& entitiesToBroadcast)
// **********************************************
{
  mEntity* ent;
  int remote_partid;
  
  int msg_size = 2*sizeof(mEntity*) + sizeof(int);

  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_tag(0);
  CM->set_fixed_msg_size(msg_size);
  void *msg_send = CM->alloc_msg(msg_size);
 
  int numPart = ParUtil::Instance()->getTgtNumParts();   

  std::vector<rc_struct_2>::const_iterator rcvIter;
  mEntity** castbuf;
  int *s_pid;
  for (rcvIter=entitiesToBroadcast.begin(); rcvIter!=entitiesToBroadcast.end();++rcvIter)
  {
    ent = rcvIter->src_ent;
    for (mEntity::RCIter rciter=ent->rcBegin();rciter!=ent->rcEnd();++rciter)
    {
      remote_partid = rciter->first;
      if (remote_partid == rcvIter->dest_partid)
        continue;
      castbuf = (mEntity**)msg_send;
      castbuf[0] = rciter->second;  // remote copy
      castbuf[1] = rcvIter->dest_ent;
      s_pid = (int*)((char*)msg_send + 2*sizeof(mEntity*));
      *s_pid = rcvIter->dest_partid;
      CM->send(remote_partid/numPart, (void*)msg_send);
    }
  }
  CM->finalize_send();
  CM->free_msg(msg_send);

  // receive phase begins
  void *msg_recv;
  int pid_from;

  while(CM->receive(msg_recv, &pid_from))
  {
    castbuf = (mEntity**)msg_recv;
    s_pid = (int*)((char*)msg_recv + 2*sizeof(mEntity*));
    castbuf[0]->addRemoteCopy(*s_pid, castbuf[1]);
#ifdef _DEBUG_
  cout<<"["<<SCUTIL_CommRank()<<"] "<<__func__<<FMDB_Ent_StrID(castbuf[0])<<"->addRemoteCopy ("
      <<*s_pid<<", "<< castbuf[1]<<")\n";
#endif
    CM->free_msg(msg_recv);
  }
}

// ***********************************************************
void removeUnusedEntities(mMesh* mesh, pmMigrationCallbacks &cb,int dim, 
			 std::map<mEntity*,int>* entitiesToRemove)
// ***********************************************************
{
  int numPart = ParUtil::Instance()->getTgtNumParts(); 

  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::All_Reduce);
  CM->set_tag(0);
  int msg_size = sizeof(mEntity*)+sizeof(int);
  CM->set_fixed_msg_size(msg_size);
  void* msg_send = CM->alloc_msg(msg_size);
  mEntity** s_ent;
  mEntity* ent; 
  int* s_id; 

  for (int i=dim; i>=0; --i)  // region to vertex
  {
    for (std::map<mEntity*, int>::iterator eiter=entitiesToRemove[i].begin(); eiter!=entitiesToRemove[i].end();++eiter) 
    {
      ent = eiter->first; 
      cb.deleteEntityData(ent);
      if (ent->getNumRemoteCopies())
      {
        for (mEntity::RCIter rcIter=ent->rcBegin(); rcIter!=ent->rcEnd();++rcIter)
        {
	  s_ent = (mEntity**)msg_send; 
	  *s_ent = rcIter->second;
	  s_id = (int*)((char*)msg_send+sizeof(mEntity*)); 
	  *s_id = eiter->second;   // ent's src partid
          CM->send((rcIter->first)/numPart, (void*)msg_send);
        }
      }
    }
  }
  CM->finalize_send();
  CM->free_msg(msg_send);

  // receive phase begins
  void *msg_recv;
  int pid_from;
  while(CM->receive(msg_recv, &pid_from))
  {
    s_ent = (mEntity**)msg_recv;
    s_id = (int*)((char*)msg_recv+sizeof(mEntity*)); 
    (*s_ent)->deleteRemoteCopy(*s_id);
    CM->free_msg(msg_recv);
  }

   void* tmp; 

  if (dim==3)
    for (std::map<mEntity*, int>::iterator eiter=entitiesToRemove[3].begin(); eiter!=entitiesToRemove[3].end();++eiter){
      if(EN_getDataPtr(eiter->first, FMDB_Util::Instance()->getPSet(), &tmp)) {
	 (eiter->first)->deleteData(FMDB_Util::Instance()->getPSet());
      }
      M_removeRegion(mesh->getPart(eiter->second%numPart), eiter->first);
    }
      
  for (std::map<mEntity*, int>::iterator eiter=entitiesToRemove[2].begin(); eiter!=entitiesToRemove[2].end();++eiter) {
      if(dim==2 && EN_getDataPtr(eiter->first, FMDB_Util::Instance()->getPSet(), &tmp)) {
	  (eiter->first)->deleteData(FMDB_Util::Instance()->getPSet()); 
       }
      M_removeFace(mesh->getPart(eiter->second%numPart), eiter->first );
  }

  for (std::map<mEntity*, int>::iterator eiter=entitiesToRemove[1].begin(); eiter!=entitiesToRemove[1].end();++eiter)
       M_removeEdge(mesh->getPart(eiter->second%numPart), eiter->first );
  
  for (std::map<mEntity*, int>::iterator eiter=entitiesToRemove[0].begin(); eiter!=entitiesToRemove[0].end();++eiter)
       M_removeVertex(mesh->getPart(eiter->second%numPart), eiter->first );

}

#endif
