/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#ifdef FMDB_PARALLEL

#include "pmModel.h"
#include "FMDB.h"

std::auto_ptr<pmModel> pmModel::instance;

pmModel::pmModel() 
{
  pmEntId = 0;
}
  
pmModel::~pmModel()
{ 
  PEIter peit=peBegin();
  for(; peit!=peEnd(); ++peit)
  {
    delete (*peit);
  }
  allPEntities.clear();
}
  
pmModel* pmModel::Instance()
{
  if(!instance.get())
    instance.reset(new pmModel);
  return instance.get();
}

pmEntity* pmModel::getPartitionEntity(set<int>& bps, int owner)
{
  if (bps.size() < 2)
     return (pmEntity*)0;
     
  PEIter it=peBegin();
  for (; it!=peEnd(); ++it)
  {
    if ((*it)->isEqual(bps))
      return (*it);
  }
  pmEntId++;
  pmEntity* newPE = new pmEntity(pmEntId, bps, 0);
  allPEntities.insert(newPE);
  if (owner!=-1)
    newPE->setOwner(owner);
  return newPE;
}

pmEntity* pmModel::getPartitionEntity(int id)
{
  PEIter it=peBegin();
  for (; it!=peEnd(); ++it)
  {
    if ((*it)->getId()==id)
      return (*it);
  }
  return (pmEntity*)0;
}

void pmModel::updateOwnership(mMesh* mesh)
{
  list<int> allSortedPids;
  list<int>::iterator pidIter;
  getAscendingPartIDWithNumPtnObj(mesh,allSortedPids);
  
  for(PEIter peit=peBegin(); peit!=peEnd(); ++peit) 
     (*peit)->setOwner(allSortedPids);
}

void pmModel::getNborPartId (mPart* part, std::set<int>& nbor_part)
{
  set<int>::iterator partid_it;
  for (PEIter peit=peBegin(); peit!=peEnd(); ++peit)
  {
    for (partid_it= (*peit)->bpBegin(); partid_it != (*peit)->bpEnd(); ++partid_it)
      if (*partid_it!= part->getId())
        nbor_part.insert(*partid_it);
  }
}

void pmModel::getNborProcRank(mMesh* mesh, std::set<int>& nbor_proc)
{
  set<int>::iterator partid_it;
  for (PEIter peit=peBegin(); peit!=peEnd(); ++peit)
  {
    for (partid_it= (*peit)->bpBegin(); partid_it != (*peit)->bpEnd(); ++partid_it)
      if ( (*partid_it)/mesh->getNumPart() != SCUTIL_CommRank())
        nbor_proc.insert((*partid_it)/mesh->getNumPart());
  }
}


void pmModel::print()
{
  PEIter it=peBegin();
  for (; it!=peEnd(); ++it)
    (*it)->print();
}

#endif /* PARALLEL */
