/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/ 
#ifdef FMDB_PARALLEL
#include "pmMigration.h"
#include "FMDB.h"
#include "mPart.h"
#include "FMDBInternals.h"
#include <list>
#include <map>

void unifyPartBdryEntList(std::list<mEntity*>& entities);

void computePtnObjToMigr(mPart* part, std::list<mEntity*>& PartBdryEnt,
		std::map<mEntity*, std::pair<int, int> >& EntToMigr, 
                std::map<mEntityGroup*, std::pair<int, int> >& SetToMigr);

void computePtnObjToMigrForSnapping(mPart* part, std::list<mEntity*>& PartBdryVtx, 
		      std::list<std::pair<mEntity*,mEntity*> > &VtxEntPair,
		      std::map<mEntity*, std::pair<int, int> >& EntToMigr,
                      std::map<mEntityGroup*, std::pair<int, int> >& SetToMigr);

// called by M_migration
// *****************************************
int pullPartBdryToOwner(mMesh* mesh, pmMigrationCallbacks &cb,
                    std::list<mEntity*>& part_bdry_entities,
                    std::vector<mEntity*>* newEnt, std::vector<mEntity*>* rmvEnt,
                    std::vector<mEntityGroup*>* newSet, std::vector<mEntityGroup*>* rmvSet)
// *****************************************
{
  if (ParUtil::Instance()->size()==1)  return SCUtil_SUCCESS;

#ifdef DEBUG
  double cur_time, prev_time=ParUtil::Instance()->getCurTime();
  double cur_mem, prev_mem=ParUtil::Instance()->getCurMem();
  double begin_time=prev_time, begin_mem=prev_mem;
#endif

// STEP 1: unify part bdry entity list among neighboring parts
  unifyPartBdryEntList(part_bdry_entities);

#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("unifying part bdry entity list...", cur_time-prev_time, cur_mem-prev_mem);
  prev_time=cur_time;
  prev_mem=cur_mem;
#endif

// STEP 2: compute partition objects to migrate
  std::map<pMeshEnt, std::pair<int, int> > EntToMigr;
  std::map<pPartSet, std::pair<int, int> > SetToMigr;
  computePtnObjToMigr(mesh->getPart(0), part_bdry_entities, EntToMigr, SetToMigr);

#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("computing part object to migrate...", cur_time-prev_time, cur_mem-prev_mem);
#endif

// STEP 3: migrate partition objects
  return FMDB_Mesh_MigrAdv (mesh, cb, EntToMigr, SetToMigr, newEnt, rmvEnt, newSet, rmvSet);
}

// called by M_migrationForSnap
// *****************************************
int pullPartBdryToOwner(mMesh* mesh, pmMigrationCallbacks &cb,
 		    std::list<mEntity*>& PartBdryVtx, 
		    std::list<std::pair<mEntity*,mEntity*> > &VtxEntPair,
                    std::vector<mEntity*>* newEnt,
                    std::vector<mEntity*>* rmvEnt,
                    std::vector<mEntityGroup*>* newSet,
                    std::vector<mEntityGroup*>* rmvSet)
// *****************************************
{

  if (ParUtil::Instance()->size()==1)  return SCUtil_SUCCESS;

#ifdef DEBUG
  double cur_time, prev_time=ParUtil::Instance()->getCurTime();
  double cur_mem, prev_mem=ParUtil::Instance()->getCurMem();
  double begin_time=prev_time, begin_mem=prev_mem;
#endif

// STEP 1: unify part bdry vertext list among neighboring parts
  unifyPartBdryEntList(PartBdryVtx);
#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("unifying part bdry entity list...", cur_time-prev_time, cur_mem-prev_mem);
  prev_time=cur_time;
  prev_mem=cur_mem;
#endif

// STEP 2: compute partition objects to migrate
  std::map<pMeshEnt, std::pair<int, int> > EntToMigr;
  std::map<pPartSet, std::pair<int, int> > SetToMigr;
  computePtnObjToMigrForSnapping(mesh->getPart(0), PartBdryVtx, VtxEntPair, EntToMigr, SetToMigr); 
#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("computing part object to migrate...", cur_time-prev_time, cur_mem-prev_mem);
#endif

// STEP 3: migrate partition objects
  return FMDB_Mesh_MigrAdv (mesh, cb, EntToMigr, SetToMigr, newEnt, rmvEnt, newSet, rmvSet);
}

// ***********************************************************
void unifyPartBdryEntList(std::list<mEntity*>& entities)
// ***********************************************************
{
  mEntity* ent;
  
  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_tag(0);
  CM->set_fixed_msg_size(sizeof(mEntity*));
  mEntity** msg_send = (mEntity**)CM->alloc_msg(sizeof(mEntity*));

  for (list<mEntity*>::iterator ent_it = entities.begin(); ent_it!=entities.end();++ent_it)
  {
    ent = (*ent_it);
    for (mEntity::RCIter rciter=ent->rcBegin();rciter!=ent->rcEnd();++rciter)
    {  
      *msg_send = rciter->second;
      CM->send(rciter->first, (void*)msg_send);
    }
  }
  CM->finalize_send();
  CM->free_msg(msg_send);

  void *msg_recv;
  int pid_from;
  while (CM->receive(msg_recv, &pid_from))
  {
    mEntity* recvEnt = *(mEntity**)msg_recv;
    if (std::find(entities.begin(), entities.end(), recvEnt)==entities.end())
      entities.push_back(recvEnt);
    CM->free_msg(msg_recv);
  }  
}

// return 1 if pid_1 is poorer than pid_2. 0 if the same. -1 if richer.
// ***********************************************************
int comparePartID (std::list<int>& pid_list, int pid_1, int pid_2) 
// ***********************************************************
{
  if (pid_1==pid_2) return 0;
  int index_1 = std::distance(pid_list.begin(), std::find(pid_list.begin(), pid_list.end(), pid_1));
  int index_2 = std::distance(pid_list.begin(), std::find(pid_list.begin(), pid_list.end(), pid_2));
  if (index_1 < index_2) return 1;
  else return -1;
}


// ***********************************************************
void detectBrokenCavityAndUpdatePartBdryEntList (
                    std::map<mEntity*, vector<mAttachableDataContainer*> >& adjPO_map, 
                    std::list<mEntity*>& PartBdryEnt)
// ***********************************************************
{
  std::vector<mEntity*> EntToRemove;

  int first_dest_partid;
  // detect broken cavity
  for (list<mEntity*>::iterator pb_ent_it = PartBdryEnt.begin(); pb_ent_it != PartBdryEnt.end(); ++pb_ent_it)
  {
    first_dest_partid = (*(adjPO_map[*pb_ent_it].begin()))->getAttachedInt(FMDB_Util::Instance()->getDestPart());
    // iterator over adj po
    for (vector<pPtnObj>::iterator adj_po_it=adjPO_map[*pb_ent_it].begin(); 
           adj_po_it!=adjPO_map[*pb_ent_it].end(); ++adj_po_it)
    {
      if (first_dest_partid!=(*adj_po_it)->getAttachedInt(FMDB_Util::Instance()->getDestPart()))
      {
        EntToRemove.push_back(*pb_ent_it);
        break; // get out of adj_po_it for loop
      }
    }
  }
 
  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_tag(0);
  CM->set_fixed_msg_size(sizeof(mEntity*));
  mEntity** msg_send = (mEntity**)CM->alloc_msg(sizeof(mEntity*));
  
  // remote part bdry entity from PartBdryEnt containeer
  for (std::vector<mEntity*>::iterator remove_it=EntToRemove.begin(); remove_it!=EntToRemove.end(); ++remove_it)
  {
    (*remove_it)->deleteData(FMDB_Util::Instance()->getDestPart());
    PartBdryEnt.erase(std::find(PartBdryEnt.begin(), PartBdryEnt.end(), *remove_it));

    for (mEntity::RCIter rciter=(*remove_it)->rcBegin();rciter!=(*remove_it)->rcEnd();++rciter)
    {
      *msg_send = rciter->second;
      CM->send(rciter->first, (void*)msg_send);
    }
  }
  CM->finalize_send();
  CM->free_msg(msg_send);

  void *msg_recv;
  int pid_from;  
  while (CM->receive(msg_recv, &pid_from))
  {  
    std::list<mEntity*>::iterator pos = std::find(PartBdryEnt.begin(), PartBdryEnt.end(), *((mEntity**)msg_recv));
    if (pos != PartBdryEnt.end())
      PartBdryEnt.erase(pos);    
    CM->free_msg(msg_recv);
  }

#ifdef DEBUG
  for (list<mEntity*>::iterator pb_ent_it = PartBdryEnt.begin(); pb_ent_it != PartBdryEnt.end(); ++pb_ent_it)
  {
    first_dest_partid = (*(adjPO_map[*pb_ent_it].begin()))->getAttachedInt(FMDB_Util::Instance()->getDestPart());
    // iterator over adj po
    for (vector<pPtnObj>::iterator adj_po_it=adjPO_map[*pb_ent_it].begin(); 
           adj_po_it!=adjPO_map[*pb_ent_it].end(); ++adj_po_it)
      assert(first_dest_partid==(*adj_po_it)->getAttachedInt(FMDB_Util::Instance()->getDestPart()));
    }
#endif
}

// ***********************************************************
void unifyDestPart (pPart part, std::list<mEntity*>& PartBdryEnt,
                    std::map<mEntity*, vector<pPtnObj> >& adjPO_map)
// ***********************************************************
{
  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_tag(0);
  int msg_size = sizeof(mEntity*) + sizeof(int);
  CM->set_fixed_msg_size(msg_size);
  void *msg_send = CM->alloc_msg(msg_size);

  mEntity** send_ent;
  int* send_int;
  int dest_partid;

  for (list<mEntity*>::iterator pb_ent_it = PartBdryEnt.begin(); pb_ent_it != PartBdryEnt.end(); ++pb_ent_it)
  {
    if ((*pb_ent_it)->getAttachedInt(FMDB_Util::Instance()->getDestPart(), &dest_partid)==0)
      continue;

    for (mEntity::RCIter rciter=(*pb_ent_it)->rcBegin();rciter!=(*pb_ent_it)->rcEnd();++rciter)
    {
      send_ent=(mEntity**)msg_send;
      *send_ent = rciter->second;
      send_int = (int*)((char*)msg_send+sizeof(mEntity*));
      *send_int = dest_partid; 
      CM->send(rciter->first, msg_send);
    }
  }  
  CM->finalize_send();
  CM->free_msg(msg_send);

  void *msg_recv;
  int pid_from;
  mEntity** remote_ent;
  int* remote_dest_partid;
  while (CM->receive(msg_recv, &pid_from))
  {
    remote_ent = (mEntity**)msg_recv;
    remote_dest_partid = (int*)((char*)msg_recv + sizeof(mEntity*));
    (*remote_ent)->attachInt(FMDB_Util::Instance()->getDestPart(), *remote_dest_partid);
    CM->free_msg(msg_recv);
  }
}

// Spec: Given the part bdry entity list, identify partition objects to migrate 
// Constraint: single part operational - input "part" handle is source part of all entities in PartBdryEnt
// ***********************************************************
void computePtnObjToMigr(mPart* part, std::list<mEntity*>& PartBdryEnt,
		std::map<mEntity*, std::pair<int, int> >& EntToMigr, 
                std::map<mEntityGroup*, std::pair<int, int> >& SetToMigr)
// ***********************************************************
{
  assert(ParUtil::Instance()->get_maxMeshDim()==3 && ParUtil::Instance()->getTgtNumParts()==1);

  int po_count, owner_partid, dest_partid;
  std::list<int> pid_list;
  getAscendingPartIDWithNumPtnObj(part->getMesh(), pid_list);
    
  mEntity* partBdry_ent, *po_ent;
  
  void* pset_ptr;
  mEntityGroup* pset;

  // STEP 1: compute adj po for each part bdry entity and store in adjPO_map
  set<mAttachableDataContainer*> adjPO;
  map<mEntity*, vector<mAttachableDataContainer*> > adjPO_map;

  for (list<mEntity*>::iterator pb_ent_it = PartBdryEnt.begin(); pb_ent_it != PartBdryEnt.end(); ++pb_ent_it)
  {
    adjPO.clear();
    (*pb_ent_it)->getAdjPtnObj (adjPO);
    for (set<mAttachableDataContainer*>::iterator adj_po_it=adjPO.begin(); 
              adj_po_it!=adjPO.end(); ++adj_po_it)
      adjPO_map[*pb_ent_it].push_back(*adj_po_it);
  }
 
  // STEP 2: compute dest part id of part bdry entity
  int pbdry_dest_partid;
  for (list<mEntity*>::iterator pb_ent_it = PartBdryEnt.begin(); pb_ent_it != PartBdryEnt.end(); ++pb_ent_it)
  {
    partBdry_ent = *pb_ent_it;
    owner_partid = partBdry_ent->getOwnPartID(part); 

    for (vector<pPtnObj>::iterator adj_po_it=adjPO_map[partBdry_ent].begin(); 
             adj_po_it!=adjPO_map[partBdry_ent].end(); ++adj_po_it)
    {
      if ((*adj_po_it)->getAttachedInt(FMDB_Util::Instance()->getDestPart(), &dest_partid)==0)
        (*adj_po_it)->attachInt(FMDB_Util::Instance()->getDestPart(), owner_partid); // attach owner_partid to adj po
      else // if dest partid attached already, check conflict
      {
        if (owner_partid==dest_partid) continue;
        if (comparePartID(pid_list, dest_partid, owner_partid)==1) // existing dest part id has priority
          (*pb_ent_it)->attachInt(FMDB_Util::Instance()->getDestPart(), dest_partid); 
        else // owner partid has priority
          (*adj_po_it)->attachInt(FMDB_Util::Instance()->getDestPart(), owner_partid); // attach owner_partid to adj po
      }
    }
  }

  // STEP 3: detect broken cavity and remove part bdry ent from PartBdryEnt list
  detectBrokenCavityAndUpdatePartBdryEntList(adjPO_map, PartBdryEnt);

  // STEP 4: unify dest part
  unifyDestPart (part, PartBdryEnt, adjPO_map);

  // STEP 5: compute SetToMigr and EntToMigr
  int ptnObj_type;
  for (list<mEntity*>::iterator pb_ent_it = PartBdryEnt.begin(); pb_ent_it != PartBdryEnt.end(); ++pb_ent_it)
  {
    if ((*pb_ent_it)->getAttachedInt(FMDB_Util::Instance()->getDestPart(), &dest_partid)==0)
      dest_partid=(*pb_ent_it)->getOwnPartID(part);
    else
      (*pb_ent_it)->deleteData(FMDB_Util::Instance()->getDestPart());

    if (dest_partid==part->getId()) continue;

    for (vector<pPtnObj>::iterator adj_po_it=adjPO_map[*pb_ent_it].begin();
           adj_po_it!=adjPO_map[*pb_ent_it].end(); ++adj_po_it)
    {
      FMDB_PtnObj_GetType(*adj_po_it, &ptnObj_type);
      if (ptnObj_type==FMDB_PSET)
      {
        if (SetToMigr.count((mEntityGroup*)(*adj_po_it))==0)
        {
          SetToMigr.insert(std::map<mEntityGroup*, std::pair<int, int> >::value_type((mEntityGroup*)(*adj_po_it), 
                   std::make_pair(part->getId(), dest_partid)));
          (*adj_po_it)->deleteData(FMDB_Util::Instance()->getDestPart());
        }
      }
      else
      {
        if (EntToMigr.count((mEntity*)(*adj_po_it))==0)
        {
          EntToMigr.insert(std::map<mEntity*, std::pair<int, int> >::value_type((mEntity*)(*adj_po_it), 
                             std::make_pair(part->getId(), dest_partid)));
          (*adj_po_it)->deleteData(FMDB_Util::Instance()->getDestPart());
        }
      }
    } // for (vector<mAttachableDataContainer*>::iterator adj_po_it)
  } // for (list<mEntity*>::iterator pb_ent_it)
}

// Spec: For part bdry vtx in PartBdryVtx, identify adjacent ptn object to migrate to owner part of vtx.
// For a pair of internal vertex on model boundary (mbdry_vtx) and an entity (vertex, edge or face), if all vertices of the 
// entity are in the part boundary, identify entity's vertices' adjacent ptn object to migrate to the part
// where mbdry_vtx exists. Vertices in PartBdryVtx have higher priority over mbdry_vtx-entity pair in VtxEntPair.
// In other words, for a pair of internal vertex (mbdry_vtx) and an entity (vertex, edge or face),
// if all of entity's vertices are found in PartBdryVtx and their remote, adjacent
// ptn object's dest. part id doesn't match the result of computePtnObjToMigr, the migration
// for mbdry_vtx-entity pair is ignored.
//
// Constraint: single part operational - input "part" handle is source part of all entities
//             in PartBdryVtx and vtxEntPair
// ***********************************************************
void computePtnObjToMigrForSnapping(mPart* part, std::list<mEntity*>& PartBdryVtx, 
		      std::list<std::pair<mEntity*,mEntity*> > &VtxEntPair,
		      std::map<mEntity*, std::pair<int, int> >& EntToMigr,
                      std::map<mEntityGroup*, std::pair<int, int> >& SetToMigr) 
// ***********************************************************
{
  // STEP 1: For each mbdry_vtx-entity pair in VtxEntPair, if entity's vertices are all part bdry,
  //         store mbdry_vtx-vertices in "OrgVtxPartBdryVtxPairToMigr" container
  map<mEntity*, vector<mEntity*> > OrgVtxPartBdryVtxPairToMigr;
  bool allvtx_on_partbdry_flag;
  for (std::list<std::pair<mEntity*,mEntity*> >::iterator pair_it=VtxEntPair.begin();
      pair_it!=VtxEntPair.end();++pair_it)
  { 
    switch (pair_it->second->getLevel())
    {
      case FMDB_VERTEX: 
              if (find(PartBdryVtx.begin(),PartBdryVtx.end(),pair_it->second)!=PartBdryVtx.end())
                 OrgVtxPartBdryVtxPairToMigr[pair_it->first].push_back(pair_it->second);
  	      break; // exit for switch
      case FMDB_EDGE: 
      case FMDB_FACE: 
             allvtx_on_partbdry_flag=true;
             for (int i=0; i<pair_it->second->size(FMDB_VERTEX); ++i)
             {
               if (std::find(PartBdryVtx.begin(),PartBdryVtx.end(), pair_it->second->get(FMDB_VERTEX, i))
	               ==PartBdryVtx.end()) // if one of the vertices is not found in PartBdryVtx
               {
                 allvtx_on_partbdry_flag=false;
                 break; // exit for
               }
             }
             if (allvtx_on_partbdry_flag)
               for (int i=0; i<pair_it->second->size(FMDB_VERTEX); ++i)
                 OrgVtxPartBdryVtxPairToMigr[pair_it->first].push_back(pair_it->second->get(FMDB_VERTEX, i));
	     break; // exit switch
      default: break;
    } // switch
  } // for pair_it
  
  cout<<"["<<SCUTIL_CommRank()<<"] "<<__func__<<" OrgVtxPartBdryVtxPairToMigr.size()="<< OrgVtxPartBdryVtxPairToMigr.size()<<endl;
  // STEP 2: determine partition object (regions and pset) to migrate as per PartBdryVtx
  computePtnObjToMigr(part, PartBdryVtx, EntToMigr, SetToMigr);

  // STEP 3: For each part bdry vtx in OrgVtxPartBdryVtxPairToMigr, SEND msg (remote_copy of vtx, orgVtx) 
  //         to remote part. On recieving msg from [org_part], remote part stores the msg
  //         into PartBdryVtxOrgVtxPair[org_part]. 
  std::vector<std::vector<pair<mEntity*, mEntity*> > > PartBdryVtxOrgVtxPair(SCUTIL_CommSize());
    
  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::Neighbors);
  CM->set_tag(0);
  CM->set_fixed_msg_size(2*sizeof(mEntity*));
  mEntity** msg_send = (mEntity**)CM->alloc_msg(2*sizeof(mEntity*));

  for (std::map<mEntity*, vector<mEntity*> >::iterator orgv_pbv_pair_it=OrgVtxPartBdryVtxPairToMigr.begin();
       orgv_pbv_pair_it!=OrgVtxPartBdryVtxPairToMigr.end();++orgv_pbv_pair_it)
  {
    for (vector<mEntity*>::iterator pbv_it=orgv_pbv_pair_it->second.begin(); 
         pbv_it!=orgv_pbv_pair_it->second.end();++pbv_it)
    {  
      for (mEntity::RCIter rciter=(*pbv_it)->rcBegin();rciter!=(*pbv_it)->rcEnd();++rciter)
      {
        msg_send[0] = rciter->second;
	msg_send[1] = orgv_pbv_pair_it->first; // orgv
        CM->send(rciter->first, (void*)msg_send);
      }
    }
  }
  CM->free_msg(msg_send);
  CM->finalize_send();

  set<pPtnObj> adjPtnObj;
  map<mEntity*, vector<pPtnObj> > adjPO_map;

  void *msg_recv;
  int mbdry_vtx_partid;
  while (CM->receive(msg_recv, &mbdry_vtx_partid))
  {
    mEntity** castbuf = (mEntity**)msg_recv;
    PartBdryVtxOrgVtxPair[mbdry_vtx_partid].push_back(pair<mEntity*, mEntity*>(castbuf[0]/*pbv*/, castbuf[1]/*orgv*/));   
    if (adjPO_map.find(castbuf[0])==adjPO_map.end())
    { // compute adj po for each part bdry vtx in and store in adjPO_map
      adjPtnObj.clear();
      (castbuf[0])->getAdjPtnObj(adjPtnObj);
      for (set<pPtnObj>::iterator adj_po_it=adjPtnObj.begin(); adj_po_it!=adjPtnObj.end(); ++adj_po_it)
        adjPO_map[castbuf[0]].push_back(*adj_po_it);
    }
    CM->free_msg(msg_recv);
  }  

  // STEP 4: For each pid from poor to rich order. For part bdry vtx in PartBdryVtxOrgVtxPair[pid], 
  //         mark adj ptn object to migrate to pid which is the part id where org vtx exists. 
  //         If any adj ptn object is marked with different dest pid, SEND msg back to orgVtx on pid
  //         so as to cancel migration for orgVtx-entity pair
  list<int> sortedPids;    // get sorted pids from poor to rich
  getAscendingPartIDWithNumPtnObj(part->getMesh(),sortedPids);

  CM->set_fixed_msg_size(sizeof(mEntity*));
  msg_send = (mEntity**)CM->alloc_msg(sizeof(mEntity*));
  CM->set_tag(0); // reset sendcounts
  int ptn_obj_type;
  mEntity* rgn;
  mEntityGroup* pset;
  bool not_to_migr_flag;  

  for (list<int>::iterator partid_it=sortedPids.begin(); partid_it!=sortedPids.end();++partid_it)
  {   
    mbdry_vtx_partid = *partid_it;
    if (mbdry_vtx_partid==SCUTIL_CommRank()) continue;

    for (vector<pair<mEntity*, mEntity*> > ::iterator pbv_orgv_it=PartBdryVtxOrgVtxPair[mbdry_vtx_partid].begin();
          pbv_orgv_it!=PartBdryVtxOrgVtxPair[mbdry_vtx_partid].end();++pbv_orgv_it)
      {
        not_to_migr_flag = false;
        for (vector<pPtnObj>::iterator adj_po_it=adjPO_map[pbv_orgv_it->first].begin(); 
                adj_po_it!=adjPO_map[pbv_orgv_it->first].end(); ++adj_po_it)
        {
          FMDB_PtnObj_GetType(*adj_po_it, &ptn_obj_type);
          if (ptn_obj_type==FMDB_PSET)
          {
            pset = (mEntityGroup*)(*adj_po_it);
            if (SetToMigr.find(pset)==SetToMigr.end())
              SetToMigr.insert(std::map<mEntityGroup*, std::pair<int, int> >::value_type
                                 (pset, std::make_pair(part->getId(), mbdry_vtx_partid)));
            else 
	    {
              /// There are no priorities anymore - first come first served, so no need to check if the region 
              /// was grabbed by a higher or lower priority pid.
              if (SetToMigr[pset].second!=mbdry_vtx_partid) 
              {
                not_to_migr_flag=true;
                break;
              }
            } // else
          } // if pset
          else
          {
            rgn = (pMeshEnt)(*adj_po_it);
            if (EntToMigr.find(rgn)==EntToMigr.end())
              EntToMigr.insert(std::map<mEntity*, std::pair<int, int> >::value_type
                                 (rgn, std::make_pair(part->getId(), mbdry_vtx_partid)));
            else 
	    {
              /// There are no priorities anymore - first come first served, so no need to check if the region 
              /// was grabbed by a higher or lower priority pid.
              if (EntToMigr[rgn].second!=mbdry_vtx_partid) 
              {
                not_to_migr_flag=true;
                break;
              }
            } // else
          } // if entity 
        } // for adj_po_it

        if (not_to_migr_flag)
        {
          for (vector<pPtnObj>::iterator adj_po_it=adjPO_map[pbv_orgv_it->first].begin(); 
                adj_po_it!=adjPO_map[pbv_orgv_it->first].end(); ++adj_po_it)
          {
            FMDB_PtnObj_GetType(*adj_po_it, &ptn_obj_type);
            if (ptn_obj_type==FMDB_PSET)
            {
              pset = (mEntityGroup*)(*adj_po_it);
     	      if (SetToMigr.find(pset)!=SetToMigr.end() && SetToMigr[pset].second==mbdry_vtx_partid)
	        SetToMigr.erase(std::map<mEntityGroup*, std::pair<int, int> >::key_type(pset));
            }
            else
            {
              rgn = (pMeshEnt)(*adj_po_it);
     	      if (EntToMigr.find(rgn)!=EntToMigr.end() && EntToMigr[rgn].second==mbdry_vtx_partid)
	        EntToMigr.erase(std::map<mEntity*, std::pair<int, int> >::key_type(rgn));
            }
          } // for adj_po_it
	  // send message to orgVtx to cancel migration related to it
          *msg_send = pbv_orgv_it->second;
          CM->send(mbdry_vtx_partid, (void*)msg_send);
        } // if not_to_migr_flag
      } // for vertices[mbdry_vtx_partid]
  }
  CM->finalize_send();

  vector<mEntity*> OrgVtxNotToMigr;
  pMeshDataId noMigrRqstPartID_tag=MD_newMeshDataId("part id requesting no migration");
  int pid_from;
  while (CM->receive(msg_recv, &pid_from))
  {
    mEntity** castbuf = (mEntity**)msg_recv;
    EN_attachDataInt(*castbuf, noMigrRqstPartID_tag, pid_from);     // mark orgVt with no migration request part id
    OrgVtxNotToMigr.push_back(*castbuf);
    CM->free_msg(msg_recv);
  }  // end of while(!AP...)

// STEP 5: for each mbdry_vtx not to migrate (collected in OrgVtxNotToMigr), send msg to all remote copies of 
//         paired part bdry vertex. For each part bdry vertex received msg, cancel migrating adjacent ptn object 
//         to org_part where mbdry_vtx exists
  CM->set_tag(0);    // reset sendcounts
  int nomigr_request_partid;
  mEntity* orgVtx;

  for (vector<mEntity*>::iterator nomigr_it = OrgVtxNotToMigr.begin(); nomigr_it!=OrgVtxNotToMigr.end(); ++nomigr_it)
  {
    orgVtx = *nomigr_it;
    EN_getDataInt(orgVtx, noMigrRqstPartID_tag, &nomigr_request_partid);
    for (vector<mEntity*>::iterator vtx_it=OrgVtxPartBdryVtxPairToMigr[orgVtx].begin(); 
            vtx_it!=OrgVtxPartBdryVtxPairToMigr[orgVtx].end();++vtx_it)
    {  
      for (mEntity::RCIter rciter=(*vtx_it)->rcBegin();rciter!=(*vtx_it)->rcEnd();++rciter)
      {
        if (rciter->first == nomigr_request_partid) continue;
        *msg_send = rciter->second;
        CM->send(rciter->first, (void*)msg_send);
      }
    }
    EN_deleteData(orgVtx, noMigrRqstPartID_tag);
  } // for NotToMigrEnt

  CM->free_msg(msg_send);
  CM->finalize_send();

  while (CM->receive(msg_recv, &mbdry_vtx_partid)) 
  {
    adjPtnObj.clear();
    (*(mEntity**)msg_recv)->getAdjPtnObj(adjPtnObj);
    for (set<pPtnObj>::iterator adj_po_it=adjPtnObj.begin(); adj_po_it!=adjPtnObj.end(); ++adj_po_it)
    {
      FMDB_PtnObj_GetType(*adj_po_it, &ptn_obj_type);
      if (ptn_obj_type==FMDB_PSET)
      {
        pset = (mEntityGroup*)(*adj_po_it);
 	if (SetToMigr.find(pset)!=SetToMigr.end() && SetToMigr[pset].second==mbdry_vtx_partid)
	  SetToMigr.erase(std::map<mEntityGroup*, std::pair<int, int> >::key_type(pset));
      }
      else
      {
        rgn = (pMeshEnt)(*adj_po_it);
        if (EntToMigr.find(rgn)!=EntToMigr.end() && EntToMigr[rgn].second==mbdry_vtx_partid)
	  EntToMigr.erase(std::map<mEntity*, std::pair<int, int> >::key_type(rgn));
      }
    } // for adj_po_it
    CM->free_msg(msg_recv);
  }  // end of while receive

  MD_deleteMeshDataId(noMigrRqstPartID_tag);
}
#endif
