/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include <iostream>
#include <map>
#include <set>
#include <algorithm>
#include <assert.h>
#include <vector>
#include "mPart.h"
#include "mEntity.h"
#include "pmEntity.h"
#include "pmModel.h"
#include "ParUtil.h"

using std::copy;
using std::map;
using std::cout;
using std::endl;
using std::set;
using std::vector;

 // to be removed: this is obsolete function based on single part assumption
int mEntity::getOwner()
{
#ifdef FMDB_PARALLEL
  if (thePClassification)
    return thePClassification->getOwner();
  else 
    return ParUtil::Instance()->rank();
#else
  return 0;
#endif    
}

int mEntity::getOwnPartID (mPart* part)
{
#ifdef FMDB_PARALLEL
  if (thePClassification)   // part boundary entity
    return thePClassification->getOwner();
  else 
#endif
    return part->getId();
}

void mEntity::setPClassification(pmEntity* pe)
{
#ifdef FMDB_PARALLEL
  if (thePClassification == pe)
    return;
  if (thePClassification != (pmEntity*)0)
    thePClassification->decNumRClasEnt();

  if (pe != (pmEntity*)0)
    pe->incNumRClasEnt();

  thePClassification = pe;
#endif  
}

pmEntity* mEntity::getPClassification()
{
#ifdef FMDB_PARALLEL
  if(thePClassification!=0) 
    return thePClassification;
  else
#endif
    return (pmEntity*)0;
}


#ifdef FMDB_PARALLEL
void mEntity::setPClassificationWithBPs()
{
  pmEntity *pe;
  if(tempBPs==NULL || tempBPs->size() <= 1) 
    pe = (pmEntity*)0;
  else
    pe = pmModel::Instance()->getPartitionEntity(*tempBPs);
  setPClassification(pe);
}

void mEntity::printBPs(int src_partid)
{
    if(tempBPs==NULL) 
      return; 
  cout<<"[part "<<src_partid<<"] BP of "<<getUid()<<"("<<this<<"): ";
  for (mEntity::BPIter bpiter=bpBegin(); bpiter!=bpEnd();++bpiter)
    cout<<*bpiter<<",";
  cout<<endl;  
}

void mEntity::printRCs()
{
  if(theRemoteCopies==NULL)
   return; 
  cout<<"["<<ParUtil::Instance()->group()<<"|"<<ParUtil::Instance()->rank()<<"] RC of "<<getUid()<<": ";
  for (mEntity::RCIter rciter=rcBegin(); rciter!=rcEnd();++rciter)
    cout<<"(P"<<rciter->first<<","<<rciter->second<<"), ";
  cout<<endl;  
}

void mEntity::getPidsExist(std::set<int>& pids)          // without current part id in MtoN
{
  int numPart = ParUtil::Instance()->getCurNumParts(); 
  int tgtNumPart = ParUtil::Instance()->getTgtNumParts();
  if(numPart==tgtNumPart && numPart==1)
    pids.insert(ParUtil::Instance()->rank());

   if(theRemoteCopies==NULL)
     return;

  for (mEntity::RCIter rcIter=rcBegin(); rcIter!=rcEnd();++rcIter)
    pids.insert(rcIter->first);
}

void mEntity::getPidsExist(std::vector<int>& pids)        // without current part id in MtoN
{
  int numPart = ParUtil::Instance()->getCurNumParts(); 
  int tgtNumPart = ParUtil::Instance()->getTgtNumParts();  
  if(numPart==tgtNumPart && numPart==1)
   pids.push_back(ParUtil::Instance()->rank());

   if(theRemoteCopies==NULL)
     return;

  for (mEntity::RCIter rcIter=rcBegin(); rcIter!=rcEnd();++rcIter)
   pids.push_back(rcIter->first);
}


mEntity* mEntity::getRemoteCopy(int pid)
{
  mEntity* remoteEnt = (mEntity*)0;

  if(theRemoteCopies==NULL)
   return remoteEnt;

  map<int,mEntity*>::iterator mapit = theRemoteCopies->find(pid);
  if (mapit!=theRemoteCopies->end())
    remoteEnt = mapit->second;
  return remoteEnt;
}

void mEntity::getRemoteCopies(map<int, mEntity*>& remoteCopies)
{
  if(theRemoteCopies==NULL)
     return;
  assert(remoteCopies.empty());
  for (RCIter it=rcBegin(); it!=rcEnd(); ++it)
    remoteCopies.insert(remoteCopyMap::value_type(it->first, it->second));
}

void mEntity::addRemoteCopy(int pid, mEntity* remoteEnt)
{
  if(theRemoteCopies==NULL)
   theRemoteCopies = new remoteCopyMap;
  theRemoteCopies->insert(remoteCopyMap::value_type(pid, remoteEnt));
}
void mEntity::deleteRemoteCopy(int pid)
{
  if(theRemoteCopies==NULL)
     return;

  theRemoteCopies->erase(remoteCopyMap::key_type(pid));

  if(theRemoteCopies->empty())
     SAFE_DELETE(theRemoteCopies);
}

void mEntity::deleteRemoteCopy(mEntity::RCIter rciter)
{
  if(theRemoteCopies==NULL)
   return;

  theRemoteCopies->erase(rciter);

  if(theRemoteCopies->empty())
   SAFE_DELETE(theRemoteCopies);
}

void mEntity::clearRemoteCopies()
{
  if(theRemoteCopies==NULL)
    return;
  theRemoteCopies->clear();
  SAFE_DELETE(theRemoteCopies);
}

#endif /* FMDB_PARALLEL */
