/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/ 
#ifdef FMDB_PARALLEL
#include "pmMigration.h"
#include <map>
#include "FMDB.h"

void static_collectEntitiesToExchg_and_reset (mMesh* mesh, int meshDim);

void createVertex (mMesh* mesh, mEntity* ent, int dest_partid);
void createEdge (mMesh* mesh, mEntity* ent, int dest_partid);
void createFace (mMesh* mesh, mEntity* ent, int dest_partid, map<mEntityGroup*, mEntityGroup*>& pset_map);
void createRegion (mMesh* mesh, mEntity* ent, int dest_partid, map<mEntityGroup*, mEntityGroup*>& pset_map);
void serial_exchangeEntities(mMesh* mesh, int meshDim, int entDim,
          map<mEntityGroup*, mEntityGroup*>& pset_map, std::vector<mEntity*>* entitiesToRemove);
void serial_updatePartBdry (mMesh*  mesh, int entDim);

// *****************************************
int serial_migratePtnObject (mMesh* mesh)
// *****************************************
{
  int dest_pid;

#ifdef _DEBUG_
    for (mPart::iterall ent_it=mesh->getPart(0)->beginall(3); ent_it!=mesh->getPart(0)->endall(3); ++ent_it)
      if ((*ent_it)->getAttachedInt(FMDB_Util::Instance()->getDestPart(), &dest_pid))
        cout<<"move "<<FMDB_Ent_StrID(*ent_it)<<" to "<<dest_pid<<endl;
#endif

  int meshDim;
  FMDB_Mesh_GetDim(mesh, &meshDim);

#ifdef DEBUG
  int before_numPeak=0;
  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
    for (int i=0; i<meshDim+1; ++i)
      before_numPeak+=(*part_it)->size(i);

  double prev_time=ParUtil::Instance()->getCurTime();
  double prev_mem=ParUtil::Instance()->getCurMem();
  double begin_time=prev_time, begin_mem=prev_mem;
#endif

  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
    for(mPart::psetIter pset_it=(*part_it)->beginPSet(); pset_it!=(*part_it)->endPSet(); ++pset_it)
    {
      if ((*pset_it)->getAttachedInt(FMDB_Util::Instance()->getDestPart(), &dest_pid))
        for (mEntityGroup::entIter ent_it=((mEntityGroup*)(*pset_it))->begin(); 
                ent_it!=((mEntityGroup*)(*pset_it))->end(); ++ent_it)
          (*ent_it)->attachInt (FMDB_Util::Instance()->getDestPart(), dest_pid);
    }  

  static_collectEntitiesToExchg_and_reset(mesh, meshDim); 

#ifdef DEBUG
  SCUTIL_DspRsrc("compute entities' dest part ids...", ParUtil::Instance()->getCurTime()-prev_time,
                         ParUtil::Instance()->getCurMem()-prev_mem);
//  for (int dim = 0; dim <= meshDim; ++dim)
//    for (mPart::iterall ent_it=mesh->getPart(0)->beginall(dim); ent_it!=mesh->getPart(0)->endall(dim); ++ent_it)
//      if ((*ent_it)->getNumBPs()>0)
//        (*ent_it)->printBPs(0);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();
#endif

// ********************************************
// STEP 2: update PClassification and collect entities to remove
// ********************************************

  for (int dim=0; dim<meshDim; ++dim)
    for (mPart::iterall ent_it=mesh->getPart(0)->beginall(dim); ent_it!=mesh->getPart(0)->endall(dim); ++ent_it)
    {
      if ((*ent_it)->getNumBPs()==0 || (*ent_it)->getNumBPs()==1 && *((*ent_it)->bpBegin())==0) continue;
      (*ent_it)->setPClassificationWithBPs();      
    }
  
#ifdef DEBUG
  SCUTIL_DspRsrc("update PClassification...", ParUtil::Instance()->getCurTime()-prev_time,
                         ParUtil::Instance()->getCurMem()-prev_mem);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();
#endif

  // ********************************************
// STEP 4: Migrate entities
// ********************************************

  mEntityGroup* new_pset;
  map<mEntityGroup*, mEntityGroup*> pset_map;
  // create pset
  for (mPart::psetIter pset_it=mesh->getPart(0)->beginPSet(); pset_it!=mesh->getPart(0)->endPSet(); ++pset_it)
  {
    if ((*pset_it)->getAttachedInt(FMDB_Util::Instance()->getDestPart(), &dest_pid))
    {
      new_pset = (mEntityGroup*)(mesh->getPart(dest_pid)->createPSet(meshDim));
      pset_map[(mEntityGroup*)(*pset_it)] = new_pset;
    }
  }

#ifdef DEBUG
  SCUTIL_DspRsrc("creating p-sets...", ParUtil::Instance()->getCurTime()-prev_time,
                         ParUtil::Instance()->getCurMem()-prev_mem);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();
#endif

  std::vector<mEntity*> entitiesToRemove[4];

  serial_exchangeEntities(mesh, meshDim, 0, pset_map, entitiesToRemove);
  serial_updatePartBdry(mesh, 0);
#ifdef DEBUG
  SCUTIL_DspRsrc("creating vertices...", ParUtil::Instance()->getCurTime()-prev_time,
                         ParUtil::Instance()->getCurMem()-prev_mem);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();
#endif

  serial_exchangeEntities(mesh, meshDim, 1, pset_map, entitiesToRemove);
  serial_updatePartBdry(mesh, 1);
#ifdef DEBUG
  SCUTIL_DspRsrc("creating edges...", ParUtil::Instance()->getCurTime()-prev_time,
                         ParUtil::Instance()->getCurMem()-prev_mem);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();
#endif

  serial_exchangeEntities(mesh, meshDim, 2, pset_map, entitiesToRemove);
  if (meshDim==3)
    serial_updatePartBdry(mesh, 2);
#ifdef DEBUG
  SCUTIL_DspRsrc("creating faces...", ParUtil::Instance()->getCurTime()-prev_time,
                         ParUtil::Instance()->getCurMem()-prev_mem);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();
#endif

  serial_exchangeEntities(mesh, meshDim, 3, pset_map, entitiesToRemove);
#ifdef DEBUG
  SCUTIL_DspRsrc("creating regions...", ParUtil::Instance()->getCurTime()-prev_time,
                         ParUtil::Instance()->getCurMem()-prev_mem);

  int mid_numPeak=0;
  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
    for (int i=0; i<meshDim+1; ++i)
      mid_numPeak+=(*part_it)->size(i);

  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();
#endif


  mEntity* ent;
  for (int dim=meshDim; dim>=0; --dim)
  {
    for (std::vector<mEntity*>::iterator ent_it=entitiesToRemove[dim].begin(); ent_it!=entitiesToRemove[dim].end(); ++ent_it)
    {
      switch(dim)
      {
        case 0: if (FMDB_Util::Instance()->tagToMigrate.size())
                {
                  int tag_type, tag_size;
                  for (std::list<unsigned int>::iterator tag_it = FMDB_Util::Instance()->tagToMigrate.begin(); 
                       tag_it!=FMDB_Util::Instance()->tagToMigrate.end(); ++tag_it)
                   {
                     tag_type=FMDB_Util::Instance()->typeMeshDataId(*tag_it);
                     tag_size=FMDB_Util::Instance()->typeMeshDataId(*tag_it);
                     switch (tag_type)
                     {
                       case SCUtil_DBL: if (tag_size>1)
                                        {
                                          double* dbl_arr;
                                          EN_getDataPtr(*ent_it, *tag_it, (void**)&dbl_arr);
                                          delete [] dbl_arr;
                                        }
                                        break;
                       case SCUtil_INT: if (tag_size>1)
                                        {
                                          int* int_arr;
                                          EN_getDataPtr(*ent_it, *tag_it, (void**)&int_arr);
                                          delete [] int_arr;
                                        }
                                        break;
                      default: break;
                    }
                    (*ent_it)->deleteData(*tag_it);
                  } // for tag_it
                } // if (FMDB_Util::Instance()->tagToMigrate.size())
                M_removeVertex(mesh->getPart(0), *ent_it); 
                break;
        case 1: M_removeEdge(mesh->getPart(0), *ent_it); break;
        case 2: M_removeFace(mesh->getPart(0), *ent_it); break;
        case 3: M_removeRegion(mesh->getPart(0), *ent_it); break;
      }
    }
  }

  for (map<mEntityGroup*, mEntityGroup*>::iterator pset_it=pset_map.begin(); pset_it!=pset_map.end(); ++pset_it)
    mesh->getPart(0)->deletePSet(pset_it->first, 0);

#ifdef DEBUG
  SCUTIL_DspRsrc("remove unnecessary entities and P-sets...", ParUtil::Instance()->getCurTime()-prev_time,
                         ParUtil::Instance()->getCurMem()-prev_mem);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();
#endif

// ********************************************
// STEP 6: update partition model
// ********************************************
  pmModel::Instance()->updateOwnership(mesh);

#ifdef DEBUG
  SCUTIL_DspRsrc("update partition model...", ParUtil::Instance()->getCurTime()-prev_time,
                         ParUtil::Instance()->getCurMem()-prev_mem);
  prev_time=ParUtil::Instance()->getCurTime();
  prev_mem=ParUtil::Instance()->getCurMem();

  int after_numPeak=0;
  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
    for (int i=0; i<meshDim+1; ++i)
      after_numPeak+=(*part_it)->size(i);

  if (!SCUTIL_CommRank())
  {
    cout<<"\nMIGRATION: Time="<<prev_time-begin_time<<" (sec), Mem="<<prev_mem - begin_mem<<" (MB)\n";
    cout<<"#entity: BEFORE migration="<<before_numPeak<<", DURING migration="<<mid_numPeak<<", AFTER migration="<<after_numPeak<<"\n\n";
  }
#endif
  return SCUtil_SUCCESS;
}

// *****************************************
void static_collectEntitiesToExchg_and_reset (mMesh* mesh, int meshDim)
// *****************************************
{
  int dest_pid;
  std::vector<mEntity*> DownEnts;
  DownEnts.resize(27);

  for (mPart::iterall ent_it=mesh->getPart(0)->beginall(meshDim); ent_it!=mesh->getPart(0)->endall(meshDim); ++ent_it)
  {
    if ((*ent_it)->getAttachedInt(FMDB_Util::Instance()->getDestPart(), &dest_pid))
    {
      if (dest_pid==0) continue;
      (*ent_it)->addBPs(dest_pid);
      DownEnts.clear();
      getDownEntities((*ent_it), DownEnts);
      for (std::vector<mEntity*>::iterator down_it=DownEnts.begin();down_it!=DownEnts.end();++down_it)
        (*down_it)->addBPs(dest_pid);
    }
    else // stay in source part
    {
      DownEnts.clear();
      getDownEntities(*ent_it, DownEnts);
      for (std::vector<mEntity*>::iterator down_it=DownEnts.begin();down_it!=DownEnts.end();++down_it)
        (*down_it)->addBPs(0);
    }
  }
}

// **********************************************
void createVertex (mMesh* mesh, mEntity* ent, int dest_partid)
// **********************************************
{
  pGEntity gent = ent->getClassification();
  short int numParam=0, gType, gTag;
  if(gent!=0) {
    gType = GEN_type(gent);
    gTag = GEN_tag(gent);
  }
  else{
    gTag = 1;
    gType = 3;
  }
  if (ent->getData(FMDB_Util::Instance()->getParametric()))
  {
    if (gType==1) // vertex classified on model edge
      numParam=1;
    else if (gType==2) // vertex classified on model face
      numParam=2;
  }

  double xyz[3];
  ((mVertex*)ent)->getCoord(xyz[0], xyz[1], xyz[2]);

  mEntity* new_ent = mesh->getPart(dest_partid)->createVertex(ent->getId(), xyz[0], xyz[1], xyz[2],
                          ent->getClassification());

  if (numParam)
  {
    SCOREC::Util::mVector vec = ent->getAttachedVector(FMDB_Util::Instance()->getParametric());
    if (numParam==1)
      new_ent->attachVector (FMDB_Util::Instance()->getParametric(),SCOREC::Util::mVector(vec[0],0,0));
    else
      new_ent->attachVector (FMDB_Util::Instance()->getParametric(),SCOREC::Util::mVector(vec[0],vec[1],0));
  }

  new_ent->setPClassification(ent->getPClassification());
  for (mEntity::BPIter bpiter=ent->bpBegin();bpiter!=ent->bpEnd();++bpiter)
    if (*bpiter!=dest_partid)
      new_ent->addBPs(*bpiter);
  // update remote copy or source and dest copy
  ent->addRemoteCopy(dest_partid, new_ent);

  // attach tag data
  if (FMDB_Util::Instance()->tagToMigrate.size())
  {
    int tag_type, tag_size;
    for (std::list<unsigned int>::iterator tag_it = FMDB_Util::Instance()->tagToMigrate.begin(); 
         tag_it!=FMDB_Util::Instance()->tagToMigrate.end(); ++tag_it)
    {
      tag_type=FMDB_Util::Instance()->typeMeshDataId(*tag_it);
      tag_size=FMDB_Util::Instance()->sizeMeshDataId(*tag_it);
      switch (tag_type)
      {
        case SCUtil_DBL: if (tag_size>1)
                         {
                           double* dbl_arr;
                           EN_getDataPtr(ent, *tag_it, (void**)&dbl_arr);

                           double* dbl_arr_back = new double[tag_size];
                           for (int i=0; i<tag_size; ++i)
                           dbl_arr_back[i] = dbl_arr[i];
                           EN_attachDataPtr(new_ent, *tag_it, (void*)dbl_arr_back);
                         }
                         else
                         {
                           double data;
                           EN_getDataDbl(ent, *tag_it, &data);
                           EN_attachDataDbl(new_ent, *tag_it, data);
                         }
                         break;
        case SCUtil_INT: if (tag_size>1)
                         {
                           int* int_arr;
                           EN_getDataPtr(ent, *tag_it, (void**)&int_arr);

                           int* int_arr_back = new int[tag_size];
                           for (int i=0; i<tag_size; ++i)
                             int_arr_back[i] = int_arr[i];
                           EN_attachDataPtr(new_ent, *tag_it, (void*)int_arr_back);
                         }
                         else
                         {
                           int data;
                           EN_getDataInt(ent, *tag_it, &data);
                           EN_attachDataInt(new_ent, *tag_it, data);
                         }
                         break;
        default: break;
      } // switch (tag_type)
    }
  }
}

// **********************************************
void createEdge (mMesh* mesh, mEntity* ent, int dest_partid)
// **********************************************
{
  pGEntity gent = ent->getClassification();
  mEntity* new_ent = M_createE(mesh->getPart(dest_partid), ent->get(0,0)->getRemoteCopy(dest_partid), ent->get(0,1)->getRemoteCopy(dest_partid), ent->getClassification());

  new_ent->setPClassification(ent->getPClassification());
  for (mEntity::BPIter bpiter=ent->bpBegin();bpiter!=ent->bpEnd();++bpiter)
    if (*bpiter!=dest_partid)
      new_ent->addBPs(*bpiter);
  // update remote copy or source and dest copy
  ent->addRemoteCopy(dest_partid, new_ent);
}

// **********************************************
void createFace (mMesh* mesh, int meshDim, mEntity* ent, int dest_partid,
            map<mEntityGroup*, mEntityGroup*>& pset_map)
// **********************************************
{
  int numEdge = ent->size(1);
  mEntity* edges[numEdge];
  for (int i=0; i<numEdge; ++i)
    edges[i] = ent->get(1,i)->getRemoteCopy(dest_partid);
  mEntity* new_ent = M_createF(mesh->getPart(dest_partid), numEdge, edges, 0, ent->getClassification());

  if (meshDim==3)
  {
    new_ent->setPClassification(ent->getPClassification());
    for (mEntity::BPIter bpiter=ent->bpBegin();bpiter!=ent->bpEnd();++bpiter)
      if (*bpiter!=dest_partid)
        new_ent->addBPs(*bpiter);
    // update remote copy or source and dest copy
    ent->addRemoteCopy(dest_partid, new_ent);
  }
  else if (meshDim==2)
  {
    void* pset;
    if (EN_getDataPtr(ent, FMDB_Util::Instance()->getPSet(), &pset))
      pset_map[(mEntityGroup*)pset]->addEntity(new_ent, ((mEntityGroup*)pset)->getEntityOrder(ent));
  }
}

// **********************************************
void createRegion (mMesh* mesh, int meshDim, mEntity* ent, int dest_partid,
            map<mEntityGroup*, mEntityGroup*>& pset_map)
// **********************************************
{
  int numFace = ent->size(2);
  mEntity* faces[numFace];
  for (int i=0; i<numFace; ++i)
    faces[i] = ent->get(2,i)->getRemoteCopy(dest_partid);

  mEntity* new_ent = (mEntity*)M_createR(mesh->getPart(dest_partid), numFace, faces, 0,
                      ent->getClassification());

  void* pset;
  if (EN_getDataPtr(ent, FMDB_Util::Instance()->getPSet(), &pset))
    pset_map[(mEntityGroup*)pset]->addEntity(new_ent, ((mEntityGroup*)pset)->getEntityOrder(ent));
}

// **********************************************
void serial_exchangeEntities(mMesh* mesh, int meshDim, int entDim,
          map<mEntityGroup*, mEntityGroup*>& pset_map, std::vector<mEntity*>* entitiesToRemove)
// **********************************************
{
  mEntity* ent;
  int dest_partid;

  for (mPart::iterall ent_it=mesh->getPart(0)->beginall(entDim); ent_it!=mesh->getPart(0)->endall(entDim); ++ent_it)
  {   
    ent = *ent_it;
    if (entDim==meshDim)
    {
      if ((*ent_it)->getAttachedInt(FMDB_Util::Instance()->getDestPart(), &dest_partid))
        entitiesToRemove[entDim].push_back(ent);
    }
    else 
    { 
      if (ent->getNumBPs()>0 && !(ent->findPidInBPs(0))) // check source partid in bps
        entitiesToRemove[entDim].push_back(ent);
    }

    if (ent->getNumBPs()==0) continue;

    for (mEntity::BPIter bpiter=ent->bpBegin();bpiter!=ent->bpEnd();++bpiter)
    {
      dest_partid = *bpiter;
      if (*bpiter==0) continue;
      switch (entDim)
      {
        case 0: createVertex (mesh, ent, dest_partid);
                break;
        case 1: createEdge (mesh, ent, dest_partid);
                break;
        case 2: createFace (mesh, meshDim, ent, dest_partid, pset_map);
                break;
        case 3: createRegion (mesh, meshDim, ent, dest_partid, pset_map);
                break;
      }
    }
  }
}

// **********************************************
void serial_updatePartBdry (mMesh*  mesh, int entDim)
// **********************************************
{
  mEntity *ent, *new_ent;
  for (mPart::iterall ent_it=mesh->getPart(0)->beginall(entDim); ent_it!=mesh->getPart(0)->endall(entDim); ++ent_it)
  {
    ent = *ent_it;
    if (ent->getNumBPs()<=1) continue;

    for (mEntity::RCIter rciter=ent->rcBegin(); rciter!=ent->rcEnd();++rciter)
    {
       new_ent = rciter->second;
       for (mEntity::BPIter bpiter=new_ent->bpBegin();bpiter!=new_ent->bpEnd();++bpiter)
         if (*bpiter!=0)
           new_ent->addRemoteCopy(*bpiter, ent->getRemoteCopy(*bpiter));
         else // (*bpiter==0)
           new_ent->addRemoteCopy(0, ent);
     }
  }
}

// **********************************************
void packEntity(mEntity *ent, int src_partid, int dest_partid, 
                void *&msg_send, int &msg_size, int meshDim)
// **********************************************
{
  static int count=0;
  int entity_size;

  int tgtNumPart = ParUtil::Instance()->getTgtNumParts(); 
  IPComMan *CM = ParUtil::Instance()->ComMan();

  short int numBP = ent->getNumBPs();
  if (numBP<2) numBP=0;
  short int numRC = ent->getNumRemoteCopies();
  short int entDim = ent->getLevel();
  short int numDE = 0;

  entity_size = 2*sizeof(int) + sizeof(packedEnt) + numRC*(sizeof(int)+sizeof(mEntity*)) + numBP*sizeof(int) ;

  int n_bytes = 0;

  pGEntity gent = ent->getClassification();
  short int numParam=0, gType, gTag;
  if(gent!=0) {
    gType = GEN_type(gent);
    gTag = GEN_tag(gent);
  }
  else{
    gTag = 1;
    gType = 3;
  }

  if (entDim == 0)
  {
    if (ent->getData(FMDB_Util::Instance()->getParametric()))
    {
      if (gType==1) // vertex classified on model edge
        numParam=1;
      else if (gType==2) // vertex classified on model face
        numParam=2;
    }

    entity_size += sizeof(packedVertex) + numParam*sizeof(double);

    msg_size = entity_size;
    msg_send = CM->alloc_msg(msg_size);

    // dest global pid and src local pid 
    int* tmp_pid; 
    tmp_pid = (int*)msg_send; 
    tmp_pid[0]=dest_partid; 
    tmp_pid[1]=src_partid; 
    n_bytes += 2*sizeof(int); 

    // fixed length data
    packedEnt *entbuf = (packedEnt*)((char*)msg_send + n_bytes);
    entbuf->numDE = numDE;
    entbuf->sender = ent;
    entbuf->gTag = gTag;
    entbuf->gType = gType;
    entbuf->numRC = numRC;
    entbuf->numBP = numBP;
    n_bytes += sizeof(packedEnt);

//    cout<<"part "<<src_partid<<" - count="<<count<<", msg_size="<<msg_size<<", entbuf->sender="<<ent->getUid()<<", numDE= "<<entbuf->numDE<<endl;

    // vertex specific parameters
    packedVertex *castbuf = (packedVertex*)((char*)msg_send + n_bytes);
    castbuf->iD = ent->getId();

    double xyz[3];
    ((mVertex*)ent)->getCoord(xyz[0], xyz[1], xyz[2]);
    for(int i=0;i<3;i++) 
       castbuf->coord[i] = xyz[i]; 
    castbuf->numParam = numParam;
    n_bytes += sizeof(packedVertex);

    if (numParam > 0)
    {
      double *Params = (double*)((char*)msg_send + n_bytes);
      SCOREC::Util::mVector vec = ent->getAttachedVector(FMDB_Util::Instance()->getParametric());
      for (int k=0; k<numParam;++k) 
      {
        Params[k] = vec(k);
      }
      n_bytes += numParam*sizeof(double);
    }
  }
  else
  {
    numDE = ent->size(0); 
    entity_size += sizeof(short int) + numDE*sizeof(mEntity*);   

  if(meshDim==ent->getLevel()) 
  { 
    entity_size = entity_size+sizeof(double)+sizeof(mEntityGroup*); 
    void* tmp_ptr; 
    if (EN_getDataPtr(ent, FMDB_Util::Instance()->getPSet(), &tmp_ptr)) 
         entity_size += sizeof(int);
   }

    msg_size = entity_size;
    msg_send = CM->alloc_msg(msg_size);

    // dest global pid and src local pid
    int* tmp_pid;
    tmp_pid = (int*)msg_send;
    tmp_pid[0]=dest_partid;
    tmp_pid[1]=src_partid;
    n_bytes += 2*sizeof(int); 

    // fixed length data
    packedEnt *entbuf = (packedEnt*)((char*)msg_send + n_bytes);
    entbuf->numDE = numDE;
    entbuf->sender = ent;
    entbuf->gTag = gTag;
    entbuf->gType = gType;
    entbuf->numRC = numRC;
    entbuf->numBP = numBP;
    n_bytes += sizeof(packedEnt);

    short int *type = (short int*)((char*)msg_send + n_bytes);
    *type = ent->getType();
    n_bytes += sizeof(short int);
 
//    cout<<"part "<<src_partid<<" - count="<<count<<", msg_size="<<msg_size<<", entbuf->sender="<<ent->getUid()<<", type="<<*type<<", numDE= "<<entbuf->numDE<<endl;

    mEntity** DEs = (mEntity**)((char*)msg_send + n_bytes);
    for (int i=0; i<numDE;++i)
    {
      DEs[i] = ent->get(0,i)->getRemoteCopy(dest_partid);
    }
    n_bytes += numDE*sizeof(mEntity*);
  }

  // variable length data - RC and BP
  if (numRC>0)
  {
    int *remote_partid;
    mEntity **RCs_ent;
    for (mEntity::RCIter rciter=ent->rcBegin();rciter!=ent->rcEnd();++rciter)
    {
      remote_partid = (int*)((char*)msg_send + n_bytes);
      *remote_partid = rciter->first;
      n_bytes += sizeof(int);
      RCs_ent = (mEntity**)((char*)msg_send + n_bytes);
      *RCs_ent = rciter->second;
      n_bytes += sizeof(mEntity*);
    }
  }

  if (numBP>1)
  {
    int i=0; 
    int* BPs = (int*)((char*)msg_send + n_bytes);
    for (mEntity::BPIter bpiter=ent->bpBegin();bpiter!=ent->bpEnd();++bpiter)
    {
      BPs[i] = *bpiter;
      i++; 
    }
    n_bytes += numBP*sizeof(int);
  }

  if(ent->getLevel()==meshDim) 
  {
     // copy weight
     double* weight=(double*)((char*)msg_send+n_bytes);
     FMDB_Ent_GetWeight(ent, weight); 
     n_bytes+=sizeof(double); 

     // copy p-set
     mEntityGroup** entGrp = (mEntityGroup**)((char*)msg_send + n_bytes); 
     void* tmp_ptr; 
     if(EN_getDataPtr(ent, FMDB_Util::Instance()->getPSet(), &tmp_ptr))                     // entity is in entity-group 
       *entGrp = (mEntityGroup*)tmp_ptr; 
     else 
       *entGrp = 0; 
       
     n_bytes += sizeof(mEntityGroup*);

     if( *entGrp!=0 ) 
     {
       int* pos = (int*)((char*)msg_send+n_bytes); 
       *pos = (*entGrp)->getEntityOrder(ent);
       n_bytes += sizeof(int); 
     }
  }
//  cout<<"part "<<src_partid<<" send msg "<<count++<<" to part "<<dest_partid<<endl;
}

// **********************************************
void unpackEntity(mPart* part, int meshDim, void *msg_recv, 
     std::map<mEntityGroup*, mEntityGroup*>& pset_map, std::vector<rc_struct_3>& entitiesToBounce)
// **********************************************
{
  int n_bytes = 0;
  short int numRC, numBP;

  mEntity *new_ent, *sender;
  int numPart = ParUtil::Instance()->getTgtNumParts(); 

// PRE-STEP: add sender to remote copy and store entity to bounce
  // dest global pid and src local pid 
  int* tmp_pid = (int*)msg_recv; 
  int dest_partid = tmp_pid[0]; 
  int src_partid = tmp_pid[1]; 
  n_bytes += 2*sizeof(int); 

  static int count=0;

  packedEnt *entbuf = (packedEnt*)((char*)msg_recv + n_bytes);
  n_bytes += sizeof(packedEnt);

  sender = entbuf->sender;
  numRC = entbuf->numRC;
  numBP = entbuf->numBP;

//  cout<<"part "<<dest_partid<<" recv msg "<<count++<<" from part "<<src_partid<<", numDE="<<entbuf->numDE<<endl;

  if (entbuf->numDE == 0)
  {
    // STEP 1: create entity
    packedVertex *castbuf = (packedVertex*)((char*)msg_recv + n_bytes);
    n_bytes += sizeof(packedVertex);

    new_ent = part->createVertex(castbuf->iD, castbuf->coord[0], castbuf->coord[1], castbuf->coord[2],
                                  part->getGEntity(entbuf->gTag, entbuf->gType));
    new_ent->attachInt (FMDB_Util::Instance()->getSourcePart(), src_partid);

    if (castbuf->numParam > 0)
    {
      double* Params = (double*)((char*)msg_recv + n_bytes);
      if (castbuf->numParam == 1)
        new_ent->attachVector (FMDB_Util::Instance()->getParametric(),SCOREC::Util::mVector(Params[0],0,0));
      else // if (castbuf.numParam==2)
        new_ent->attachVector (FMDB_Util::Instance()->getParametric(),SCOREC::Util::mVector(Params[0],Params[1],0));
      n_bytes += castbuf->numParam*sizeof(double);
    }
  }
  else
  {
    short int *type = (short int*)((char*)msg_recv + n_bytes);
    n_bytes += sizeof(short int);

    pGEntity g = part->getGEntity(entbuf->gTag, entbuf->gType);

// STEP 1: create an entity
    short int numDE;
    mEntity** DEs = (mEntity**)((char*)msg_recv + n_bytes);
    switch (*type)
    {
      case mEntity::EDGE:
      {
        numDE=2;
#ifdef _DEBUG_
        if (!(mVertex*)DEs[0]) 
          cout<<"["<<SCUTIL_CommRank()<<"] src part "<<src_partid<<", dest partid "<<dest_partid<<" failed getting vertex v1"<<v1<<endl;
        if (!(mVertex*)DEs[1]) 
          cout<<"["<<SCUTIL_CommRank()<<"] src part "<<src_partid<<", dest partid "<<dest_partid<<" failed getting vertex v2"<<v2<<endl;
#endif
        new_ent = M_createE(part, (mVertex*)DEs[0], (mVertex*)DEs[1], g);
      } break;

      case mEntity::TRI:
      {
        numDE=3;
        mEntity* edges[3];
        edges[0]=(mEntity*)E_exist((mVertex*)DEs[0],(mVertex*)DEs[1]);
        edges[1]=(mEntity*)E_exist((mVertex*)DEs[1],(mVertex*)DEs[2]);
        edges[2]=(mEntity*)E_exist((mVertex*)DEs[0],(mVertex*)DEs[2]);
#ifdef _DEBUG_
        if (!edges[0]) 
          cout<<"["<<SCUTIL_CommRank()<<"] part "<<dest_partid<<" failed getting edge 0 "<<edges[0]<<endl;
        if (!edges[1]) 
          cout<<"["<<SCUTIL_CommRank()<<"] part "<<dest_partid<<" failed getting edge 1 "<<edges[1]<<endl;
        if (!edges[2]) 
          cout<<"["<<SCUTIL_CommRank()<<"] part "<<dest_partid<<" failed getting edge 2 "<<edges[2]<<endl;
#endif
        new_ent = M_createF(part, 3, edges, 0, g);
      } break;

      case mEntity::QUAD:
      {
	numDE=4; 
	mEntity* edges[4];
	edges[0]=(mEntity*)E_exist((mVertex*)DEs[0],(mVertex*)DEs[1]);
	edges[1]=(mEntity*)E_exist((mVertex*)DEs[1],(mVertex*)DEs[2]);
	edges[2]=(mEntity*)E_exist((mVertex*)DEs[2],(mVertex*)DEs[3]);
	edges[3]=(mEntity*)E_exist((mVertex*)DEs[3],(mVertex*)DEs[0]);               //does the order of vertex matter?
	new_ent = M_createF(part, 4, edges, 0, g);	
      } break;

      case mEntity::TET:
      {
        numDE=4;
        mEntity* faces[4];    
        faces[0]=(mEntity*)F_exist(0,DEs[0],DEs[1],DEs[2],(mEntity*)0);
        faces[1]=(mEntity*)F_exist(0,DEs[0],DEs[1],DEs[3],(mEntity*)0);
        faces[2]=(mEntity*)F_exist(0,DEs[1],DEs[2],DEs[3],(mEntity*)0);
        faces[3]=(mEntity*)F_exist(0,DEs[0],DEs[2],DEs[3],(mEntity*)0);
        new_ent=(mEntity*)M_createR(part, 4, faces, 0, g);
      } break;

      case mEntity::HEX: 
      {
	numDE=8; 
		
	mEntity* faces[6];
	faces[0]=(mEntity*)F_exist(0,DEs[0],DEs[3],DEs[2],DEs[1]);
	faces[1]=(mEntity*)F_exist(0,DEs[0],DEs[1],DEs[5],DEs[4]);
	faces[2]=(mEntity*)F_exist(0,DEs[1],DEs[2],DEs[6],DEs[5]);
	faces[3]=(mEntity*)F_exist(0,DEs[2],DEs[3],DEs[7],DEs[6]);
	faces[4]=(mEntity*)F_exist(0,DEs[3],DEs[0],DEs[4],DEs[7]);
	faces[5]=(mEntity*)F_exist(0,DEs[4],DEs[5],DEs[6],DEs[7]);
	new_ent=(mEntity*)M_createR(part, 6, faces, 0, g);
      } break;
   
      case mEntity::PRISM: 
      {
	numDE=6; 
	mEntity* faces[5];
	faces[0]=(mEntity*)F_exist(0,DEs[0],DEs[1],DEs[2],(mEntity*)NULL);
	faces[1]=(mEntity*)F_exist(0,DEs[0],DEs[1],DEs[4],DEs[3]);
	faces[2]=(mEntity*)F_exist(0,DEs[1],DEs[2],DEs[5],DEs[4]);
	faces[3]=(mEntity*)F_exist(0,DEs[2],DEs[0],DEs[3],DEs[5]);
	faces[4]=(mEntity*)F_exist(0,DEs[3],DEs[4],DEs[5],(mEntity*)NULL);
	new_ent=(mEntity*)M_createR(part, 5, faces, 0, g);
      } break; 

      case mEntity::PYRAMID: 
      {
	numDE=5; 

	mEntity* faces[5];
	faces[0]=(mEntity*)F_exist(0,DEs[0],DEs[1],DEs[2],DEs[3]);
	faces[1]=(mEntity*)F_exist(0,DEs[0],DEs[1],DEs[4],(mEntity*)0);
	faces[2]=(mEntity*)F_exist(0,DEs[1],DEs[2],DEs[4],(mEntity*)0);
	faces[3]=(mEntity*)F_exist(0,DEs[2],DEs[3],DEs[4],(mEntity*)0);
	faces[4]=(mEntity*)F_exist(0,DEs[3],DEs[0],DEs[4],(mEntity*)0);
	
	new_ent=(mEntity*)M_createR(part, 5, faces, 0, g);
      } break;

      default :
        cout<<"FMDB FATAL: part "<<dest_partid<<" - invalid entity type ("<<*type<<") received from source part "<<src_partid<<endl;
    }
    n_bytes += numDE*sizeof(mEntity*);
  }
#ifdef _DEBUG_
  cout<<__func__<<": src_part "<<src_partid<<", dest_part "<<dest_partid<<" created "<<FMDB_Ent_StrID(new_ent)<<"(mem: "<<new_ent<<")\n";
#endif
// STEP 3: copy Remote Copies of sender
  if (numRC>0)
  {
    int remote_partid;
    mEntity *RCs_ent;
    for (short int i=0; i < numRC; i++)
    {
      remote_partid = *(int*)((char*)msg_recv + n_bytes);
      n_bytes += sizeof(int);
      RCs_ent = *(mEntity**)((char*)msg_recv + n_bytes);
      n_bytes += sizeof(mEntity*);
      new_ent->addRemoteCopy(remote_partid, RCs_ent);
#ifdef _DEBUG_
  cout<<"["<<SCUTIL_CommRank()<<"] part "<<dest_partid<<": "<<FMDB_Ent_StrID(new_ent)<<"->addRemoteCopy ("
      <<remote_partid<<", "<<RCs_ent<<")\n";
#endif

    }
  }

// STEP 4: copy bounding partition iDs and set PClassification
  if (numBP > 1)
  {
    int* BPs = (int*)((char*)msg_recv + n_bytes);
    set<int> BPset;
    for (short int i=0; i < numBP; ++i)
      BPset.insert(BPs[i]);
    pmEntity* pe = pmModel::Instance()->getPartitionEntity(BPset);
    new_ent->setPClassification(pe);
    n_bytes += numBP*sizeof(int);
  }

  if (new_ent->getLevel()==meshDim)
  {
    // get weight
    double* weight = (double*)((char*)msg_recv + n_bytes);
    FMDB_Ent_SetWeight(new_ent, *weight);
    n_bytes+=sizeof(double);
    
    // get p-set
    void** tmp = (void**)((char*)msg_recv + n_bytes);
    n_bytes += sizeof(mEntityGroup*); 
    
    if((*tmp)!=0)
    {                                          // in entity-group 
      int* pos = (int*)((char*)msg_recv + n_bytes);
      n_bytes += sizeof(int);
      pset_map[(mEntityGroup*)(*tmp)]->addEntity(new_ent, *pos);
    } // if ((*tmp)!=0) 
  } 
  else
    entitiesToBounce.push_back(rc_struct_3(dest_partid, new_ent, src_partid, sender));  
}


// **********************************************
void bounceEntitiesToSource (std::vector<rc_struct_3>& entitiesToBounce)
// **********************************************
{
  int msg_size = sizeof(int)+2*sizeof(mEntity*);

  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::All_Reduce);

  CM->set_tag(0);
  CM->set_fixed_msg_size(msg_size);
  void* msg_send = CM->alloc_msg(msg_size);

  std::vector<rc_struct_3>::const_iterator rcvIter;
  for (rcvIter=entitiesToBounce.begin(); rcvIter!=entitiesToBounce.end();++rcvIter)
  { 
    mEntity** s_ent = (mEntity**)msg_send; 
    s_ent[0] = rcvIter->src_ent;  // new_ent
    s_ent[1] = rcvIter->dest_ent; // sender entity
    int* s_id = (int*)((char*)msg_send+sizeof(mEntity*)*2);  
    *s_id = rcvIter->src_partid; // new_ent partid
    CM->send(rcvIter->dest_partid/(ParUtil::Instance()->getTgtNumParts()) , (void*)msg_send);
#ifdef _DEBUG
  cout<<"["<<SCUTIL_CommRank()<<"] "<<__func__<<" send "<<FMDB_Ent_StrID(rcvIter->src_ent)<<"from part "
      <<rcvIter->src_partid<<" to part "<<rcvIter->dest_partid<<"\n";
#endif
  }
  CM->finalize_send();
  CM->free_msg(msg_send);

  // receive phase begins
  void *msg_recv;
  int pid_from;
  while(CM->receive(msg_recv, &pid_from))
  {
    mEntity** s_ent = (mEntity**)msg_recv;
    int* s_id = (int*)((char*)msg_recv+sizeof(mEntity*)*2); 
    s_ent[1]->addRemoteCopy(*s_id, s_ent[0]);
#ifdef _DEBUG
  cout<<"["<<SCUTIL_CommRank()<<"] "<<__func__<<FMDB_Ent_StrID(s_ent[1])<<"->addRemoteCopy ("
      <<*s_id<<", "<<s_ent[0]<<")\n";
#endif
    CM->free_msg(msg_recv);
  }
}

// **********************************************
void static_exchangeEntities(mMesh* mesh, int meshDim, int entDim,
          map<mEntityGroup*, mEntityGroup*>& pset_map, std::vector<mEntity*>* entitiesToRemove)
// **********************************************
{
  mEntity* ent;
  int dest_partid, source_partid;
  int numpart = ParUtil::Instance()->getTgtNumParts(); 

  std::vector<rc_struct_3> entitiesToBounce;

  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::All_Reduce);
  CM->set_tag(0);
  CM->set_fixed_msg_size(0);

  for (int i=0; i<numpart; ++i)
  {
    source_partid=mesh->getPart(i)->getId();
    for (mPart::iterall ent_it=mesh->getPart(i)->beginall(entDim); ent_it!=mesh->getPart(i)->endall(entDim); ++ent_it)
    {   
      ent = *ent_it;
      // collect entities to remove
      if (entDim==meshDim)
      {
        if ((*ent_it)->getAttachedInt(FMDB_Util::Instance()->getDestPart(), &dest_partid))
          entitiesToRemove[entDim].push_back(ent);
      }
      else 
      { 
        if (ent->getNumBPs()>0 && !(ent->findPidInBPs(source_partid))) // check source partid in bps
          entitiesToRemove[entDim].push_back(ent);
      }
      // exchange entities
      if (ent->getNumBPs()==0) continue;

      for (mEntity::BPIter bpiter=ent->bpBegin();bpiter!=ent->bpEnd();++bpiter)
      {
        if (*bpiter==source_partid) continue;
        dest_partid = *bpiter;
        if (FMDB_PartID_CommRank(dest_partid)==SCUTIL_CommRank())
        {
          switch (entDim)
          {
          case 0: createVertex (mesh, ent, dest_partid);
                  break;
          case 1: createEdge (mesh, ent, dest_partid);
                  break;
          case 2: createFace (mesh, meshDim, ent, dest_partid, pset_map);
                  break;
          case 3: createRegion (mesh, meshDim, ent, dest_partid, pset_map);
                  break;
          } // switch
        } //if (FMDB_PartID_CommRank(dest_partid)==SCUtil_CommRank())
        else // send entity to remote process
        {
          void *msg_send;
          int msg_size; 
          packEntity(ent,  source_partid, dest_partid, msg_send, msg_size, meshDim);
          CM->send(dest_partid/numpart, msg_send, msg_size);
          CM->free_msg(msg_send); 
        }
      } // for (mEntity::BPIter)
    } // for (mPart::iterall ent_it)
  } // for (int i=0; i<numpart; ++i)
  CM->finalize_send();

   // receive phase begins
  void *msg_recv;
  int pid_from;
  mPart* dest_part; 
  while(CM->receive(msg_recv, &pid_from))
  {
    // get the dest part id
    int* tmp_ptr = (int*)msg_recv;  
    dest_partid = *tmp_ptr; 
    dest_part = mesh->getPart(FMDB_PartID_LocalID(dest_partid)); 	
    unpackEntity(dest_part, meshDim, msg_recv, pset_map, entitiesToBounce);
    CM->free_msg(msg_recv);
  }
  bounceEntitiesToSource(entitiesToBounce);
}

// **********************************************
void static_updatePartBdry (mMesh*  mesh, int entDim)
// **********************************************
{
  if (SCUTIL_CommSize()==1)
  {
    serial_updatePartBdry (mesh, entDim);
    return;
  } 

  mEntity* ent;
  int remote_partid;
  
  int msg_size = 2*sizeof(mEntity*) + sizeof(int);

  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_tag(0);
  CM->set_fixed_msg_size(msg_size);
  void *msg_send = CM->alloc_msg(msg_size);
 
  int numpart = ParUtil::Instance()->getTgtNumParts();   

  mEntity** s_entity;
  int *s_partid;

  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
  {
    for (mPart::iterall ent_it=(*part_it)->beginall(entDim); ent_it!=(*part_it)->endall(entDim); ++ent_it)
    {
      ent = *ent_it;
      if (ent->getNumBPs()<=1 || ent->getNumRemoteCopies()==0) continue;
      for (mEntity::RCIter rciter=ent->rcBegin(); rciter!=ent->rcEnd();++rciter)
      {
        for (mEntity::BPIter bpiter=ent->bpBegin();bpiter!=ent->bpEnd();++bpiter)
        {
          if (*bpiter==rciter->first) continue;
          s_entity = (mEntity**)msg_send;
          s_entity[0] = rciter->second;  // remote copy
          if (*bpiter==(*part_it)->getId())
            s_entity[1] = ent;
          else
            s_entity[1] = ent->getRemoteCopy(*bpiter);
          s_partid = (int*)((char*)msg_send + 2*sizeof(mEntity*));
          *s_partid = *bpiter;
          CM->send(rciter->first/numpart, (void*)msg_send);
        } // (mEntity::BPIter
      } // for (mEntity::RCIter)
    } // for (mPart::iterall ent_it
  } // for (mMesh::partIter part_it
  CM->finalize_send();
  CM->free_msg(msg_send);

  // receive phase begins
  void *msg_recv;
  int pid_from;

  while(CM->receive(msg_recv, &pid_from))
  {
    s_entity = (mEntity**)msg_recv;
    s_partid = (int*)((char*)msg_recv + 2*sizeof(mEntity*));
    s_entity[0]->addRemoteCopy(*s_partid, s_entity[1]);
#ifdef _DEBUG
  cout<<"["<<SCUTIL_CommRank()<<"] "<<__func__<<FMDB_Ent_StrID(s_entity[0])<<"->addRemoteCopy ("
      <<*s_partid<<", "<< s_entity[1]<<")\n";
#endif
    CM->free_msg(msg_recv);
  }
}

// **********************************************
int check_static_migration_prep (mMesh* mesh)
// **********************************************
{
  int ent_dest_pid, pset_dest_pid, valid=1, mesh_dim, numV=0, numE=0, numF=0, numR=0;
  FMDB_Mesh_GetDim(mesh, &mesh_dim);

  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
  {
    for(mPart::psetIter pset_it=(*part_it)->beginPSet(); pset_it!=(*part_it)->endPSet(); ++pset_it)
    {
      ((mEntityGroup*)(*pset_it))->getNumEntity(&numV, &numE, &numF, &numR);
      if ((mesh_dim==3 && numV+numE+numF) || (mesh_dim==2 && numV+numE))
      {
        cout<<"part "<<(*part_it)->getId()<<": pset with non-po entity detected\n";
        valid=0;
      }
    }
  }

  void* pset;
  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
  {
    for (mPart::iterall ent_it=(*part_it)->beginall(mesh_dim); ent_it!=(*part_it)->endall(mesh_dim); ++ent_it)
    {
      if ((*ent_it)->getAttachedInt(FMDB_Util::Instance()->getDestPart(), &ent_dest_pid) && 
         EN_getDataPtr(*ent_it, FMDB_Util::Instance()->getPSet(), &pset))
      {  // check dest_partid of entity and pset
        if (((mEntityGroup*)pset)->getAttachedInt(FMDB_Util::Instance()->getDestPart(), &pset_dest_pid))
        {
          if (ent_dest_pid!=pset_dest_pid)
          {
            cout<<"part "<<(*part_it)->getId()<<": dest part id mismatch of "
                <<(*ent_it)->getUid()<<" and its beloning pset\n";
            valid=0;
          }
        }
        else // belonging pset doesn't marked to migrate
        {
          cout<<"part "<<(*part_it)->getId()<<": attempt to migrate "
              <<(*ent_it)->getUid()<<" without migrating its beloning pset\n";
          valid=0;
        } 
      } // entity belongs to pset
    } // each entity
  } // each part  
  if (valid) return SCUtil_SUCCESS;
  else return SCUtil_FAILURE;
}

// *****************************************
int static_migratePtnObject (mMesh* mesh)
// *****************************************
{
  int dest_pid;
  int numpart = ParUtil::Instance()->getTgtNumParts();

#ifdef _DEBUG_
  if (!SCUTIL_CommRank())
  {
    if (!check_static_migration_prep(mesh))
      cout<<"proc "<<SCUTIL_CommRank()<<": check_static_migration_prep SUCCEEDED\n\n";
    else
      cout<<"proc "<<SCUTIL_CommRank()<<": check_static_migration_prep FAILED\n\n";
  }
#endif

  int meshDim;
  FMDB_Mesh_GetDim(mesh, &meshDim);

#ifdef DEBUG
/*  int before_numPeak=0;
  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
    for (int i=0; i<meshDim+1; ++i)
      before_numPeak+=(*part_it)->size(i);
*/
  double cur_time, prev_time=ParUtil::Instance()->getCurTime();
  double cur_mem, prev_mem=ParUtil::Instance()->getCurMem();
  double begin_time=prev_time, begin_mem=prev_mem;
#endif
  vector<mEntityGroup*> psetToRemove;
  int count=0;
  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
    for(mPart::psetIter pset_it=(*part_it)->beginPSet(); pset_it!=(*part_it)->endPSet(); ++pset_it)
    {
      if ((*pset_it)->getAttachedInt(FMDB_Util::Instance()->getDestPart(), &dest_pid))
      {
        psetToRemove.push_back((mEntityGroup*)(*pset_it));
        (*pset_it)->attachInt(FMDB_Util::Instance()->getSourcePart(), (*part_it)->getId());
        for (mEntityGroup::entIter ent_it=((mEntityGroup*)(*pset_it))->begin(); 
                ent_it!=((mEntityGroup*)(*pset_it))->end(); ++ent_it)
          (*ent_it)->attachInt (FMDB_Util::Instance()->getDestPart(), dest_pid);
      }
    }  

  static_collectEntitiesToExchg_and_reset(mesh, meshDim); 

#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("compute entities' dest part ids...", cur_time-prev_time, cur_mem-prev_mem);
//  for (int dim = 0; dim <= meshDim; ++dim)
//    for (mPart::iterall ent_it=mesh->getPart(0)->beginall(dim); ent_it!=mesh->getPart(0)->endall(dim); ++ent_it)
//      if ((*ent_it)->getNumBPs()>0)
//        (*ent_it)->printBPs(0);
  prev_time=cur_time;
  prev_mem=cur_mem;
#endif

// ********************************************
// STEP 2: update PClassification and collect entities to remove
// ********************************************

  for (int dim=0; dim<meshDim; ++dim)
  {
    for (int i=0; i<numpart; ++i)
    {
      for (mPart::iterall ent_it=mesh->getPart(i)->beginall(dim); ent_it!=mesh->getPart(i)->endall(dim); ++ent_it)
      {
        if ((*ent_it)->getNumBPs()==0 || (*ent_it)->getNumBPs()==1 && *((*ent_it)->bpBegin())==0) 
          continue;  // internal entity on part 0
        (*ent_it)->setPClassificationWithBPs();      
      }
    }
  }
  
#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("update PClassification...", cur_time-prev_time, cur_mem-prev_mem);
  prev_time=cur_time;
  prev_mem=cur_mem;
#endif

// ********************************************
// STEP 3: Migrate p-sets
// ********************************************

  mEntityGroup* new_pset;
  map<mEntityGroup*, mEntityGroup*> pset_map;

  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::All_Reduce);
  CM->set_tag(0);

  int msg_size = sizeof(int)+sizeof(mEntityGroup*);
  CM->set_fixed_msg_size(msg_size);
  void* msg_send = CM->alloc_msg(msg_size);
  int* s_int;
  mEntityGroup** s_pset;

  for (int i=0; i<numpart; ++i)
  {
    if (mesh->getPart(i)->numPSet()==0) continue;
    for (mPart::psetIter pset_it=mesh->getPart(i)->beginPSet(); pset_it!=mesh->getPart(i)->endPSet(); ++pset_it)
    {
      if (!((*pset_it)->getAttachedInt(FMDB_Util::Instance()->getDestPart(), &dest_pid))) continue;
      if (FMDB_PartID_CommRank(dest_pid)!=SCUTIL_CommRank())
      {
        s_int =  (int*)((char*)msg_send);
        s_pset = (mEntityGroup**)((char*)msg_send+sizeof(int));
        *s_int = dest_pid;
        *s_pset = (mEntityGroup*)(*pset_it);
        CM->send(dest_pid/numpart, (void*)msg_send);
      }
      else // (FMDB_PartID_CommRank(dest_pid)==SCUTIL_CommRank()
      { 
        new_pset = (mEntityGroup*)(mesh->getPart(FMDB_PartID_LocalID(dest_pid))->createPSet(meshDim));
        pset_map[(mEntityGroup*)(*pset_it)] = new_pset;
      }
    } // for (mPart::iterall ent_it=mesh->getPart(i)->beginall(entDim))
  } // for (int i=0; i<numpart; ++i)

  CM->finalize_send();
  CM->free_msg(msg_send);

  void* msg_recv;
  int pid_from;
  // receive phase
  while(CM->receive(msg_recv, &pid_from))
  {
    s_int = (int*)((char*)msg_recv);
    s_pset = (mEntityGroup**)((char*)msg_recv+sizeof(int));
    new_pset = (mEntityGroup*)(mesh->getPart(FMDB_PartID_LocalID(*s_int))->createPSet(meshDim));
    pset_map[*s_pset] = new_pset;
    CM->free_msg(msg_recv);
  } //  while (CM->receive(msg_recv, &pid_from))

#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("creating p-sets...", cur_time-prev_time, cur_mem-prev_mem);
  prev_time=cur_time;
  prev_mem=cur_mem;
#endif

// ********************************************
// STEP 3: Migrate entities
// ********************************************
  std::vector<mEntity*> entitiesToRemove[4];

  static_exchangeEntities(mesh, meshDim, FMDB_VERTEX, pset_map, entitiesToRemove);
  static_updatePartBdry(mesh, 0);
#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("creating vertices...", cur_time-prev_time, cur_mem-prev_mem);
  prev_time=cur_time;
  prev_mem=cur_mem;
#endif

  static_exchangeEntities(mesh, meshDim, FMDB_EDGE, pset_map, entitiesToRemove);
  static_updatePartBdry(mesh, 1);
#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("creating edges...", cur_time-prev_time, cur_mem-prev_mem);
  prev_time=cur_time;
  prev_mem=cur_mem;
#endif

  static_exchangeEntities(mesh, meshDim, FMDB_FACE, pset_map, entitiesToRemove);
  if (meshDim==3)
    static_updatePartBdry(mesh, 2);
#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("creating faces...", cur_time-prev_time, cur_mem-prev_mem);
  prev_time=cur_time;
  prev_mem=cur_mem;
#endif

  static_exchangeEntities(mesh, meshDim, FMDB_REGION, pset_map, entitiesToRemove);
#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("creating regions...", cur_time-prev_time, cur_mem-prev_mem);
  prev_time=cur_time;
  prev_mem=cur_mem;

  int mid_numPeak=0;
  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
    for (int i=0; i<meshDim+1; ++i)
      mid_numPeak+=(*part_it)->size(i);
#endif

 count=0;
 int source_partid;
  for (vector<mEntityGroup*>::iterator pset_it=psetToRemove.begin(); pset_it!=psetToRemove.end(); ++pset_it)
  {
    (*pset_it)->getAttachedInt(FMDB_Util::Instance()->getSourcePart(), &source_partid);
    mesh->getPart(FMDB_PartID_LocalID(source_partid))->deletePSet(*pset_it, 0);
    ++count;
  }

  mEntity* ent;
  for (int dim=meshDim; dim>=0; --dim)
  {
    for (std::vector<mEntity*>::iterator ent_it=entitiesToRemove[dim].begin(); ent_it!=entitiesToRemove[dim].end(); ++ent_it)
    {
      switch(dim)
      {
        case 0: if (FMDB_Util::Instance()->tagToMigrate.size())
                {
                  int tag_type, tag_size;
                  for (std::list<unsigned int>::iterator tag_it = FMDB_Util::Instance()->tagToMigrate.begin(); 
                       tag_it!=FMDB_Util::Instance()->tagToMigrate.end(); ++tag_it)
                   {
                     tag_type=FMDB_Util::Instance()->typeMeshDataId(*tag_it);
                     tag_size=FMDB_Util::Instance()->typeMeshDataId(*tag_it);
                     switch (tag_type)
                     {
                       case SCUtil_DBL: if (tag_size>1)
                                        {
                                          double* dbl_arr;
                                          EN_getDataPtr(*ent_it, *tag_it, (void**)&dbl_arr);
                                          delete [] dbl_arr;
                                        }
                                        break;
                       case SCUtil_INT: if (tag_size>1)
                                        {
                                          int* int_arr;
                                          EN_getDataPtr(*ent_it, *tag_it, (void**)&int_arr);
                                          delete [] int_arr;
                                        }
                                        break;
                      default: break;
                    }
                    (*ent_it)->deleteData(*tag_it);
                  } // for tag_it
                } // if (FMDB_Util::Instance()->tagToMigrate.size())
                M_removeVertex(mesh->getPart(0), *ent_it); 
                break;
        case 1: M_removeEdge(mesh->getPart(0), *ent_it); break;
        case 2: M_removeFace(mesh->getPart(0), *ent_it); break;
        case 3: M_removeRegion(mesh->getPart(0), *ent_it); break;
      }
    }
  }

#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("remove unnecessary entities and p-sets...", cur_time-prev_time, cur_mem-prev_mem);
  prev_time=cur_time;
  prev_mem=cur_mem;
#endif

// ********************************************
// STEP 6: update partition model
// ********************************************
  pmModel::Instance()->updateOwnership(mesh);

#ifdef DEBUG
  cur_mem = ParUtil::Instance()->getCurMem();
  cur_time = ParUtil::Instance()->getCurTime();
  SCUTIL_DspRsrc("update partition model...", cur_time-prev_time, cur_mem-prev_mem);
  prev_time=cur_time;
  prev_mem=cur_mem;
/*
  int after_numPeak=0;
  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
    for (int i=0; i<meshDim+1; ++i)
      after_numPeak+=(*part_it)->size(i);
*/
  if (!SCUTIL_CommRank())
  {
    cout<<"\nMIGRATION: Time="<<prev_time-begin_time<<" (sec), Mem="<<prev_mem - begin_mem<<" (MB)\n";
//    cout<<"#entity: BEFORE migration="<<before_numPeak<<", DURING migration="<<mid_numPeak<<", AFTER migration="<<after_numPeak<<"\n\n";
  }
#endif
  return SCUtil_SUCCESS;
}
#endif

