/****************************************************************************** 
  
  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "oldFMDB.h"
#include "FMDB_Iterator.h"
#include "iUtil.h"
#include <assert.h>
#include <list>

#include "FMDB_cint.h"
#include "FMDB_Internals.h"
#include "mMesh.h"

using namespace std;

void ITAPS_Util::addIterator(pIterator iter)
{
  allIterators.push_back(iter);
}

void ITAPS_Util::eraseIterator(pIterator& iter)
{
  list<pIterator>::iterator loc=find(allIterators.begin(),
  allIterators.end(),iter);
  if(iter->type==1)                   // entity set
  {
    if(iter->isList==1)
       delete ((pEntSetUIterator) iter->iter); 
    if(iter->isList==0)
       delete ((pEntSetOIterator) iter->iter);
  }
  if(iter->type==2)
    delete ((pMeshIterator) iter->iter); 

  if(iter->type==3){                   // the ptr inside iter is a structure
    delete (part_pid_struct*)((pMeshIterator)(iter->iter))->getPtr(); 
    delete (pMeshIterator) (iter->iter); 
  }
  SAFE_DELETE(iter);
  allIterators.erase(loc++);
}


// type 1 iterator
void createIterator(mEntitySet* eset, pIterator& iter, int type, int topo,
int ws)
{
  iter= new Iterator; 
  iter->type=1; 
  iter->size=ws;

  if(eset->isList())
  {
     iter->isList=1; 
     iter->iter = new mEntSetUIterator(((mEntitySetUnordered*)eset)->begin(), ((mEntitySetUnordered*)eset)->end(),type,topo, (void*)eset, &processingEntitySetUFilter);
  }
  else
  {
     iter->isList=0; 
     iter->iter = new mEntSetOIterator(((mEntitySetOrdered*)eset)->begin(), ((mEntitySetOrdered*)eset)->end() ,type,topo, (void*)eset, &processingEntitySetOFilter );
  }
   ITAPS_Util::Instance()->addIterator((pIterator)iter);
}

// type 2 iterator
void createIterator(mPart* part, pIterator& iter, int type, int topo, int ws)
{
  iter= new Iterator;
  iter->type=2;
  iter->isList=-1;
  iter->size=ws;
  if(type>=4)  // all types
  {
     switch (topo)
     {
      case FMDB_POINT: 
	 iter->iter = new mPartIterator(part->beginall(0), part->endall(0), 0, topo, (void*)part, &processingMeshFilter); 
	 break; 
      case FMDB_LINE:
	 iter->iter = new mPartIterator(part->beginall(1), part->endall(1), 1, topo, (void*)part, &processingMeshFilter);
	 break;
      case FMDB_POLYGON: 
      case FMDB_TRI:
      case FMDB_QUAD:
	 iter->iter = new mPartIterator(part->beginall(2), part->endall(2), 2, topo, (void*)part, &processingMeshFilter);
	 break;
      case FMDB_POLYHEDRON:
      case FMDB_TET:
      case FMDB_HEX:
      case FMDB_PYRAMID: 
      case FMDB_PRISM:
      case FMDB_SEPTA:
	 iter->iter = new mPartIterator(part->beginall(3), part->endall(3), 3, topo, (void*)part, &processingMeshFilter);
	 break;
      default: // all topologies
       {
         iter->iter = new mPartIterator(part->beginall(0), part->endall(part->getDimension()), 4, topo, (void*)part, &processingMeshFilter);
       }
       break;
     }
  }
  else // specific entity type iterator
  {
     iter->iter = new mPartIterator(part->beginall(type), part->endall(type), type, topo, (void*)part, &processingMeshFilter);
  }

  ITAPS_Util::Instance()->addIterator((pIterator)iter);
}

#ifdef FMDB_PARALLEL
void createPartBdryIterator(mPart* part, pIterator& iter, int tgt_pid, int type, 
		    int topo, int ws)
{
  iter= new Iterator;
  iter->type=3;                         // structure other than mesh and entityset
  iter->isList=-1; 
  iter->size=ws;

  part_pid_struct* part_pid = new part_pid_struct; 
  part_pid->part = part; 
  part_pid->pid =tgt_pid; 
  
   if(type == FMDB_ALLTYPE)
   {
     if (topo==FMDB_ALLTOPO)
       iter->iter = new mPartIterator(part->beginall(FMDB_VERTEX), part->endall(part->getDimension()), 
              FMDB_ALLTYPE, FMDB_ALLTOPO, (void*)part_pid, &processingPartBdryEntFilter);
     else // specific topo
       iter->iter = new mPartIterator(part->beginall(FMDB_Topo_GetType(topo)), part->endall(FMDB_Topo_GetType(topo)),
              FMDB_Topo_GetType(topo), topo, (void*)part_pid, &processingPartBdryEntFilter);
   }
   else
     iter->iter = new mPartIterator(part->beginall(type), part->endall(type), type, topo,
                             (void*)part_pid, &processingPartBdryEntFilter);

  ITAPS_Util::Instance()->addIterator((pIterator)iter);
}
#endif

void deleteIterator(pIterator& iter)
{
   ITAPS_Util::Instance()->eraseIterator((pIterator&)iter);
}

bool iterEnd(Iterator* iter)
{
   if(iter->type==1)
   {
      if(iter->isList)
      {
	 ((pEntSetUIterator)iter->iter)->valid_check();
        return ((pEntSetUIterator)iter->iter)->end(); 
      }
      else
	return ((pEntSetOIterator)iter->iter)->end();
   }
   
   if(iter->type==2 || iter->type==3)
   {
      return ((pMeshIterator)iter->iter)->end();
   }
   std::cerr << "impossible iterator type " << iter->type << '\n';
   return NULL;
}

pEntity iterData(Iterator* iter)
{
   if(iter->type==1)
   { 
      if(iter->isList)
	 return *(*((pEntSetUIterator)iter->iter)); 
      else
	 return *(*((pEntSetOIterator)iter->iter));
   }
   
   if(iter->type==2 || iter->type==3)
   {
      return  *(*((pMeshIterator)iter->iter)); 
   }
   std::cerr << "impossible iterator type " << iter->type << '\n';
   return NULL;
}

void iterNext(Iterator* iter)
{
   if(iter->type==1)
   {
       if(iter->isList)
         ((pEntSetUIterator)iter->iter)->next();
       else
	 ((pEntSetOIterator)iter->iter)->next();
  }	 
   else if(iter->type==2 || iter->type==3)
      ((pMeshIterator)iter->iter)->next();    
}

void iterReset(Iterator* iter)
{
   if(iter->type==1)
   {
      if(iter->isList)
         ((pEntSetUIterator)iter->iter)->reset();
      else
	 ((pEntSetOIterator)iter->iter)->reset();
   }
   else if(iter->type==2 || iter->type==3)
      ((pMeshIterator)iter->iter)->reset();
}

int wrkSize(Iterator* iter)
{
  return ((pIterator)iter)->size;
}

