/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include <string.h>
#include "mFMDB.h"
#include "mMeshIO.h"
#include "mPart.h"
#include "mVertex.h"
#include "mEdge.h"
#include "mFace.h"
#include "mTet.h"
#include "mException.h"
#include "mMesh.h"
#include "ParUtil.h"
#include "mMeshIO.h"
#include "modeler.h"

#include <utility>
#include <iostream>
#include<iomanip>
#include <fstream>
#include <stdio.h>

using std::vector;
using std::cout;
using std::ofstream;
using std::ostream;
using std::istream;
using std::endl;
using namespace std;


bool isEqual(double x, double y)
{  
  if (fabs(x)<accuracy && fabs(y)<accuracy) return true;
  if (fabs((x-y)/(0.5*(x+y))) < accuracy) return true;
  return false;
}

  std::auto_ptr<FMDB_Util> FMDB_Util::instance;

  FMDB_Util::FMDB_Util()
  {
#ifdef FMDB_PARALLEL
    procGrpSize = ParUtil::Instance()->size();
#else
    procGrpSize = 1;
#endif
    _ATTPARAMETRIC = lookupMeshDataId ("_parametric");
    _ATTSOURCEPART = lookupMeshDataId ("_sourcepart");
    _ATTDESTPART = lookupMeshDataId   ("_destpart");
    _ATTFMOD = lookupMeshDataId   ("_fmod");
    _ATTEMOD = lookupMeshDataId   ("_emod");
    _ATTID   = lookupMeshDataId   ("_id");
    _ATTDN   = lookupMeshDataId   ("_dn");
    _ATTSIZE = lookupMeshDataId   ("_size");
    _ATTPSET = lookupMeshDataId   ("_pset");
    _ATTWEIGHT = lookupMeshDataId   ("_weight");
    _ATTNODE = lookupMeshDataId   ("_node");
    _ATTCHECKMARK = lookupMeshDataId   ("_checkmark");
  }

  FMDB_Util::~FMDB_Util()
  {
    deleteMeshDataId(_ATTPARAMETRIC);
    deleteMeshDataId(_ATTSOURCEPART);
    deleteMeshDataId(_ATTDESTPART);
    deleteMeshDataId(_ATTFMOD);
    deleteMeshDataId(_ATTEMOD);
    deleteMeshDataId(_ATTID);
    deleteMeshDataId(_ATTDN);
    deleteMeshDataId(_ATTSIZE);
    deleteMeshDataId(_ATTPSET);
    deleteMeshDataId(_ATTWEIGHT);
    deleteMeshDataId(_ATTNODE);
    deleteMeshDataId(_ATTCHECKMARK);
  }

  FMDB_Util* FMDB_Util::Instance()
  {
    if(!instance.get())
      {
	instance.reset(new FMDB_Util);
      }
    return instance.get();
  }

  int FMDB_Util::import_mesh (mMesh* mesh, const char *fileName)
  {
    char ext[6];
    int err = SCUtil_SUCCESS;
    strcpy(ext,fileName+(strlen(fileName)-3));
    pPart part;
    FMDB_Mesh_GetPart (mesh, 0, part);

    if(!strcmp(ext,"cdf"))
      importNCDF(part, fileName);
#ifdef NCDF64
    else if(!strcmp(ext, ".nc"))
      importNCDFBin(part, fileName, 1);
#endif
    else if(!strcmp(ext,"vtk"))
      err = importVTK(mesh, fileName);
    else
      err = importSMS(mesh, fileName);
    if(err != SCUtil_SUCCESS)
      return err;
    int local_dim = (mesh->getPart(0)->size(3)? 3: 2), global_dim=0;
#ifdef FMDB_PARALLEL
    MPI_Allreduce(&local_dim, &global_dim,1,MPI_INT,MPI_MAX,ParUtil::Instance()->getComm());
#else
    global_dim=local_dim;
#endif
    ParUtil::Instance()->set_maxMeshDim(global_dim);
   return SCUtil_SUCCESS;
  }
  
  void FMDB_Util::export_mesh (mMesh *mesh, const char *fName)
  {
    char ext[6];
    char text[256];
    strcpy(ext,fName+(strlen(fName)-4));

    if(!strcmp(ext,".sms"))
      exportSMS(mesh, fName);      
    else if(!strcmp(ext,".vtk"))
      exportMeshToVTK(mesh, fName);
    else
    {
      sprintf(text,"unknown extension %s in file %s",ext,fName);
      throw mException(__LINE__,__FILE__,text);
    }
  }
