/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include <iostream>
#include "pmGraph.h"
#include "mPart.h"
#include "mEntity.h"
#include "FMDB_LoadBalancer.h"            
#include "ParUtil.h"
#include "mFMDB.h"

#include <stdio.h>
#include <algorithm>
#include <map>

using namespace std;

  /*
    Graph based on a certain level of refinement of the octree
    this is certaily better than the other way because the unbalance
    was much too large due to the strategy of keeping all subtrees
    on the same processor. Note that we will have to take into account 
    edge weights as well as vertex weights.
  */

 
  FMDB_graph::FMDB_graph (mPart *m, int from, int to , bool w, int level)
  {
    int n = from;
    int nbTot = 0;    
    int numnod = 0;
  
    unsigned int tagId = FMDB_Util::Instance()->getId();
    unsigned int tagDn = FMDB_Util::Instance()->getDn();
    unsigned int tagAtt1 = FMDB_Util::Instance()->getCheckMark();    
 
    { // for win32 compiler
      for(mPart::iterall it = m->beginall(n) ; it != m->endall(n) ; ++it)
	{
	  mEntity *e = *it;
	  // we only take this level
	  if(m->getRefinementLevel(e) == 0)
	    {	    
	      //	    nbTot++;
	      numnod++;
	      // access entities of dimension to (i.e. graph edges)
	      for(int i=0;i<e->size(to);i++)
		{
		  nbTot ++;
		}
	    }
	}
    }

    //  printf("sss proc %d %d %d\n",ParUtil::Instance()->rank(),numnod,nbTot);
    nn = numnod;

    /* allocation of the graph*/
    xadj   = new int[numnod+2];
    adjncy = new int[nbTot+1];   
    adjpid = new int[nbTot+1];      
    adjwgt = new int[nbTot+1];    

    /* temp vector */
    int icount = 0;  
    int ecount = 0;  
    { 
      for(mPart::iterall it = m->beginall(n) ; it!=m->endall(n) ; ++it)
	{
	  mEntity *e = *it;
	  if(m->getRefinementLevel(e) == 0)
	    {	    
	      std::set<int> adj;
	      std::map<int,int> wgt;
	      std::map<int,int> pid;             // object id and process id
	      int id1 =  e->getAttachedInt(tagId);
	      for(int i=0;i<e->size(to);i++)
		{
		  mEntity *down = e->get(to,i);
		  list<mEntity*> leaves;
		  down->getLeaves(leaves);		

		  for(list<mEntity*>::const_iterator itl = leaves.begin();itl != leaves.end();++itl)
		    {
		      mEntity *leaf = *itl;
		      if(leaf->getData(tagDn))
			{
			  int idproc = leaf->getAttachedInt(tagDn);
			  int procId = leaf->getAttachedInt(tagAtt1);    
			  adj.insert(idproc);
			  if(procId!=-1)                  // we use -1 to represent 0.  
			    pid[idproc]=procId;
			  else
			    pid[idproc]=0; 
			  wgt[idproc] ++; 
			}
		      for(mAdjacencyContainer::iter it2 = leaf->begin(n) ; it2!=leaf->end(n) ; ++it2)
			{
			  mEntity *up = *it2;
			  int procId =  up->getOwner();                
			  int id2 =  up->getAttachedInt(tagId) ;
			  if(id1 != id2)
			    {
			      adj.insert( id2 );
			      pid[id2]=procId;
			      wgt[id2] ++;                     
			    }
			}
		    }
		}
	      xadj[ecount++] = icount+1;
	      for(std::set<int>::iterator its = adj.begin() ; its != adj.end() ; ++its) 
		{
		  adjncy[icount] = *its; 
		  adjpid[icount] = pid[*its];
		  adjwgt[icount] = wgt[*its];
		  icount ++;
		}
	    }
	}
      xadj[ecount] = icount+1;
    }

    //  printf("proc %d nn %d icount %d ecount %d nbTot %d %d\n",ParUtil::Instance()->rank(),nn,icount,ecount,nbTot,numnod);

    vwgt=0;
  }



  /** ------------------------------------------------------------------------
     This constructor is the one prensently used.     
   ---------------------------------------------------------------------------*/

  FMDB_graph::FMDB_graph (mPart *m, int from, int to , bool w)
  {
    int n = from;

    unsigned int tagId = FMDB_Util::Instance()->getId();
    unsigned int tagDn = FMDB_Util::Instance()->getDn();

    int nbTot = 0;
    int numnod = 0;
    { // for win32 compiler

      // iterator on mesh entities of dimension n (i.e. graph vertices)
      for(mPart::iter it = m->begin(n) ; it!=m->end(n) ; ++it)
	{
	  // access what is in the iterator
	  mEntity *e = *it;
	  numnod++;
	  // access entities of dimension to (i.e. graph edges)
	  for(int i=0;i<e->size(to);i++)
	    {
	      // down is ith the entity
	      mEntity *down = e->get(to,i);
	      list<mEntity*> leaves;
	      // it can be an octree-like structure, we get all the leaves !!!
	      down->getLeaves(leaves);
	      for(list<mEntity*>::const_iterator itl = leaves.begin();itl != leaves.end();++itl)
		{
		  mEntity *leaf = *itl;
		  if(leaf->getData(tagDn))nbTot++;
		  for(mAdjacencyContainer::iter it2 = leaf->begin(n) ; it2!=leaf->end(n) ; ++it2)
		    {
		      mEntity *up = *it2;
		      if(up != e)
			{
			  nbTot ++;
			}
		    }
		}
	    }
	}
    }
    
    nn = numnod;
    
    xadj   = new int[numnod+2];
    adjncy = new int[nbTot+1];    
    adjwgt = new int[nbTot+1];    
    
    int icount = 0;
    int ecount = 0;
    
    { 
      for(mPart::iter it = m->begin(n) ; it!=m->end(n) ; ++it)
	{
	  mEntity *e = *it;
	  std::set<int> adj;
	  std::map<int,int> wgt;
	  int id1 =  e->getAttachedInt(tagId);
	  //      printf("node %d ",e->getAttachedInt("id"));e->print();
	  for(int i=0;i<e->size(to);i++)
	    {
	      mEntity *down = e->get(to,i);
	      list<mEntity*> leaves;
	      down->getLeaves(leaves);

	      for(list<mEntity*>::const_iterator itl = leaves.begin();itl != leaves.end();++itl)
		{
		  // for parallel purpose, we add the id of distributed  neighbor here
		  mEntity *leaf = *itl;

		  if(leaf->getData(tagDn))
		    {
		      int idproc = leaf->getAttachedInt(tagDn);
		      adj.insert(idproc);
		      wgt[idproc] = 1;
		      //		      printf("inter proc found %d %d \n",idproc,id1);
		    }

		  for(mAdjacencyContainer::iter it2 = leaf->begin(n) ; it2!=leaf->end(n) ; ++it2)
		    {
		      mEntity *up = *it2;
		      int id2 =  up->getAttachedInt(tagId) ;
		      if(id1 != id2)
			{
			  adj.insert( id2 );
			  if(up->root() == e->root())
			    wgt[id2]=10000;
			  else
			    wgt[id2]=1;
			}
		    }
		}
	    }
	  xadj[ecount++] = icount+1;      
	  for(std::set<int>::iterator its = adj.begin() ; its != adj.end() ; ++its) 
	    {
	      adjncy[icount] = *its;
	      adjwgt[icount] = wgt[*its];
	      icount ++;
	    }
	}
    }
    xadj[ecount] = icount+1;

    //  printf("proc %d nn %d icount %d ecount %d nbTot %d %d\n",ParUtil::Instance()->rank(),nn,icount,ecount,nbTot,numnod);

    //  print(std::cout);
    vwgt=0;
  }

  FMDB_graph :: ~FMDB_graph()
  {
    if(vid)                  // add if clauses, 0812, xie
      delete [] vid;
    if(adjpid)
      delete [] adjpid; 
    if(xadj) 
      delete [] xadj;
    if(adjncy)
      delete [] adjncy;
    if(vwgt)delete [] vwgt;
    if(adjwgt)delete [] adjwgt;
  }

  void FMDB_graph::print( std::ostream & o)       // Dont print vid, adjpid.             
  {
    o << "graph with " << nn << " nodes\n";

    o << "xadj ";
    for(int i=0;i<nn+1;i++)o << xadj[i] << " ";
    o << "\n";
    o << "adjncy ";
    for(int j=0; j<xadj[nn]-1;j++)
      {
	o << adjncy[j] << " ";
      }
    o << "\n";
    if(vwgt)
      {
	o << "vwgt ";
	for(int j=0; j<nn;j++)
	  {
	    o << vwgt[j] << " ";
	  }
	o << "\n";
      }
    if(adjwgt)
      {
	o << "adjwgt ";
	for(int j=0; j<xadj[nn]-1;j++)
	  {
	    o << adjwgt[j] << " ";
	  }
	o << "\n";
      }
  }


  FMDB_distributed_graph :: FMDB_distributed_graph(mPart *m, int from, int to,
						   FMDB_LoadBalancerCallbacks *dm,
						   int level)  
  {
    /** Hypothesis is that nodes of the graph are mesh
	entities of dimension from and edges are of dimension
	to */
    int n = from;
    vtxdist = new int [ParUtil::Instance()->size()+1];
    int *nbNodesPerProc = new int [ParUtil::Instance()->size()];

    unsigned int tagId = FMDB_Util::Instance()->getId();
    unsigned int tagDn = FMDB_Util::Instance()->getDn();

    //printf("computing distributed graph (p%d)\n",ParUtil::Instance()->rank());

    /// We compute how many nodes of the graph per processor
    ParUtil::Instance()->Barrier(__LINE__,__FILE__);
    computeNbNodesPerProc(m,nbNodesPerProc,from,level);

    /// We then compute a vector that tells the range
    /// of nodes on processor i : 
    /// range of proc i is [vtxdist[i] ... vtxdist[i+1]]
    int init = 1;
    vtxdist[0] = 1;
    int total = 1;
    for (int i=0;i<ParUtil::Instance()->size();i++)
      {
	total += nbNodesPerProc[i];
	vtxdist[i+1] = total;      
      }
    
    delete [] nbNodesPerProc;
  

    /// Print vtxdist on the stdout
    /*if(ParUtil::Instance()->master())
      {
	ParUtil::Instance()->Msg(ParUtil::WARNING,"vtxdist(%d) = (",ParUtil::Instance()->rank());
	for (int i=0;i<=ParUtil::Instance()->size();i++)
	  {
	    if(i)
	     ParUtil::Instance()->Msg(ParUtil::WARNING,",%d",vtxdist[i]);
	    else
	      ParUtil::Instance()->Msg(ParUtil::WARNING,"%d",vtxdist[i]);
	  }
	ParUtil::Instance()->Msg(ParUtil::WARNING,")\n");	   
      }
    */
    /// We will tag all nodes with id's in this range
    init =vtxdist[ParUtil::Instance()->rank()];  
    /**
       if level < 0 then the graph is composed of all leaves
       of the mesh.

       else the graph is composed of elements of refinement
       level level. level=0 means all roots !!
    */

    if(level < 0)
      {
	for(mPart::iter it = m->begin(n) ; it!=m->end(n) ; ++it)
	  {	  
	    (*it)->attachInt(tagId,init++);
	    if ((*it)->getAttachedInt(tagId) != init-1)throw;
	  }
      }
    else
      {
	for(mPart::iterall it = m->beginall(n) ; it!=m->endall(n) ; ++it)
	  {
	    mEntity *e = *it;
	    // we only take this level	  
	    if(m->getRefinementLevel(e) == level)
	      {	    
		int id = init++;
		e->attachInt(tagId,id);
		list<mEntity*> leaves;
		e->getLeaves(leaves);		
		for(list<mEntity*>::const_iterator itl = leaves.begin();itl != leaves.end();++itl)
		  {
		    mEntity *leaf = *itl;
		    leaf->attachInt(tagId,id);
		  }
	      }
	  }
      }

/*
    /// add neighbors through inter processor boundaries  
#ifdef PARALLEL
    Graph_DataExchanger myCallback(from,to);
    m->exchangeDataOnPartBdrys (myCallback);
#endif
*/
    
    // create the serial graph with inter processor links !!!
    if(level < 0)
      theGraph = new FMDB_graph (m,from,to,false);    
    else
      theGraph = new FMDB_graph (m,from,to,false,level);    
  


    /// add vertex weights to the graph
    {
      int initial_id = vtxdist[ParUtil::Instance()->rank()];
      if(theGraph->vwgt)delete [] theGraph->vwgt; 
      theGraph->vwgt = new int [theGraph->nn];
      for(int i=0;i<theGraph->nn;i++)theGraph->vwgt[i] = 1;
      for(mPart::iter it = m->begin(n) ; it!=m->end(n) ; ++it)
	{
	  int id = (*it)->getAttachedInt(tagId) - initial_id;	
	  if(dm->useWeights())
	    theGraph->vwgt[id] *= (int) (dm->getWeight((*it)));
	}
      int sumw = 0;
      {      for(int i=0;i<theGraph->nn;i++)sumw += theGraph->vwgt[i];}
      //      ParUtil::Instance()->Msg(ParUtil::WARNING,"prw = %d\n",sumw);
    }
    


    //printf(" (%d) mGraph done...\n",ParUtil::Instance()->rank());
    for(mPart::iter it = m->begin(n-1) ; it!=m->end(n-1) ; ++it)(*it)->deleteData(tagDn);
  }


  void FMDB_distributed_graph :: print(ostream &o)
  {
    printf("on processor %d :\n",ParUtil::Instance()->rank());
    theGraph->print(o);
    o << "vtxdist : ";
    for(int i=0;i<=ParUtil::Instance()->size();i++)o << vtxdist[i] << " ";
    o << "\n";
  }



  FMDB_distributed_graph :: ~FMDB_distributed_graph()
  {
    delete theGraph;
    delete [] vtxdist;
  }

  void computeNbNodesPerProc(mPart *m, int *nbNodesPerProc, int n, int level)
  {
    int *sendcounts = new int[ParUtil::Instance()->size()];
    for(int i=0;i<ParUtil::Instance()->size();i++)sendcounts[i]=0;
  
    int numnod = 0;
    if(level < 0)
      {
	for(mPart::iter it = m->begin(n) ; it!=m->end(n) ; ++it)
	  {
	    numnod ++;      	
	  }
      }
    else
      {
	for(mPart::iterall it = m->beginall(n) ; it!=m->endall(n) ; ++it)
	  {
	    if(m->getRefinementLevel(*it) == 0)
	      numnod ++;      	
	  }
      }
#ifndef FMDB_PARALLEL
    nbNodesPerProc[0] = numnod;
#endif
  }
