/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/

#include "FMDB_Iterator.h"

void SingleTypeFilter(part_iter& it_begin, part_iter& it_end)
{
  return;
}

void SingleTypeTopoFilter(part_iter& it_begin, part_iter& it_end, int topo)
{
  for (part_iter it= it_begin; it!=it_end;++it)
    if ((*it)->getTopo()==topo)
    { it_begin = it; return ; }
  it_begin = it_end;
}

void AllTypeFilter(part_iter& it_begin, part_iter& it_end, pPart part)
{
  int part_dim = FMDB_FACE;
  if (part->size(FMDB_REGION)) part_dim=FMDB_REGION;

  part_iter it=it_begin;
  if (it==part->endall(0)) 
    it_begin=part->beginall(1);
  else if (it==part->endall(1)) 
    it_begin=part->beginall(2);
  else if (it==part->endall(2) && part_dim==3) 
    it_begin=part->beginall(3);
  else if (it==part->endall(3) || (it==part->endall(2) && part_dim==2)) 
    it_begin=it_end; 

  return;
}

#ifdef FMDB_PARALLEL
void PartBdrySingleTypeFilter(part_iter& it_begin, part_iter& it_end, int target_pid)
{
  for (part_iter it = it_begin; it!=it_end;++it) 
  {
    if ((*it)->getPClassification()) 
    {
      if (target_pid != -1 ) // specific target part
      {
	for (mEntity::RCIter rcIter=(*it)->rcBegin(); rcIter!=(*it)->rcEnd();++rcIter)
	{
	  if ( (*rcIter).first== target_pid ) 
	  {  it_begin = it;  return;  }
	}
      }	
      else
      {  it_begin = it;  return;  }
    } 
  }
  it_begin = it_end;
}

void PartBdrySingleTypeTopoFilter(part_iter& it_begin, part_iter& it_end, int target_pid, int topo)
{
  for (part_iter it= it_begin; it!=it_end;++it)
  {
    if ((*it)->getTopo()!=topo || !((*it)->getPClassification()))
      continue;
    if (target_pid != -1 )   //specific target part
    {
      for (mEntity::RCIter rcIter=(*it)->rcBegin(); rcIter!=(*it)->rcEnd();++rcIter)
      {
	if ( (*rcIter).first == target_pid ) 
	{  it_begin = it;  return;  }
      } // for (mEntity::RCIter 
    } 
    else // if ( pid == -1 ) 
    {  it_begin = it;  return;  }
  } // for (it= it_begin; it!=it_end;++it)
  it_begin = it_end;
}

void PartBdryAllTypeFilter(part_iter& it_begin, part_iter& it_end, pPart part, int target_pid)
{
  int cur_dim, part_dim = FMDB_FACE;
  if (part->size(FMDB_REGION)) part_dim=FMDB_REGION;

  part_iter it=it_begin;
  if (it==part->endall(0)) it=part->beginall(1);
  else if (it==part->endall(1) && part_dim==3) it=part->beginall(2);
  else if (it==part->endall(2) || (it==part->endall(1) && part_dim==2)) { it_begin=it_end; return; }

  cur_dim = (*it)->getLevel(); 
  while(1)
  {
    for(; it!=part->endall(cur_dim); ++it)
    {	
      if (!((*it)->getPClassification())) continue;
      if ( target_pid != -1 )  // specific target part id
      {
        for (mEntity::RCIter rcIter=(*it)->rcBegin(); rcIter!=(*it)->rcEnd();++rcIter)
        {
	  if ( (*rcIter).first== target_pid ) 
	  { it_begin = it; return; }
	}
      }
      else
      {  it_begin = it;  return;  }
    } // failed to find part bdry ent in cur_dim
    ++cur_dim; 
    if (cur_dim==part_dim) break;
    it = part->beginall(cur_dim); // replace it with beginall(next dim) and continue while loop
  } // while
  it_begin=it_end;
}
#endif

void RevClasSingleTypeFilter(part_iter& it_begin, part_iter& it_end, pGeomEnt gent)
{
  for (part_iter it = it_begin; it!=it_end;++it)
    if ((*it)->getClassification()==gent)
    {  it_begin = it;  return;  }
  it_begin = it_end;
}

void RevClasAllTypeFilter(part_iter& it_begin, part_iter& it_end, pPart part, pGeomEnt gent)
{
  int cur_dim, part_dim = FMDB_FACE;
  if (part->size(FMDB_REGION)) part_dim=FMDB_REGION;

  part_iter it=it_begin;  
  if (it==part->endall(0))
    it=part->beginall(1);
  else if (it==part->endall(1))
    it=part->beginall(2);
  else if (it==part->endall(2) && part_dim==3)
    it=part->beginall(3);
  else if (it==part->endall(3) || (it==part->endall(2) && part_dim==2))
  { it_begin=it_end; return; }

  cur_dim = (*it)->getLevel();
  while(1)
  {
    for(; it!=part->endall(cur_dim); ++it)
    {
      if ((*it)->getClassification()==gent)
      {  it_begin = it;  return;  }
    } // failed to find part bdry ent in cur_dim
    ++cur_dim;
    if (cur_dim==part_dim+1) break;
    it = part->beginall(cur_dim); // replace it with beginall(next dim) and continue while loop
  } // while
  it_begin=it_end;
}

