/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include <iostream>
#include <stdio.h>
#include <string.h>

#include "mPart.h"
#include "mVertex.h"
#include "mEdge.h"
#include "mFace.h"
#include "mTet.h"
#include "mHex.h"
#include "mPrism.h"
#include "mPyramid.h"
#include "mPoint.h"
#include "mException.h"
#include "ParUtil.h"
#include "mBuildAdj.h"
#include "mFMDB.h"
#include "pmModel.h"
#include "FMDB.h"
#include "oldFMDB.h"
#include "FMDBInternals.h"

// added by Jie and Eunyoung, 09/2003
#include "NullModel.h"
#include "NullVertex.h"
#include "NullEdge.h"
#include "NullFace.h"
#include "NullRegion.h"
#include "modeler.h"
#include "iUtil.h"

using std::list;
using std::set;
using std::istream;
using std::ostream;
using std::cout;
using std::endl;
using std::find; 

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
  //////////////////////////////////////////////////////////////////////

  mPart::mPart(int id, pGModel model, mMesh* mesh)
    : mesh_instance(mesh), theSolidModel (model),
      iD(id), 
      dimension(0),
      representationFlag(0),
      classif_fp(GM_entityByTag)
  {
  for(int i = 0; i < 12; i++)
    numTopo[i]=0;
  }

  mPart::~mPart()
  {
    clearPSet(); 
    for(int i=3;i>=0;i--)
      {
	iterall it  = beginall(i);
	iterall ite = endall(i);
	for(;it != ite;++it)
	  {
	    delete *it;
	  }
      }
    if (theSolidModel && !strcmp ((char*)(theSolidModel->modeler ()), "null"))
      delete static_cast<NullModel*>(theSolidModel);
  }
  
  int mPart::size(int what) const
  { return allEntities.size(what);}

  mPart::iterall mPart::beginall(int what) const
  {
    return allEntities.begin(what);
  }

  mPart::iterall mPart::endall(int what) const
  {
    return allEntities.end(what);
  }

  mPart::iter mPart::begin(int what) const
  {
    return mLeavesIterator(beginall(what),endall(what));
  }

  mPart::iter mPart::end(int what) const
  {
    return mLeavesIterator(endall(what),endall(what));
  }

  mClassIterator mPart::begin(int what, int which, int onwhat) const
  {
    return mClassIterator(beginall(what),endall(what),which, onwhat);
  }

  mClassIterator mPart::end(int what, int which, int onwhat) const
  {
    return mClassIterator(endall(what),endall(what),which, onwhat);
  }

  mVertex *mPart::getVertex (int theId)
  {
  //  mEntity* ent;
//     for (mPart::iterall it=beginall(0); it!=endall(0); ++it) {
//       ent = *it;
//       if(ent->getId()==theId)
//         return (mVertex*)ent;
//     }                    // not efficient 

    static mVertex v(0,SCOREC::Util::mPoint(0,0,0),0);
    v.setId(theId);
    return (mVertex*)find(&v);
  }

  mEdge *mPart::getEdge (mVertex *v1, mVertex *v2)
  {
    static mEdge e(0,0,0);
    e.setVertices(v1,v2);
    return (mEdge*)find(&e);
  }
               
  mFace *mPart::getTri (mVertex *v1, mVertex *v2, mVertex *v3)
  {
    mFace f(v1,v2,v3,0);
    return (mFace*)find(&f);
  }


  mFace *mPart::getQuad (mVertex *v1, mVertex *v2, mVertex *v3, mVertex *v4)
  {
    mFace f(v1,v2,v3,v4,0);
    return (mFace*)find(&f);
  }


  mTet *mPart::getTet (mVertex *v1, mVertex *v2, mVertex *v3, mVertex *v4)
  {
    mTet t(v1,v2,v3,v4,0);
    return (mTet*)find(&t);
  }


  mHex *mPart::getHex (mVertex *v1, mVertex *v2, mVertex *v3, mVertex *v4,
		       mVertex *v5, mVertex *v6, mVertex *v7, mVertex *v8)
  {
    mHex h(v1,v2,v3,v4,v5,v6,v7,v8,0);
    return (mHex*)find(&h);
  }

  mVertex* mPart::createVertex (int theId, const double x, const double y, const double z, 
  			pGEntity classif)
  {
   mVertex* theNewVertex = new mVertex(theId,SCOREC::Util::mPoint(x,y,z),classif);
   add(theNewVertex);
    incTopo(FMDB_POINT);
   theIdGenerator.setMaxValue(theId);
   return theNewVertex;	  
  }

  mVertex* mPart::createVertex_noUpdateId(int theId, const double x, const double y, const double z, 
  			pGEntity classif)
  {
    mVertex* theNewVertex = new mVertex(theId,SCOREC::Util::mPoint(x,y,z),classif);
    add(theNewVertex);
    incTopo(FMDB_POINT);
    return theNewVertex;
  }


  // create entity-group
  EntitySet<mEntity>* mPart :: createPSet(int dim)
  {
    mEntityGroup* eg = new mEntityGroup(dim); 
    allPSet.push_back(eg);
    return eg; 
  }

  void  mPart :: deletePSet(EntitySet<mEntity>* PSet, int force) 
  {
    psetIter egIter = std::find(allPSet.begin(), allPSet.end(), PSet);
    if(egIter!=allPSet.end())
       allPSet.erase(egIter); 
    
    if(force==1) {               
    // delete attached data on interior entities 
      for(mEntityGroup::iter entIt=((mEntityGroup*)PSet)->begin(); entIt!=((mEntityGroup*)PSet)->end(); ++entIt) {
        mEntity* ent = *entIt;
        ent->deleteData(FMDB_Util::Instance()->getPSet());
      }
    }
    
    delete (mEntityGroup*)PSet;
  }

  void mPart::clearPSet ()
  {
     EntitySet<mEntity>* PSet; 
     mEntity* ent; 
     if(allPSet.size()==0) 
      return; 

     psetIter egIter = allPSet.begin(); 
     while (egIter != allPSet.end() )
     {
       PSet = *egIter; 
       for(mEntityGroup::iter entIt=((mEntityGroup*)PSet)->begin(); entIt!=((mEntityGroup*)PSet)->end(); ++entIt) {
         ent = *entIt;
	 ent->deleteData(FMDB_Util::Instance()->getPSet());
       }
       delete (mEntityGroup*)PSet; 
       ++egIter; 
     }
     allPSet.clear(); 
     MD_deleteMeshDataId(FMDB_Util::Instance()->getPSet()); 
  }

  int mPart :: getNewVertexId()
  {
    return theIdGenerator.generateId();
  }

  void mPart :: setNewVertexId(int theId)
  {
    theIdGenerator.setMaxValue(theId);
  }

#ifdef FMDB_PARALLEL
  void mPart :: pmSetNewVertexId(int theId)
  {
    theIdGenerator.pmSetMaxValue(theId);
  }
#endif

  mVertex *mPart::createVertex (const double x, const double y, const double z, 
  			pGEntity classif)
  {
    return createVertex(getNewVertexId(),x,y,z,classif);
  }

  mEdge *mPart::createEdge (int i1, int i2, pGEntity classif)
  {
    mVertex *v1 = getVertex(i1);
    mVertex *v2 = getVertex(i2);
    if(!v1 || !v2)
      {
	char text[256];
	sprintf(text,"unknown verex id's %d(%p) %d(%p) in new edge creation\n",i1,v1,i2,v2);
	throw mException (__LINE__,__FILE__,text);
      }
    return createEdge(v1,v2,classif);
  }

  void mPart :: DEL (mEntity *e)
  {
    decTopo(e->getTopo());
    del(e);
    delete e;
  } 

  mEdge *mPart::createEdge (mVertex *v1, mVertex *v2, pGEntity classif)
  {
    mEdge *theNewEdge = new mEdge(v1,v2,classif);
    add(theNewEdge);
    incTopo(FMDB_LINE);
    return theNewEdge;
  }


mFace* mPart::createFaceWithVertices (int i1, int i2, int i3, pGEntity classif)
  {
    mVertex *v1 = getVertex(i1);
    mVertex *v2 = getVertex(i2);
    mVertex *v3 = getVertex(i3);
    if(!v1 || !v2 || !v3)
      {
	throw mException (__LINE__,__FILE__,"unknown verex id's in new face creation");
      }
    return createFaceWithVertices(v1,v2,v3,classif);
  }

  mFace *mPart::createFaceWithVertices (mVertex *v1, mVertex *v2, mVertex *v3, 
  			pGEntity classif)
  {
    mFace *theNewFace = new mFace(v1,v2,v3,classif);
    add(theNewFace);
    incTopo(FMDB_TRI);
    return theNewFace;
  }

  mFace *mPart::createFaceWithVertices (int i1, int i2, int i3, int i4, 
  			pGEntity classif)
  {
    mVertex *v1 = getVertex(i1);
    mVertex *v2 = getVertex(i2);
    mVertex *v3 = getVertex(i3);
    mVertex *v4 = getVertex(i4);
    if(!v1 || !v2 || !v3 || !v4)
      {
	char text[256];
	sprintf(text,"trying to create a quad with vertices %d %d %d %d, search gives %p %p %p %p\n", 
		i1,i2,i3,i4,v1,v2,v3,v4);
	throw mException (__LINE__,__FILE__,text);
      }

    return createFaceWithVertices(v1,v2,v3,v4,classif);
  }

  mFace *mPart::createFaceWithVertices (mVertex *v1, mVertex *v2, mVertex *v3, 
  			mVertex *v4, pGEntity classif)
  {
    if(!v1 || !v2 || !v3 || !v4)
      {
	char text[256];
	sprintf(text,"trying to create a quad with %p %p %p %p\n",
		v1,v2,v3,v4);
	throw mException (__LINE__,__FILE__,text);
      }
    

    mFace *theNewFace = new mFace(v1,v2,v3,v4,classif);
    add(theNewFace);
    incTopo(FMDB_QUAD);
    return theNewFace;
  }

  mTet *mPart::createTetWithVertices (mVertex*v1, mVertex*v2, mVertex*v3, mVertex*v4,
				      pGEntity classif)
  {
    mTet *theNewTet;
    theNewTet = new mTet(v1,v2,v3,v4,classif);
    add(theNewTet);
    incTopo(FMDB_TET);
    return theNewTet;
  }

  mTet *mPart::createTetWithVertices (int i1, int i2, int i3, int i4, pGEntity classif)
  {
    mVertex *v1 = getVertex(i1);
    mVertex *v2 = getVertex(i2);
    mVertex *v3 = getVertex(i3);
    mVertex *v4 = getVertex(i4);
    if(!v1 || !v2 || !v3 || !v4)
      {
	throw mException (__LINE__,__FILE__,"unknown verex id's in new face creation");
      }
    return createTetWithVertices(v1,v2,v3,v4,classif);
  }


  mTet *mPart::createTetWithFaces (mFace *f1, mFace *f2, mFace *f3, mFace *f4, 
  			pGEntity classif)
  {
    mTet *theNewTet;
    theNewTet = new mTet(f1,f2,f3,f4,classif);
    add(theNewTet);
    incTopo(FMDB_TET);
    return theNewTet;
  }

  mFace *mPart::createFaceWithEdges (mEdge *e1, mEdge *e2, mEdge *e3, mEdge *e4, 
  			pGEntity classif, int *dir)
  {
    mFace *theNewFace;
    theNewFace = new mFace(e1,e2,e3,e4,classif,dir);
    add(theNewFace);
    incTopo(FMDB_QUAD);
    return theNewFace;
  }

  mFace *mPart::createFaceWithEdges (mEdge *e1, mEdge *e2, mEdge *e3, 
  			pGEntity classif, int *dir)
  {
    mFace *theNewFace;
    theNewFace = new mFace(e1,e2,e3,classif,dir);
    add(theNewFace);
    incTopo(FMDB_TRI);
    return theNewFace;
  }

  mHex *mPart::createHexWithVertices (mVertex *v1, mVertex *v2, mVertex *v3, mVertex *v4,
	      		mVertex *v5, mVertex *v6, mVertex *v7, mVertex *v8, pGEntity
			classif)
  {
    mHex *theNewHex = new mHex(v1,v2,v3,v4,v5,v6,v7,v8,classif);
    add(theNewHex);
    incTopo(FMDB_HEX);
    return theNewHex;
  }
   
  mHex *mPart::createHexWithFaces (mFace *f1, mFace *f2, mFace *f3, mFace *f4, mFace *f5, mFace *f6, pGEntity
			classif)
  {
    mHex *theNewHex = new mHex(f1,f2,f3,f4,f5,f6,classif);
    add(theNewHex);
    incTopo(FMDB_HEX);
    return theNewHex;
  }


  mPrism *mPart::createPrismWithVertices (mVertex*v1, mVertex*v2, mVertex*v3, mVertex*v4,
			mVertex*v5, mVertex*v6, pGEntity classif)
  {
    mPrism *theNewPrism = new mPrism(v1,v2,v3,v4,v5,v6,classif);
    add(theNewPrism);
    incTopo(FMDB_PRISM);
    return theNewPrism;
  }
  
  mPrism *mPart::createPrismWithFaces (mFace *f1, mFace *f2, mFace *f3, mFace *f4, mFace *f5, pGEntity
			classif)
  {
    mPrism *theNewPrism = new mPrism(f1,f2,f3,f4,f5,classif);
    add(theNewPrism);
    incTopo(FMDB_PRISM);
    return theNewPrism;
  }

  mPyramid *mPart::createPyramidWithVertices (mVertex*v1, mVertex*v2, mVertex*v3, mVertex*v4,
			mVertex*v5,pGEntity classif)
  {
    mPyramid *theNewPyramid = new mPyramid(v1,v2,v3,v4,v5,classif);
    add(theNewPyramid);
    incTopo(FMDB_PYRAMID);
    return theNewPyramid;
  }
  
  mPyramid *mPart::createPyramidWithFaces (mFace *f1, mFace *f2, mFace *f3, mFace *f4, mFace *f5, pGEntity
			classif)
  {
    mPyramid *theNewPyramid = new mPyramid(f1,f2,f3,f4,f5,classif);
    add(theNewPyramid);
    incTopo(FMDB_PYRAMID);
    return theNewPyramid;
  }

  mHex *mPart::createHexWithVertices (int i1, int i2, int i3, int i4, 
		      	int i5, int i6, int i7, int i8, pGEntity classif)
  {
    mVertex *v1 = getVertex(i1);
    mVertex *v2 = getVertex(i2);
    mVertex *v3 = getVertex(i3);
    mVertex *v4 = getVertex(i4);
    mVertex *v5 = getVertex(i5);
    mVertex *v6 = getVertex(i6);
    mVertex *v7 = getVertex(i7);
    mVertex *v8 = getVertex(i8);
    if(!v1 || !v2 || !v3 || !v4 ||
       !v5 || !v6 || !v7 || !v8)
      {
	throw mException (__LINE__,__FILE__,"unknown verex id's in new hex creation");
      }
#ifdef _DEBUG_
#endif
    return createHexWithVertices (v1,v2,v3,v4,v5,v6,v7,v8,classif);
  }

  mPrism *mPart::createPrismWithVertices (int i1, int i2, int i3, int i4, 
		  	int i5, int i6, pGEntity classif)
  {
    mVertex *v1 = getVertex(i1);
    mVertex *v2 = getVertex(i2);
    mVertex *v3 = getVertex(i3);
    mVertex *v4 = getVertex(i4);
    mVertex *v5 = getVertex(i5);
    mVertex *v6 = getVertex(i6);
    if(!v1 || !v2 || !v3 || !v4 ||
       !v5 || !v6)
      {
	throw mException (__LINE__,__FILE__,"unknown verex id's in new prism creation");
      }
    return createPrismWithVertices (v1,v2,v3,v4,v5,v6,classif);
  }
  
  mPyramid *mPart::createPyramidWithVertices (int i1, int i2, int i3, int i4, 
		  	int i5,  pGEntity classif)
  {
    mVertex *v1 = getVertex(i1);
    mVertex *v2 = getVertex(i2);
    mVertex *v3 = getVertex(i3);
    mVertex *v4 = getVertex(i4);
    mVertex *v5 = getVertex(i5);
    if(!v1 || !v2 || !v3 || !v4 ||
       !v5 )
      {
	throw mException (__LINE__,__FILE__,"unknown verex id's in new pyramid creation");
      }
    return createPyramidWithVertices (v1,v2,v3,v4,v5,classif);
  }
  

pGEntity mPart::getGEntity (int tag,int type)
{
  return getGEntity(tag, type,classif_fp);
}

pGEntity mPart::getGEntity(int tag, int type,pGEntity (*pf)(pGModel,int,int))
{
  pGEntity theSolidModelEntity;
  bool flag_NullModel =false;
  if (!theSolidModel)
     theSolidModel =new NullModel (); 
  if (!strcmp ((char*)(theSolidModel->modeler ()), "null")) 
    flag_NullModel =true;  

#ifdef _POINTER2INT_
   mEntity::theSolidModel = theSolidModel;
#endif
 
  if (!flag_NullModel)
  {
    theSolidModelEntity = pf(theSolidModel,type,tag);
    if (!theSolidModelEntity) {
      printf ("model entity with dim %d and tag %d not found...\n", type,tag);
      throw;
    }
  }
  else
  {
    theSolidModelEntity = GM_entityByTag(theSolidModel,type,tag);
    if (!theSolidModelEntity) {
      // if the model entity has not been defined, define it
      switch (type) {
	case 0:    // model vertex
	  theSolidModelEntity =new NullVertex((NullModel*)theSolidModel, tag);
	  break;
	case 1:    // model edge
	  theSolidModelEntity =new NullEdge((NullModel*)theSolidModel, tag);
	  break;
	case 2:    // model face
	  theSolidModelEntity =new NullFace((NullModel*)theSolidModel, tag);
	  break;
	case 3:    // model region
	  theSolidModelEntity =new NullRegion((NullModel*)theSolidModel, tag);
	  break;
      }
    }
  }
  return theSolidModelEntity;
}

  mEntity* mPart::find(mEntity *e)
  {
    return allEntities.find(e);
  } 

  void mPart::add(mEntity *e)
  {
    allEntities.add(e);
  } 

  void mPart::del(mEntity *e)
  {
    allEntities.del(e);
  } 

  mEntity* mPart::getEntity (mEntity::mType type, mVertex *v[], pGEntity g)
  {  
    mEntity *e;
    switch(type)
      {
      case mEntity::HEX  : 
	e = (mEntity*)getHex(v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7]);
	if(!e)e = (mEntity*) createHexWithVertices (v[0],v[1],v[2],v[3],v[4],
				v[5],v[6],v[7],g);
	break;
      case mEntity::TET  : 
	e = (mEntity*)getTet(v[0],v[1],v[2],v[3]);
	if(!e)e = (mEntity*) createTetWithVertices (v[0],v[1],v[2],v[3],g);
	break;
      case mEntity::QUAD : 
	e = (mEntity*)getQuad(v[0],v[1],v[2],v[3]);
	if(!e)e = (mEntity*) createFaceWithVertices (v[0],v[1],v[2],v[3],g);
	break;
      case mEntity::TRI  : 
	e = (mEntity*)getTri(v[0],v[1],v[2]);
	if(!e)e = (mEntity*) createFaceWithVertices (v[0],v[1],v[2],g);
	break;
      case mEntity::EDGE : 
	e = (mEntity*) getEdge (v[0],v[1]);
	if(!e)e = (mEntity*) createEdge(v[0],v[1],g);
	break;      
        case mEntity::VERTEX:
        case mEntity::PRISM:
        case mEntity::PYRAMID:
        case mEntity::MIRROR:
        case mEntity::FACE:
             std::cerr << "impossible situation in mPart::getEntity\n";
            break;
      }
    return e;  
  }

  int mPart::getDimension() const
  {
    if(size(3))return 3;
    if(size(2))return 2;
    if(size(1))return 1;
    return dimension;
  }

  SCOREC::Util::mPoint mPart::closestPoint (const SCOREC::Util::mPoint & pt , int dim , int tag ) const
  {
    mClassIterator it = begin (dim, tag, dim);
    mClassIterator ite = end  (dim, tag, dim);

    if (it == ite)throw;

    double mind = 1.e20;

    SCOREC::Util::mPoint pp;

    /// model edge 
    if (dim == 1)
      {
	while (it != ite)
	  {
	    mEdge *e = (mEdge *) *it;
	    SCOREC::Util::mVector v1 (e->vertex(1)->point()-e->vertex(0)->point());
	    SCOREC::Util::mVector v2 (pt-e->vertex(0)->point());
	    double nv2 = v2.normValue();
	    SCOREC::Util::mVector v1Xv2 (v1 % v2);
	    double d = fabs(v1Xv2.normValue() * nv2);
	    if (d < mind)
	      {
		double xx = fabs (v1 * v2);
		mind = d;
		pp = e->vertex(0)->point() * (1.-xx) + e->vertex(1)->point() * xx;
	      }
	    ++it;
	  }
      }
    else
      {
	throw;
      }
    return pp;
  }

  void mPart::setSolidModel (pGModel m)
  {
    theSolidModel=m;
  }
/***************************GHOST INFORMATION*********************/

void mPart::addGhost(mGhost* ghost)
{
  ghosts.push_back(ghost);
}

int mPart::getGhostVecSize()
{
  return ghosts.size();
}

void mPart::clearGhostVec()
{
  ghosts.clear();
}
std::vector<mGhost*> mPart::getGhostVec()
{
  return ghosts;
}

mGhost* mPart::getGhost(int index)
{
  if(index < 0)
   return NULL;
  
  if(((size_t)index) > ghosts.size())
    cout << "\n Index out of bound ghosts size: " << ghosts.size() << " index: " << index << endl; 

  return ghosts[index];
}

// ***********************************************
//            ENTITY TOPO COUNT 
// ***********************************************
void mPart::incTopo(int topo)
{
 ++numTopo[topo]; 
}

void mPart::decTopo(int topo)
{
  --numTopo[topo];
}

int mPart::getNumTopo(int topo)
{
  if (topo==FMDB_ALLTOPO)
  {
    int sum=0;
    for (int i=FMDB_POINT; i<FMDB_ALLTOPO; ++i)
      sum+=numTopo[i];
    return sum;
  }  
  else
    return numTopo[topo]; 
}

// ***********************************************
//            PRINTING PART INFO
// ***********************************************
void mPart::printAll()
{
#ifdef FMDB_PARALLEL
  if (getId()%getMesh()->getNumPart()==0)
  for (pmModel::PEIter peiter=pmModel::Instance()->peBegin(); peiter!=pmModel::Instance()->peEnd();++peiter)
    (*peiter)->print();
#endif

  printPSet();

  for (int dim=0; dim<=getDimension();++dim)
    for (iterall ent_it = beginall(dim); ent_it!=endall(dim);++ent_it)
      FMDB_Ent_DspInfo(this, (*ent_it));
}

void mPart::printPSet()
{
  int pset_count=0;
  for (psetIter pset_it=beginPSet(); pset_it!=endPSet(); ++pset_it)
    (*pset_it)->attachInt(FMDB_Util::Instance()->getId(), pset_count++);

  for (psetIter pset_it=beginPSet(); pset_it!=endPSet(); ++pset_it)
    ((mEntityGroup*)(*pset_it))->print();      
}

void mPart::printNumEntities()
{
  unsigned int numPE = ParUtil::Instance()->size();
  vector<int> pset_output;
  pset_output.resize(numPE);
#ifdef FMDB_PARALLEL
  int *in = new int;
  int *out = new int[numPE];
  *in = numPSet();
  MPI_Allgather(in, 1, MPI_INT, out, 1, MPI_INT, ParUtil::Instance()->getComm());
  for (unsigned i = 0; i < numPE; ++i)
    pset_output[i] = out[i];
#else
   pset_output.push_back(numPSet());
#endif
  if (!ParUtil::Instance()->rank())
  {
    int sum=0;
    for (int j=0; j<SCUTIL_CommSize();++j)
      sum+=pset_output[j]; 
    cout<<"[grp"<<SCUTIL_CommGrp()<<"|rank"<<SCUTIL_CommRank()<<"] Part "<<getId()
        <<">> #p-set:  "<<sum<<" (";
    for (int j=0; j<SCUTIL_CommSize();++j)
    {  
      cout<<pset_output[j];
      if (j!=ParUtil::Instance()->size()-1)
        cout<<",";
    }
    cout<<")"<<endl;
  }

  // compute # owned entities per part
  for (int dim=0; dim<=3; ++dim)
  {
    vector<int> ent_output;
#ifdef FMDB_PARALLEL

    ent_output.resize(numPE);
    mEntity* ent;
    int numEnt=0;
    for (mPart::iterall it=beginall(dim);it!=endall(dim);++it)
    {
      ent = *it;
      if (ent->getOwnPartID(this)==getId())
      numEnt++;
    }

    *in = numEnt;
    MPI_Allgather(in, 1, MPI_INT, out, 1, MPI_INT, ParUtil::Instance()->getComm());
    for (unsigned i = 0; i < numPE; i++)
      ent_output[i] = out[i];
#else
    ent_output.push_back(size(dim));
#endif

  // print # owned entities per part/proc
    if (!ParUtil::Instance()->rank())
    {
      int sum=0;
      for (int j=0; j<SCUTIL_CommSize();++j)
        sum+=ent_output[j];  
      cout<<"[grp"<<SCUTIL_CommGrp()<<"|rank"<<SCUTIL_CommRank()<<"] Part "<<getId()
          <<">> #ent["<<dim<<"]: "<<sum<<" (";
      for (int j=0; j<SCUTIL_CommSize();++j)
      {  
         cout<<ent_output[j];
         if (j!=ParUtil::Instance()->size()-1)
	   cout<<",";
      }
      cout<<")"<<endl;
    } // if (!ParUtil::Instance()->rank())
  } // for dim
#ifdef FMDB_PARALLEL
  delete in;
  delete [] out;
#endif
}

int mPart::numUniqueEntities(int type)
{
  if (ParUtil::Instance()->size()==1)
    return size(type);
  int mypid=ParUtil::Instance()->rank();
  int numEnt=0;
  for (iterall it=beginall(type); it!=endall(type); ++it)
    if ((*it)->getOwner()==mypid)
      numEnt++;
  return numEnt;
}
