/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include <iostream>
#include "mPrism.h"
#include "mEntityContainer.h"
#include "mException.h"
#include "mVertex.h"
#include "mEdge.h"
#include "mFace.h"
#include "function_template.h"

#include <stdio.h>

  mPrism::mPrism()
    : mRegion() 
  {
  }
  
  mPrism::mPrism(mVertex *v1, mVertex *v2, mVertex *v3, mVertex *v4,
		 mVertex *v5, mVertex *v6, 
		 GEntity *classification)
    : mRegion() 
  {
    down_adj_dim = 0; 
     down_adj[0]=(mEntity*)v1;
     down_adj[1]=(mEntity*)v2;
     down_adj[2]=(mEntity*)v3;
     down_adj[3]=(mEntity*)v4;
     down_adj[4]=(mEntity*)v5;
     down_adj[5]=(mEntity*)v6;

#ifndef _POINTER2INT_
  theClassification = classification;
#else
   gType = GEN_type(classification);
   gTag = GEN_tag(classification);
#endif

  }
  
  
  mPrism::mPrism(mFace *f1, mFace *f2, mFace *f3, mFace *f4, mFace *f5, GEntity *classification)
  : mRegion ()  
 {

   down_adj_dim=2; 
   down_adj[0]=(mEntity*)f1;
   down_adj[1]=(mEntity*)f2;
   down_adj[2]=(mEntity*)f3;
   down_adj[3]=(mEntity*)f4;
   down_adj[4]=(mEntity*)f5;

#ifndef _POINTER2INT_
  theClassification = classification;
#else
   gType = GEN_type(classification);
   gTag = GEN_tag(classification);
#endif

 }

  int mPrism:: getNbTemplates (int what) const
  {
    switch(what)
      {
      case 0: return 6;
      case 1: return 9;
      case 2: return 5;
	
      default : 
	{
	  char text[256];
	  sprintf(text,"trying to ask for adjacency list of level %d for a prism when it's not created",what);
	  throw new mException (__LINE__,__FILE__,text);
	}
      }
  }

  short int mPrism::pmev[9][2] = {{0,1},{1,2},{2,0},{0,3},{1,4},{2,5},
                          {3,4},{4,5},{5,3}};

  short int mPrism::pmfv[5][4] =   {{0,1,2,-1},
                            {0,1,4,3},
                            {1,2,5,4},
                            {2,0,3,5},
                            {3,4,5,-1}};

  short int mPrism::pmfe[5][4] =   {{0,1,2,-1},
                            {0,4,6,3},
                            {1,4,7,5},
                            {2,3,8,5},
                            {6,7,8,-1}};

  short int mPrism::pmve[6][2] = {{0,2},{0,1},{1,2},{6,8},{6,7},{7,8}};
  short int mPrism::pmvf[6][3] = {{0,1,3},{0,1,2},{0,2,3},{1,3,4},{1,2,4},{2,3,4}};
  short int mPrism::pmef[9][2] = {{0,1},{0,2},{0,3},
                           {1,3},{1,2},{2,3},
                           {1,4},{2,4},{3,4}};

  mEntity *mPrism::getTemplate(int ith, int what, int with)const
  {

#ifndef _POINTER2INT_
  GEntity* classification = theClassification;
#else
  GEntity* classification = getClassification();  
#endif

    switch(what)
      {
      case 0:
	 if(down_adj_dim==0)
	       return down_adj[ith];

	 if(down_adj_dim==1)
	 {
	    mEdge *e1 = (mEdge*)get(1,pmve[ith][0]);
	    mEdge *e2 = (mEdge*)get(1,pmve[ith][1]);
	    return e1->commonVertex(e2);
	 }
	 else if (down_adj_dim==2)
	  {
	    mFace *f1 = (mFace*)get(2,pmvf[ith][0]);
	    mFace *f2 = (mFace*)get(2,pmvf[ith][1]);
	    mFace *f3 = (mFace*)get(2,pmvf[ith][2]);
	    return f1->commonVertex(f2,f3);
	  }
	break;
      case 1:
        if(down_adj_dim==2)
	  {
	    mFace *f1 = (mFace*)get(2,pmef[ith][0]);
	    mFace *f2 = (mFace*)get(2,pmef[ith][1]);
	    mEdge *e =  f1->commonEdge(f2);
	    if(e) return e;
	  }

	 return new mEdge ((mVertex*)get(0, pmev[ith][0]),
	        	   (mVertex*)get(0, pmev[ith][1]),
		            classification);
	break;
      case 2:
	if(pmfv[ith][3] == -1)
	  {
	    if(with == 0) {
		return new mFace ((mVertex*)get(0,pmfv[ith][0]),
				  (mVertex*)get(0, pmfv[ith][1]),
				  (mVertex*)get(0, pmfv[ith][2]),
				  classification);
	    }
	    else if (with == 1)
	      {
		return new mFace ((mEdge*)get(1, pmfe[ith][0]),
				  (mEdge*)get(1, pmfe[ith][1]),
				  (mEdge*)get(1, pmfe[ith][2]),
				  classification);
	      }
	    else throw new mException (__LINE__,__FILE__,"weird adjacency asked");
	  }
	else
	  {
	    if(with == 0) {
		return new mFace ((mVertex*)get(0,pmfv[ith][0]),
				  (mVertex*)get(0,pmfv[ith][1]),
				  (mVertex*)get(0,pmfv[ith][2]),
				  (mVertex*)get(0,pmfv[ith][3]),
				  classification);
	    }
	    else if (with == 1)
	      {
		return new mFace ((mEdge*)get(1, pmfe[ith][0]),
				  (mEdge*)get(1, pmfe[ith][1]),
				  (mEdge*)get(1, pmfe[ith][2]),
				  (mEdge*)get(1, pmfe[ith][3]),
				  classification);
	      }
	    else throw new mException (__LINE__,__FILE__,"weird adjacency asked");
	  }
      }
    return 0;
  }
  
  mPrism::~mPrism()
  {
    
  }
 
    // adjacency related
   int mPrism::size (int what)const
   {
      return getNbTemplates(what);
   }

   mEntity* mPrism::get(int what, int ith)const
   {
      if(what==down_adj_dim)
         return down_adj[ith];

      return getTemplate(ith, what, 0);
   }

   // void mPrism::getAppendPList(int what, pPList pl) const;

   mEntity * mPrism::find(mEntity* m) const
   {
      return find_entity_in_array<mEntity*>((mEntity**)down_adj, 5, m);
   }

