/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/

#include <iostream>
#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <iomanip>
#include "FMDB.h"
#include "mMesh.h"

char _string_LBMethod[][30] = {"RCB", "RIB", "GRAPH", "HYPERGRAPH", "ParMETIS"};
char _string_LBApproach[][30] = {"Partition", "Repartition", "Refine", "PartKway", "PartGeom", "PartGeomKWay", "AdaptiveRepart", "RefineKway"};

int TEST_PART_BOUNDARY(pMeshMdl mesh);
int TEST_MIGRATION (pMeshMdl mesh, int numPart, int numMigr, int call_serial);
int TEST_ENT_MIGRATION(pMeshMdl mesh, int numMigr);

void CHECK(const char *msg, int oK, int myrank=0)
{     
  if (myrank) return;
  if (oK)
    fprintf(stdout,"%s OK \n",msg);
  else
    fprintf(stdout,"%s FAILED\n",msg);
  fflush(stdout);
}

int  LB_method=FMDB_PARMETIS, LB_approach=PartKway;
double imbal_tol=1.03;
void Construct_Pset(pMeshMdl);

int main (int argc, char *argv[])
{
  SCUTIL_Init(MPI_COMM_WORLD);
  int numMigr=0, numPart=1, numIMigr=0;

  if (argc<5)
  {
    if (!SCUTIL_CommRank())
    {
      cout<<__func__<<" ERROR: Invalid input \n       ./main mesh_file distributed_file(no:0, yes:1), #parts (default: 1), #migrations (default:0), #individual migrations (default:0), LB_Method (0-4), LB_Approach (0-7), imbalance_tolerance (enter 1.xx for xx%, default 1.03), construct_pset(no:0, yes:1), call_serialGlobPtn (no:0, yes:1) \n";
      cout<<"\tFMDB_LBMethod {FMDB_RCB=0, /*1*/ FMDB_RIB, /*2*/ FMDB_GRAPH, /*3*/ FMDB_HYPERGRAPH, /*4*/ FMDB_PARMETIS (default)}\n";
      cout<<"\tFMDB_LBApproach {FMDB_PARTITION=0, /*1*/ FMDB_REPARTITION, /*2*/ FMDB_REFINE, /*3*/ PartKway (default), /*4*/ PartGeom, /*5*/ PartGeomKWay, /*6*/ AdaptiveRepart, /*7*/ RefineKway}\n";
    }
    return 1;
  }
  if (atoi(argv[3])>0)
    numPart = atoi(argv[3]);

  numMigr = atoi(argv[4]);

  if (argc>5 && atoi(argv[5])>0)
    numIMigr = atoi(argv[5]);

  if (argc>6 && atoi(argv[6])>=0 && atoi(argv[6])<=4)
    LB_method = atoi(argv[6]);

  if (argc>7 && atoi(argv[7])>=0 && atoi(argv[7])<=7)
    LB_approach = atoi(argv[7]);

  if (argc>8 && atoi(argv[8])>=0.0)
    imbal_tol = atof(argv[8]);

  int construct_pset=0;
  if (argc>9 && atoi(argv[9])>0)
    construct_pset=atoi(argv[9]);

  int call_serial=0;
  if (argc>10 && atoi(argv[10])>0)
    call_serial=1;
 
  int print_all=0;
  if (argc>11 && atoi(argv[11])>0)
    print_all=1;

  char mesh_file[1024];
  strcpy(mesh_file, argv[1]);

  if (!SCUTIL_CommRank()) 
  {
    cout<<"************************************************************************\n";
    cout<<"[INPUT]\n";
    cout<<"\tdistributed mesh? ";
    if (atoi(argv[2])) cout<<"YES\n";
    else cout<<"NO\n";
    cout<<"\t#parts per proc: "<<numPart<<endl;
    cout<<"\t# migration steps: "<<numMigr<<endl;
    cout<<"\t# individual migration steps: "<<numIMigr<<endl;
    cout<<"\tLB_method="<<_string_LBMethod[LB_method]<<endl;
    cout<<"\tLB_approach="<<_string_LBApproach[LB_approach]<<endl;
    cout<<setprecision(5)<<"\timbalance_tol="<<imbal_tol<<endl;
    cout<<"\tconstruct p-set? ";
    if (construct_pset) cout<<"YES\n";
    else cout<<"NO\n";

    SCUTIL_DspSysInfo();
    cout<<"************************************************************************\n\n";  
  }

  // **************************************************************
  // TEST 1: MESH FILE I/O
  // **************************************************************

  if (!SCUTIL_CommRank())  cout<<"\n***** MESH/PART FUNCTIONS *****\n"; 

  pGModel model=NULL;
  pMeshMdl mesh;
  FMDB_Mesh_Create (model, mesh);
  SCUTIL_DspCurMem("INITIAL COST: ");

  SCUTIL_ResetRsrc();

  FMDB_Mesh_SetPtnParam (mesh, LB_method, LB_approach, imbal_tol, 0);

  if (FMDB_Mesh_LoadFromFile (mesh, mesh_file,atoi(argv[2])))
  {
    if (!SCUTIL_CommRank()) cout<<"FAILED MESH LOADING - check mesh file or argv[2]\n";
    FMDB_Mesh_Del(mesh);
    SCUTIL_Finalize();
    return SCUtil_FAILURE;
  }

  if (!SCUTIL_CommRank()) 
    cout<<endl;
  SCUTIL_DspRsrcDiff("MESH LOADING: ");
  FMDB_Mesh_DspStat(mesh);
 
  int isValid=0;
  FMDB_Mesh_Verify(mesh, &isValid);
  if (!isValid)
  {
    FMDB_Mesh_Del(mesh);
    SCUTIL_Finalize();
    return SCUtil_FAILURE;
  }

  if (construct_pset)
  {
    Construct_Pset(mesh);
    FMDB_Mesh_DspNumEnt(mesh);
  }

  pMeshDataId double_arr_2  = MD_createMeshDataId("2 double array", SCUtil_DBL, 2);
  pMeshDataId double_arr_4  = MD_createMeshDataId("4 double array", SCUtil_DBL, 4);
  assert (FMDB_Util::Instance()->typeMeshDataId(double_arr_2)==SCUtil_DBL && FMDB_Util::Instance()->sizeMeshDataId(double_arr_2)==2);
  assert (FMDB_Util::Instance()->typeMeshDataId(double_arr_4)==SCUtil_DBL && FMDB_Util::Instance()->sizeMeshDataId(double_arr_4)==4);
 
  if (SCUTIL_CommSize()==1)
  {
    // attach double array
    for(mPart::iterall vtx_it = mesh->getPart(0)->beginall(0) ; vtx_it!=mesh->getPart(0)->endall(0) ; ++vtx_it)
    {
      double* dbl_arr = new double[2];
      dbl_arr[0] = ((double)((*vtx_it)->getId()))*1.1;
      dbl_arr[1] = ((double)((*vtx_it)->getId()))*2.2;

      EN_attachDataPtr(*vtx_it, double_arr_2 , (void*)dbl_arr);
  
      double* dbl_arr_4 = new double[4];
      for (int i=0; i<4; ++i)
        dbl_arr_4[i] = (double)((i+1)*2.2);

      EN_attachDataPtr(*vtx_it, double_arr_4 , (void*)dbl_arr_4);
    }
    MD_migrateMeshDataId(double_arr_2, FMDB_VERTEX, 1);
    MD_migrateMeshDataId(double_arr_2, FMDB_VERTEX, 0);
    MD_migrateMeshDataId(double_arr_2, FMDB_VERTEX, 1);
    MD_migrateMeshDataId(double_arr_4, FMDB_VERTEX, 1);
  }

  // **************************************************************
  // TEST 2: PART BOUNDARY CHECK
  // **************************************************************
//  CHECK("\n* Checking part boundary...", !TEST_PART_BOUNDARY (mesh), SCUTIL_CommRank());
//  SCUTIL_Sync();

  // **************************************************************
  // TEST 3: MIGRATION TESTING WITH SINGLE PART
  // **************************************************************
  CHECK("* TESTING MIGRATION ...", !TEST_MIGRATION(mesh, numPart, numMigr, call_serial), SCUTIL_CommRank());
 
  if (print_all)
  {
    for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
      (*part_it)->printAll();
  }


  // **************************************************************
  // TEST 4: INDIVIDUAL MIGRATION TESTING
  // **************************************************************
  if (numIMigr)
    CHECK( "\n* TESTING INDIVIDUAL MIGRATION...", !TEST_ENT_MIGRATION(mesh, numIMigr), SCUTIL_CommRank());

  // **************************************************************
  // TEST 5: TESTING AUTO TAG MIGRATION - DOUBLE ARRAY
  // **************************************************************

  if (SCUTIL_CommSize()==1)
  {
    double* dbl_arr_back = new double[2];
    int arr_size;
    for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
      for(mPart::iterall vtx_it = (*part_it)->beginall(0) ; vtx_it!=(*part_it)->endall(0) ; ++vtx_it)
      {
        assert(EN_getDataPtr(*vtx_it,  double_arr_2, (void**)&dbl_arr_back));
        assert(EN_getDataPtr(*vtx_it,  double_arr_4, (void**)&dbl_arr_back));
    }

    for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
      for(mPart::iterall vtx_it = (*part_it)->beginall(0) ; vtx_it!=(*part_it)->endall(0) ; ++vtx_it)
      {
        EN_getDataPtr(*vtx_it,  double_arr_2, (void**)&dbl_arr_back);
        delete [] dbl_arr_back;
        (*vtx_it)->deleteData(double_arr_2);

        EN_getDataPtr(*vtx_it,  double_arr_4, (void**)&dbl_arr_back);
        delete [] dbl_arr_back;
        (*vtx_it)->deleteData(double_arr_4);
    }

    for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
      for(mPart::iterall vtx_it = (*part_it)->beginall(0) ; vtx_it!=(*part_it)->endall(0) ; ++vtx_it)
      {
        assert(!EN_getDataPtr(*vtx_it,  double_arr_2, (void**)&dbl_arr_back));
        assert(!EN_getDataPtr(*vtx_it,  double_arr_4, (void**)&dbl_arr_back));
      }
  }
  FMDB_Mesh_Del (mesh);
  SCUTIL_Finalize();
  system ("rm -rf *part_p*sms");
  return 0;
}

void Construct_Pset(pMeshMdl mesh)
{
  int mesh_dim, isEnd, num;
  FMDB_Mesh_GetDim(mesh, &mesh_dim);

  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
    for(mPart::iterall rgn_it = (*part_it)->beginall(mesh_dim) ; rgn_it!=(*part_it)->endall(mesh_dim) ; ++rgn_it)
       FMDB_Ent_SetWeight(*rgn_it,  2.0);

  pMeshEnt ent;
  pEntSet set;

  for (mMesh::partIter pit=mesh->partBegin(); pit!=mesh->partEnd(); ++pit)
  {
    pPart part = *pit;
    pEntSet psets[part->getId()+1];
    for (int i=0; i<part->getId()+1; ++i)
      FMDB_Set_Create (mesh, part, FMDB_PSET, psets[i]);

    vector<pEntSet> vec_set;
    FMDB_Part_GetSet(part, vec_set);
    assert (vec_set.size()==part->getId()+1);

    pPartEntIter entIter;  
    isEnd = FMDB_PartEntIter_Init (part, mesh_dim, FMDB_ALLTOPO, entIter);
    while (!isEnd)
    {
      isEnd = FMDB_PartEntIter_GetNext(entIter, ent);
      if(isEnd) break; 
      if (rand()%4==1) // random picking of partition objects up to approx. 1/4 of po-ent
        FMDB_Set_AddEnt(psets[rand()%(part->getId()+1)], ent);
    }
    FMDB_PartEntIter_Del (entIter);
  } // for mMesh::partIter

  double weight;
  for (mMesh::partIter pit=mesh->partBegin(); pit!=mesh->partEnd(); ++pit)
  {
    pPart part = *pit;
    pPartSetIter pset_iter;
    isEnd = FMDB_PartSetIter_Init (part, pset_iter); 
    while (!isEnd)
    {
      isEnd = FMDB_PartSetIter_GetNext(part, pset_iter, set);
      if(isEnd) break; 
      FMDB_Set_GetWeight(set, &weight);
      FMDB_Set_GetNumEnt (set, &num);
      assert(weight == num*2.0);
    }
    FMDB_PartSetIter_Del (pset_iter);
  }
}

int TEST_MIGRATION (pMeshMdl mesh, int numPart, int numMigr, int call_serial)
{
  MigrCB cb;
  int meshDim, maxNumPart,isValid=0;;    
  FMDB_Mesh_GetDim (mesh, &meshDim);
  FMDB_Mesh_GetNumPart(mesh, &maxNumPart);

//  if (numPart>1)
  { 
    SCUTIL_ResetRsrc();
    if (numPart>1) FMDB_Mesh_SetNumPart(mesh, numPart);  // prev: PM_setMaxNumParts
    if (numMigr==-1)
    {
      if (!SCUTIL_CommRank()) 
        cout<<"\nFMDB_Topo_SetWeight(mesh, FMDB_PYRAMID, 2.0) & FMDB_Topo_SetWeight(mesh, FMDB_PRISM, 3.0)\n";
      FMDB_Topo_SetWeight(mesh, FMDB_PYRAMID, 2.0);
      FMDB_Topo_SetWeight(mesh, FMDB_PRISM, 3.0);
    }
    if (SCUTIL_CommSize()>1 || !call_serial)
      FMDB_Mesh_GlobPtn (mesh, cb, LB_method, LB_approach, imbal_tol, 0);
    else
      FMDB_Mesh_SerialGlobPtn (mesh, LB_method, LB_approach, imbal_tol, 0);
    SCUTIL_DspRsrcDiff("MESH REDISTRIBUTION: ");
    
    int pyramid_count=0, prism_count=0,tet_count=0, weight_sum=0;
    for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
    {
      weight_sum=0;
      if (numMigr==-1)
      {
        for (mPart::iterall ent_it=(*part_it)->beginall(3); ent_it!=(*part_it)->endall(3); ++ent_it)
        {
          if ((*ent_it)->getTopo() == FMDB_PYRAMID)
          {
             weight_sum+=2;
            ++pyramid_count;
          }
          else if ((*ent_it)->getTopo() == FMDB_PRISM)
          {
             weight_sum+=3;
            ++prism_count;
          }
          else 
          {
            weight_sum+=1;
            ++tet_count;
          }
        }
      }
      else
        weight_sum=(*part_it)->size(3);
    }
    if (numMigr==-1)
      cout<<"#tet="<<tet_count<<", #pyramid="<<pyramid_count<<", #prism="<<prism_count<<endl;
    //FMDB_Mesh_Merge(mesh, cb);
    //FMDB_Mesh_Verify(mesh, &isValid);
    //CHECK("\n* Checking Re-distributed Mesh Validity...", isValid, SCUTIL_CommRank());
    //FMDB_Mesh_GlobPtn (mesh, cb, LB_method, LB_approach, imbal_tol, 0);
    FMDB_Mesh_DspStat(mesh);
    FMDB_Mesh_WriteToFile (mesh, "partitioned.sms", 1); 
  }

  if (numMigr<1) return 0;

  FMDB_Mesh_GetNumPart(mesh, &maxNumPart);

  if (!SCUTIL_CommRank()) cout<<"\n **** START MIGRATION TEST ***\n";

  int mesh_dim;
  FMDB_Mesh_GetDim(mesh, &mesh_dim);
  // get the PO to migrate (we do neighborhood migration)
  void* tmp_ptr; 
  for (int iter=0; iter<numMigr;++iter) // loop over niter times
  {
      if (!SCUTIL_CommRank()) 
        cout<<"\n["<<SCUTIL_CommGrp()<<"|"<<SCUTIL_CommRank()<<"] *** "<<iter+1<<"'th migration ***\n";
      std::map<pMeshEnt, std::pair<int, int> > EntToMigr;
      std::map<pPartSet, std::pair<int, int> > SetToMigr;
      pMeshEnt ent;
     
      // randomly pick entities and p-sets
      for (mMesh::partIter pit=mesh->partBegin(); pit!=mesh->partEnd(); ++pit)
      {
        pPart part = *pit;
        if (part->size(mesh_dim)==0) continue; // empty part

        int rand_partid; 
  
        for(mPart::psetIter it=(*pit)->beginPSet(); it!=(*pit)->endPSet(); ++it)
        {
          rand_partid=rand()%(SCUTIL_CommSize()*maxNumPart);
          if (rand_partid==FMDB_Part_GlobID(part)) continue;
          if (rand()%4==1) // random picking of partition objects up to approx. 1/3 of mesh
            SetToMigr.insert(std::map<mEntityGroup*, std::pair<int, int> >::value_type((mEntityGroup*)(*it), std::make_pair(FMDB_Part_GlobID(part), rand_partid))); // 0 .. # total 
       }

        pPartEntIter entIter;
        int isEnd = FMDB_PartEntIter_Init (part, meshDim, FMDB_ALLTOPO, entIter);
        while (!isEnd)
        {
          isEnd = FMDB_PartEntIter_GetNext(entIter, ent);
          if(isEnd) break; 
          if (EN_getDataPtr(ent, FMDB_Util::Instance()->getPSet(), &tmp_ptr)) // in p-set
            continue;
          rand_partid=rand()%(SCUTIL_CommSize()*maxNumPart);
          if (rand_partid==FMDB_Part_GlobID(part)) continue;
          if (rand()%4==1) // random picking of partition objects up to approx. 1/3 of mesh
            EntToMigr.insert(std::map<mEntity*, std::pair<int, int> >::value_type(ent, std::make_pair(FMDB_Part_GlobID(part), rand_partid))); // 0 .. # total 
        }
        FMDB_PartEntIter_Del (entIter);
      } // for mMesh::partIter

#ifdef _DEBUG_
  SCUTIL_Sync();
  FMDB_Mesh_DspInfo (mesh);
  SCUTIL_Sync();
#endif

      FMDB_Mesh_Migr(mesh, cb, EntToMigr, SetToMigr);
      SCUTIL_Sync();
      FMDB_Mesh_Verify(mesh, &isValid);
      CHECK("\n* Checking Randomly Migrated Mesh Validity...", isValid, SCUTIL_CommRank());

      if (iter==numMigr-1)
      {
        if (!SCUTIL_CommRank()) 
          cout<<"\n["<<SCUTIL_CommGrp()<<"|"<<SCUTIL_CommRank()<<"] *** LOADBALANCE *** \n";

        if (SCUTIL_CommSize()>1 || !call_serial)
          FMDB_Mesh_GlobPtn (mesh, cb, LB_method, LB_approach, imbal_tol, 0);
        else
           FMDB_Mesh_SerialGlobPtn (mesh, LB_method, LB_approach, imbal_tol, 0);
        FMDB_Mesh_Verify(mesh, &isValid);
        CHECK("\n* Checking Re-distributed Mesh Validity...", isValid, SCUTIL_CommRank());
        FMDB_Mesh_DspStat(mesh);
      } // if (iter==numMigr-1)
    }  // for iter
    SCUTIL_Sync();
    
    return 0;
}

int TEST_PART_BOUNDARY(pMeshMdl mesh)
{
  pPart part;
  pMeshEnt pEnt;
  int isOnPrtBdry, counter=0;
  std::vector<pPart> vecPart;
  FMDB_Mesh_GetAllPart (mesh, SCUTIL_CommRank(), vecPart);
  CHECK("\n* Checking #parts on local proc...", vecPart.size(), SCUTIL_CommRank());
  
  /* Mesh/Part traversal (part boundary iterator)*/
  /// For each part on the meshModel, get part boundary entities of different types 
  int iIsEnd = 0;
  int a2iPartBdryEntCount[5][12];
  int a2iEntCount[5][12];
  for (int i=0; i<5; ++i)
    for (int j=0; j<12; ++j)
    {
       a2iEntCount[i][j]=0;
       a2iPartBdryEntCount[i][j]=0;
    }
  for (vector<pPart>::iterator It = vecPart.begin(); It != vecPart.end(); It++ )
  {
    part = (pPart)(*It);
     
    set<pMeshEnt> bdryEnts;
    bool bOK_valid=true;
    for(int type = FMDB_VERTEX; type <= FMDB_ALLTYPE; type++)
    {
      for (int topo=FMDB_POINT; topo<=FMDB_ALLTOPO; ++topo)
      {
        counter=0;
        if (!checkTypeTopo(type, topo))
        {  a2iEntCount[type][topo]=0;
           continue;
        }

        pPartEntIter pIter; int numEnt;
        iIsEnd = FMDB_PartEntIter_Init(part, type, topo, pIter);
        while(!iIsEnd)
        {
          iIsEnd = FMDB_PartEntIter_GetNext(pIter, pEnt);
          if(iIsEnd) break;
          /* Cross check if the entity is on part boundary */
          ++counter;
        }
        FMDB_PartEntIter_Del (pIter);
        a2iEntCount[type][topo]=counter;
        FMDB_Part_GetNumEnt(part, type, topo, &numEnt);
        if (numEnt!=counter) bOK_valid=false;
      } // for topo
    } // for type
    
    CHECK("\n* Checking Part Iterator...", a2iEntCount[FMDB_VERTEX][FMDB_ALLTOPO]+
           a2iEntCount[FMDB_EDGE][FMDB_ALLTOPO]+a2iEntCount[FMDB_FACE][FMDB_ALLTOPO]+a2iEntCount[FMDB_REGION][FMDB_ALLTOPO]
           ==a2iEntCount[FMDB_ALLTYPE][FMDB_ALLTOPO] && bOK_valid, SCUTIL_CommRank());
    
    bOK_valid=true;
    for(int type = FMDB_VERTEX; type <= FMDB_ALLTYPE; type++)
    {
      for (int topo=FMDB_POINT; topo<=FMDB_ALLTOPO; ++topo)
      {
        counter=0;
        if (!checkTypeTopo(type, topo))
        {  a2iPartBdryEntCount[type][topo]=0;         
           continue; 
        }

        pPartEntIter pBdryIter;      
        iIsEnd = FMDB_PartEntIter_InitPartBdry(part, -1, type, topo, pBdryIter); 
        while(!iIsEnd)
        {
  	  iIsEnd = FMDB_PartEntIter_GetNext(pBdryIter, pEnt);
          if(iIsEnd) break; 
          /* Cross check if the entity is on part boundary */
          ++counter;
          FMDB_Ent_IsOnPartBdry (pEnt, &isOnPrtBdry);
   	  if (isOnPrtBdry != 1) bOK_valid=false;

  	  /* Get next entity on part boundary having dimension type. Store the boundary entities*/
  	  bdryEnts.insert(pEnt);
        }       
        FMDB_PartEntIter_Del (pBdryIter);
        a2iPartBdryEntCount[type][topo]=counter;
      } // for topo
    } // for type
    CHECK("\n* Checking Part Boundary Iterator...", a2iPartBdryEntCount[FMDB_VERTEX][FMDB_ALLTOPO]+
           a2iPartBdryEntCount[FMDB_EDGE][FMDB_ALLTOPO]+a2iPartBdryEntCount[FMDB_FACE][FMDB_ALLTOPO]
           ==a2iPartBdryEntCount[FMDB_ALLTYPE][FMDB_ALLTOPO] && bOK_valid, SCUTIL_CommRank());
  
    /* Distributed entity functions (ownership, remote copies, ghosts)*/
    std::set<pMeshEnt>::iterator itSet;
    for(itSet = bdryEnts.begin(); itSet != bdryEnts.end(); itSet++)
    {
      cout<<"["<<SCUTIL_CommRank()<<"] Bdry Ent"<<FMDB_Ent_StrID(*itSet);
      pMeshEnt pEnt = *itSet;
      
      /* Now for each part boundary entity in the set, get its owning part id */
      int iOwnerPrt;
      FMDB_Ent_GetOwnPartID(pEnt, part, &iOwnerPrt);

      /* If the entity is a copy on local part then get its owning entity */
     
      if(iOwnerPrt != FMDB_Part_GlobID(part))
      {
        pMeshEnt pOwnerEnt;
        FMDB_Ent_GetOwnEnt(pEnt, part, pOwnerEnt);
        assert(pOwnerEnt != NULL);
      }

      /* Get all remote copies of the part boundary entity */    
      std::vector<std::pair<int, pMeshEnt> > vecRmtCpy;
      FMDB_Ent_GetAllRmt(pEnt, vecRmtCpy);
      
      /* Now iterate through the remote copies of each entity.*/
      std::vector<std::pair<int, pMeshEnt> >::iterator itVec;

      for(itVec = vecRmtCpy.begin(); itVec != vecRmtCpy.end(); itVec++ )
      {
        int pid = itVec->first;
	pMeshEnt pRmt = itVec->second;

        /* There should exist a remote copy on pid then, Cross-check with function 
           GetRmt(A O(1) operation for getting remote copy on a target part id) */
    	pMeshEnt pRmtCopy;
        FMDB_Ent_GetRmt(pEnt, pid, pRmtCopy);
        assert(pRmtCopy == pRmt);
      }
  
      /*Now get the number of remote copies for each entity */
      int iNumRmtCopies;
      FMDB_Ent_GetNumRmt(pEnt, &iNumRmtCopies);
      assert(((size_t)iNumRmtCopies) == vecRmtCpy.size());

      /* Clear all remote copies */       
      FMDB_Ent_ClrRmt(pEnt);
    
      // Now get number of remote copies again, should be zero now
      int iNumRmtEmpty;
      FMDB_Ent_GetNumRmt(pEnt, &iNumRmtEmpty);
      assert(iNumRmtEmpty == 0);     

      // Now add the remote copies again
      for(itVec = vecRmtCpy.begin(); itVec != vecRmtCpy.end(); itVec++ )
      {
	int pid = itVec->first;
        pMeshEnt pRmt = itVec->second;
        FMDB_Ent_SetRmt(pEnt, pid, pRmt);
      }
  
      // Get the number of remote copies again
      int iNumRmtCopiesN;
      FMDB_Ent_GetNumRmt(pEnt, &iNumRmtCopiesN);
      assert(iNumRmtCopiesN == iNumRmtCopies);

      vecRmtCpy.clear();
    } 
    bdryEnts.clear();
  }
  return SCUtil_SUCCESS;
}

int TEST_ENT_MIGRATION(pMeshMdl mesh, int numMigr)
{
  MigrCB cb;
  FMDB_Mesh_DspNumEnt(mesh);

/*
  SCUTIL_Sync();
  if (!SCUTIL_CommRank()) 
    cout<<"\n **** START ENTITY MIGRATION TEST || TOTAL ROUNDS = " << numMigr << " ***\n";
  SCUTIL_Sync();
*/

  int mesh_dim, maxNumPart, type;
  int num_migr = 0, num_add = 0, num_rmv = 0;
  FMDB_Mesh_GetDim(mesh, &mesh_dim);
  FMDB_Mesh_GetNumPart(mesh, &maxNumPart);

  std::map<pMeshEnt, std::pair<int, int> > entsToMigr;
  std::vector<pMeshEnt> newEnts[mesh_dim+1], oldEnts[mesh_dim+1];
  std::vector<MPI_Request*> reqs;
  pMeshEnt ent = NULL;
  int dest_pid, count, src_pid, numEnts;
  int has_ent, isBdry, numBdry, i = 0, entIndex;

  int this_pid = SCUTIL_CommRank();
  pPart part;
  FMDB_Mesh_GetPart(mesh, this_pid % mesh->getNumPart(), part);

  // repeat for numMigr times
  for (int count = 0; count < numMigr; count++)
  {
    dest_pid = -1;
    src_pid = -1;
    numBdry = -1;
    numEnts = -1;
    ent = NULL;
    entsToMigr.clear();

    FMDB_Part_GetNumEnt(part, FMDB_ALLTYPE, FMDB_ALLTOPO, &numEnts);
    if (numEnts > 0)
    {
      src_pid = this_pid;

      // randomly pick a type to choose from
      type = rand() % (mesh_dim+1);
      
      // randomly pick a boundary part to migrate to
      bool validSrc = false;
      while (!validSrc)
      {
        dest_pid = rand() % (SCUTIL_CommSize() * maxNumPart);
        if (dest_pid == src_pid)
          continue;
        validSrc = true;
      }

      // randomly pick an entity to migrate
      entIndex = rand() % numEnts;

      pPartEntIter entIter;
      int isEnd = FMDB_PartEntIter_Init(part, FMDB_ALLTYPE, FMDB_ALLTOPO, entIter);
      while (!isEnd)
      {
        isEnd = FMDB_PartEntIter_GetNext(entIter, ent);
        if (entIndex == i) 
          break;
        i++;
      }
      FMDB_PartEntIter_Del (entIter);
    
      num_migr += 1;
/*
      cout << this_pid << ": dest_pid = " << dest_pid << endl;
      cout << this_pid << ": type     = " << type << endl;
      cout << this_pid << ": src_pid  = " << src_pid << endl;
      cout << this_pid << ": ent      = " << FMDB_Ent_StrID(ent) << endl;
*/
    }

    // clear all reqs, oldEnt, newEnt
    for (int i = FMDB_VERTEX; i <= mesh_dim; i++)
    {
      newEnts[i].clear();
      oldEnts[i].clear();
    }
    reqs.clear();

    FMDB_Ent_Migr(mesh, cb, ent, src_pid, dest_pid, newEnts, oldEnts, reqs);
    
    for (int i = FMDB_VERTEX; i <= mesh_dim; i++)
      num_add += newEnts[i].size();
    for (int i = FMDB_VERTEX; i <= mesh_dim; i++)
      num_rmv += oldEnts[i].size();

    // reset IPComMan neighbors
    std::vector<int> nbr_proc;
    std::set<int> set_nbr_proc;
    FMDB_Mesh_GetNborProcID(mesh, nbr_proc);
    std::copy(nbr_proc.begin(), nbr_proc.end(), inserter(set_nbr_proc, set_nbr_proc.end()));
  
    ParUtil::Instance()->ComMan()->set_Nbrs(set_nbr_proc);
    ParUtil::Instance()->ComMan()->set_tag(0);
    ParUtil::Instance()->ComMan()->set_fixed_msg_size(0);

    // wait for all requests to complete
    for (int j = 0; j < reqs.size(); j ++)
    {
      MPI_Status status;
      MPI_Wait(reqs[j], &status);
    }
   SCUTIL_Sync();

 //   FMDB_Mesh_DspInfo (mesh);

    int isValid;
    FMDB_Mesh_Verify(mesh, &isValid);
    
    CHECK("\n* Checking Migrated Entity Mesh Validity...", isValid, SCUTIL_CommRank());
    //FMDB_Mesh_DspNumEnt (mesh);

    int hasEnt, isBdry;
    // verify the entity to be migrated is on destination part, and does not have a remote copy on src part
    if (this_pid == dest_pid)
    {
      FMDB_Part_FindEnt(part, ent, &hasEnt);
      if (!hasEnt)
        cout << this_pid << ": ERROR: " << FMDB_Ent_StrID(ent) << " not found on destination part." << endl;
      assert(hasEnt);

      std::vector<std::pair<int, pMeshEnt> > rcs;
      FMDB_Ent_GetAllRmt(ent, rcs);

      for (int j = 0; j < rcs.size(); ++j)
      {
        if (rcs[j].first == src_pid)
          cout << this_pid << ": ERROR: " << FMDB_Ent_StrID(ent) << " was found on source part." << endl;
        assert(rcs[j].first != src_pid);
      }
    }

    // verify that all newEnts are on the part
    for (int i = FMDB_VERTEX; i <= mesh_dim; i++)
    {
      for (int j = 0; j < newEnts[i].size(); j++) 
      {
        FMDB_Part_FindEnt(part, newEnts[i][j], &hasEnt);
        if (!hasEnt)
          cout << this_pid << ": " << FMDB_Ent_StrID(newEnts[i][j]) << " not found" << endl;
        assert(hasEnt);
      }
    }

    
  }
  cout << this_pid << ": MIGRATED " << num_migr << " entities." << endl;
  cout << this_pid << ": ADDED    " << num_add << " entities." << endl;
  cout << this_pid << ": REMOVED  " << num_rmv << " entities." << endl;
  return SCUtil_SUCCESS;
}
